// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.metrics.reporter;

import com.yammer.metrics.Metrics;
import com.yammer.metrics.core.MetricsRegistry;

import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.ClusterResource;
import org.apache.kafka.common.ClusterResourceListener;
import org.apache.kafka.common.MetricName;
import org.apache.kafka.common.errors.InterruptException;
import org.apache.kafka.common.metrics.KafkaMetric;
import org.apache.kafka.common.protocol.SecurityProtocol;
import org.apache.kafka.common.utils.Utils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.TimeUnit;
import java.util.regex.Pattern;

import io.confluent.metrics.record.ConfluentMetric;
import io.confluent.metrics.record.ConfluentMetric.KafkaMeasurable;
import io.confluent.metrics.record.ConfluentMetric.MetricsMessage;
import io.confluent.serializers.ProtoSerde;
import kafka.server.KafkaConfig;
import kafka.utils.ZkUtils;

import static io.confluent.metrics.record.ConfluentMetric.MetricType.BROKER;
import static io.confluent.metrics.record.ConfluentMetric.MetricType.CONSUMER;
import static io.confluent.metrics.record.ConfluentMetric.MetricType.PRODUCER;
import static io.confluent.metrics.record.ConfluentMetric.MetricType.UNKNOWN;


public class ConfluentMetricsReporter implements org.apache.kafka.common.metrics.MetricsReporter, ClusterResourceListener {
    private static final Logger log = LoggerFactory.getLogger(ConfluentMetricsReporter.class);
    //Yammer metric registry
    private static final MetricsRegistry metricsRegistry = Metrics.defaultRegistry();

    private ConfluentMetricsReporterConfig metricsReporterConfig;

    private final ProtoSerde<MetricsMessage> metricsMessageSerde = new ProtoSerde<>(MetricsMessage.getDefaultInstance());

    // Kafka metrics
    private final Map<MetricName, KafkaMetric> metricMap = new ConcurrentHashMap<>();

    private final ScheduledThreadPoolExecutor executor = new ScheduledThreadPoolExecutor(1);
    private KafkaProducer<byte[], byte[]> producer;

    private SecurityProtocol securityProtocol = SecurityProtocol.PLAINTEXT;

    private long reportIntervalMs;
    private String publishTopic;
    private int publishTopicPartitions;
    private int publishTopicReplicas;
    private Pattern pattern = null;
    private volatile String clusterId = null;
    private int brokerId;
    private String clientId;
    private String groupId;
    private ConfluentMetric.MetricType metricType;

    @Override
    public void onUpdate(ClusterResource clusterResource) {
        this.clusterId = clusterResource.clusterId();

        log.info("Starting Confluent metrics reporter for cluster id {} with an interval of {} ms", clusterId, reportIntervalMs);
        executor.scheduleAtFixedRate(new MetricReportRunnable(),
                                     reportIntervalMs,
                                     reportIntervalMs,
                                     TimeUnit.MILLISECONDS);
    }

    /**
     * Configure this class with the given key-value pairs
     */
    public void configure(Map<String, ?> configs) {
        Properties producerProperties = ConfluentMetricsReporterConfig.getProducerProperties(configs);
        if (producerProperties.containsKey(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG))
            securityProtocol = SecurityProtocol.forName(producerProperties.getProperty(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG));
        metricsReporterConfig = new ConfluentMetricsReporterConfig(configs);
        reportIntervalMs = metricsReporterConfig.getLong(ConfluentMetricsReporterConfig.PUBLISH_PERIOD_CONFIG);
        publishTopic = metricsReporterConfig.getString(ConfluentMetricsReporterConfig.TOPIC_CONFIG);
        publishTopicPartitions = metricsReporterConfig.getInt(ConfluentMetricsReporterConfig.TOPIC_PARTITIONS_CONFIG);
        publishTopicReplicas = metricsReporterConfig.getInt(ConfluentMetricsReporterConfig.TOPIC_REPLICAS_CONFIG);
        String regexString = metricsReporterConfig.getString(ConfluentMetricsReporterConfig.WHITELIST_CONFIG).trim();
        pattern = regexString.isEmpty() ? null : Pattern.compile(regexString);

        if (configs.containsKey(KafkaConfig.BrokerIdProp())){
            metricType = BROKER;
        } else if (configs.containsKey(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG)){
            metricType = CONSUMER;
        } else if (configs.containsKey(ProducerConfig.KEY_SERIALIZER_CLASS_CONFIG)){
            metricType = PRODUCER;
        } else {
            metricType = UNKNOWN;
        }
        brokerId = metricType == BROKER ? Integer.parseInt((String) configs.get(KafkaConfig.BrokerIdProp())) : -1;

        boolean isClient = metricType == CONSUMER || metricType == PRODUCER;
        clientId = isClient && configs.containsKey(CommonClientConfigs.CLIENT_ID_CONFIG) ? (String) configs.get(CommonClientConfigs.CLIENT_ID_CONFIG) : "";
        groupId = (metricType == CONSUMER) && configs.containsKey(ConsumerConfig.GROUP_ID_CONFIG) ? (String) configs.get(ConsumerConfig.GROUP_ID_CONFIG) : "";

        producer = new KafkaProducer<>(producerProperties);
    }

    private boolean createTopicIfNotPresent(SecurityProtocol protocol, String publishTopic) {
        String zkUrl = metricsReporterConfig.getString(ConfluentMetricsReporterConfig.ZK_URL_CONFIG);
        long retentionMs = metricsReporterConfig.getLong(ConfluentMetricsReporterConfig.TOPIC_RETENTION_MS_CONFIG);
        long retentionBytes = metricsReporterConfig.getLong(ConfluentMetricsReporterConfig.TOPIC_RETENTION_BYTES_CONFIG);
        long rollMs = metricsReporterConfig.getLong(ConfluentMetricsReporterConfig.TOPIC_ROLL_MS_CONFIG);

        ZkUtils zkUtils = KafkaAdminHelper.zkUtils(zkUrl);

        boolean isTopicCreated = false;
        try {
            Set<String> topics = new HashSet<>();
            topics.add(publishTopic);
            isTopicCreated = KafkaAdminHelper.createTopicsIfNotPresent(zkUtils, protocol, topics,
                publishTopicPartitions, publishTopicReplicas, retentionMs, retentionBytes, rollMs);
        } finally {
            zkUtils.close();
        }

        return isTopicCreated;
    }

    /**
     * Called when the metrics repository is closed.
     */
    public void close() {
        log.info("Stopping Confluent metrics reporter");
        executor.shutdownNow();
        if (producer != null) {
            synchronized (producer) {
                producer.close(0, TimeUnit.MILLISECONDS);
            }
        }
    }

    /**
     * This is called when the reporter is first registered to initially register all existing metrics
     * @param metrics All currently existing metrics
     */
    public void init(List<KafkaMetric> metrics) {
        for (KafkaMetric m : metrics)
            metricMap.put(m.metricName(), m);

        executor.setContinueExistingPeriodicTasksAfterShutdownPolicy(false);
        executor.setExecuteExistingDelayedTasksAfterShutdownPolicy(false);
        executor.setThreadFactory(new ThreadFactory() {
            public Thread newThread(Runnable runnable) {
                return Utils.newThread("confluent-metrics-reporter-scheduler", runnable, true);
            }
        });
    }

    /**
     * This is called whenever a metric is updated or added
     * @param metric
     */
    public void metricChange(KafkaMetric metric) {
        metricMap.put(metric.metricName(), metric);
    }

    /**
     * This is called whenever a metric is removed
     * @param metric
     */
    public void metricRemoval(KafkaMetric metric) {
        log.debug("removing kafka metric : {}", metric.metricName());
        metricMap.remove(metric.metricName());
    }

    private class MetricReportRunnable implements Runnable {
        private final MetricsMessage.Builder metricsMessageBuilder = MetricsMessage.newBuilder();
        private boolean isTopicCreated = false;

        public MetricReportRunnable() {
        }

        public void run() {
            try {
                if (!isTopicCreated)
                    isTopicCreated = createTopicIfNotPresent(securityProtocol, publishTopic);

                // if topic can't be created, skip the rest
                if (!isTopicCreated)
                    return;

                log.debug("Begin publishing metrics");

                MetricsMessage metricsMessage = genMetricsMessage();
                log.debug("Generated metric message : {}", metricsMessage);

                synchronized (producer) {
                    // producer may already be closed if we are shutting down
                    if (!Thread.currentThread().isInterrupted()) {
                        producer.send(new ProducerRecord<byte[], byte[]>(
                            publishTopic,
                            null,
                            metricsMessageSerde.serialize(metricsMessage))
                        );
                    }
                }
            } catch (InterruptException e) {
                // broker is shutting shutdown, interrupt flag is taken care of by InterruptException constructor
            } catch (Throwable t) {
                log.warn("Failed to publish metrics message in the reporter", t);
            }
        }

        private MetricsMessage genMetricsMessage() {
            long now = System.currentTimeMillis();

            metricsMessageBuilder.clear();

            // add top level fields
            metricsMessageBuilder.setMetricType(metricType);
            metricsMessageBuilder.setBrokerId(brokerId);
            metricsMessageBuilder.setClientId(clientId);
            metricsMessageBuilder.setGroupId(groupId);
            metricsMessageBuilder.setClusterId(clusterId == null ? "" : clusterId);
            metricsMessageBuilder.setTimestamp(now);

            // add Kafka metrics
            List<KafkaMeasurable> kafkaMeasurables = KafkaMetricsHelper.collectKafkaMetrics(metricMap, pattern);
            metricsMessageBuilder.addAllKafkaMeasurable(kafkaMeasurables);

            // add Yammer metrics if the source is a broker, the clients don't produce Yammer metrics
            if (metricType == BROKER) {
                YammerMetricsHelper.YammerMetricsResult yammerMetricsResult = YammerMetricsHelper.collectYammerMetrics(metricsRegistry, pattern);
                metricsMessageBuilder.addAllYammerGauge(yammerMetricsResult.gauges);
                metricsMessageBuilder.addAllYammerMeter(yammerMetricsResult.meters);
                metricsMessageBuilder.addAllYammerHistogram(yammerMetricsResult.histograms);
                metricsMessageBuilder.addAllYammerTimer(yammerMetricsResult.timers);
            }

            return metricsMessageBuilder.build();
        }
    }
}
