// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.metrics.reporter;

import io.confluent.monitoring.common.MonitoringProducerDefaults;
import io.confluent.monitoring.common.TimeBucket;
import kafka.server.KafkaConfig;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.common.config.AbstractConfig;
import org.apache.kafka.common.config.ConfigDef;
import org.apache.kafka.common.config.ConfigException;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

public class ConfluentMetricsReporterConfig extends AbstractConfig {

    public static final String METRICS_REPORTER_PREFIX = "confluent.metrics.reporter.";

    private static final ConfigDef config;

    public static final String BOOTSTRAP_SERVERS_CONFIG = METRICS_REPORTER_PREFIX + ProducerConfig.BOOTSTRAP_SERVERS_CONFIG;
    public static final String BOOTSTRAP_SERVERS_DOC = "Bootstrap servers for publish metrics.";

    public static final String ZK_URL_CONFIG = METRICS_REPORTER_PREFIX + KafkaConfig.ZkConnectProp();
    public static final String ZK_URL_DOC = "Zookeeper url for the Kafka cluster to publish metrics. This is needed to create the metrics topic.";

    public static final String TOPIC_CONFIG = METRICS_REPORTER_PREFIX + "topic";
    public static final String DEFAULT_TOPIC_CONFIG = "_confluent-metrics";
    public static final String TOPIC_DOC = "Topic on which metrics data will be written.";

    public static final String TOPIC_PARTITIONS_CONFIG = METRICS_REPORTER_PREFIX + "topic.partitions";
    public static final int DEFAULT_TOPIC_PARTITIONS_CONFIG = 10;
    public static final String TOPIC_PARTITIONS_DOC = "Number of partitions in the metric topic.";

    public static final String TOPIC_REPLICAS_CONFIG = METRICS_REPORTER_PREFIX + "topic.replicas";
    public static final int DEFAULT_TOPIC_REPLICAS_CONFIG = 3;
    public static final String TOPIC_REPLICAS_DOC = "Number of replicas in the metric topic.";

    public static final String TOPIC_RETENTION_MS_CONFIG = METRICS_REPORTER_PREFIX + "topic.retention.ms";
    public static final long DEFAULT_TOPIC_RETENTION_MS_CONFIG = 4 /*hours*/ * 3600 * 1000L;
    public static final String TOPIC_RETENTION_MS_DOC = "Retention time for the metrics topic.";

    public static final String TOPIC_RETENTION_BYTES_CONFIG = METRICS_REPORTER_PREFIX + "topic.retention.bytes";
    public static final long DEFAULT_TOPIC_RETENTION_BYTES_CONFIG = -1L;
    public static final String TOPIC_RETENTION_BYTES_DOC = "Retention bytes for the metrics topic.";

    public static final String TOPIC_ROLL_MS_CONFIG = METRICS_REPORTER_PREFIX + "topic.roll.ms";
    public static final long DEFAULT_TOPIC_ROLL_MS_CONFIG = 4 /*hours*/ * 3600 * 1000L;
    public static final String TOPIC_ROLL_MS_DOC = "Log rolling time for the metrics topic.";

    public static final String PUBLISH_PERIOD_CONFIG = METRICS_REPORTER_PREFIX + "publish.ms";
    public static final Long DEFAULT_PUBLISH_PERIOD = TimeBucket.SIZE;
    public static final String PUBLISH_PERIOD_DOC =
            "Period the metrics reporter should use to publish messages to " + METRICS_REPORTER_PREFIX +
                    TOPIC_CONFIG + ".";

    public static final String WHITELIST_CONFIG = METRICS_REPORTER_PREFIX + "whitelist";
    public static final String DEFAULT_WHITELIST_PERIOD = ".*MaxLag.*|kafka.log:type=Log,name=Size.*";
    public static final String WHITELIST_DOC =
            "Regex matching the yammer metric mbean name or Kafka metric name.";

    static {
        config = new ConfigDef()
                .define(
                        BOOTSTRAP_SERVERS_CONFIG,
                        ConfigDef.Type.STRING,
                        ConfigDef.Importance.HIGH,
                        BOOTSTRAP_SERVERS_DOC
                ).define(
                        ZK_URL_CONFIG,
                        ConfigDef.Type.STRING,
                        ConfigDef.Importance.HIGH,
                        ZK_URL_DOC
                ).define(
                        TOPIC_CONFIG,
                        ConfigDef.Type.STRING,
                        DEFAULT_TOPIC_CONFIG,
                        ConfigDef.Importance.MEDIUM,
                        TOPIC_DOC
                ).define(
                        TOPIC_PARTITIONS_CONFIG,
                        ConfigDef.Type.INT,
                        DEFAULT_TOPIC_PARTITIONS_CONFIG,
                        ConfigDef.Importance.MEDIUM,
                        TOPIC_PARTITIONS_DOC
                ).define(
                        TOPIC_REPLICAS_CONFIG,
                        ConfigDef.Type.INT,
                        DEFAULT_TOPIC_REPLICAS_CONFIG,
                        ConfigDef.Importance.MEDIUM,
                        TOPIC_REPLICAS_DOC
                ).define(
                        TOPIC_RETENTION_MS_CONFIG,
                        ConfigDef.Type.LONG,
                        DEFAULT_TOPIC_RETENTION_MS_CONFIG,
                        ConfigDef.Importance.MEDIUM,
                        TOPIC_RETENTION_MS_DOC
                ).define(
                        TOPIC_RETENTION_BYTES_CONFIG,
                        ConfigDef.Type.LONG,
                        DEFAULT_TOPIC_RETENTION_BYTES_CONFIG,
                        ConfigDef.Importance.MEDIUM,
                        TOPIC_RETENTION_BYTES_DOC
                ).define(
                        TOPIC_ROLL_MS_CONFIG,
                        ConfigDef.Type.LONG,
                        DEFAULT_TOPIC_ROLL_MS_CONFIG,
                        ConfigDef.Importance.MEDIUM,
                        TOPIC_ROLL_MS_DOC
                ).define(
                        PUBLISH_PERIOD_CONFIG,
                        ConfigDef.Type.LONG,
                        DEFAULT_PUBLISH_PERIOD,
                        ConfigDef.Importance.MEDIUM,
                        PUBLISH_PERIOD_DOC
                ).define(
                        WHITELIST_CONFIG,
                        ConfigDef.Type.STRING,
                        DEFAULT_WHITELIST_PERIOD,
                        ConfigDef.Importance.MEDIUM,
                        WHITELIST_DOC
                );
    }

    public ConfluentMetricsReporterConfig(Properties props) {
        super(config, props);
    }

    public ConfluentMetricsReporterConfig(Map<String, ?> clientConfigs) {
        super(config, clientConfigs);
    }

    public static void main(String[] args) {
        System.out.println(config.toHtmlTable());
    }

    private static Map<String, Object> producerConfigDefaults()  {
        Map<String, Object> defaults = new HashMap<>();
        defaults.put(ProducerConfig.ACKS_CONFIG, "1");
        defaults.put(ProducerConfig.KEY_SERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.ByteArraySerializer");
        defaults.put(ProducerConfig.VALUE_SERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.ByteArraySerializer");
        defaults.put(ProducerConfig.COMPRESSION_TYPE_CONFIG, MonitoringProducerDefaults.COMPRESSION_TYPE_CONFIG);
        defaults.put(ProducerConfig.LINGER_MS_CONFIG, MonitoringProducerDefaults.LINGER_MS_CONFIG);
        defaults.put(ProducerConfig.INTERCEPTOR_CLASSES_CONFIG, "");
        defaults.put(ProducerConfig.CLIENT_ID_CONFIG, "confluent-metrics-reporter");
        return defaults;
    }

    public static Properties getProducerProperties(Map<String, ?> clientConfigs) {
        Properties props = new Properties();
        for (Map.Entry<String, ?> entry : clientConfigs.entrySet()) {
            if (entry.getKey().startsWith(METRICS_REPORTER_PREFIX))
                props.put(entry.getKey().substring(METRICS_REPORTER_PREFIX.length()), entry.getValue());
        }
        for (Map.Entry<String, ?> entry : producerConfigDefaults().entrySet())
            props.put(entry.getKey(), entry.getValue());

        // we require bootstrap servers
        Object bootstrap = props.get(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG);
        if (bootstrap == null)
            throw new ConfigException("Missing required property " + METRICS_REPORTER_PREFIX + ProducerConfig.BOOTSTRAP_SERVERS_CONFIG);

        return props;
    }
}