// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.metrics.reporter;

import kafka.admin.AdminUtils;
import kafka.admin.RackAwareMode;
import kafka.log.LogConfig$;
import kafka.utils.ZkUtils;
import org.apache.kafka.common.protocol.Errors;
import org.apache.kafka.common.protocol.SecurityProtocol;
import org.apache.kafka.common.requests.MetadataResponse;
import org.apache.kafka.common.security.JaasUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashSet;
import java.util.Properties;
import java.util.Set;

/**
 *  Code copied from KafkaHelper to avoid dragging more dependencies
 */
public class KafkaAdminHelper {
    private static final Logger log = LoggerFactory.getLogger(KafkaAdminHelper.class);

    public static ZkUtils zkUtils(String zkUrl) {
        return ZkUtils.apply(zkUrl, 30000, 7000, JaasUtils.isZkSecurityEnabled());
    }

    private static Set<MetadataResponse.TopicMetadata> fetchTopicMetadataFromZk(Set<String> topics, ZkUtils zkUtils, SecurityProtocol protocol) {
        return scala.collection.JavaConversions.setAsJavaSet(
                AdminUtils.fetchTopicMetadataFromZk(scala.collection.JavaConversions.asScalaSet(topics), zkUtils, protocol)
        );
    }

    private static void createTopic(ZkUtils zkUtils, String topic, int partitions, int replication, Properties props, RackAwareMode.Enforced$ module$) {
        AdminUtils.createTopic(zkUtils, topic, partitions, replication, props, module$);
    }

    public static boolean createTopicsIfNotPresent(ZkUtils zkUtils, SecurityProtocol protocol, Set<String> topics, int partitions, int replication, long retentionMs, long retentionBytes, long rollingMs) {
        Set<MetadataResponse.TopicMetadata> k = fetchTopicMetadataFromZk(topics, zkUtils, protocol);

        if (k.size() != topics.size()) {
            log.error("expected count={} topics actual={}", topics.size(), k.size());
            return false;
        }

        Set<String> invalidTopics = new HashSet<>();
        boolean topicInError = false;
        for (MetadataResponse.TopicMetadata topicMetadata : k) {
            if (Errors.UNKNOWN_TOPIC_OR_PARTITION.equals(topicMetadata.error())) {
                invalidTopics.add(topicMetadata.topic());
                continue;
            } else if (topicMetadata.error() != null && !Errors.NONE.equals(topicMetadata.error())) {
                log.error("topic={} in error={}", topicMetadata.topic(), topicMetadata.error());
                topicInError = true;
                continue;
            }
            if (!checkPartitions(topicMetadata, partitions)) {
                log.warn("found topic={} with partitions={} instead of expectedPartitions={}", topicMetadata.topic(), topicMetadata.partitionMetadata().size(), partitions);
            }
            for (MetadataResponse.PartitionMetadata partitionMetadata : topicMetadata.partitionMetadata()) {
                if (!checkReplicas(partitionMetadata, replication))
                    log.warn("found topic={} with replication={} instead of expectedReplication={}", topicMetadata.topic(), partitionMetadata.replicas().size(), replication);
            }
        }
        // TODO: We should further validate if the configurations are expected.

        if (topicInError) {
            return false;
        }
        for (String topic : invalidTopics) {
            try {
                Properties props = getProperties(retentionMs, retentionBytes, rollingMs);
                createTopic(zkUtils, topic, partitions, replication, props, RackAwareMode.Enforced$.MODULE$);
                log.info("Created topic={} partitions={} replication={} retentionMs={} retentionBytes={} rollingMs={}", topic, partitions, replication, retentionMs, retentionBytes, rollingMs);
            } catch (Throwable e) {
                topicInError = true;
                log.error("Failed to create topic={} partitions={} replication={} retentionMs={} retentionBytes={} rollingMs={}. If the Kafka cluster has fewer than 3 brokers, make sure that {} is set to number of brokers.",
                          topic, partitions, replication, retentionMs, retentionBytes, rollingMs, ConfluentMetricsReporterConfig.TOPIC_REPLICAS_CONFIG, e);
            }
        }
        return !topicInError;
    }

    private static boolean checkPartitions(MetadataResponse.TopicMetadata topicMetadata, int partitions) {
        return topicMetadata.partitionMetadata().size() == partitions;
    }

    private static boolean checkReplicas(MetadataResponse.PartitionMetadata partitionMetadata, int replicas) {
        return partitionMetadata.replicas().size() == replicas;
    }

    private static Properties getProperties(long retentionMs, long retentionBytes, long rollingMs) {
        Properties props = new Properties();
        props.setProperty(LogConfig$.MODULE$.RetentionMsProp(), "" + retentionMs);
        props.setProperty(LogConfig$.MODULE$.RetentionBytesProp(), "" + retentionBytes);
        props.setProperty(LogConfig$.MODULE$.SegmentMsProp(), "" + rollingMs);
        return props;
    }
}
