// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.metrics.reporter;

import com.yammer.metrics.core.*;
import io.confluent.metrics.record.ConfluentMetric.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

public class YammerMetricsHelper {
    private static final Logger log = LoggerFactory.getLogger(ConfluentMetricsReporter.class);

    private static final YammerMetricName.Builder metricNameBuilder = YammerMetricName.newBuilder();
    private static final YammerGauge.Builder gaugeBuilder = YammerGauge.newBuilder();
    private static final YammerMeter.Builder meterBuilder = YammerMeter.newBuilder();
    private static final YammerHistogram.Builder histogramBuilder = YammerHistogram.newBuilder();
    private static final YammerTimer.Builder timerBuilder = YammerTimer.newBuilder();

    private YammerMetricsHelper() {
        // static class; can't instantiate
    }

    public static class YammerMetricsResult {
        public final List<YammerGauge> gauges;
        public final List<YammerMeter> meters;
        public final List<YammerHistogram> histograms;
        public final List<YammerTimer> timers;

        public YammerMetricsResult(List<YammerGauge> gauges,
                                   List<YammerMeter> meters,
                                   List<YammerHistogram> histograms,
                                   List<YammerTimer> timers) {
            this.gauges = gauges;
            this.meters = meters;
            this.histograms = histograms;
            this.timers = timers;
        }
    }

    public static YammerMetricsResult collectYammerMetrics(MetricsRegistry metricsRegistry, Pattern pattern) {
        Set<Map.Entry<MetricName, Metric>> metrics = metricsRegistry.allMetrics().entrySet();
        List<YammerGauge> yammerGauges = new ArrayList<>();
        List<YammerMeter> yammerMeters = new ArrayList<>();
        List<YammerHistogram> yammerHistograms = new ArrayList<>();
        List<YammerTimer> yammerTimers = new ArrayList<>();

        for (Map.Entry<MetricName, Metric> entry : metrics) {
            MetricName metricName = entry.getKey();
            if (pattern != null && !pattern.matcher(metricName.getMBeanName()).matches())
                continue;
            try {
                metricNameBuilder.clear();
                metricNameBuilder.setGroup(Utils.notNullOrEmpty(metricName.getGroup()));
                metricNameBuilder.setName(Utils.notNullOrEmpty(metricName.getName()));
                metricNameBuilder.setType(Utils.notNullOrEmpty(metricName.getType()));
                metricNameBuilder.setScope(Utils.notNullOrEmpty(metricName.getScope()));
                metricNameBuilder.setMBeanName(Utils.notNullOrEmpty(metricName.getMBeanName()));

                Metric metric = entry.getValue();
                if (metric instanceof Gauge) {
                    Gauge<Object> gauge = (Gauge<Object>) metric;
                    gaugeBuilder.clear();
                    gaugeBuilder.setMetricName(metricNameBuilder.build());
                    Object value = gauge.value();
                    gaugeBuilder.setValue(value == null ? Utils.EMPTY_STRING : gauge.value().toString());
                    yammerGauges.add(gaugeBuilder.build());
                } else if (metric instanceof Meter) {
                    Meter meter = (Meter) metric;
                    meterBuilder.clear();
                    meterBuilder.setMetricName(metricNameBuilder.build());
                    meterBuilder.setCount(meter.count());
                    meterBuilder.setOneMinuteRate(meter.oneMinuteRate());
                    meterBuilder.setFiveMinuteRate(meter.fiveMinuteRate());
                    meterBuilder.setFifteenMinuteRate(meter.fifteenMinuteRate());
                    meterBuilder.setMeanRate(meter.meanRate());
                    yammerMeters.add(meterBuilder.build());
                } else if (metric instanceof Histogram) {
                    Histogram histogram = (Histogram) metric;
                    histogramBuilder.clear();
                    histogramBuilder.setMetricName(metricNameBuilder.build());
                    histogramBuilder.setCount(histogram.count());
                    histogramBuilder.setMax(histogram.max());
                    histogramBuilder.setMin(histogram.min());
                    histogramBuilder.setMean(histogram.mean());
                    histogramBuilder.setStdDev(histogram.stdDev());
                    histogramBuilder.setSum(histogram.sum());
                    histogramBuilder.setMedian(histogram.getSnapshot().getMedian());
                    histogramBuilder.setPercentile75Th(histogram.getSnapshot().get75thPercentile());
                    histogramBuilder.setPercentile95Th(histogram.getSnapshot().get95thPercentile());
                    histogramBuilder.setPercentile98Th(histogram.getSnapshot().get98thPercentile());
                    histogramBuilder.setPercentile99Th(histogram.getSnapshot().get99thPercentile());
                    histogramBuilder.setPercentile999Th(histogram.getSnapshot().get999thPercentile());
                    histogramBuilder.setSize(histogram.getSnapshot().size());
                    yammerHistograms.add(histogramBuilder.build());
                } else if (metric instanceof Timer) {
                    Timer timer = (Timer) metric;
                    timerBuilder.clear();
                    timerBuilder.setMetricName(metricNameBuilder.build());
                    timerBuilder.setCount(timer.count());
                    timerBuilder.setMax(timer.max());
                    timerBuilder.setMin(timer.min());
                    timerBuilder.setMean(timer.mean());
                    timerBuilder.setStdDev(timer.stdDev());
                    timerBuilder.setSum(timer.sum());
                    timerBuilder.setMedian(timer.getSnapshot().getMedian());
                    timerBuilder.setPercentile75Th(timer.getSnapshot().get75thPercentile());
                    timerBuilder.setPercentile95Th(timer.getSnapshot().get95thPercentile());
                    timerBuilder.setPercentile98Th(timer.getSnapshot().get98thPercentile());
                    timerBuilder.setPercentile99Th(timer.getSnapshot().get99thPercentile());
                    timerBuilder.setPercentile999Th(timer.getSnapshot().get999thPercentile());
                    timerBuilder.setSize(timer.getSnapshot().size());
                    meterBuilder.setOneMinuteRate(timer.oneMinuteRate());
                    timerBuilder.setFiveMinuteRate(timer.fiveMinuteRate());
                    timerBuilder.setFifteenMinuteRate(timer.fifteenMinuteRate());
                    timerBuilder.setMeanRate(timer.meanRate());
                    yammerTimers.add(timerBuilder.build());
                }
                // Kafka current doesn't use other types of Yammer metrics.
            }
            catch (Exception e) {
                // Due to KAFKA-4286, we could get NullPointerException when reading a Yammer metric. It's better
                // to continue so that we can collect other metrics.
                log.warn("Unexpected error in processing Yammer metric {}", metricName, e);
            }
        }

        return new YammerMetricsResult(yammerGauges, yammerMeters, yammerHistograms, yammerTimers);
    }
}