// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing;

import io.airlift.airline.Cli;
import io.airlift.airline.Cli.CliBuilder;
import io.airlift.airline.Command;
import io.airlift.airline.Help;
import io.airlift.airline.Option;
import io.airlift.airline.ParseException;
import io.confluent.kafka.databalancing.exception.NoRebalanceInProgressException;
import io.confluent.kafka.databalancing.exception.RebalanceInProgressException;
import io.confluent.kafka.databalancing.exception.ValidationException;
import io.confluent.kafka.databalancing.license.DefaultLicenseValidator;
import io.confluent.kafka.databalancing.license.LicenseValidator;
import io.confluent.kafka.databalancing.report.RebalanceReport;
import io.confluent.kafka.databalancing.topology.ReplicaAssignment;
import io.confluent.kafka.databalancing.topology.TopologyUtils;
import io.confluent.kafka.databalancing.view.DefaultRebalancerFactory;
import kafka.common.AdminCommandFailedException;
import org.apache.kafka.common.utils.SystemTime;
import org.apache.kafka.common.utils.Utils;
import kafka.admin.AdminOperationException;
import scala.Console;

import java.io.Closeable;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Properties;

public class ConfluentRebalancerCommand {

    public static void main(String... args) {
        try {
            run(new DefaultRebalancerFactory(), new DefaultLicenseValidator(new SystemTime()), args);
        } catch (ParseException | ValidationException | AdminOperationException | AdminCommandFailedException e) {
            System.err.println(e.getMessage());
            System.exit(1);
        } catch (Throwable t) {
            System.err.println("Unexpected exception: " + t.getMessage());
            System.err.println(Utils.stackTrace(t));
            System.exit(1);
        }
    }

    public static void run(RebalancerFactory rebalancerFactory, LicenseValidator licenseValidator, String ... args) {
        CliBuilder<Runnable> builder = Cli.<Runnable>builder("confluent-rebalancer")
                .withDescription("Confluent Rebalancer Tool for Kafka")
                .withDefaultCommand(Help.class)
                .withCommands(Help.class, ProposedAssignment.class, Execute.class, Status.class, Finish.class);
        Cli<Runnable> cli = builder.build();
        Runnable runnable = cli.parse(args);

        // this is ugly, but not clear how to do better without a DI library/framework
        if (runnable instanceof BaseRebalanceCommand) {
            BaseRebalanceCommand command = (BaseRebalanceCommand) runnable;
            command.setRebalancerFactory(rebalancerFactory);
            command.setLicenseValidator(licenseValidator);
            try {
                command.run();
            } finally {
                command.close();
            }
        }
        else
            runnable.run();
    }

    abstract static class BaseRebalanceCommand implements Runnable, Closeable {
        @Option(name = "--zookeeper", required = true, description = "The connection string for the zookeeper " +
                "connection in the form host:port. Multiple URLS can be given to allow fail-over.")
        protected String zkConnect;

        @Option(name = "--config-file", description = "Configuration file in properties format")
        protected String configFile;

        protected RebalancerConfig rebalancerConfig;

        private Rebalancer rebalancer;

        private RebalancerFactory rebalancerFactory;

        private LicenseValidator licenseValidator;

        protected Rebalancer getOrCreateRebalancer() {
            if (rebalancer == null) {
                if (rebalancerFactory == null)
                    throw new IllegalStateException("rebalancerFactory has not been set");
                rebalancer = rebalancerFactory.create(zkConnect);
            }
            return rebalancer;
        }

        public void setLicenseValidator(LicenseValidator validator) {
            licenseValidator = validator;
        }


        public void setRebalancerFactory(RebalancerFactory rebalancerFactory) {
            this.rebalancerFactory = rebalancerFactory;
        }

        @Override
        public final void run() {
            rebalancerConfig = rebalancerConfig();
            try {
                if (rebalancerConfig.isTrial()) {
                    long timeRemaining = licenseValidator.validateTrialPeriod(zkConnect);
                    if (timeRemaining == 0) {
                        throw new ValidationException("Your trial period has expired. Please purchase a license"
                                + " to continue using this product.");
                    }
                } else {
                    licenseValidator.validateLicense(rebalancerConfig.licenseString());
                }
            } catch (LicenseValidator.ValidationFailedException e) {
                throw new ValidationException("Could not validate your license: " + e.getMessage());
            }
            doRun();
        }

        protected abstract void doRun();

        protected RebalancerConfig rebalancerConfig() {
            try {
                Properties rebalancerProps;
                if (configFile == null)
                    rebalancerProps = new Properties();
                else
                    rebalancerProps = Utils.loadProps(configFile);
                return new RebalancerConfig(rebalancerProps);
            } catch (IOException e) {
                throw new ValidationException("Could not load configuration properties: " + e.getMessage());
            }
        }

        public void close() {
            if (rebalancer != null)
                rebalancer.close();
        }

    }

    /** Base class for commands that compute a new partition assignment */
    abstract static class BaseReassignCommand extends BaseRebalanceCommand {
        @Option(name = "--metrics-bootstrap-server", required = true,
                description = "A list of host:port pairs to use for establishing the initial connection to the " +
                        "metrics Kafka cluster")
        protected String metricBootstrapServers;

        @Option(name = "--remove-broker-ids", description = "Partitions will be moved away from the specified brokers (comma-separated). " +
                "This option can be used to decommission brokers.")
        private String removeBrokerIds;

        @Override
        protected RebalancerConfig rebalancerConfig() {
            try {
                Properties rebalancerProps;
                if (configFile == null)
                    rebalancerProps = new Properties();
                else
                    rebalancerProps = Utils.loadProps(configFile);
                rebalancerProps.put(RebalancerConfig.METRICS_BOOTSTRAP_SERVERS_CONFIG, metricBootstrapServers);
                return new RebalancerConfig(rebalancerProps);
            } catch (IOException e) {
                throw new ValidationException("Could not load configuration properties: " + e.getMessage());
            }
        }

        protected List<Integer> removeBrokerIds() {
            if (removeBrokerIds == null)
                return Collections.emptyList();

            String[] values = removeBrokerIds.split(",");
            if (values.length == 0)
                throw removeBrokerIdsException();

            List<Integer> brokerIds = new ArrayList<>();
            for (String value : values) {
                try {
                    brokerIds.add(Integer.parseInt(value.trim()));
                } catch (NumberFormatException e) {
                    throw removeBrokerIdsException();
                }
            }
            return brokerIds;
        }

        private ValidationException removeBrokerIdsException() {
            return new ValidationException("Expected a comma-separated list of numeric broker ids for option " +
                    "'--remove-broker-ids', but received " + removeBrokerIds);
        }

    }

    @Command(name = "proposed-assignment",
            description = "Generate the proposed partition assignment and output it in JSON format")
    public static class ProposedAssignment extends BaseReassignCommand {

        @Override
        protected void doRun() {
            Rebalancer rebalancer = getOrCreateRebalancer();
            ProposedRebalance proposedRebalance = rebalancer.proposeRebalance(removeBrokerIds(), rebalancerConfig());
            String json = proposedRebalance.proposedAssignment().toJson();
            System.out.println(json);
        }

    }

    @Command(name = "execute",
            description = "Kick off a cluster-wide rebalance operation")
    public static class Execute extends BaseReassignCommand {

        @Option(name = "--throttle", required = true,
                description = "The maximum bandwidth, in bytes per second, allocated to moving replicas.")
        protected long replicationQuota;

        @Option(name = "--force", description = "Suppress prompts if true")
        protected boolean force;

        @Option(name = "--verbose", description = "Also output before/after per broker information (disk space usage, "
                + "leader count, replica count, topic partitions)")
        private boolean verbose = false;

        @Override
        protected void doRun() {
            Rebalancer rebalancer = getOrCreateRebalancer();

            RebalanceStatus status = rebalancer.status();

            if (rebalancer.maybeUpdateReplicationQuota(status.inProgressAssignment(), replicationQuota))
                return;

            System.out.println("Computing the rebalance plan (this may take a while) ...");

            ProposedRebalance proposedRebalance = rebalancer.proposeRebalance(removeBrokerIds(), rebalancerConfig());

            if (!proposedRebalance.shouldRebalance()) {
                if (rebalancer.disengageThrottle())
                    System.out.println("The throttle was removed.");
                System.out.println("The cluster is already balanced, exiting.");
                return;
            }

            RebalanceReport report = proposedRebalance.report();
            report.print(System.out, verbose);

            if (force)
                startRebalance(rebalancer, proposedRebalance);
            else {
                Boolean startRebalance = null;
                do {
                    System.out.print("Would you like to continue? (y/n): ");
                    String line = Console.readLine();
                    if (line != null) {
                        if (line.equalsIgnoreCase("y"))
                            startRebalance = true;
                        else if (line.equalsIgnoreCase("n"))
                            startRebalance = false;
                    }
                } while (startRebalance == null);

                System.out.println();
                if (startRebalance)
                    startRebalance(rebalancer, proposedRebalance);
                else
                    System.out.println("OK, exiting");
            }

        }

        private void startRebalance(Rebalancer rebalancer, ProposedRebalance proposedRebalance) {
            ReplicaAssignment currentAssignment = proposedRebalance.currentAssignment();
            ReplicaAssignment proposedAssignment = proposedRebalance.proposedAssignment();
            rebalancer.startRebalance(currentAssignment, proposedAssignment, replicationQuota);
            System.out.println("The rebalance has been started, run `status` to check progress.");
            System.out.println(String.format("\nWarning: You must run the `status` or `finish` command periodically, until the " +
                    "rebalance completes, to ensure the throttle is removed. You can also alter the throttle by re-running " +
                    "the execute command passing a new value.\n"));
        }
    }

    @Command(name = "status",
            description = "Show the status of the current rebalance operation (if there is one)")
    public static class Status extends BaseRebalanceCommand {
        @Override
        protected void doRun() {
            Rebalancer rebalancer = getOrCreateRebalancer();
            RebalanceStatus status = rebalancer.status();

            if (status.inProgressAssignment().isEmpty())
                throw new NoRebalanceInProgressException("No rebalance is currently in progress. If you have called `status` after " +
                        "a rebalance was started successfully, the rebalance has completed. Run the `execute` command to check if " +
                        "the cluster is balanced.");

            System.out.println("Partitions being rebalanced:");
            for (String line : TopologyUtils.partitionsByTopicToLines(status.inProgressAssignment().topicPartitions()))
                System.out.println("\t" + line);
        }
    }

    @Command(name = "finish",
            description = "Performs clean-up activities (like disabling replication throttling) as the rebalance completes")
    public static class Finish extends BaseRebalanceCommand {
        @Override
        protected void doRun() {
            Rebalancer rebalancer = getOrCreateRebalancer();
            RebalanceStatus status = rebalancer.status();

            if (!status.inProgressAssignment().isEmpty())
                throw new RebalanceInProgressException(status.inProgressAssignment().size() + " partitions are being rebalanced");

            System.out.println("The rebalance has completed and throttling has been disabled");
        }
    }

}

