// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing;

import io.confluent.kafka.databalancing.exception.ValidationException;
import io.confluent.kafka.databalancing.metric.Metrics;
import io.confluent.kafka.databalancing.topology.Broker;
import kafka.admin.BrokerMetadata;
import org.apache.kafka.common.TopicPartition;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class DefaultRebalanceContext implements MutableRebalanceContext {

    private final Map<String, Integer> replicationFactors;
    private final Map<Broker, String> brokerToRack;
    private final Map<String, List<Broker>> rackToBrokers;
    private final Set<Broker> brokersToBeRemoved;
    private final Map<TopicPartition, Long> partitionToSize;

    /* The variables below can be mutated by move, addReplica, etc. */
    private final Map<TopicPartition, List<Broker>> partitionToBrokers;
    private final Map<Broker, List<TopicPartition>> brokerToPartitions;
    private final Map<Broker, List<TopicPartition>> brokerToLeaders;
    private final Map<Broker, List<TopicPartition>> brokerToFollowers;

    public DefaultRebalanceContext(Set<Broker> allBrokers,
                                   Map<TopicPartition, List<Broker>> partitionToBrokers,
                                   Map<Broker, String> brokerToRack,
                                   Map<String, Integer> replicationFactors,
                                   Metrics metrics,
                                   Set<Broker> brokersToBeRemoved) {
        this.replicationFactors = replicationFactors;
        this.partitionToBrokers = partitionToBrokers;
        this.brokerToRack = brokerToRack;
        this.partitionToSize = Collections.unmodifiableMap(metrics.partitionToSize());
        this.brokersToBeRemoved = Collections.unmodifiableSet(brokersToBeRemoved);

        Map<Broker, List<TopicPartition>> brokerToPartitions = new HashMap<>();
        Map<Broker, List<TopicPartition>> brokerToFollowers = new HashMap<>();
        Map<Broker, List<TopicPartition>> brokerToLeaders = new HashMap<>();
        for (Map.Entry<TopicPartition, List<Broker>> entry : partitionToBrokers.entrySet()) {
            List<Broker> replicas = entry.getValue();
            if (!replicas.isEmpty()) {
                Broker leader = replicas.get(0);
                for (Broker broker : entry.getValue()) {
                    TopicPartition topicPartition = entry.getKey();
                    addOrUpdate(brokerToPartitions, topicPartition, broker);
                    if (broker.equals(leader))
                        addOrUpdate(brokerToLeaders, topicPartition, broker);
                    else
                        addOrUpdate(brokerToFollowers, topicPartition, broker);
                }
            }
        }

        Map<String, List<Broker>> rackToBrokers = new HashMap<>();
        for (Map.Entry<Broker, String> entry : brokerToRack.entrySet()) {
            String rack = entry.getValue();
            List<Broker> brokers = rackToBrokers.get(rack);
            if (brokers == null) {
                brokers = new ArrayList<>();
                rackToBrokers.put(rack, brokers);
            }
            Broker broker = entry.getKey();
            brokers.add(broker);
        }

        for (Broker broker : allBrokers) {
            putIfAbsent(brokerToPartitions, broker, Collections.<TopicPartition>emptyList());
            putIfAbsent(brokerToFollowers, broker, Collections.<TopicPartition>emptyList());
            putIfAbsent(brokerToLeaders, broker, Collections.<TopicPartition>emptyList());
        }

        this.rackToBrokers = rackToBrokers;
        this.brokerToPartitions = brokerToPartitions;
        this.brokerToFollowers = brokerToFollowers;
        this.brokerToLeaders = brokerToLeaders;
    }

    private static <K, V> void putIfAbsent(Map<K, V> map, K key, V value) {
        if (!map.containsKey(key))
            map.put(key, value);
    }

    private static void addOrUpdate(Map<Broker, List<TopicPartition>> brokerToPartitions, TopicPartition topicPartition,
                             Broker broker) {
        List<TopicPartition> partitions = brokerToPartitions.get(broker);
        if (partitions == null) {
            partitions = new ArrayList<>();
            brokerToPartitions.put(broker, partitions);
        }
        partitions.add(topicPartition);
    }

    @Override
    public List<Broker> brokers(TopicPartition topicPartition, String rack) {
        List<Broker> brokers = partitionToBrokers.get(topicPartition);
        if (brokers == null)
            return Collections.emptyList();
        if (rack == null)
            return Collections.unmodifiableList(brokers);

        List<Broker> brokersOnRack = new ArrayList<>(brokers.size());
        for (Broker broker : brokers) {
            if (rack.equals(brokerToRack.get(broker)))
                brokersOnRack.add(broker);
        }
        return Collections.unmodifiableList(brokersOnRack);
    }

    @Override
    public List<Broker> brokers(TopicPartition topicPartition) {
        List<Broker> brokers = partitionToBrokers.get(topicPartition);
        return brokers == null ? Collections.<Broker>emptyList() : Collections.unmodifiableList(brokers);
    }

    @Override
    public Set<String> allRacks() {
        return Collections.unmodifiableSet(rackToBrokers.keySet());
    }

    @Override
    public Collection<Broker> brokersOnRack(String rack) {
        if (rack == null)
            return allBrokers();

        List<Broker> brokers = rackToBrokers.get(rack);
        return brokers == null ? Collections.<Broker>emptyList() : Collections.unmodifiableList(brokers);
    }

    @Override
    public Collection<Broker> brokersOnRacks(Collection<String> racks) {
        Set<Broker> brokers = new HashSet<>();
        for (String rack : racks) {
            List<Broker> brokersOnRack = rackToBrokers.get(rack);
            if (brokersOnRack != null)
                brokers.addAll(brokersOnRack);
        }
        return Collections.unmodifiableCollection(brokers);
    }

    @Override
    public Collection<Broker> allBrokers() {
        return Collections.unmodifiableSet(brokerToPartitions.keySet());
    }

    @Override
    public Map<TopicPartition, List<Broker>> partitionToBrokers() {
        return Collections.unmodifiableMap(partitionToBrokers);
    }

    @Override
    public int replicationFactor(String topic) {
        Integer rf = replicationFactors.get(topic);
        if (rf == null)
            throw new IllegalArgumentException("Unknown topic " + topic);
        return rf;
    }

    @Override
    public Collection<TopicPartition> allPartitions() {
        return Collections.unmodifiableCollection(partitionToBrokers.keySet());
    }

    @Override
    public Broker leader(TopicPartition topicPartition) {
        List<Broker> replicas = brokers(topicPartition);
        return replicas.isEmpty() ? null : replicas.get(0);
    }

    @Override
    public List<TopicPartition> followers(Broker broker) {
        List<TopicPartition> followers = brokerToFollowers.get(broker);
        return followers == null ? Collections.<TopicPartition>emptyList() : Collections.unmodifiableList(followers);
    }

    @Override
    public int brokerTopicReplicaCount(Broker broker, String topic) {
        int result = 0;
        for (TopicPartition tp : replicas(broker)) {
            if (topic == null || tp.topic().equals(topic))
                result += 1;
        }
        return result;
    }

    @Override
    public List<TopicPartition> replicas(Broker broker) {
        List<TopicPartition> topicPartitions = brokerToPartitions.get(broker);
        return topicPartitions == null ? Collections.<TopicPartition>emptyList() : Collections.unmodifiableList(topicPartitions);
    }

    @Override
    public List<TopicPartition> leaders(Broker broker) {
        List<TopicPartition> leaders = brokerToLeaders.get(broker);
        return leaders == null ? Collections.<TopicPartition>emptyList() : Collections.unmodifiableList(leaders);
    }

    @Override
    public String brokerRack(Broker broker) {
        return brokerToRack.get(broker);
    }

    @Override
    public Collection<Broker> brokersToBeRemoved() {
        return brokersToBeRemoved;
    }

    @Override
    public Collection<String> topics() {
        Set<String> topics = new HashSet<>();
        for (TopicPartition tp : allPartitions())
            topics.add(tp.topic());
        return topics;
    }

    public int brokerTopicLeaderCount(Broker broker, String topic) {
        int result = 0;
        for (TopicPartition tp : leaders(broker)) {
            if (topic == null || tp.topic().equals(topic))
                result += 1;
        }
        return result;
    }

    @Override
    public long brokerSize(Broker broker) {
        return brokerTopicSize(broker, null);
    }

    @Override
    public long brokerTopicSize(Broker broker, String topic) {
        long result = 0;
        for (TopicPartition partition : replicas(broker)) {
            if (topic == null || partition.topic().equals(topic))
                result += partitionSize(partition);
        }
        return result;
    }

    @Override
    public long brokerLeadersSize(Broker broker) {
        return brokerTopicLeadersSize(broker, null);
    }

    @Override
    public long brokerTopicLeadersSize(Broker broker, String topic) {
        long result = 0;
        for (TopicPartition partition : leaders(broker)) {
            if (topic == null || partition.topic().equals(topic))
                result += partitionSize(partition);
        }
        return result;
    }

    @Override
    public long partitionSize(TopicPartition topicPartition) {
        Long size = partitionToSize.get(topicPartition);
        if (size == null)
            throw new IllegalArgumentException("Unknown partition " + topicPartition);
        return size;
    }

    public void addReplica(TopicPartition topicPartition, Broker destinationBroker) {
        List<Broker> brokers = new ArrayList<>(partitionToBrokers.get(topicPartition));
        if (brokers.contains(destinationBroker))
            throw new IllegalArgumentException("Broker " + destinationBroker + " already contains " + topicPartition);

        if (brokers.isEmpty())
            updateBrokerToPartitionsMap(brokerToLeaders, topicPartition, null, destinationBroker);
        else
            updateBrokerToPartitionsMap(brokerToFollowers, topicPartition, null, destinationBroker);

        brokers.add(destinationBroker);
        partitionToBrokers.put(topicPartition, brokers);

        List<TopicPartition> partitions = new ArrayList<>(brokerToPartitions.get(destinationBroker));
        if (partitions.contains(topicPartition))
            throw new AssertionError("Inconsistent state between partitionToBrokers and brokerToPartitions with regards to " + topicPartition + " and " + destinationBroker);
        partitions.add(topicPartition);
        brokerToPartitions.put(destinationBroker, partitions);
    }

    @Override
    public void removeReplica(TopicPartition topicPartition, Broker broker) {
        List<Broker> brokers = new ArrayList<>(partitionToBrokers.get(topicPartition));
        int index = brokers.indexOf(broker);
        if (index == -1)
            throw new IllegalArgumentException("Broker " + broker + " does not contain " + topicPartition);

        brokers.remove(index);
        partitionToBrokers.put(topicPartition, brokers);

        if (index == 0) {
            Broker newLeader = null;
            if (!brokers.isEmpty()) {
                newLeader = brokers.get(0);
                updateBrokerToPartitionsMap(brokerToFollowers, topicPartition, newLeader, null);
            }
            updateBrokerToPartitionsMap(brokerToLeaders, topicPartition, broker, newLeader);
        }
        else
            updateBrokerToPartitionsMap(brokerToFollowers, topicPartition, broker, null);

        List<TopicPartition> partitions = new ArrayList<>(brokerToPartitions.get(broker));
        if (!partitions.contains(topicPartition))
            throw new AssertionError("Inconsistent state between partitionToBrokers and brokerToPartitions with regards to " + topicPartition + " and " + broker);

        partitions.remove(topicPartition);
        brokerToPartitions.put(broker, partitions);
    }

    public static MutableRebalanceContext create(List<BrokerMetadata> allBrokers,
                                                 Map<TopicPartition, List<Integer>> partitionToReplicaIds,
                                                 Map<String, Integer> replicationFactors,
                                                 Metrics metrics,
                                                 List<Broker> brokersToRemove) {
        Set<Broker> allBrokersSet = new HashSet<>(allBrokers.size());
        Map<Broker, String> brokerToRack = new HashMap<>(allBrokers.size());
        for (BrokerMetadata broker : allBrokers) {
            allBrokersSet.add(new Broker(broker.id()));
            if (broker.rack().isDefined())
                brokerToRack.put(new Broker(broker.id()), broker.rack().get());
        }

        Map<TopicPartition, List<Broker>> partitionToBrokers = new HashMap<>(partitionToReplicaIds.size());
        for (Map.Entry<TopicPartition, List<Integer>> entry : partitionToReplicaIds.entrySet()) {
            TopicPartition tp = entry.getKey();

            if (!metrics.partitionToSize().containsKey(tp))
                throw new IllegalArgumentException("`partitionToSize` is missing an entry for partition " + tp);
            if (!replicationFactors.containsKey(tp.topic()))
                throw new IllegalArgumentException("`replicationFactors` is missing an entry for topic " + tp.topic());

            List<Integer> brokerIds = entry.getValue();
            List<Broker> brokers = new ArrayList<>(brokerIds.size());
            for (Integer brokerId : brokerIds) {
                Broker broker = new Broker(brokerId);
                if (!(allBrokersSet.contains(broker) || brokersToRemove.contains(broker))) {
                    // brokerId exists in the current assignment but not in the list of online brokers in zookeeper and
                    // is not being removed. Make a suggestion to the user to rebalance data away from this broker.
                    throw new ValidationException("Broker with id " + brokerId + " is not currently online. Please " +
                            " add this broker to the removal list so partitions can be moved away from it.");
                }

                brokers.add(broker);
            }
            partitionToBrokers.put(tp, brokers);
        }

        return new DefaultRebalanceContext(allBrokersSet, partitionToBrokers, brokerToRack, replicationFactors,
                metrics, new HashSet<>(brokersToRemove));
    }

    @Override
    public Set<String> racks(List<Broker> brokers) {
        Set<String> racks = new HashSet<>();
        for (Broker broker : brokers) {
            String rack = brokerToRack.get(broker);
            if (rack != null)
                racks.add(rack);
        }
        return Collections.unmodifiableSet(racks);
    }

    public void movePartition(TopicPartition topicPartition, Broker from, Broker to) {
        updateBrokerToPartitionsMap(brokerToPartitions, topicPartition, from, to);

        List<Broker> brokers = new ArrayList<>(partitionToBrokers.get(topicPartition));
        int index = brokers.indexOf(from);
        if (index == -1)
            throw new AssertionError("Inconsistent state between partitionToBrokers and brokerToPartitions with regards to " + topicPartition + " and " + from);
        if (brokers.contains(to))
            throw new AssertionError("Inconsistent state between partitionToBrokers and brokerToPartitions with regards to " + topicPartition + " and " + to);
        if (index == 0)
            updateBrokerToPartitionsMap(brokerToLeaders, topicPartition, from, to);
        else
            updateBrokerToPartitionsMap(brokerToFollowers, topicPartition, from, to);
        brokers.set(index, to);
        partitionToBrokers.put(topicPartition, brokers);
    }

    public void makeLeader(TopicPartition topicPartition, Broker newLeader) {
        List<Broker> replicas = new ArrayList<>(partitionToBrokers.get(topicPartition));
        Broker currentLeader = replicas.get(0);
        if (currentLeader.equals(newLeader))
            throw new IllegalArgumentException("Broker " + newLeader + " is already the leader for partition " + topicPartition);
        if (!replicas.contains(newLeader))
            throw new IllegalArgumentException("Broker " + newLeader + " does not hold partition " + topicPartition);
        replicas.remove(newLeader);
        replicas.add(0, newLeader);
        partitionToBrokers.put(topicPartition, replicas);

        updateBrokerToPartitionsMap(brokerToLeaders, topicPartition, currentLeader, newLeader);
        updateBrokerToPartitionsMap(brokerToFollowers, topicPartition, newLeader, currentLeader);
    }

    private static void updateBrokerToPartitionsMap(Map<Broker, List<TopicPartition>> map, TopicPartition topicPartition,
                                                    Broker toRemove, Broker toAdd) {
        if (toRemove != null) {
            List<TopicPartition> partitionsToRemove = new ArrayList<>(map.get(toRemove));
            if (!partitionsToRemove.contains(topicPartition))
                throw new IllegalArgumentException(toRemove + " does not hold partition " + topicPartition);
            partitionsToRemove.remove(topicPartition);
            map.put(toRemove, partitionsToRemove);
        }

        if (toAdd != null) {
            List<TopicPartition> partitionsToAdd = new ArrayList<>(map.get(toAdd));
            if (partitionsToAdd.contains(topicPartition))
                throw new IllegalArgumentException(toAdd + " already holds partition " + topicPartition);
            partitionsToAdd.add(topicPartition);
            map.put(toAdd, partitionsToAdd);
        }
    }
}
