// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing;

import io.confluent.kafka.databalancing.metric.MetricsCollector;
import io.confluent.kafka.databalancing.metric.Metrics;
import io.confluent.kafka.databalancing.throttle.Throttle;
import io.confluent.kafka.databalancing.throttle.Throttler;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.ReplicaAssignment;
import io.confluent.kafka.databalancing.topology.TopologyUtils;
import kafka.admin.AdminUtils;
import kafka.admin.BrokerMetadata;
import kafka.admin.RackAwareMode;
import kafka.common.AdminCommandFailedException;
import kafka.common.TopicAndPartition;
import kafka.controller.ReassignedPartitionsContext;
import kafka.utils.ZkUtils;
import org.I0Itec.zkclient.exception.ZkNodeExistsException;
import org.apache.kafka.common.errors.TimeoutException;
import org.apache.kafka.common.security.JaasUtils;
import org.apache.kafka.common.TopicPartition;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import scala.Option;
import scala.collection.JavaConversions;
import scala.collection.Seq;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

public class DefaultRebalancer extends AbstractRebalancer {
    private static final Logger logger = LoggerFactory.getLogger(DefaultRebalancer.class);

    private final ZkUtils zkUtils;
    private final RebalancePolicy rebalancePolicy = new MovesOptimisedRebalancePolicy();
    private final Throttle throttle;

    public DefaultRebalancer(String zkUrl) {
        this(ZkUtils.apply(zkUrl, 30000, 7000, JaasUtils.isZkSecurityEnabled()));
    }

    DefaultRebalancer(ZkUtils zkUtils) {
        this(zkUtils, new Throttler(zkUtils));
    }

    DefaultRebalancer(ZkUtils zkUtils, Throttler throttler) {
        this.zkUtils = zkUtils;
        throttle = throttler;
    }

    @Override
    public ProposedRebalance proposeRebalance(List<Integer> brokerIdsToRemove, RebalancerConfig config) {
        ReplicaAssignment currentAssignment = currentAssignment();

        Metrics metrics;
        try {
            metrics = metrics(config, currentAssignment.topicPartitions());
        } catch (TimeoutException e) {
            throw new AdminCommandFailedException("Failed to retrieve metrics data. " + e.getMessage());
        }

        List<BrokerMetadata> allBrokers = allBrokers();
        List<Broker> brokersToRemove = brokersToRemove(allBrokers, currentAssignment, brokerIdsToRemove);
        ReplicaAssignment proposedAssignment = rebalancePolicy.rebalancePartitions(allBrokers, currentAssignment,
                currentAssignment.replicationFactors(), metrics, brokersToRemove);

        return new ProposedRebalance(allBrokers, currentAssignment, proposedAssignment, metrics, brokersToRemove);
    }

    public void startRebalance(ReplicaAssignment currentAssignment, ReplicaAssignment proposedAssignment, long replicationQuota) {
        throttle.engage(replicationQuota, currentAssignment, proposedAssignment);
        try {
            //TODO Validate partitions by checking if they exist in ZK. This is done in the open-source tool as
            //it processes user input while this tool does not.

            // The controller logs errors if assignedReplicas == proposedReplicas even though it ignores such partitions
            ReplicaAssignment proposedAssignmentChanges = retainChangedAssignments(currentAssignment, proposedAssignment);

            zkUtils.createPersistentPath(ZkUtils.ReassignPartitionsPath(), proposedAssignmentChanges.toJson(),
                    zkUtils.DefaultAcls());
        } catch (ZkNodeExistsException e) {
            System.out.println("The throttle rate was updated to " + replicationQuota + " bytes/sec.");
            throw partitionsBeingReassignedException();
        }
    }

    @Override
    public boolean maybeUpdateReplicationQuota(ReplicaAssignment inProgressAssignment, long replicationQuota) {
        if (!inProgressAssignment.isEmpty()) {
            throttle.engage(replicationQuota, currentAssignment(), inProgressAssignment);
            System.out.println("The throttle rate was updated to " + replicationQuota + " bytes/sec.");
            System.out.println("A rebalance is currently in progress for:");
            for (String line : TopologyUtils.partitionsByTopicToLines(inProgressAssignment.topicPartitions()))
                System.out.println("\t" + line);
            return true;
        }
        return false;
    }

    public ReplicaAssignment retainChangedAssignments(ReplicaAssignment currentAssignment, ReplicaAssignment proposedAssignment) {
        Map<TopicPartition, List<Integer>> result = new HashMap<>();
        for (TopicPartition tp : proposedAssignment.topicPartitions()) {
            List<Integer> currentReplicas = currentAssignment.replicas(tp);
            List<Integer> proposedReplicas = proposedAssignment.replicas(tp);
            if (!currentReplicas.equals(proposedReplicas))
                result.put(tp, proposedReplicas);
        }
        return new ReplicaAssignment(result);
    }

    @Override
    public boolean disengageThrottle() {
        return throttle.disengage();
    }

    @Override
    public RebalanceStatus status() {
        Map<TopicAndPartition, ReassignedPartitionsContext> partitionToReassignContext =
                JavaConversions.mapAsJavaMap(zkUtils.getPartitionsBeingReassigned());
        Map<TopicPartition, List<Integer>> partitionToReplicas = new HashMap<>();
        for (Map.Entry<TopicAndPartition, ReassignedPartitionsContext> entry : partitionToReassignContext.entrySet()) {
            TopicPartition tp = new TopicPartition(entry.getKey().topic(), entry.getKey().partition());
            List<Integer> brokerIds = ScalaUtils.fromScalaSeqInt(entry.getValue().newReplicas());
            partitionToReplicas.put(tp, brokerIds);
        }

        if (partitionToReplicas.isEmpty())
            disengageThrottle();

        return new RebalanceStatus(new ReplicaAssignment(partitionToReplicas));
    }

    private Metrics metrics(RebalancerConfig config, Collection<TopicPartition> allPartitions) throws TimeoutException {
        Properties consumerProps = new Properties();
        consumerProps.putAll(config.originalsWithPrefix(RebalancerConfig.METRICS_PREFIX));
        consumerProps.remove(RebalancerConfig.METRICS_TOPIC_CONFIG);
        String metricsTopic = config.getString(RebalancerConfig.METRICS_TOPIC_CONFIG);
        String clusterId = zkUtils.getClusterId().isEmpty() ? null : zkUtils.getClusterId().get();
        return MetricsCollector.collectMetrics(metricsTopic, consumerProps, allPartitions, clusterId);
    }

    private ReplicaAssignment currentAssignment() {
        return ReplicaAssignment.fromScala(zkUtils.getReplicaAssignmentForTopics(zkUtils.getAllTopics()));
    }

    private List<BrokerMetadata> allBrokers() {
        Seq<BrokerMetadata> brokers = AdminUtils.getBrokerMetadatas(zkUtils, RackAwareMode.Enforced$.MODULE$,
                Option.<Seq<Object>>empty());
        return JavaConversions.seqAsJavaList(brokers);
    }

    private AdminCommandFailedException partitionsBeingReassignedException() {
        Map<TopicAndPartition, ReassignedPartitionsContext> reassignmentMap =
                JavaConversions.mapAsJavaMap(zkUtils.getPartitionsBeingReassigned());

        Set<TopicPartition> partitionsBeingReassigned = new HashSet<>();
        for (TopicAndPartition tp : reassignmentMap.keySet())
            partitionsBeingReassigned.add(new TopicPartition(tp.topic(), tp.partition()));

        return partitionsBeingReassignedException(partitionsBeingReassigned);
    }

    @Override
    public void close() {
        try {
            zkUtils.close();
        } catch (Throwable t) {
            logger.warn("An exception was thrown while closing zkUtils", t);
        }
    }

}
