// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing;

import io.confluent.kafka.databalancing.constraint.Constraints;
import io.confluent.kafka.databalancing.constraint.RebalanceConstraints;
import io.confluent.kafka.databalancing.exception.ValidationException;
import io.confluent.kafka.databalancing.metric.Metrics;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.Replica;
import io.confluent.kafka.databalancing.topology.ReplicaAssignment;
import io.confluent.kafka.databalancing.topology.TopologyUtils;
import io.confluent.kafka.databalancing.view.BrokerCountFairView;
import io.confluent.kafka.databalancing.view.BrokerFairView;
import io.confluent.kafka.databalancing.view.BrokerSizeFairView;
import io.confluent.kafka.databalancing.view.ClusterView;
import io.confluent.kafka.databalancing.view.RackCountFairView;
import io.confluent.kafka.databalancing.view.RackSizeFairView;
import io.confluent.kafka.databalancing.view.TopicViewSupplier;
import kafka.admin.BrokerMetadata;
import org.apache.kafka.common.TopicPartition;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import static io.confluent.kafka.databalancing.topology.TopologyUtils.brokerLeaderCounts;
import static io.confluent.kafka.databalancing.topology.TopologyUtils.brokerReplicaCounts;
import static io.confluent.kafka.databalancing.topology.TopologyUtils.rackLeaderCounts;
import static io.confluent.kafka.databalancing.topology.TopologyUtils.rackReplicaCounts;
import static java.util.Arrays.asList;

/**
 * A policy which takes a cluster topology, as a set of partitions->brokerIds, and rebalances it using
 * a supplied notion of fairness.
 *
 * The policy is incremental, meaning it will move the minimum number of replicas required to achieve
 * replica and leader fairness across the cluster.
 *
 * The algorithm is strictly rack aware with respect to replica placement. This means that if racks are
 * not assigned brokers equally the number of replicas on each broker may be skewed.
 *
 * However leaders will be balanced equally amongst brokers, regardless of what rack they are on.
 *
 * There are a few other things worthy of note:
 *
 * (1) The algorithm has five stages: re-replication, moving of replicas from brokers that are to be decommissioned,
 * fair topics, fair-racks, fair-brokers, fair-leaders
 *
 * (2) Each stage (ignoring the first two) is wrapped in an outer loop. This continues until no move can be made
 * for a single batch. Moves are then tested to ensure they improve fairness. So the algorithm will continues until
 * no fairness improvement can be made. This allows the algorithm to terminate should it not be able to complete
 * (certain configurations cannot create fairness, due to the rack and partition constraints).
 *
 * (3) The definition of fairness is different for above/below par values, when the value is not an integer
 * BelowPar is always defined with ceil(). AbovePar is always defined with floor(). So, where fairness is
 * a fractional value, there will be an overlap between abovePar and belowPar. This is required as the
 * algorithm is not symmetric (it starts at with abovePar brokers then searches belowPar brokers). Consider
 * an arrangement where the replica counts on each broker are (1,2,2,3,3). The desired, fair, assignment
 * would be (2,2,2,2,3) but this would not be achieved if fairness: ceil(11/5)=3.
 *
 * TBC
 */
public class MovesOptimisedRebalancePolicy implements RebalancePolicy {

    private static final Logger logger = LoggerFactory.getLogger(MovesOptimisedRebalancePolicy.class);

    // TODO This is a tactical implementation which has a sub-optimal structure due to the late addition of the
    // requirement to optimise for topic fairness. This adds some complexity because each step must now find both a
    // fair distribution of replicas/leaders across brokers/racks (by-broker), as well as a fair distribution of
    // replicas/leaders for each topic across brokers/racks (by-topic). The approach here executes as if it were
    // optimising by-broker, solving the additional by-topic by adding a fairness predicate as a 'constraint'.
    // There should be a better factored solution to this problem that encapsulates the fairness evaluation inside the
    // views rather than in the policy, but this is left to future work.

    // FIXME Replace these counters by code that compares source and target replica assignments. The motivation is
    // that these counters can overstate the number of moves. For example, if the algorithm moves a partition from
    // broker 0 to broker 1 and then to broker 2, we will count it as 2 moves while, in reality, it's a single move
    // from broker 0 to broker 2.
    private int replicasMoved = 0;
    private int leadersMoved = 0;

    @Override
    public ReplicaAssignment rebalancePartitions(List<BrokerMetadata> brokers,
                                                 ReplicaAssignment currentAssignment,
                                                 Map<String, Integer> replicationFactors,
                                                 Metrics metrics,
                                                 List<Broker> brokersToRemove) {
        MutableRebalanceContext context = DefaultRebalanceContext.create(brokers, currentAssignment.asMap(),
                replicationFactors, metrics, brokersToRemove);
        Constraints constraints = new Constraints(context, null);

        List<String> racks = Utils.sorted(context.allRacks());
        List<String> processedRacks = racks;
        if (racks.isEmpty()) {
            processedRacks = asList(new String[]{null});
            logger.info("Executing cluster-wide as no racks configured");
        }

        int step = 1;

        // Stage 1
        logger.info("**** {}. Ensure no under-replicated partitions ****", step++);
        fullyReplicate(context, constraints);

        // Stage 2
        logger.info("**** {}. Move replicas from brokers to be removed ****", step++);
        moveReplicasFromRemovedBrokers(context);

        List<String> topics = Utils.sorted(context.topics());

        // Stage 3.1: Ensure topic replica fairness across racks
        if (!racks.isEmpty()) {
            for (String topic : topics) {
                logger.info("**** {}. Ensure balanced replica count across racks for topic: {} ****", step++, topic);
                replicaTopicFairness(context, new RackCountFairView(context, topic));

                logger.info("**** {}. Ensure balanced replica size across racks for topic: {} ****", step++, topic);
                replicaTopicFairness(context, new RackSizeFairView(context, topic));
            }
        }

        // Stage 3.2: Ensure topic replica fairness on each rack separately
        for (String rack : processedRacks) {
            String rackLogMessageFragment = rackLogMessageFragment(rack);

            for (String topic : Utils.sorted(context.topics())) {
                logger.info("**** {}. Ensure balanced replica count for topic: {} {} ****", step++,
                        topic, rackLogMessageFragment);
                replicaTopicFairness(context, new BrokerCountFairView(context, rack, topic));

                logger.info("**** {}. Ensure balanced replica size for topic: {} {} ****", step++,
                        topic, rackLogMessageFragment);
                replicaTopicFairness(context, new BrokerSizeFairView(context, rack, topic));
            }
        }

        // Stage 3.3: Ensure topic leader fairness cluster-wide
        for (String topic : topics) {
            logger.info("**** {}. Ensure balanced leaders for topic: {} {} ****", step++, topic, maybeClusterWide(racks));
            leaderTopicFairness(context, new BrokerCountFairView(context, null, topic));
        }

        // Stage 4: Ensure replica fairness across racks
        if (!racks.isEmpty()) {
            logger.info("**** {}. Ensure balanced replica count across racks ****", step++);
            replicaFairness(context, new RackCountFairView(context, null), TopicViewSupplier.rackCount(context));

            logger.info("**** {}. Ensure balanced replica size across racks ****", step++);
            replicaSizeFairness(context, new RackSizeFairView(context, null), TopicViewSupplier.rackSize(context));
        }

        // Stage 5: Ensure replica fairness for brokers, on each rack separately
        for (String rack : processedRacks) {
            String rackLogMessageFragment = rackLogMessageFragment(rack);
            logger.info("**** {}. Ensure balanced replica count {} ****", step++, rackLogMessageFragment);
            replicaFairness(context, new BrokerCountFairView(context, rack, null),
                    TopicViewSupplier.brokerCount(context, rack));

            logger.info("**** {}. Ensure balanced replica size {} ****", step++, rackLogMessageFragment);
            replicaSizeFairness(context, new BrokerSizeFairView(context, rack, null),
                    TopicViewSupplier.brokerSize(context, rack));
        }

        // Stage 6: Ensure leader fairness for brokers cluster-wide
        logger.info("**** {}. Ensure balanced leaders for brokers {} ****", step++, maybeClusterWide(racks));
        BrokerFairView brokerView = new BrokerCountFairView(context, null, null);
        leaderFairness(context, brokerView, TopicViewSupplier.brokerCount(context, null));

        log(context);

        return new ReplicaAssignment(partitionToReplicaIds(context));
    }

    private String maybeClusterWide(List<String> racks) {
        //This is reported once at the start in the case that there are no racks.
        return racks.isEmpty() ? "" : "(cluster-wide)";
    }

    private String rackLogMessageFragment(String rack) {
        if (rack == null)
            return "";
        else
            return "across brokers on rack " + rack;
    }

    /**
     * This is similar to `replicaFairness`, but it is simpler in that it only takes into account topic fairness when
     * deciding whether a move is allowed or not.
     */
    private void replicaTopicFairness(final MutableRebalanceContext context, ClusterView topicView) {

        // Proceed in batches, terminating if no move is found (as constraints can cause termination without fairness being achieved)
        // Each batch contains above par replicas. We check each of them is still above par (as fairness may change as we move replicas)
        // as an optimisation
        // We use a fairness predicate to decide if a move should be allowed to allow for different implementations
        // (e.g. replicaFairness). This is a temporary measure until we find a good way to move such logic to
        // the view classes

        boolean moveFound = true;
        while (moveFound) {
            moveFound = false;
            for (final Replica abovePar : topicView.replicasOnAboveParBrokers()) {
                if (topicView.replicasOnAboveParBrokers().contains(abovePar)) {
                    final ClusterView tView = topicView;
                    Predicate<Broker> fairnessPredicate = new Predicate<Broker>() {
                        @Override
                        public boolean test(Broker belowPar) {
                            return tView.compareReplicaFairness(abovePar.topicPartition(), abovePar.broker(),
                                    belowPar).isGreater();
                        }
                    };
                    List<Broker> belowParBrokers = topicView.brokersWithBelowParReplicaFairness();
                    if (moveReplicaToBelowParBroker(context, topicView.constraints(), abovePar, belowParBrokers,
                            fairnessPredicate)) {
                        moveFound = true;
                        topicView = topicView.refresh(context);
                    }
                }
            }
        }
    }

    /**
     * The reason why we have a separate `replicaSizeFairness` to complement `replicaFairness` is that we also try
     * to improve fairness by doing two moves (i.e. switch partitions), which helps in cases where partition sizes
     * are uneven. This is not necessary for the replica count case.
     */
    void replicaSizeFairness(MutableRebalanceContext context, ClusterView clusterView, final TopicViewSupplier<?> topicViewSupplier) {
        ClusterView view = replicaFairness(context, clusterView, topicViewSupplier);

        // After we can't find any single moves that improve fairness, we try to improve by doing two
        // moves (i.e. switch partitions)
        
        // Proceed in batches, terminating if no move is found (as constraints can cause termination without fairness being achieved)
        // Each batch contains above par replicas. We check each of them is still above par (as fairness may change as we move replicas)
        // as an optimisation
        // We use a fairness predicate to decide if a move should be allowed to allow for different implementations
        // This is a temporary measure until we find a good way to move such logic to the view classes

        boolean moveFound = true;
        while (moveFound) {
            moveFound = false;
            for (final Replica abovePar : view.replicasOnAboveParBrokers()) {
                if (view.replicasOnAboveParBrokers().contains(abovePar)) {
                    final ClusterView cView = view;
                    Predicate<Replica> fairnessPredicate = new Predicate<Replica>() {
                        private final ClusterView aboveParTopicView = topicViewSupplier.get(abovePar.topicPartition().topic());
                        @Override
                        public boolean test(Replica belowParReplica) {
                            Broker belowParBroker = belowParReplica.broker();
                            ClusterView belowParTopicView = topicViewSupplier.get(belowParReplica.topicPartition().topic());
                            boolean clusterFairness = cView.compareReplicaFairness(abovePar.broker(), abovePar.topicPartition(),
                                belowParBroker, belowParReplica.topicPartition()).isGreater();
                            boolean topicFairnessOut = aboveParTopicView.compareReplicaFairness(abovePar.broker(),
                                abovePar.topicPartition(), belowParBroker, belowParReplica.topicPartition()).isGreaterOrEqual();
                            boolean topicFairnessBack = belowParTopicView.compareReplicaFairness(abovePar.broker(),
                                abovePar.topicPartition(), belowParBroker, belowParReplica.topicPartition()).isGreaterOrEqual();
                            return clusterFairness && topicFairnessOut && topicFairnessBack;
                        }
                    };

                    List<Replica> belowParReplicas = new ArrayList<>();
                    for (Broker belowParBroker : view.brokersWithBelowParReplicaFairness()) {
                        for (TopicPartition belowParPartition : TopologyUtils.sortPartitionsBySize(context, context.replicas(belowParBroker)))
                            belowParReplicas.add(new Replica(belowParPartition, belowParBroker));
                    }

                    if (switchReplicaWithBelowParBrokerReplica(context, view.constraints(), abovePar, belowParReplicas,
                            fairnessPredicate)) {
                        moveFound = true;
                        view = view.refresh(context);
                        topicViewSupplier.clear();
                    }
                }
            }
        }
    }

    private boolean switchReplicaWithBelowParBrokerReplica(MutableRebalanceContext context,
                                                           RebalanceConstraints constraints, Replica abovePar,
                                                           Collection<Replica> belowParReplicas,
                                                           Predicate<Replica> fairnessPredicate) {
        logger.debug("Moving {}, # belowPar replica candidates: {} ", abovePar, belowParReplicas.size());

        for (Replica belowParReplica : belowParReplicas) {
            Broker belowParBroker = belowParReplica.broker();
            TopicPartition belowParPartition = belowParReplica.topicPartition();
            boolean obeysPartitionOut = constraints.obeysPartitionConstraint(abovePar.topicPartition(), belowParBroker);
            boolean obeysPartitionBack = constraints.obeysPartitionConstraint(belowParPartition, abovePar.broker());
            boolean obeysRackOut = constraints.obeysRackConstraint(abovePar.topicPartition(), abovePar.broker(),
                    belowParBroker);
            boolean obeysRackBack = constraints.obeysRackConstraint(belowParPartition, belowParBroker,
                    abovePar.broker());
            boolean fairness = fairnessPredicate.test(belowParReplica);
            if (obeysPartitionOut && obeysPartitionBack && fairness && obeysRackOut && obeysRackBack) {
                movePartition(context, abovePar.topicPartition(), abovePar.broker(), belowParBroker);
                movePartition(context, belowParReplica.topicPartition(), belowParBroker, abovePar.broker());
                return true;
            }
        }

        logger.debug("Replica {} could not be moved despite attempting {} different brokers", abovePar,
                belowParReplicas.size());
        return false;
    }


    void moveReplicasFromRemovedBrokers(MutableRebalanceContext context) {
        Constraints constraints = new Constraints(context, null);
        for (Broker broker : context.brokersToBeRemoved()) {
            for (TopicPartition tp : context.replicas(broker)) {
                Replica replica = new Replica(tp, broker);
                if (!moveReplicaFromRemovedBroker(context, constraints, replica))
                    throw new ValidationException("ERROR: Could not move replica from removed broker due to " +
                            "either rack or partition " +  "constraints. Thus broker " + broker.id() +
                            " will still have partition " + tp + ".");
            }
        }
    }

    private boolean moveReplicaFromRemovedBroker(MutableRebalanceContext context, Constraints constraints, Replica replica) {
        String rack = context.brokerRack(replica.broker());
        List<Broker> candidateDestinationBrokers = TopologyUtils.leastLoadedBrokersPreferringTheseRacks(context,
                rack == null ? Collections.<String>emptySet() : Collections.singleton(rack));
        logger.debug("Moving {} from removed broker, # candidates: {} ", replica, candidateDestinationBrokers.size());

        for (Broker destination : candidateDestinationBrokers) {
            boolean obeysPartition = constraints.obeysPartitionConstraint(replica.topicPartition(), destination);
            boolean obeysRack = constraints.obeysRackConstraint(replica.topicPartition(), replica.broker(), destination);

            if (obeysRack && obeysPartition) {
                movePartition(context, replica.topicPartition(), replica.broker(), destination);
                return true;
            }
            logger.debug("Move to {} failed due to rack/partition constraints: {}, {}", destination, obeysRack,
                    obeysPartition);
        }
        logger.debug("Replica {} could not be moved despite attempting {} different brokers", replica,
                candidateDestinationBrokers.size());
        return false;
    }

    /**
     * Create new replicas for any under-replicated partitions on a least loaded broker. If a partition cannot
     * be fully replicated, due to there not being a valid broker available, the algorithm will progress regardless
     * but it will output a warning.
     */
    void fullyReplicate(MutableRebalanceContext context, Constraints constraints) {
        for (TopicPartition tp : context.allPartitions()) {
            int replicationFactor = context.replicationFactor(tp.topic());
            Set<String> racks = TopologyUtils.racksFor(context, tp);

            while (replicationFactor > context.brokers(tp).size()) {
                if (!addReplica(context, constraints, racks, tp)) {
                    logger.warn("WARNING: Could not create replica due to either rack or partition constraints. Thus " +
                            "this partition will remain under-replicated {}", tp);
                    break;
                }
            }
        }
    }

    private boolean addReplica(MutableRebalanceContext context, Constraints constraints, Set<String> racks, TopicPartition tp) {
        List<Broker> leastLoadedBrokers = TopologyUtils.leastLoadedBrokersPreferringOtherRacks(context, racks);
        for (Broker destinationBroker : leastLoadedBrokers) {
            if (constraints.obeysPartitionConstraint(tp, destinationBroker) &&
                    constraints.obeysRackConstraint(tp, null, destinationBroker)) {
                context.addReplica(tp, destinationBroker);
                return true;
            }
        }
        return false;
    }

    /**
     * Move replicas on above-par brokers/racks to below-par brokers/racks if they obey partition/rack constraints and
     * improve broker/rack fairness while not making topic fairness worse.
     *
     * @see #replicaSizeFairness(MutableRebalanceContext, ClusterView, TopicViewSupplier)
     * @see #replicaTopicFairness(MutableRebalanceContext, ClusterView)
     */
    private ClusterView replicaFairness(MutableRebalanceContext context, ClusterView clusterView,
                                        final TopicViewSupplier<?> topicViewSupplier) {

        // Proceed in batches, terminating if no move is found (as constraints can cause termination without fairness being achieved)
        // Each batch contains above par replicas. We check each of them is still above par (as fairness may change as we move replicas)
        // as an optimisation
        // We use a fairness predicate to decide if a move should be allowed to allow for different implementations
        // (e.g. replicaTopicFairness). This is a temporary measure until we find a good way to move such logic to
        // the view classes

        boolean moveFound = true;
        while (moveFound) {
            moveFound = false;
            for (final Replica abovePar : clusterView.replicasOnAboveParBrokers()) {
                if (clusterView.replicasOnAboveParBrokers().contains(abovePar)) {
                    final ClusterView view = clusterView;
                    Predicate<Broker> fairnessPredicate = new Predicate<Broker>() {
                        private final ClusterView topicView = topicViewSupplier.get(abovePar.topicPartition().topic());
                        @Override
                        public boolean test(Broker belowPar) {
                            boolean clusterFairness = view.compareReplicaFairness(abovePar.topicPartition(), abovePar.broker(),
                                    belowPar).isGreater();
                            boolean topicFairness = topicView.compareReplicaFairness(abovePar.topicPartition(),
                                    abovePar.broker(), belowPar).isGreaterOrEqual();
                            return clusterFairness && topicFairness;
                        }
                    };
                    List<Broker> belowParBrokers = clusterView.brokersWithBelowParReplicaFairness();
                    if (moveReplicaToBelowParBroker(context, clusterView.constraints(), abovePar, belowParBrokers,
                            fairnessPredicate)) {
                        moveFound = true;
                        clusterView = clusterView.refresh(context);
                        topicViewSupplier.clear();
                    }
                }
            }
        }

        return clusterView;
    }

    private boolean moveReplicaToBelowParBroker(MutableRebalanceContext context, RebalanceConstraints constraints,
                                                Replica abovePar, Collection<Broker> belowParBrokers,
                                                Predicate<Broker> fairnessPredicate) {
        logger.debug("Moving {}, # belowPar candidates: {} ", abovePar, belowParBrokers.size());

        for (Broker belowPar : belowParBrokers) {
            boolean obeysPartition = constraints.obeysPartitionConstraint(abovePar.topicPartition(), belowPar);
            boolean obeysRack = constraints.obeysRackConstraint(abovePar.topicPartition(), abovePar.broker(), belowPar);
            boolean fairness = fairnessPredicate.test(belowPar);

            if (obeysRack && obeysPartition && fairness) {
                movePartition(context, abovePar.topicPartition(), abovePar.broker(), belowPar);
                return true;
            }
            logger.debug("Move to {} failed due to rack/partition/fairness constraints: {}, {}, {}", belowPar, obeysRack,
                    obeysPartition, fairness);
        }

        logger.debug("Replica {} could not be moved despite attempting {} different brokers", abovePar,
                belowParBrokers.size());
        return false;
    }

    private void movePartition(MutableRebalanceContext context, TopicPartition partition, Broker from, Broker to) {
        if (to.equals(from))
            logger.debug("Movement was not made as both source and destination broker are the same: {}", from.id());
        else {
            context.movePartition(partition, from, to);
            replicasMoved += 1;

            if (logger.isDebugEnabled())
                logger.debug("Moved partition {} ({} MB) from broker {} ({} MB) to broker {} ({} MB)",
                        partition, TopologyUtils.formattedPartitionSize(context, partition),
                        from.id(), TopologyUtils.formattedBrokerSize(context, from),
                        to.id(), TopologyUtils.formattedBrokerSize(context, to));
        }
    }

    /**
     * Similar to `leaderFairness`, but simpler as it only considers topic fairness when deciding whether a move should
     * be allowed.
     */
    private void leaderTopicFairness(final MutableRebalanceContext context, ClusterView topicView) {
        // Proceed in batches, terminating if no move is found (as constraints can cause termination without fairness being achieved)
        // Each batch contains above par leaders. We check each of them is still above par (as fairness may change as we move replicas)
        // as an optimisation
        // We use a fairness predicate to decide if a move should be allowed to allow for different implementations
        // (e.g. leaderFairness). This is a temporary measure until we find a good way to move such logic to
        // the view classes

        boolean moveFound = true;
        while (moveFound) {
            moveFound = false;
            for (final TopicPartition aboveParLeader : topicView.leadersOnAboveParBrokers()) {
                if (topicView.leadersOnAboveParBrokers().contains(aboveParLeader)) {
                    final ClusterView tView = topicView;
                    BiPredicate<Broker> fairnessPredicate = new BiPredicate<Broker>() {
                        @Override
                        public boolean test(Broker leader, Broker follower) {
                            return tView.compareLeaderFairness(aboveParLeader, leader, follower).isGreater();
                        }
                    };
                    if (changeLeadership(context, aboveParLeader, fairnessPredicate)) {
                        moveFound = true;
                        topicView = topicView.refresh(context);
                    }
                }
            }
        }

        moveFound = true;
        while (moveFound) {
            moveFound = false;
            for (final TopicPartition aboveParLeader : topicView.leadersOnAboveParBrokers()) {
                if (topicView.leadersOnAboveParBrokers().contains(aboveParLeader)) {

                    final Collection<Broker> belowParBrokers = topicView.brokersWithBelowParLeaderFairness();
                    final ClusterView tView = topicView;

                    Predicate<Replica> fairnessPredicate = new Predicate<Replica>() {
                        private final Broker aboveParLeaderBroker = context.brokers(aboveParLeader).get(0);
                        @Override
                        public boolean test(Replica belowParReplica) {
                            Broker belowParBroker = belowParReplica.broker();
                            TopicPartition belowParFollower = belowParReplica.topicPartition();
                            boolean obeysPartitionOut = tView.constraints().obeysPartitionConstraint(aboveParLeader, belowParBroker);
                            boolean obeysPartitionBack = tView.constraints().obeysPartitionConstraint(belowParFollower, aboveParLeaderBroker);
                            boolean fairness = tView.compareLeaderFairness(aboveParLeader, aboveParLeaderBroker, belowParBroker).isGreater();
                            return obeysPartitionOut && obeysPartitionBack && fairness;
                        }
                    };

                    if (switchLeaderToOtherBroker(context, aboveParLeader, belowParBrokers, fairnessPredicate)) {
                        moveFound = true;
                        topicView = topicView.refresh(context);
                    }
                }
            }
        }
    }

    /**
     * Move leadership to from above-par to below-par brokers. If a valid follower replica exists for this partition,
     * leadership is simply switched (without a resulting data movement). If this cannot be achieved a replica from a
     * different partition, on a above-par broker, will be selected, and the two replicas will be swapped (i.e. two
     * way data movement) allowing leadership to be moved to a below-par broker.
     *
     * @see #leaderTopicFairness(MutableRebalanceContext, ClusterView)
     */
    private void leaderFairness(final MutableRebalanceContext context, ClusterView clusterView, TopicViewSupplier<?> topicViewSupplier) {
        // Proceed in batches, terminating if no move is found (as constraints can cause termination without fairness being achieved)
        // Each batch contains above par leaders. We check each of them is still above par (as fairness may change as we move replicas)
        // as an optimisation
        // We use a fairness predicate to decide if a move should be allowed to allow for different implementations
        // (e.g. leaderTopicFairness). This is a temporary measure until we find a good way to move such logic to
        // the view classes

        boolean moveFound = true;
        while (moveFound) {
            moveFound = false;
            for (final TopicPartition aboveParLeader : clusterView.leadersOnAboveParBrokers()) {
                if (clusterView.leadersOnAboveParBrokers().contains(aboveParLeader)) {
                    final ClusterView cView = clusterView;
                    final ClusterView tView = topicViewSupplier.get(aboveParLeader.topic());
                    BiPredicate<Broker> fairnessPredicate = new BiPredicate<Broker>() {
                        @Override
                        public boolean test(Broker leader, Broker follower) {
                            return cView.compareLeaderFairness(aboveParLeader, leader, follower).isGreater() &&
                                    tView.compareLeaderFairness(aboveParLeader, leader, follower).isGreaterOrEqual();
                        }
                    };
                    if (changeLeadership(context, aboveParLeader, fairnessPredicate)) {
                        moveFound = true;
                        clusterView = clusterView.refresh(context);
                        topicViewSupplier.clear();
                    }
                }
            }
        }

        moveFound = true;
        while (moveFound) {
            moveFound = false;
            for (final TopicPartition aboveParLeader : clusterView.leadersOnAboveParBrokers()) {
                if (clusterView.leadersOnAboveParBrokers().contains(aboveParLeader)) {

                    final Collection<Broker> belowParBrokers = clusterView.brokersWithBelowParLeaderFairness();
                    final ClusterView topicView = topicViewSupplier.get(aboveParLeader.topic());
                    final ClusterView cView = clusterView;

                    Predicate<Replica> fairnessPredicate = new Predicate<Replica>() {
                        private final Broker aboveParLeaderBroker = context.brokers(aboveParLeader).get(0);
                        @Override
                        public boolean test(Replica belowParReplica) {
                            Broker belowParBroker = belowParReplica.broker();
                            TopicPartition belowParFollower = belowParReplica.topicPartition();
                            boolean obeysPartitionOut = cView.constraints().obeysPartitionConstraint(aboveParLeader, belowParBroker);
                            boolean obeysPartitionBack = cView.constraints().obeysPartitionConstraint(belowParFollower, aboveParLeaderBroker);
                            boolean clusterFairness = cView.compareLeaderFairness(aboveParLeader, aboveParLeaderBroker, belowParBroker).isGreater();
                            boolean topicFairness = topicView.compareLeaderFairness(aboveParLeader, aboveParLeaderBroker,
                                    belowParBroker).isGreaterOrEqual();
                            return obeysPartitionOut && obeysPartitionBack && clusterFairness && topicFairness;
                        }
                    };

                    if (switchLeaderToOtherBroker(context, aboveParLeader, belowParBrokers, fairnessPredicate)) {
                        moveFound = true;
                        clusterView = clusterView.refresh(context);
                        topicViewSupplier.clear();
                    }
                }
            }
        }
    }

    private boolean changeLeadership(MutableRebalanceContext context, TopicPartition aboveParLeader,
                                     BiPredicate<Broker> fairnessPredicate) {
        // Attempt to switch leadership within partitions to achieve fairness (i.e. no data movement)
        List<Broker> replicas = context.brokers(aboveParLeader);
        Broker leader = replicas.get(0);
        for (Broker follower : replicas.subList(1, replicas.size())) {
            //it's a follower replica on a below par broker
            if (fairnessPredicate.test(leader, follower)) {
                makeLeader(context, aboveParLeader, follower);
                return true;
            }
        }
        return false;
    }

    private boolean switchLeaderToOtherBroker(final MutableRebalanceContext context, TopicPartition aboveParLeader,
                                              Collection<Broker> belowParBrokers, Predicate<Replica> fairnessPredicate) {
        // No luck changing leadership within the partition so pick a completely different partition, which is on a below
        // par broker, and physically swap replicas so this leader ends up on a different broker.
        Broker aboveParLeaderBroker = context.brokers(aboveParLeader).get(0);
        List<Replica> belowParFollowers = new ArrayList<>();
        for (Broker belowParBroker : belowParBrokers) {
            for (TopicPartition belowParFollower : context.followers(belowParBroker))
                belowParFollowers.add(new Replica(belowParFollower, belowParBroker));
        }

        final long leaderSize = context.partitionSize(aboveParLeader);

        // sort followers by follower partition size deviation from leader partition size to try and maintain replica size
        // balance
        Collections.sort(belowParFollowers, new Comparator<Replica>() {
            @Override
            public int compare(Replica r1, Replica r2) {
                long s1 = context.partitionSize(r1.topicPartition());
                long s2 = context.partitionSize(r2.topicPartition());
                int result = Long.compare(Math.abs(leaderSize - s1), Math.abs(leaderSize - s2));
                if (result != 0)
                    return result;
                if (s1 != s2)
                    return Long.compare(s1, s2);
                if (r1.broker().id() != r2.broker().id())
                    return Integer.compare(r1.broker().id(), r2.broker().id());
                return TopologyUtils.topicPartitionComparator.compare(r1.topicPartition(), r2.topicPartition());
            }
        });

        for (Replica replica : belowParFollowers) {
            if (fairnessPredicate.test(replica)) {
                Broker belowParBroker = replica.broker();
                movePartition(context, aboveParLeader, aboveParLeaderBroker, belowParBroker);
                movePartition(context, replica.topicPartition(), belowParBroker, aboveParLeaderBroker);
                return true;
            }
        }

        return false;
    }

    void makeLeader(MutableRebalanceContext context, TopicPartition topicPartition, Broker newLeader) {
        Broker currentLeader = context.leader(topicPartition);

        if (!newLeader.equals(currentLeader)) {
            context.makeLeader(topicPartition, newLeader);
            logger.debug("Leadership moved brokers: [{} -> {}] for partition {}:{}", currentLeader, newLeader,
                    topicPartition, context.brokers(topicPartition));
            leadersMoved += 1;
        }
        else
            System.out.println("Leadership change was not made as " + newLeader + " was already the leader for partition " +
                    topicPartition + " - see: " + context.brokers(topicPartition));
    }

    private Map<TopicPartition,List<Integer>> partitionToReplicaIds(RebalanceContext context) {
        Map<TopicPartition, List<Integer>> partitionToReplicaIds = new HashMap<>(context.partitionToBrokers().size());
        for (Map.Entry<TopicPartition, List<Broker>> entry : context.partitionToBrokers().entrySet()) {
            List<Integer> replicaIds = new ArrayList<>(entry.getValue().size());
            for (Broker broker : entry.getValue())
                replicaIds.add(broker.id());
            partitionToReplicaIds.put(entry.getKey(), replicaIds);
        }
        return partitionToReplicaIds;
    }

    public int replicasMoved() {
        return replicasMoved;
    }

    public int leadersMoved() {
        return leadersMoved;
    }

    private void log(RebalanceContext context) {
        logger.debug("Racks to replica Counts {}", rackReplicaCounts(context, null, null));
        logger.debug("Racks to leader Counts {}", rackLeaderCounts(context, null, null));
        logger.debug("Broker to replica Counts {}", brokerReplicaCounts(context, context.allBrokers(), null));
        logger.debug("Broker to leader Counts {}",  brokerLeaderCounts(context, context.allBrokers(), null));
        logger.debug("Number of replicas moves {}", replicasMoved);
        logger.debug("Number of leader moves {}", leadersMoved);
    }

    interface Predicate<T> {
        boolean test(T t);
    }

    interface BiPredicate<T> {
        boolean test(T t1, T t2);
    }
}
