// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.constraint;

import io.confluent.kafka.databalancing.RebalanceContext;
import io.confluent.kafka.databalancing.topology.Broker;
import org.apache.kafka.common.TopicPartition;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

public class Constraints implements RebalanceConstraints {

    private static final Logger logger = LoggerFactory.getLogger(Constraints.class);

    private final RebalanceContext context;
    private final String rack;
    private final int rackCount;

    public Constraints(RebalanceContext context, String rack) {
        this.context = context;
        this.rack = rack;
        this.rackCount = rack == null ? context.allRacks().size() : 1;
    }

    private <T> void replaceAddOrRemove(List<T> list, T toBeReplaced, T toReplace) {
        int index = toBeReplaced == null ? -1 : list.indexOf(toBeReplaced);
        if (index != -1)
            if (toReplace == null)
                list.remove(index);
            else
                list.set(index, toReplace);
        else if (toReplace != null)
            list.add(toReplace);
    }

    @Override
    public boolean obeysRackConstraint(TopicPartition partition, Broker brokerFrom, Broker brokerTo) {
        if (brokerFrom == null && brokerTo == null)
            throw new IllegalArgumentException("At least one of `brokerFrom` and `brokerTo` should not be null");

        int minRacksSpanned = Math.min(context.replicationFactor(partition.topic()), rackCount);

        //get replicas for partition, replacing brokerFrom with brokerTo
        List<Broker> proposedReplicas = new ArrayList<>(context.brokers(partition, rack));
        replaceAddOrRemove(proposedReplicas, brokerFrom, brokerTo);
        int racksSpanned = context.racks(proposedReplicas).size();

        boolean result = racksSpanned >= minRacksSpanned;

        if (!result)
            logger.debug("failing rack constraint {}:{}, {} -> {} ", partition, proposedReplicas, brokerFrom, brokerTo);

        return result;
    }

    @Override
    public boolean obeysPartitionConstraint(TopicPartition partition, Broker brokerMovingTo) {
        Objects.requireNonNull(brokerMovingTo, "brokerMovingTo should not be null");

        List<Broker> replicas = context.brokers(partition, rack);
        boolean result = !replicas.contains(brokerMovingTo);

        if (!result)
            logger.debug("Failing partition constraint for {}:{} -> {}", partition, replicas, brokerMovingTo);

        return result;
    }

}
