// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.fairness;

import io.confluent.kafka.databalancing.CompareResult;
import io.confluent.kafka.databalancing.RebalanceContext;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.TopologyUtils;
import org.apache.kafka.common.TopicPartition;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;

import static io.confluent.kafka.databalancing.Utils.sorted;

/**
 * Provides a reasonable default implementation for `Fairness` where subclasses have to provide:
 *
 * 1. Broker to value map
 * 2. Rack to value map
 * 3. A method that returns the value of a topic partition
 *
 * Impact of moves on fairness is based on the impact on the variance of the values from the mean.
 */
public abstract class BaseFairness implements Fairness {

    protected final RebalanceContext context;
    protected final String rack;
    protected final String topic;

    private final Map<String, Long> rackValueMap;
    private final Map<Broker, Long> brokerValueMap;

    private final double brokerFairValue;
    private final double rackFairValue;

    // Created lazily
    private List<String> aboveParRacks;
    private List<String> belowParRacks;
    private List<Broker> aboveParBrokers;
    private List<Broker> belowParBrokers;

    protected BaseFairness(RebalanceContext context, String rack, String topic) {
        this.context = context;
        this.topic = topic;
        this.rack = rack;
        this.brokerValueMap = createBrokerValueMap(context, rack, topic);
        this.rackValueMap = createRackValueMap(context, rack, topic);

        long replicaTotal = 0;
        for (long value : brokerValueMap.values())
            replicaTotal += value;

        this.brokerFairValue = replicaTotal / (double) brokerValueMap.size();
        this.rackFairValue = replicaTotal / (double) (rack == null ? context.allRacks().size() : 1);
    }

    protected abstract Map<Broker, Long> createBrokerValueMap(RebalanceContext context, String rack, String topic);

    protected abstract Map<String, Long> createRackValueMap(RebalanceContext context, String rack, String topic);

    protected long brokerValue(Broker broker) {
        return brokerValueMap.get(broker);
    }

    protected long rackValue(Broker broker) {
        return rackValueMap.get(context.brokerRack(broker));
    }

    public double rackFairValue() {
        return rackFairValue;
    }

    public double brokerFairValue() {
        return brokerFairValue;
    }

    @Override
    public List<String> aboveParRacks() {
        if (aboveParRacks == null)
            createAboveAndBelowParRacks();
        return aboveParRacks;
    }

    @Override
    public List<String> belowParRacks() {
        if (belowParRacks == null)
            createAboveAndBelowParRacks();
        return belowParRacks;
    }

    private void createAboveAndBelowParRacks() {
        List<String> aboveParRacks = new ArrayList<>();
        List<String> belowParRacks = new ArrayList<>();

        Comparator<String> comparator;
        if (topic == null)
            comparator = TopologyUtils.rackComparator(rackValueMap);
        else
            comparator = TopologyUtils.rackTopicComparator(rackValueMap, createRackValueMap(context, rack, null));

        List<String> sortedRacks = sorted(rackValueMap.keySet(), comparator);
        for (String r : sortedRacks) {
            long value = rackValueMap.get(r);
            if (value > Math.floor(rackFairValue))
                aboveParRacks.add(r);
            if (value < Math.ceil(rackFairValue))
                belowParRacks.add(r);
        }
        Collections.reverse(aboveParRacks);
        this.aboveParRacks = Collections.unmodifiableList(aboveParRacks);
        this.belowParRacks = Collections.unmodifiableList(belowParRacks);
    }

    @Override
    public List<Broker> aboveParBrokers() {
        if (aboveParBrokers == null)
            createAboveAndBelowParBrokers();
        return aboveParBrokers;
    }

    @Override
    public List<Broker> belowParBrokers() {
        if (belowParBrokers == null)
            createAboveAndBelowParBrokers();
        return belowParBrokers;
    }

    private void createAboveAndBelowParBrokers() {
        List<Broker> aboveParBrokers = new ArrayList<>();
        List<Broker> belowParBrokers = new ArrayList<>();

        Comparator<Broker> comparator;
        if (topic == null)
            comparator = TopologyUtils.brokerComparator(brokerValueMap);
        else
            comparator = TopologyUtils.brokerTopicComparator(brokerValueMap, createBrokerValueMap(context, rack, null));

        List<Broker> sortedBrokers = sorted(brokerValueMap.keySet(), comparator);
        for (Broker broker : sortedBrokers) {
            long value = brokerValueMap.get(broker);
            if (value > Math.floor(brokerFairValue))
                aboveParBrokers.add(broker);
            if (value < Math.ceil(brokerFairValue))
                belowParBrokers.add(broker);
        }
        Collections.reverse(aboveParBrokers);
        this.aboveParBrokers = Collections.unmodifiableList(aboveParBrokers);
        this.belowParBrokers = Collections.unmodifiableList(belowParBrokers);
    }

    @Override
    public CompareResult compareBrokerFairness(TopicPartition tp, Broker b1, Broker b2) {
        return compareBrokerFairness(b1, tp, b2, null);
    }

    protected abstract long partitionValue(TopicPartition tp);

    @Override
    public CompareResult compareBrokerFairness(Broker b1, TopicPartition tp1, Broker b2, TopicPartition tp2) {
        long tp1Size = partitionValue(tp1);
        long tp2Size = tp2 == null ? 0 : partitionValue(tp2);

        // optimise the case where the partitions being moved don't affect this fairness class (i.e. neither is for
        // `topic`)
        if (tp1Size == 0 && tp2Size == 0)
            return CompareResult.EQUAL;

        long b1Size = brokerValue(b1);
        long b2Size = brokerValue(b2);
        double mean = brokerFairValue();
        return TopologyUtils.compareSizeFairness(b1Size, tp1Size, b2Size, tp2Size, mean);
    }

    @Override
    public CompareResult compareRackFairness(TopicPartition tp, Broker b1, Broker b2) {
        return compareRackFairness(b1, tp, b2, null);
    }

    @Override
    public CompareResult compareRackFairness(Broker b1, TopicPartition tp1, Broker b2, TopicPartition tp2) {
        long tp1Size = partitionValue(tp1);
        long tp2Size = tp2 == null ? 0 : partitionValue(tp2);

        // optimise the case where the partitions being moved don't affect this fairness class (i.e. neither is for
        // `topic`)
        if (tp1Size == 0 && tp2Size == 0)
            return CompareResult.EQUAL;

        long r1Size = rackValue(b1);
        long r2Size = rackValue(b2);
        double mean = rackFairValue();
        return TopologyUtils.compareSizeFairness(r1Size, tp1Size, r2Size, tp2Size, mean);
    }

}
