// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.license;

import com.google.common.annotations.VisibleForTesting;
import io.confluent.license.License;
import org.apache.kafka.common.utils.SystemTime;
import org.apache.kafka.common.utils.Time;
import org.jose4j.jwt.JwtClaims;
import org.jose4j.jwt.consumer.InvalidJwtException;

import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.spec.InvalidKeySpecException;
import java.text.SimpleDateFormat;
import java.util.Date;

/**
 * Validate the license string supplied by the user
 */
public class DefaultLicenseValidator implements LicenseValidator {
    private final Time time;

    public DefaultLicenseValidator(Time time) {
        this.time = time;
    }

    @Override
    public void validateLicense(String licenseStr) throws ValidationFailedException {
        PublicKey publicKey;
        try {
            publicKey = License.loadPublicKey();
        } catch (NoSuchAlgorithmException | IOException | InvalidKeySpecException e) {
            // This can happen for two reasons:
            //  1. The public key was not packaged with the jar.
            //  2. The public key which was packaged was corrupt.
            throw new IllegalStateException("Internal license validation error", e);
        }
        validateLicense(licenseStr, publicKey);
   }

    @Override
    public long validateTrialPeriod(String zkConnectStr) {
        ZkTrialPeriod trialPeriod = new ZkTrialPeriod(zkConnectStr, time);
        return trialPeriod.startOrVerify();
    }

    @VisibleForTesting
    void validateLicense(String licenseStr, PublicKey publicKey) throws ValidationFailedException {
        JwtClaims claims;
        try {
            claims = License.verify(publicKey, licenseStr);
        } catch (InvalidJwtException e) {
            throw new ValidationFailedException("Invalid License.", e);
        }

        long expirationTimeMs = License.getExpiration(claims);
        if (time.milliseconds() > expirationTimeMs) {
            SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd");
            throw new ValidationFailedException("License expired on : "
                    + dateFormat.format(new Date(expirationTimeMs)));
        }
    }
}


