// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.metric;

import io.confluent.kafka.databalancing.RebalancerConfig;
import io.confluent.metrics.record.ConfluentMetric;
import io.confluent.serializers.ProtoSerde;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.clients.consumer.KafkaConsumer;
import org.apache.kafka.clients.consumer.OffsetAndTimestamp;
import org.apache.kafka.common.PartitionInfo;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.errors.TimeoutException;
import org.apache.kafka.common.serialization.ByteArrayDeserializer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

/**
 * Extracts metrics of interest from the metrics topic. Will instantiate specific @MetricReader classes to extract
 * values from each @ConfluenMetric.MetricsMessage.
 */
public class MetricsCollector {
    public static final Logger LOG = LoggerFactory.getLogger(MetricsCollector.class);
    public static final long CONSUMER_REPLAY_PERIOD_MS = 30000L;
    public static final long COLLECTION_TIMEOUT_MS = 60000L;
    public static final long METRIC_TOPIC_METADATA_POLL_RETRY_MS = 200L;

    /**
     * Collect metrics.
     * @param metricsTopic The name of the topic where metrics are published to.
     * @param consumerProps The properties of the consumer of the metrics topic. This includes the metrics bootstrap servers.
     * @param partitions The TopicPartitions whose metrics need to be collected.
     * @param clusterId The id of the cluster for whom metrics are to be collected. This cannot be null.
     * @return The @Metrics for the given cluster and topic partitions.
     * @throws IllegalArgumentException if the cluster id is null.
     */
    public static Metrics collectMetrics(String metricsTopic, Properties consumerProps,
                                         Collection<TopicPartition> partitions, String clusterId)
            throws TimeoutException {
        if (clusterId == null) {
            throw new IllegalArgumentException("Did not receive a cluster id to filter on. Please ensure that you are" +
            " using Kafka >= 0.10.1, and that you have some brokers running in the cluster.");
        }
        Map<TopicPartition, Long> logSizes = new HashMap<>();
        Set<TopicPartition> partitionsToCollect = new HashSet<>(partitions);
        if (!consumerProps.containsKey(ConsumerConfig.GROUP_ID_CONFIG))
            consumerProps.put(ConsumerConfig.GROUP_ID_CONFIG, "confluent-rebalancer-metrics-consumer");

        if (!consumerProps.containsKey(ConsumerConfig.METADATA_MAX_AGE_CONFIG)) {
            consumerProps.put(ConsumerConfig.METADATA_MAX_AGE_CONFIG, Long.toString(COLLECTION_TIMEOUT_MS / 4));
        }

        if (!consumerProps.containsKey(ConsumerConfig.REQUEST_TIMEOUT_MS_CONFIG)) {
            consumerProps.put(ConsumerConfig.REQUEST_TIMEOUT_MS_CONFIG, "20000");
        }

        KafkaConsumer<byte[], byte[]> consumer = new KafkaConsumer<>(consumerProps, new ByteArrayDeserializer(),
                new ByteArrayDeserializer());

        ProtoSerde<ConfluentMetric.MetricsMessage> serdes = new ProtoSerde<>(ConfluentMetric.MetricsMessage.getDefaultInstance());
        LogSizeMetricReader logSizeMetricReader = new LogSizeMetricReader();

        if (metricsTopic == null)
            metricsTopic = RebalancerConfig.DEFAULT_METRICS_TOPIC;

        long startTime = System.currentTimeMillis();
        Collection<TopicPartition> topicPartitions = partitionsForTopic(consumer, metricsTopic);
        consumer.assign(topicPartitions);
        // Seek to offsets CONSUMER_REPLAY_PERIOD_MS ago.
        long time = System.currentTimeMillis() - CONSUMER_REPLAY_PERIOD_MS;
        seekToOffsetsByTime(consumer, topicPartitions, time);

        try {
            LOG.info("Going to wait for at most " + COLLECTION_TIMEOUT_MS + "ms to collect metrics for "
                    + partitionsToCollect.size() + " partitions.");
            while (!partitionsToCollect.isEmpty()) {
                if (COLLECTION_TIMEOUT_MS < (System.currentTimeMillis() - startTime)) {
                    throw new TimeoutException("Could not collect all required metrics after "
                            + COLLECTION_TIMEOUT_MS + "ms.");
                }
                ConsumerRecords<byte[], byte[]> records = consumer.poll(100);
                for (ConsumerRecord<byte[], byte[]> record : records) {
                    ConfluentMetric.MetricsMessage message = serdes.deserialize(record.value());
                    if (!message.getClusterId().equals(clusterId))
                        continue;
                    Map<TopicPartition, Long> currentLogSizes = logSizeMetricReader.metricValues(message);
                    for (TopicPartition topicPartition : currentLogSizes.keySet()) {
                        if (partitionsToCollect.remove(topicPartition)) {
                            logSizes.put(topicPartition, currentLogSizes.get(topicPartition));
                        } else if (logSizes.containsKey(topicPartition)) {
                            // Store the latest value we read.
                            logSizes.put(topicPartition, currentLogSizes.get(topicPartition));
                        }
                    }
                }
            }
        }
        finally {
            consumer.close();
        }

        Metrics metrics = new Metrics(logSizes);
        LOG.debug("Collected: {}", metrics.toString());
        return metrics;
    }

    /**
     * Polls the given topic for partitions upto for at most COLLECTION_TIMEOUT_MS. If no partitions are received in time
     * it will throw a TimeoutException. Otherwise will return a the list of topic partitions received.
     * @return A list of TopicPartitions for the given topic.
     */
    private static Collection<TopicPartition> partitionsForTopic(KafkaConsumer<byte[], byte[]> consumer,
                                                                 String topic) throws TimeoutException {
        long startTime = System.currentTimeMillis();
        List<PartitionInfo> partitions = consumer.partitionsFor(topic);
        while (partitions == null || partitions.isEmpty()) {
            try {
                Thread.sleep(METRIC_TOPIC_METADATA_POLL_RETRY_MS);
            } catch (InterruptedException e) {
                // Swallow it.
            }
            if (COLLECTION_TIMEOUT_MS < (System.currentTimeMillis() - startTime)) {
                throw new TimeoutException("No partition information received for " + topic + " in "
                        + COLLECTION_TIMEOUT_MS + "ms. Aborting.");
            }
            partitions = consumer.partitionsFor(topic);
        }
        List<TopicPartition> topicPartitions = new ArrayList<>();
        for (PartitionInfo partitionInfo : partitions) {
            topicPartitions.add(new TopicPartition(topic, partitionInfo.partition()));
        }
       return topicPartitions;
    }

    private static void seekToOffsetsByTime(KafkaConsumer<byte[], byte[]> consumer,
                                     Collection<TopicPartition> topicPartitions, long time) {
        Map<TopicPartition, Long> partitionToTimestamp = new HashMap<>();
        for (TopicPartition topicPartition : topicPartitions)
            partitionToTimestamp.put(topicPartition, time);
        Map<TopicPartition, OffsetAndTimestamp> offsets = consumer.offsetsForTimes(partitionToTimestamp);
        for (TopicPartition topicPartition : offsets.keySet()) {
            OffsetAndTimestamp offsetAndTimestamp = offsets.get(topicPartition);
            if (offsetAndTimestamp == null) {
                // there is no offset with a timestamp greater or equal than `seekToTime` so just seek to the end
                // this means that we will have to wait until the metrics publisher produces to this partition
                // by default, it publishes the metric every 15 secs, but users may adjust it
                consumer.seekToEnd(Collections.singletonList(topicPartition));
            } else {
                long offset = offsetAndTimestamp.offset();
                consumer.seek(topicPartition, offset);
            }
        }
    }
}
