// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.throttle;

import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.ReplicaAssignment;
import org.apache.kafka.common.TopicPartition;

import java.util.*;

class AssignmentPair {

    private final ReplicaAssignment current;
    private final ReplicaAssignment proposed;

    AssignmentPair(ReplicaAssignment current, ReplicaAssignment proposed) {
        this.current = current;
        this.proposed = proposed;
    }

    Map<TopicPartition, List<Integer>> moveDestinations(String topic) {
        //For each partition in the proposed list, remove any replica that did not move
        Map<TopicPartition, List<Integer>> moveDestinationReplicas = new TreeMap<>(topicPartitionComparator());
        for (TopicPartition tp : proposed.asMap().keySet()) {
            if (topic.equals(tp.topic())) {
                moveDestinationReplicas.put(tp, movingReplicasFor(tp));
            }
        }
        return moveDestinationReplicas;
    }

    Map<TopicPartition, List<Integer>> moveSources(String topic) {
        //Throttle all existing replicas (as any one might be a leader), so we just need to filter out any which aren't moving.
        Map<TopicPartition, List<Integer>> existingPartitionsThatMove = new TreeMap<>(topicPartitionComparator());
        for (TopicPartition tp : proposed.asMap().keySet()) {
            if (topic.equals(tp.topic())) {
                if (!movingReplicasFor(tp).isEmpty()) {
                    //then ignore the moving replicas, instead use all original ones
                    existingPartitionsThatMove.put(tp, current.asMap().get(tp));
                }
            }
        }
        return existingPartitionsThatMove;
    }

    private List<Integer> movingReplicasFor(TopicPartition tp) {
        List<Integer> proposedReplicas = new ArrayList(proposed.asMap().get(tp));
        List<Integer> existingReplicas = current.asMap().get(tp);
        proposedReplicas.removeAll(existingReplicas);
        return proposedReplicas;
    }

    Collection<Broker> brokers() {
        Set allBrokers = new HashSet();
        allBrokers.addAll(current.brokers());
        allBrokers.addAll(proposed.brokers());
        return allBrokers;
    }

    private static Comparator<TopicPartition> topicPartitionComparator() {
        return new Comparator<TopicPartition>() {
            @Override
            public int compare(TopicPartition tp1, TopicPartition tp2) {
                if (tp1 == null && tp2 == null) return 0;
                if (tp2 == null) return -1;
                if (tp1 == null) return 1;
                int topicComparison = tp1.topic().compareTo(tp2.topic());
                return topicComparison == 0 ? Integer.compare(tp1.partition(), tp2.partition()) : topicComparison;
            }
        };
    }

    Collection<String> topics() {
        return proposed.topics();
    }
}
