// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.topology;

import io.confluent.kafka.databalancing.CompareResult;
import io.confluent.kafka.databalancing.RebalanceContext;
import kafka.admin.BrokerMetadata;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.utils.Utils;

import java.text.DecimalFormat;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Deque;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import static io.confluent.kafka.databalancing.Utils.sorted;

public final class TopologyUtils {

    private TopologyUtils() {}

    public static BrokerMetadata findBrokerMetadata(List<BrokerMetadata> brokers, int brokerId) {
        for (BrokerMetadata broker : brokers) {
            if (broker.id() == brokerId)
                return broker;
        }
        return null;
    }

    /** Excludes `context.brokersToBeRemoved` from the result. */
    public static List<Broker> availableBrokers(RebalanceContext context, Collection<Broker> brokers) {
        List<Broker> result = new ArrayList<>(brokers);
        result.removeAll(context.brokersToBeRemoved());
        return result;
    }

    public static List<Broker> availableBrokersOnRack(RebalanceContext context, String rack) {
        return availableBrokers(context, context.brokersOnRack(rack));
    }

    public static List<Broker> allAvailableBrokers(RebalanceContext context) {
        return availableBrokers(context, context.allBrokers());
    }

    /** Returns available brokers sorted by least loaded. */
    public static List<Broker> leastLoadedBrokers(RebalanceContext context) {
        return sortBrokersBySize(context, allAvailableBrokers(context), null);
    }

    public static List<Broker> sortBrokersByReplicaCount(RebalanceContext context, Collection<Broker> brokers, String topic) {
        Map<Broker, Long> brokerTopicCounts = topic == null ? null : brokerReplicaCounts(context, brokers, topic);
        Map<Broker, Long> brokerCounts = brokerReplicaCounts(context, brokers, null);
        return sorted(brokers, brokerTopicComparator(brokerTopicCounts, brokerCounts));
    }

    public static List<Broker> sortBrokersByLeaderCount(RebalanceContext context, Collection<Broker> brokers, String topic) {
        Map<Broker, Long> brokerTopicCounts = topic == null ? null : brokerLeaderCounts(context, brokers, topic);
        Map<Broker, Long> brokerCounts = brokerLeaderCounts(context, brokers, null);
        return sorted(brokers, brokerTopicComparator(brokerTopicCounts, brokerCounts));
    }

    public static List<Broker> sortBrokersBySize(RebalanceContext context, Collection<Broker> brokers, String topic) {
        Map<Broker, Long> brokerTopicCounts = topic == null ? null : brokerSizes(context, brokers, topic);
        Map<Broker, Long> brokerCounts = brokerSizes(context, brokers, null);
        return sorted(brokers, brokerTopicComparator(brokerTopicCounts, brokerCounts));
    }

    public static List<Broker> sortBrokersByLeadersSize(RebalanceContext context, Collection<Broker> brokers, String topic) {
        Map<Broker, Long> brokerTopicCounts = topic == null ? null : brokerLeadersSizes(context, brokers, topic);
        Map<Broker, Long> brokerCounts = brokerLeadersSizes(context, brokers, null);
        return sorted(brokers, brokerTopicComparator(brokerTopicCounts, brokerCounts));
    }

    public static List<TopicPartition> sortPartitionsBySize(RebalanceContext context, Collection<TopicPartition> partitions) {
        Map<TopicPartition, Long> partitionToSize = new HashMap<>();
        for (TopicPartition topicPartition : partitions) {
            if (!partitionToSize.containsKey(topicPartition))
                partitionToSize.put(topicPartition, context.partitionSize(topicPartition));
        }
        return sorted(partitions, partitionSizeComparator(partitionToSize));
    }

    /**
     * Find the least loaded brokers, but push those on the supplied racks to the top of the list.
     */
    public static List<Broker> leastLoadedBrokersPreferringTheseRacks(RebalanceContext context, Set<String> racks) {
        List<Broker> leastLoaded = leastLoadedBrokers(context);
        uprank(leastLoaded, context.brokersOnRacks(racks));
        return leastLoaded;
    }

    private static void uprank(List<Broker> all, Collection<Broker> toUprank) {
        all.removeAll(toUprank);
        all.addAll(0, toUprank);
    }

    /**
     * Find the least loaded brokers, but push those on the supplied racks to the bottom of the list.
     */
    public static List<Broker> leastLoadedBrokersPreferringOtherRacks(RebalanceContext context, Set<String> racks) {
        List<Broker> leastLoaded = leastLoadedBrokers(context);
        downrank(leastLoaded, context.brokersOnRacks(racks));
        return leastLoaded;
    }

    public static void downrank(List<Broker> all, Collection<Broker> toDownrank) {
        List<Broker> downranked = new ArrayList<>();
        for (Iterator<Broker> it = all.iterator(); it.hasNext(); ) {
            Broker broker = it.next();
            if (toDownrank.contains(broker)) {
                it.remove();
                downranked.add(broker);
            }
        }
        all.addAll(downranked);
    }

    public static Set<String> racksFor(RebalanceContext context, TopicPartition topicPartition) {
        return context.racks(context.brokers(topicPartition));
    }

    public static Map<Broker, Long> brokerLeaderCounts(RebalanceContext context, Collection<Broker> brokers, String topic) {
        Map<Broker, Long> brokerToLeaderCount = new HashMap<>(brokers.size());
        for (Broker broker : brokers)
            brokerToLeaderCount.put(broker, (long) context.brokerTopicLeaderCount(broker, topic));
        return brokerToLeaderCount;
    }

    public static Map<Broker, Long> brokerReplicaCounts(RebalanceContext context, Collection<Broker> brokers, String topic) {
        Map<Broker, Long> brokerToReplicaCount = new HashMap<>(brokers.size());
        for (Broker broker : brokers)
            brokerToReplicaCount.put(broker, (long) context.brokerTopicReplicaCount(broker, topic));
        return brokerToReplicaCount;
    }

    public static Map<String, Long> rackLeaderCounts(RebalanceContext context, String rack, String topic) {
        Collection<Broker> brokers = context.brokersOnRack(rack);
        Map<String, Long> rackToLeaderCount = new HashMap<>();
        for (Broker broker : brokers) {
            int brokerLeaderCount = context.brokerTopicLeaderCount(broker, topic);
            String brokerRack = rack == null ? context.brokerRack(broker) : rack;
            if (brokerRack != null) {
                Long rackLeaderCount = rackToLeaderCount.get(brokerRack);
                if (rackLeaderCount == null)
                    rackLeaderCount = 0L;
                rackToLeaderCount.put(brokerRack, rackLeaderCount + brokerLeaderCount);
            }
        }
        return rackToLeaderCount;
    }

    public static Map<String, Long> rackReplicaCounts(RebalanceContext context, String rack, String topic) {
        Collection<Broker> brokers = context.brokersOnRack(rack);
        Map<String, Long> rackToReplicaCount = new HashMap<>();
        for (Broker broker : brokers) {
            int brokerReplicaCount = context.brokerTopicReplicaCount(broker, topic);
            String brokerRack = rack == null ? context.brokerRack(broker) : rack;
            if (brokerRack != null) {
                Long rackReplicaCount = rackToReplicaCount.get(brokerRack);
                if (rackReplicaCount == null)
                    rackReplicaCount = 0L;
                rackToReplicaCount.put(brokerRack, rackReplicaCount + brokerReplicaCount);
            }
        }
        return rackToReplicaCount;
    }

    public static Map<Broker, Long> brokerLeadersSizes(RebalanceContext context, Collection<Broker> brokers, String topic) {
        Map<Broker, Long> result = new HashMap<>(brokers.size());
        for (Broker broker : brokers)
            result.put(broker, context.brokerTopicLeadersSize(broker, topic));
        return result;
    }

    public static Map<Broker, Long> brokerSizes(RebalanceContext context, Collection<Broker> brokers, String topic) {
        Map<Broker, Long> result = new HashMap<>(brokers.size());
        for (Broker broker : brokers)
            result.put(broker, context.brokerTopicSize(broker, topic));
        return result;
    }

    public static Map<String, Long> rackLeadersSizes(RebalanceContext context, String rack, String topic) {
        Collection<Broker> brokers = context.brokersOnRack(rack);
        Map<String, Long> result = new HashMap<>();
        for (Broker broker : brokers) {
            long brokerSize = context.brokerTopicLeadersSize(broker, topic);
            String brokerRack = rack == null ? context.brokerRack(broker) : rack;
            if (brokerRack != null) {
                Long rackSize = result.get(brokerRack);
                if (rackSize == null)
                    rackSize = 0L;
                result.put(brokerRack, rackSize + brokerSize);
            }
        }
        return result;
    }

    public static Map<String, Long> rackSizes(RebalanceContext context, String rack, String topic) {
        Collection<Broker> brokers = context.brokersOnRack(rack);
        Map<String, Long> result = new HashMap<>();
        for (Broker broker : brokers) {
            long brokerSize = context.brokerTopicSize(broker, topic);
            String brokerRack = rack == null ? context.brokerRack(broker) : rack;
            if (brokerRack != null) {
                Long rackSize = result.get(brokerRack);
                if (rackSize == null)
                    rackSize = 0L;
                result.put(brokerRack, rackSize + brokerSize);
            }
        }
        return result;
    }

    public static List<Replica> replicas(RebalanceContext context, Collection<Broker> brokers, String topic) {
        List<Replica> replicas = new ArrayList<>();
        for (Broker broker : brokers) {
            List<TopicPartition> topicPartitions = context.replicas(broker);
            for (TopicPartition tp : topicPartitions) {
                if (topic == null || tp.topic().equals(topic))
                    replicas.add(new Replica(tp, broker));
            }
        }
        return replicas;
    }

    public static List<TopicPartition> leaders(RebalanceContext context, List<Broker> brokers, String topic) {
        List<TopicPartition> leaders = new ArrayList<>();
        for (Broker broker : brokers) {
            List<TopicPartition> brokerLeaders = context.leaders(broker);
            List<TopicPartition> filteredLeaders;
            if (topic == null)
                filteredLeaders = brokerLeaders;
            else {
                filteredLeaders = new ArrayList<>();
                for (TopicPartition tp : brokerLeaders) {
                    if (tp.topic().equals(topic))
                        filteredLeaders.add(tp);
                }
            }
            leaders.addAll(sorted(filteredLeaders, topicPartitionComparator));
        }
        return leaders;
    }

    public static List<Replica> availableReplicasOnRack(RebalanceContext context, String rack, String topic) {
        return replicas(context, availableBrokersOnRack(context, rack), topic);
    }

    public static List<TopicPartition> availableLeadersOnRack(RebalanceContext context, String rack, String topic) {
        return leaders(context, availableBrokersOnRack(context, rack), topic);
    }

    /* Replicas returned are interleaved so that we visit each broker once before we visit any of them a second time. */
    public static List<Replica> interleavedReplicas(RebalanceContext context, List<Broker> brokers, String topic) {
        List<Replica> result = new ArrayList<>();
        // Order of `brokers` is important
        Map<Broker, Deque<TopicPartition>> brokerToReplicas = new LinkedHashMap<>(brokers.size());
        for (Broker broker : brokers) {
            List<TopicPartition> replicas = sorted(context.replicas(broker), topicPartitionComparator);
            Deque<TopicPartition> filteredReplicas;
            if (topic == null)
                filteredReplicas = new ArrayDeque<>(replicas);
            else {
                filteredReplicas = new ArrayDeque<>();
                for (TopicPartition tp : replicas) {
                    if (tp.topic().equals(topic))
                        filteredReplicas.add(tp);
                }
            }
            if (!filteredReplicas.isEmpty())
                brokerToReplicas.put(broker, filteredReplicas);
        }
        while (!brokerToReplicas.isEmpty()) {
            for (Iterator<Map.Entry<Broker, Deque<TopicPartition>>> it = brokerToReplicas.entrySet().iterator(); it.hasNext(); ) {
                Map.Entry<Broker, Deque<TopicPartition>> entry = it.next();
                Broker broker = entry.getKey();
                Deque<TopicPartition> replicas = entry.getValue();
                TopicPartition partition = replicas.pollFirst();
                if (replicas.isEmpty())
                    it.remove();
                result.add(new Replica(partition, broker));

            }
        }
        return result;
    }

    public static String formattedPartitionSize(RebalanceContext context, TopicPartition tp) {
        long partitionSize = context.partitionSize(tp);
        return io.confluent.kafka.databalancing.Utils.bytesToFormattedMb(partitionSize);
    }

    public static String formattedBrokerSize(RebalanceContext context, Broker broker) {
        long brokerSize = context.brokerSize(broker);
        return io.confluent.kafka.databalancing.Utils.bytesToFormattedMb(brokerSize);
    }

    public static CompareResult compareSizeFairness(long size1, long tp1Size, long size2, long tp2Size, double mean) {
        double current = square(size1 - mean) + square(size2 - mean);
        double candidate = square(size1 - tp1Size + tp2Size - mean) + square(size2 + tp1Size - tp2Size - mean);
        return CompareResult.fromInt(Double.compare(current, candidate));
    }

    private static double square(double value) {
        return value * value;
    }

    public static Comparator<Broker> brokerComparator(final Map<Broker, Long> brokerValueMap) {
        return new Comparator<Broker>() {
            @Override
            public int compare(Broker b1, Broker b2) {
                int result = Long.compare(brokerValueMap.get(b1), brokerValueMap.get(b2));
                // Ensure deterministic behaviour
                if (result == 0)
                    return Long.compare(b1.id(), b2.id());
                return result;
            }
        };
    }

    public static Comparator<Broker> brokerTopicComparator(final Map<Broker, Long> brokerTopicValueMap,
                                                           final Map<Broker, Long> brokerValueMap) {
        return new Comparator<Broker>() {
            @Override
            public int compare(Broker b1, Broker b2) {
                int result = brokerTopicValueMap == null ? 0 :
                        Long.compare(brokerTopicValueMap.get(b1), brokerTopicValueMap.get(b2));
                if (result != 0)
                    return result;
                // If the number of brokers is larger than the number of partitions per topic, there could be many
                // brokers without any partition for that topic. This comparison is important for this case
                result = Long.compare(brokerValueMap.get(b1), brokerValueMap.get(b2));
                if (result != 0)
                    return result;
                // Ensure deterministic behaviour
                return Long.compare(b1.id(), b2.id());
            }
        };
    }

    public static Comparator<String> rackTopicComparator(final Map<String, Long> rackTopicValueMap, final Map<String, Long> rackValueMap) {
        return new Comparator<String>() {
            @Override
            public int compare(String rack1, String rack2) {
                int result = rackTopicValueMap == null ? 0 :
                        Long.compare(rackTopicValueMap.get(rack1), rackTopicValueMap.get(rack2));
                if (result != 0)
                    return result;

                result = Long.compare(rackValueMap.get(rack1), rackValueMap.get(rack2));
                if (result != 0)
                    return result;

                // Ensure deterministic behaviour
                return rack1.compareTo(rack2);
            }
        };
    }

    public static Comparator<String> rackComparator(final Map<String, Long> rackValueMap) {
        return new Comparator<String>() {
            @Override
            public int compare(String rack1, String rack2) {
                int result = Long.compare(rackValueMap.get(rack1), rackValueMap.get(rack2));
                if (result != 0)
                    return result;

                // Ensure deterministic behaviour
                return rack1.compareTo(rack2);
            }
        };
    }

    public static Comparator<TopicPartition> partitionSizeComparator(final Map<TopicPartition, Long> partitionValueMap) {
        return new Comparator<TopicPartition>() {
            @Override
            public int compare(TopicPartition tp1, TopicPartition tp2) {
                int result = Long.compare(partitionValueMap.get(tp1), partitionValueMap.get(tp2));
                if (result != 0)
                    return result;
                // Ensure deterministic behaviour
                return topicPartitionComparator.compare(tp1, tp2);
            }
        };
    }

    public static final Comparator<TopicPartition> topicPartitionComparator = new Comparator<TopicPartition>() {
        @Override
        public int compare(TopicPartition tp1, TopicPartition tp2) {
            int result = tp1.topic().compareTo(tp2.topic());
            if (result == 0)
                return Integer.compare(tp1.partition(), tp2.partition());
            return result;
        }
    };

    public static List<String> partitionsByTopicToLines(Set<TopicPartition> topicPartitions) {
        Map<String, List<Integer>> topicToPartitions = new HashMap<>();
        for (TopicPartition partition : topicPartitions) {
            List<Integer> partitionsForTopic = topicToPartitions.get(partition.topic());
            if (partitionsForTopic == null) {
                partitionsForTopic = new ArrayList<>();
                topicToPartitions.put(partition.topic(), partitionsForTopic);
            }
            partitionsForTopic.add(partition.partition());
        }

        List<String> result = new ArrayList<>();
        for (String topic : Utils.sorted(topicToPartitions.keySet())) {
            List<Integer> partitions = Utils.sorted(topicToPartitions.get(topic));
            String line = "Topic " + topic + ": " + Utils.mkString(partitions, ",");
            result.add(line);
        }
        return result;
    }
}
