// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.view;

import io.confluent.kafka.databalancing.CompareResult;
import io.confluent.kafka.databalancing.RebalanceContext;
import io.confluent.kafka.databalancing.constraint.Constraints;
import io.confluent.kafka.databalancing.constraint.RebalanceConstraints;
import io.confluent.kafka.databalancing.fairness.LeaderFairness;
import io.confluent.kafka.databalancing.fairness.ReplicaFairness;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.Replica;
import org.apache.kafka.common.TopicPartition;

import java.util.Collections;
import java.util.List;

public abstract class BrokerFairView implements ClusterView {

    protected final RebalanceContext context;
    protected final String rack;
    protected final String topic;
    private final ReplicaFairness replicaFairness;
    private final LeaderFairness leaderFairness;
    private final Constraints constraints;

    // Created lazily
    private List<TopicPartition> leadersOnAboveParBrokers;
    private List<Replica> replicasOnAboveParBrokers;

    /**
     * @param context required
     * @param rack nullable
     * @param topic nullable
     */
    public BrokerFairView(RebalanceContext context, String rack, String topic) {
        this.context = context;
        this.rack = rack;
        this.topic = topic;

        this.replicaFairness = createReplicaFairness(this.context, rack, topic);
        this.leaderFairness = createLeaderFairness(this.context, rack, topic);
        this.constraints = new Constraints(this.context, rack);
    }

    protected abstract List<TopicPartition> createLeadersOnAboveParBrokers(RebalanceContext context,
                                                                           LeaderFairness fairness, String topic);

    protected abstract List<Replica> createReplicasOnAboveParBrokers(RebalanceContext context, ReplicaFairness fairness,
                                                                     String topic);

    protected abstract ReplicaFairness createReplicaFairness(RebalanceContext context, String rack, String topic);

    protected abstract LeaderFairness createLeaderFairness(RebalanceContext context, String rack, String topic);

    @Override
    public CompareResult compareLeaderFairness(TopicPartition topicPartition, Broker b1, Broker b2) {
        return leaderFairness.compareBrokerFairness(topicPartition, b1, b2);
    }

    @Override
    public CompareResult compareReplicaFairness(TopicPartition topicPartition, Broker b1, Broker b2) {
        return replicaFairness.compareBrokerFairness(topicPartition, b1, b2);
    }

    @Override
    public CompareResult compareReplicaFairness(Broker b1, TopicPartition tp1, Broker b2, TopicPartition tp2) {
        return replicaFairness.compareBrokerFairness(b1, tp1, b2, tp2);
    }

    @Override
    public List<Replica> replicasOnAboveParBrokers() {
        if (replicasOnAboveParBrokers == null)
            replicasOnAboveParBrokers = Collections.unmodifiableList(createReplicasOnAboveParBrokers(context,
                    replicaFairness, topic));
        return replicasOnAboveParBrokers;
    }

    @Override
    public List<Broker> brokersWithBelowParReplicaFairness() {
        return replicaFairness.belowParBrokers();
    }

    @Override
    public List<TopicPartition> leadersOnAboveParBrokers() {
        if (leadersOnAboveParBrokers == null)
            leadersOnAboveParBrokers = Collections.unmodifiableList(createLeadersOnAboveParBrokers(context,
                    leaderFairness, topic));
        return leadersOnAboveParBrokers;
    }

    @Override
    public List<Broker> brokersWithBelowParLeaderFairness() {
        return leaderFairness.belowParBrokers();
    }

    @Override
    public RebalanceConstraints constraints() {
        return constraints;
    }

    public ReplicaFairness replicaFairness() {
        return replicaFairness;
    }

    public LeaderFairness leaderFairness() {
        return leaderFairness;
    }

}
