// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.view;

import io.confluent.kafka.databalancing.RebalanceContext;
import io.confluent.kafka.databalancing.fairness.LeaderFairness;
import io.confluent.kafka.databalancing.fairness.LeaderSizeFairness;
import io.confluent.kafka.databalancing.fairness.ReplicaFairness;
import io.confluent.kafka.databalancing.fairness.ReplicaSizeFairness;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.Replica;
import io.confluent.kafka.databalancing.topology.TopologyUtils;
import org.apache.kafka.common.TopicPartition;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

public class BrokerSizeFairView extends BrokerFairView {

    /**
     * @param context required
     * @param rack nullable
     * @param topic nullable
     */
    public BrokerSizeFairView(RebalanceContext context, String rack, String topic) {
        super(context, rack, topic);
    }

    @Override
    protected List<TopicPartition> createLeadersOnAboveParBrokers(RebalanceContext context, LeaderFairness fairness,
                                                                  String topic) {
        List<TopicPartition> result = new ArrayList<>();
        for (Broker broker : fairness.aboveParBrokers()) {
            List<TopicPartition> leaders = createSortedBrokerLeaders(context, fairness, broker, topic, context.leaders(broker));
            result.addAll(leaders);
        }
        return result;
    }

    /**
     * Sorts broker leader partitions by the ones that have the most impact in reducing the variance. In the optimal case
     * we move a leader that causes the broker size to be the same as the broker fair value. In case of ties, we
     * choose the smaller partition (less data movement). If there are still ties at this point, we compare topic names
     * and partition ids to ensure determinism.
     */
    private static List<TopicPartition> createSortedBrokerLeaders(final RebalanceContext context, LeaderFairness fairness,
                                                                  Broker broker, String topic,
                                                                  List<TopicPartition> brokerPartitions) {
        List<TopicPartition> sorted = new ArrayList<>();
        for (TopicPartition topicPartition : brokerPartitions) {
            if (topic == null || topicPartition.topic().equals(topic))
                sorted.add(topicPartition);
        }

        final double deviation = context.brokerTopicLeadersSize(broker, topic) - fairness.brokerFairValue();

        Collections.sort(sorted, new Comparator<TopicPartition>() {
            @Override
            public int compare(TopicPartition tp1, TopicPartition tp2) {
                long c1 = context.partitionSize(tp1);
                long c2 = context.partitionSize(tp2);
                int result = Double.compare(Math.abs(c1 - deviation), Math.abs(c2 - deviation));
                if (result != 0)
                    return result;
                if (c1 != c2)
                    return Long.compare(c1, c2);
                return TopologyUtils.topicPartitionComparator.compare(tp1, tp2);
            }
        });
        return sorted;
    }

    @Override
    protected List<Replica> createReplicasOnAboveParBrokers(RebalanceContext context, ReplicaFairness fairness,
                                                            String topic) {
        List<Replica> result = new ArrayList<>();
        for (Broker broker : fairness.aboveParBrokers()) {
            List<Replica> replicas = createSortedBrokerReplicas(context, fairness, broker, topic, context.replicas(broker));
            result.addAll(replicas);
        }
        return result;
    }

    /**
     * Sorts broker replicas by the ones that have the most impact in reducing the variance. In the optimal case
     * we move a replica that causes the broker size to be the same as the broker fair value. In case of ties, we
     * choose the smaller partition (less data movement). If there are still ties at this point, we compare broker ids,
     * topic names and partition ids to ensure determinism.
     */
    private static List<Replica> createSortedBrokerReplicas(final RebalanceContext context,
                                                            ReplicaFairness fairness, Broker broker, String topic,
                                                            List<TopicPartition> brokerPartitions) {

        List<Replica> replicas = new ArrayList<>();
        for (TopicPartition topicPartition : brokerPartitions) {
            if (topic == null || topicPartition.topic().equals(topic))
                replicas.add(new Replica(topicPartition, broker));
        }

        final double deviation = context.brokerTopicSize(broker, topic) - fairness.brokerFairValue();

        Collections.sort(replicas, new Comparator<Replica>() {
            @Override
            public int compare(Replica r1, Replica r2) {
                TopicPartition tp1 = r1.topicPartition();
                TopicPartition tp2 = r2.topicPartition();
                long c1 = context.partitionSize(tp1);
                long c2 = context.partitionSize(tp2);
                int result = Double.compare(Math.abs(c1 - deviation), Math.abs(c2 - deviation));
                if (result != 0)
                    return result;
                if (c1 != c2)
                    return Long.compare(c1, c2);
                return TopologyUtils.topicPartitionComparator.compare(tp1, tp2);
            }
        });

        return replicas;
    }

    @Override
    public ClusterView refresh(RebalanceContext context) {
        return new BrokerSizeFairView(context, rack, topic);
    }

    @Override
    protected ReplicaFairness createReplicaFairness(RebalanceContext context, String rack, String topic) {
        return new ReplicaSizeFairness(context, rack, topic);
    }

    @Override
    protected LeaderFairness createLeaderFairness(RebalanceContext context, String rack, String topic) {
        return new LeaderSizeFairness(context, rack, topic);
    }

}
