// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.view;

import io.confluent.kafka.databalancing.RebalanceContext;
import io.confluent.kafka.databalancing.fairness.LeaderCountFairness;
import io.confluent.kafka.databalancing.fairness.LeaderFairness;
import io.confluent.kafka.databalancing.fairness.ReplicaCountFairness;
import io.confluent.kafka.databalancing.fairness.ReplicaFairness;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.Replica;
import io.confluent.kafka.databalancing.topology.TopologyUtils;
import org.apache.kafka.common.TopicPartition;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import static io.confluent.kafka.databalancing.topology.TopologyUtils.sortBrokersByLeaderCount;
import static io.confluent.kafka.databalancing.topology.TopologyUtils.sortBrokersByReplicaCount;

public class RackCountFairView extends RackFairView {

    public RackCountFairView(RebalanceContext context, String topic) {
        super(context, topic);
    }

    @Override
    protected List<TopicPartition> createLeadersOnAboveParRacks(RebalanceContext context, LeaderFairness fairness,
                                                                String topic) {
        List<TopicPartition> result = new ArrayList<>();
        for (String rack : fairness.aboveParRacks()) {
            List<Broker> brokers = sortBrokersByLeaderCount(context, TopologyUtils.availableBrokersOnRack(context, rack), topic);
            Collections.reverse(brokers);
            result.addAll(TopologyUtils.leaders(context, brokers, topic));
        }
        return result;
    }

    @Override
    protected List<Replica> createReplicasOnAboveParRacks(RebalanceContext context, ReplicaFairness fairness,
                                                          String topic) {
        List<Replica> result = new ArrayList<>();
        for (String rack : fairness.aboveParRacks()) {
            List<Broker> brokers = sortBrokersByReplicaCount(context, TopologyUtils.availableBrokersOnRack(context, rack), topic);
            Collections.reverse(brokers);
            result.addAll(TopologyUtils.interleavedReplicas(context, brokers, topic));
        }
        return result;
    }

    @Override
    protected List<Broker> createBrokersWithBelowParLeaderMetric(RebalanceContext context, LeaderFairness fairness, String topic) {
        List<Broker> result = new ArrayList<>();
        for (String rack : fairness.belowParRacks())
            result.addAll(sortBrokersByLeaderCount(context, TopologyUtils.availableBrokersOnRack(context, rack), topic));
        return result;
    }

    @Override
    protected List<Broker> createBrokersWithBelowParReplicaMetric(RebalanceContext context, ReplicaFairness fairness, String topic) {
        List<Broker> result = new ArrayList<>();
        for (String rack : fairness.belowParRacks())
            result.addAll(sortBrokersByReplicaCount(context, TopologyUtils.availableBrokersOnRack(context, rack), topic));
        return result;
    }

    @Override
    protected ReplicaFairness createReplicaFairness(RebalanceContext context, String topic) {
        return new ReplicaCountFairness(context, null, topic);
    }

    @Override
    protected LeaderFairness createLeaderFairness(RebalanceContext context, String topic) {
        return new LeaderCountFairness(context, null, topic);
    }

    @Override
    public ClusterView refresh(RebalanceContext context) {
        return new RackCountFairView(context, topic);
    }

}
