// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.view;

import io.confluent.kafka.databalancing.CompareResult;
import io.confluent.kafka.databalancing.RebalanceContext;
import io.confluent.kafka.databalancing.constraint.Constraints;
import io.confluent.kafka.databalancing.constraint.RebalanceConstraints;
import io.confluent.kafka.databalancing.fairness.LeaderFairness;
import io.confluent.kafka.databalancing.fairness.ReplicaFairness;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.Replica;
import org.apache.kafka.common.TopicPartition;

import java.util.Collections;
import java.util.List;

public abstract class RackFairView implements ClusterView {

    protected final RebalanceContext context;
    protected final String topic;

    private final ReplicaFairness replicaFairness;
    private final LeaderFairness leaderFairness;
    private final Constraints constraints;

    // Created lazily
    private List<Replica> replicasOnAboveParBrokers;
    private List<TopicPartition> leadersOnAboveParBrokers;
    private List<Broker> brokersWithBelowParReplicaFairness;
    private List<Broker> brokersWithBelowParLeaderFairness;

    public RackFairView(RebalanceContext context, String topic) {
        this.context = context;
        this.topic = topic;

        this.replicaFairness = createReplicaFairness(context, topic);
        this.leaderFairness = createLeaderFairness(context, topic);
        this.constraints = new Constraints(context, null);
    }

    protected abstract ReplicaFairness createReplicaFairness(RebalanceContext context, String topic);

    protected abstract LeaderFairness createLeaderFairness(RebalanceContext context, String topic);

    protected abstract List<TopicPartition> createLeadersOnAboveParRacks(RebalanceContext context, LeaderFairness fairness,
                                                                         String topic);

    protected abstract List<Replica> createReplicasOnAboveParRacks(RebalanceContext context, ReplicaFairness fairness,
                                                                   String topic);

    protected abstract List<Broker> createBrokersWithBelowParLeaderMetric(RebalanceContext context,
                                                                          LeaderFairness fairness, String topic);

    protected abstract List<Broker> createBrokersWithBelowParReplicaMetric(RebalanceContext context,
                                                                           ReplicaFairness fairness, String topic);


    @Override
    public CompareResult compareLeaderFairness(TopicPartition topicPartition, Broker b1, Broker b2) {
        return leaderFairness.compareRackFairness(topicPartition, b1, b2);
    }

    @Override
    public CompareResult compareReplicaFairness(TopicPartition topicPartition, Broker b1, Broker b2) {
        return replicaFairness.compareRackFairness(topicPartition, b1, b2);
    }

    @Override
    public CompareResult compareReplicaFairness(Broker b1, TopicPartition tp1, Broker b2, TopicPartition tp2) {
        return replicaFairness.compareRackFairness(b1, tp1, b2, tp2);
    }

    public ReplicaFairness replicaFairness() {
        return replicaFairness;
    }

    public LeaderFairness leaderFairness() {
        return leaderFairness;
    }

    @Override
    public List<Replica> replicasOnAboveParBrokers() {
        if (replicasOnAboveParBrokers == null)
            replicasOnAboveParBrokers = Collections.unmodifiableList(createReplicasOnAboveParRacks(context,
                    replicaFairness, topic));
        return replicasOnAboveParBrokers;
    }

    @Override
    public List<Broker> brokersWithBelowParReplicaFairness() {
        if (brokersWithBelowParReplicaFairness == null)
            brokersWithBelowParReplicaFairness = Collections.unmodifiableList(createBrokersWithBelowParReplicaMetric(context,
                    replicaFairness, topic));
        return brokersWithBelowParReplicaFairness;
    }

    @Override
    public List<TopicPartition> leadersOnAboveParBrokers() {
        if (leadersOnAboveParBrokers == null)
            leadersOnAboveParBrokers = Collections.unmodifiableList(createLeadersOnAboveParRacks(context,
                    leaderFairness, topic));
        return leadersOnAboveParBrokers;
    }

    @Override
    public List<Broker> brokersWithBelowParLeaderFairness() {
        if (brokersWithBelowParLeaderFairness == null)
            brokersWithBelowParLeaderFairness = Collections.unmodifiableList(createBrokersWithBelowParLeaderMetric(context,
                    leaderFairness, topic));
        return brokersWithBelowParLeaderFairness;
    }

    @Override
    public RebalanceConstraints constraints() {
        return constraints;
    }

}
