// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing;

import io.confluent.kafka.databalancing.constraint.Constraints;
import io.confluent.kafka.databalancing.topology.Broker;
import kafka.admin.BrokerMetadata;
import org.apache.kafka.common.TopicPartition;
import org.junit.Test;

import java.util.List;
import java.util.Map;

import static io.confluent.kafka.databalancing.TestUtils.bk;
import static io.confluent.kafka.databalancing.TestUtils.concat;
import static io.confluent.kafka.databalancing.TestUtils.createBrokers;
import static io.confluent.kafka.databalancing.TestUtils.createMap;
import static io.confluent.kafka.databalancing.TestUtils.createRebalanceContext;
import static io.confluent.kafka.databalancing.TestUtils.entry;
import static java.util.Arrays.asList;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

public class ConstraintsTest {

    private static final String topic = "topic";

    private static TopicPartition tp(int partition) {
        return new TopicPartition(topic, partition);
    }

    private static Map<String, Integer> replicationFactors(int replicationFactor) {
        return TestUtils.replicationFactors(asList(topic), replicationFactor);
    }

    @Test
    public void shouldAllowReplicaMoveIfDoesNotBreakRackConstraint() {
        //Given
        List<BrokerMetadata> brokers = concat(
                createBrokers(100, 102, "rack1"),
                createBrokers(102, 103, "rack2")
        );
        Map<TopicPartition, List<Integer>> partitions = createMap(entry(tp(0), asList(100, 101)));

        //When
        Broker brokerFrom = new Broker(100);
        Broker brokerTo = new Broker(102);
        Constraints cluster = new Constraints(createRebalanceContext(brokers, partitions, replicationFactors(2)), null);

        //Then
        assertTrue(cluster.obeysRackConstraint(tp(0), brokerFrom, brokerTo));
    }

    @Test
    public void shouldAllowReplicaCreationIfDoesNotBreakRackConstraint() {
        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(entry(tp(0), asList(100)));

        //When
        Broker brokerFrom = null; // doesn't exist (i.e. we're under-replicated
        Broker brokerTo = new Broker(101);
        Constraints cluster = new Constraints(createRebalanceContext(brokers, partitions, replicationFactors(2)), null);

        //Then
        assertTrue(cluster.obeysRackConstraint(tp(0), brokerFrom, brokerTo));
    }

    @Test
    public void shouldFailIfReplicaCreationBreaksRackConstraint() {
        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(entry(tp(0), asList(100)));

        //When
        Broker brokerFrom = null; // doesn't exist (i.e. we're under-replicated
        Broker brokerTo = new Broker(100);
        Constraints cluster = new Constraints(createRebalanceContext(brokers, partitions, replicationFactors(2)), null);

        //Then
        assertFalse(cluster.obeysRackConstraint(tp(0), brokerFrom, brokerTo));
    }

    @Test
    public void shouldFailPartitionConstraintIfReplicaAlreadyExistsOnTargetForMove() {
        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(entry(tp(0), asList(100)));

        //When
        Constraints cluster = new Constraints(createRebalanceContext(brokers, partitions, replicationFactors(2)), null);

        //Then
        assertFalse(cluster.obeysPartitionConstraint(tp(0), new Broker(100)));
    }

    @Test
    public void shouldPassPartitionConstraintIfReplicaDoesntExistsOnTargetForMove() {
        //Given
        List<BrokerMetadata> brokers = asList(bk(100, "rack1"), bk(101, "rack2"));
        Map<TopicPartition, List<Integer>> partitions = createMap(entry(tp(0), asList(100)));

        //When
        Constraints cluster = new Constraints(createRebalanceContext(brokers, partitions, replicationFactors(2)), null);

        //Then
        assertTrue(cluster.obeysPartitionConstraint(tp(0), new Broker(101)));
    }

}
