// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing;

import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.view.BrokerFairView;
import io.confluent.kafka.databalancing.view.BrokerCountFairView;
import io.confluent.kafka.databalancing.view.RackCountFairView;
import io.confluent.kafka.databalancing.view.RackFairView;
import kafka.admin.BrokerMetadata;
import org.apache.kafka.common.TopicPartition;
import org.junit.Test;

import java.util.Collections;
import java.util.List;

import static io.confluent.kafka.databalancing.TestUtils.*;
import static java.util.Arrays.asList;
import static org.junit.Assert.assertEquals;

public class CountFairnessTest {

    private static final String topic = "topic";

    private static TopicPartition tp(int partition) {
        return new TopicPartition(topic, partition);
    }

    @Test
    public void shouldCalculateReplicaFairnessForRacks() {
        List<BrokerMetadata> brokers = concat(createBrokers(100, 102, "rack1"), createBrokers(102, 104, "rack2"));

        //1 replica, 2 racks. rack1 is Below Par, rack2 is onPar.
        RackFairView view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(101))), Collections.singletonMap("topic", 1)), null);
        assertEquals(1, (int) Math.ceil(view.replicaFairness().rackFairValue()));
        assertEquals(0, (int) Math.floor(view.replicaFairness().rackFairValue()));
        assertEquals(asList("rack1"), view.replicaFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.replicaFairness().belowParRacks());

        //2 replicas, 2 racks, both replicas on one rack
        view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(103, 102))), Collections.singletonMap("topic", 1)), null);
        assertEquals(1, (int) Math.ceil(view.replicaFairness().rackFairValue()));
        assertEquals(1, (int) Math.floor(view.replicaFairness().rackFairValue()));
        assertEquals(asList("rack2"), view.replicaFairness().aboveParRacks());
        assertEquals(asList("rack1"), view.replicaFairness().belowParRacks());

        //2 replicas, 2 racks, replicas on different racks
        view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100, 101)), entry(tp(1), asList(102, 103))),
                Collections.singletonMap("topic", 1)), null);
        assertEquals(2, (int) Math.ceil(view.replicaFairness().rackFairValue()));
        assertEquals(2, (int) Math.floor(view.replicaFairness().rackFairValue()));
        assertEquals(asList(), view.replicaFairness().aboveParRacks());
        assertEquals(asList(), view.replicaFairness().belowParRacks());

        //3 replicas, 2 racks, replicas spread 2/1
        view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100, 101, 103))), Collections.singletonMap("topic", 1)), null);
        assertEquals(2, (int) Math.ceil(view.replicaFairness().rackFairValue()));
        assertEquals(1, (int) Math.floor(view.replicaFairness().rackFairValue()));
        assertEquals(asList("rack1"), view.replicaFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.replicaFairness().belowParRacks());

        //3 replicas, 2 racks, replicas spread 3/0
        view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100, 101)), entry(tp(1), asList(100))),
                Collections.singletonMap("topic", 1)), null);
        assertEquals(2, (int) Math.ceil(view.replicaFairness().rackFairValue()));
        assertEquals(1, (int) Math.floor(view.replicaFairness().rackFairValue()));
        assertEquals(asList("rack1"), view.replicaFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.replicaFairness().belowParRacks());

        //4 replicas, 2 racks, spread 2/2
        view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(103, 102, 101, 100))), Collections.singletonMap("topic", 1)), null);
        assertEquals(2, (int) Math.ceil(view.replicaFairness().rackFairValue()));
        assertEquals(2, (int) Math.floor(view.replicaFairness().rackFairValue()));
        assertEquals(asList(), view.replicaFairness().aboveParRacks());
        assertEquals(asList(), view.replicaFairness().belowParRacks());

        //4 replicas, 2 racks, spread 3/1
        view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(103))),
                Collections.singletonMap("topic", 1)), null);
        assertEquals(2, (int) Math.ceil(view.replicaFairness().rackFairValue()));
        assertEquals(2, (int) Math.floor(view.replicaFairness().rackFairValue()));
        assertEquals(asList("rack1"), view.replicaFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.replicaFairness().belowParRacks());

        //4 replicas, 2 racks, spread 4/0
        view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(100))),
                Collections.singletonMap("topic", 1)), null);
        assertEquals(2, (int) Math.ceil(view.replicaFairness().rackFairValue()));
        assertEquals(2, (int) Math.floor(view.replicaFairness().rackFairValue()));
        assertEquals(asList("rack1"), view.replicaFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.replicaFairness().belowParRacks());
    }

    @Test
    public void shouldCalculateReplicaFairnessForBrokers() {
        List<BrokerMetadata> brokers = createBrokers(100, 102, "rack1");
        Broker broker0 = new Broker(brokers.get(0).id());
        Broker broker1 = new Broker(brokers.get(1).id());

        //1 replica, 2 brokers. bk1 is Below Par, bk2 is onPar.
        BrokerFairView view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(101))), Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(1, (int) Math.ceil(view.replicaFairness().brokerFairValue()));
        assertEquals(0, (int) Math.floor(view.replicaFairness().brokerFairValue()));
        assertEquals(asList(broker1), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(broker0), view.replicaFairness().belowParBrokers());

        //2 replicas, 2 brokers, both replicas on one broker
        view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(100))),
                Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(1, (int) Math.ceil(view.replicaFairness().brokerFairValue()));
        assertEquals(1, (int) Math.floor(view.replicaFairness().brokerFairValue()));
        assertEquals(asList(broker0), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.replicaFairness().belowParBrokers());

        //2 replicas, 2 brokers, replicas on each
        view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(101))),
                Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(1, (int) Math.ceil(view.replicaFairness().brokerFairValue()));
        assertEquals(1, (int) Math.floor(view.replicaFairness().brokerFairValue()));
        assertEquals(asList(), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(), view.replicaFairness().belowParBrokers());

        //3 replicas, 2 brokers, replicas spread 2/1
        view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100, 101)), entry(tp(1), asList(100))),
                Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(2, (int) Math.ceil(view.replicaFairness().brokerFairValue()));
        assertEquals(1, (int) Math.floor(view.replicaFairness().brokerFairValue()));
        assertEquals(asList(broker0), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.replicaFairness().belowParBrokers());

        //3 replicas, 2 brokers, replicas spread 3/0
        view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(100)), entry(tp(2), asList(100))),
                Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(2, (int) Math.ceil(view.replicaFairness().brokerFairValue()));
        assertEquals(1, (int) Math.floor(view.replicaFairness().brokerFairValue()));
        assertEquals(asList(broker0), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.replicaFairness().belowParBrokers());

        //4 replicas, 2 brokers, spread 2/2
        view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100, 101)), entry(tp(1), asList(100, 101))),
                Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(2, (int) Math.ceil(view.replicaFairness().brokerFairValue()));
        assertEquals(2, (int) Math.floor(view.replicaFairness().brokerFairValue()));
        assertEquals(asList(), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(), view.replicaFairness().belowParBrokers());

        //4 replicas, 2 brokers, spread 3/1
        view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(101))),
                Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(2, (int) Math.ceil(view.replicaFairness().brokerFairValue()));
        assertEquals(2, (int) Math.floor(view.replicaFairness().brokerFairValue()));
        assertEquals(asList(broker0), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.replicaFairness().belowParBrokers());

        //4 replicas, 2 brokers, spread 4/0
        view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(100))),
                Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(2, (int) Math.ceil(view.replicaFairness().brokerFairValue()));
        assertEquals(2, (int) Math.floor(view.replicaFairness().brokerFairValue()));
        assertEquals(asList(broker0), view.replicaFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.replicaFairness().belowParBrokers());
    }

    @Test
    public void shouldCalculateLeaderFairnessForRacks() {
        List<BrokerMetadata> brokers = concat(createBrokers(100, 102, "rack1"), createBrokers(102, 104, "rack2"));

        //1 leader, 2 racks. rack1 is Below Par, rack2 is onPar.
        RackFairView view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(101))), Collections.singletonMap("topic", 1)), null);
        assertEquals(1, (int) Math.ceil(view.leaderFairness().rackFairValue()));
        assertEquals(0, (int) Math.floor(view.leaderFairness().rackFairValue()));
        assertEquals(asList("rack1"), view.leaderFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.leaderFairness().belowParRacks());

        //2 leaders, 2 racks, both leaders on one rack
        view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(102)), entry(tp(1), asList(102))),
                Collections.singletonMap("topic", 1)), null);
        assertEquals(1, (int) Math.ceil(view.leaderFairness().rackFairValue()));
        assertEquals(1, (int) Math.floor(view.leaderFairness().rackFairValue()));
        assertEquals(asList("rack2"), view.leaderFairness().aboveParRacks());
        assertEquals(asList("rack1"), view.leaderFairness().belowParRacks());

        //2 leaders, 2 racks, leaders on different racks
        view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(102))),
                Collections.singletonMap("topic", 1)), null);
        assertEquals(1, (int) Math.ceil(view.leaderFairness().rackFairValue()));
        assertEquals(1, (int) Math.floor(view.leaderFairness().rackFairValue()));
        assertEquals(asList(), view.leaderFairness().aboveParRacks());
        assertEquals(asList(), view.leaderFairness().belowParRacks());

        //3 leaders, 2 racks, replicas spread 2/1
        view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(100)), entry(tp(2), asList(102))),
                Collections.singletonMap("topic", 1)), null);
        assertEquals(2, (int) Math.ceil(view.leaderFairness().rackFairValue()));
        assertEquals(1, (int) Math.floor(view.leaderFairness().rackFairValue()));
        assertEquals(asList("rack1"), view.leaderFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.leaderFairness().belowParRacks());

        //3 replicas, 2 racks, replicas spread 3/0
        view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(100)), entry(tp(2), asList(100))),
                Collections.singletonMap("topic", 1)), null);
        assertEquals(2, (int) Math.ceil(view.leaderFairness().rackFairValue()));
        assertEquals(1, (int) Math.floor(view.leaderFairness().rackFairValue()));
        assertEquals(asList("rack1"), view.leaderFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.leaderFairness().belowParRacks());

        //4 leader, 2 racks, spread 2/2
        view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(103)),
                        entry(tp(3), asList(103))),
                Collections.singletonMap("topic", 1)), null);
        assertEquals(2, (int) Math.ceil(view.leaderFairness().rackFairValue()));
        assertEquals(2, (int) Math.floor(view.leaderFairness().rackFairValue()));
        assertEquals(asList(), view.leaderFairness().aboveParRacks());
        assertEquals(asList(), view.leaderFairness().belowParRacks());

        //4 leader, 2 racks, spread 3/1
        view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(103))),
                Collections.singletonMap("topic", 1)), null);
        assertEquals(2, (int) Math.ceil(view.leaderFairness().rackFairValue()));
        assertEquals(2, (int) Math.floor(view.leaderFairness().rackFairValue()));
        assertEquals(asList("rack1"), view.leaderFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.leaderFairness().belowParRacks());

        //4 leader, 2 racks, spread 4/0
        view = new RackCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(100))),
                Collections.singletonMap("topic", 1)), null);
        assertEquals(2, (int) Math.ceil(view.leaderFairness().rackFairValue()));
        assertEquals(2, (int) Math.floor(view.leaderFairness().rackFairValue()));
        assertEquals(asList("rack1"), view.leaderFairness().aboveParRacks());
        assertEquals(asList("rack2"), view.leaderFairness().belowParRacks());
    }

    @Test
    public void shouldCalculateLeaderFairnessForBrokers() {
        List<BrokerMetadata> brokers = createBrokers(100, 102, "rack1");
        Broker broker0 = new Broker(brokers.get(0).id());
        Broker broker1 = new Broker(brokers.get(1).id());

        //1 leader, 2 brokers. bk1 is Below Par, bk2 is onPar.
        BrokerFairView view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(101))), Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(1, (int) Math.ceil(view.leaderFairness().brokerFairValue()));
        assertEquals(0, (int) Math.floor(view.leaderFairness().brokerFairValue()));
        assertEquals(asList(broker1), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(broker0), view.leaderFairness().belowParBrokers());

        //2 leaders, 2 brokers, both replicas on one broker
        view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(100))),
                Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(1, (int) Math.ceil(view.leaderFairness().brokerFairValue()));
        assertEquals(1, (int) Math.floor(view.leaderFairness().brokerFairValue()));
        assertEquals(asList(broker0), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.leaderFairness().belowParBrokers());

        //2 leaders, 2 brokers, replicas on each
        view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(101))),
                Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(1, (int) Math.ceil(view.leaderFairness().brokerFairValue()));
        assertEquals(1, (int) Math.floor(view.leaderFairness().brokerFairValue()));
        assertEquals(asList(), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(), view.leaderFairness().belowParBrokers());

        //3 leaders, 2 brokers, replicas spread 2/1
        view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(100)), entry(tp(2), asList(101))),
                Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(2, (int) Math.ceil(view.leaderFairness().brokerFairValue()));
        assertEquals(1, (int) Math.floor(view.leaderFairness().brokerFairValue()));
        assertEquals(asList(broker0), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.leaderFairness().belowParBrokers());

        //3 leaders, 2 brokers, replicas spread 3/0
        view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)), entry(tp(1), asList(100)), entry(tp(2), asList(100))),
                Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(2, (int) Math.ceil(view.leaderFairness().brokerFairValue()));
        assertEquals(1, (int) Math.floor(view.leaderFairness().brokerFairValue()));
        assertEquals(asList(broker0), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.leaderFairness().belowParBrokers());

        //4 leaders, 2 brokers, spread 2/2
        view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(101)),
                        entry(tp(3), asList(101))),
                Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(2, (int) Math.ceil(view.leaderFairness().brokerFairValue()));
        assertEquals(2, (int) Math.floor(view.leaderFairness().brokerFairValue()));
        assertEquals(asList(), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(), view.leaderFairness().belowParBrokers());

        //4 leaders, 2 brokers, spread 3/1
        view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(101))),
                Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(2, (int) Math.ceil(view.leaderFairness().brokerFairValue()));
        assertEquals(2, (int) Math.floor(view.leaderFairness().brokerFairValue()));
        assertEquals(asList(broker0), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.leaderFairness().belowParBrokers());

        //4 leaders, 2 brokers, spread 4/0
        view = new BrokerCountFairView(createRebalanceContext(brokers,
                createMap(entry(tp(0), asList(100)),
                        entry(tp(1), asList(100)),
                        entry(tp(2), asList(100)),
                        entry(tp(3), asList(100))),
                Collections.singletonMap(topic, 1)), "rack1", null);
        assertEquals(2, (int) Math.ceil(view.leaderFairness().brokerFairValue()));
        assertEquals(2, (int) Math.floor(view.leaderFairness().brokerFairValue()));
        assertEquals(asList(broker0), view.leaderFairness().aboveParBrokers());
        assertEquals(asList(broker1), view.leaderFairness().belowParBrokers());
    }

}
