// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing;

import io.confluent.kafka.databalancing.metric.Metrics;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.ReplicaAssignment;
import kafka.admin.BrokerMetadata;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.protocol.SecurityProtocol;
import scala.Option;
import scala.collection.JavaConversions;
import scala.collection.Seq;

import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Properties;
import java.util.Iterator;


import static java.util.Arrays.asList;
import static org.junit.Assert.fail;

public class TestUtils {

    public static List<BrokerMetadata> filterBrokersByRack(List<BrokerMetadata> brokers, String rack) {
        List<BrokerMetadata> result = new ArrayList<>();
        for (BrokerMetadata broker : brokers)
            if (rack.equals(broker.rack().get()))
                result.add(broker);
        return result;
    }

    public static <K, V> Map<K, V> merged(Map<K, V> map1, Map<K, V> map2) {
        Map<K, V> result = new HashMap<>(map1);
        result.putAll(map2);
        return result;
    }


    public static Map<TopicPartition, List<Integer>> createPartitions(List<String> topics, int partitionCount, List<Integer> replicas) {
        Map<TopicPartition, List<Integer>> result = new HashMap<>();
        for (String topic : topics)
            for (int p = 0; p < partitionCount; ++p)
                result.put(new TopicPartition(topic, p), new ArrayList<>(replicas));
        return result;
    }

    public static List<BrokerMetadata> createBrokers(int startBrokerId, int endBrokerId, String rack) {
        List<BrokerMetadata> result = new ArrayList<>();
        for (int b = startBrokerId; b < endBrokerId; ++b)
            result.add(new BrokerMetadata(b, Option.apply(rack)));
        return result;
    }

    public static Map<String, Integer> replicationFactors(List<String> topics, int replicationFactor) {
        Map<String, Integer> replicationFactors = new HashMap<>();
        for (String topic : topics)
            replicationFactors.put(topic, replicationFactor);
        return replicationFactors;
    }

    public static List<TopicPartition> brokerPartitions(Map<TopicPartition, List<Integer>> assignment, int brokerId) {
        List<TopicPartition> result = new ArrayList<>();
        for (Map.Entry<TopicPartition, List<Integer>> entry : assignment.entrySet()) {
            if (entry.getValue().contains(brokerId))
                result.add(entry.getKey());
        }
        return result;
    }

    public static List<Integer> brokers(Map<TopicPartition, List<Integer>> assignment) {
        List<Integer> result = new ArrayList<>();
        for (Map.Entry<TopicPartition, List<Integer>> entry : assignment.entrySet()) {
            TopicPartition tp = entry.getKey();
            List<Integer> replicas = entry.getValue();
            if (new HashSet<>(replicas).size() != replicas.size())
                fail("Duplicate broker ids in replicas for partition " + tp + ", broker ids " + replicas);
            result.addAll(replicas);
        }
        return result;
    }

    public static <T> List<T> filter(Collection<T> list, T element) {
        List<T> result = new ArrayList<>(list);
        result.retainAll(Collections.singleton(element));
        return result;
    }

    public static <K, V> Map<K, V> getAll(Map<K, V> map, Collection<K> keys) {
        Map<K, V> result = new HashMap<>();
        for (K key : keys) {
            V value = map.get(key);
            if (value != null)
                result.put(key, value);
        }
        return result;
    }

    public static Map<TopicPartition, Integer> leaders(Map<TopicPartition, List<Integer>> assignment) {
        Map<TopicPartition, Integer> result = new HashMap<>();
        for (Map.Entry<TopicPartition, List<Integer>> entry : assignment.entrySet())
            result.put(entry.getKey(), entry.getValue().get(0));
        return result;
    }

    public static <T> List<T> concat(Collection<T> list1, Collection<T> list2) {
        List<T> result = new ArrayList<>(list1);
        result.addAll(list2);
        return result;
    }

    public static <K, V> Map<K, V> createMap(Map.Entry<K, V>... entries) {
        Map<K, V> result = new HashMap<>();
        for (Map.Entry<K, V> entry : entries)
            result.put(entry.getKey(), entry.getValue());
        return result;
    }

    public static <K, V> Map.Entry<K, V> entry(K key, V value) {
        return new AbstractMap.SimpleImmutableEntry<K, V>(key, value);
    }

    public static BrokerMetadata bk(int brokerId, String rack) {
        return new BrokerMetadata(brokerId, Option.apply(rack));
    }

    public static MutableRebalanceContext createRebalanceContext(List<BrokerMetadata> allBrokers,
                                                                 Map<TopicPartition, List<Integer>> partitionToReplicaIds,
                                                                 Map<String, Integer> replicationFactors,
                                                                 Map<TopicPartition, Long> partitionToSize) {
        Metrics metrics = new Metrics(partitionToSize);
        return DefaultRebalanceContext.create(allBrokers, partitionToReplicaIds, replicationFactors, metrics,
                Collections.<Broker>emptyList());
    }

    public static MutableRebalanceContext createRebalanceContext(List<BrokerMetadata> allBrokers,
                                                                 Map<TopicPartition, List<Integer>> partitionToReplicaIds,
                                                                 Map<String, Integer> replicationFactors) {
        Metrics metrics = createSimpleMetrics(partitionToReplicaIds.keySet());
        return DefaultRebalanceContext.create(allBrokers, partitionToReplicaIds, replicationFactors, metrics,
                Collections.<Broker>emptyList());
    }

    /**
     * Create a Metrics instance where each partition has size 1 (effectively making the replica size optimisation a
     * no-op).
     */
    public static Metrics createSimpleMetrics(Collection<TopicPartition> allPartitions) {
        Map<TopicPartition, Long> partitionToSize = new HashMap<>();
        for (TopicPartition tp : allPartitions)
            partitionToSize.put(tp, 1L);
        return new Metrics(partitionToSize);
    }

    public static Map<Integer, List<TopicPartition>> brokerToPartitions(Map<TopicPartition, List<Integer>> partitionToReplicas) {
        Map<Integer, List<TopicPartition>> result = new HashMap<>();
        for (Map.Entry<TopicPartition, List<Integer>> entry : partitionToReplicas.entrySet()) {
            TopicPartition tp = entry.getKey();
            for (Integer broker : entry.getValue()) {
                List<TopicPartition> partitions = result.get(broker);
                if (partitions == null) {
                    partitions = new ArrayList<>();
                    result.put(broker, partitions);
                }
                partitions.add(tp);
            }
        }
        return result;
    }

    public static long partitionSizeSum(Map<TopicPartition, Long> partitionToSize, Collection<TopicPartition> partitions) {
        long result = 0;
        for (TopicPartition tp : partitions) {
            result += partitionToSize.get(tp);
        }
        return result;
    }

    public static long sum(Collection<Long> values) {
        long result = 0;
        for (long value : values)
            result += value;
        return result;
    }

    public static ReplicaAssignment assignment(List<String> topics, int partitionCount, List<Integer> replicas) {
        return new ReplicaAssignment(createPartitions(topics, partitionCount, replicas));
    }

    public static Seq<Object> toSeq(Object val) {
        return JavaConversions.asScalaBuffer(asList(val));
    }

    public static Seq<kafka.cluster.Broker> toBrokerSeq(int... ids) {
        List<kafka.cluster.Broker> list = new ArrayList<>();
        for (int id : ids)
            list.add(new kafka.cluster.Broker(id,"some-host", 1000, SecurityProtocol.PLAINTEXT));
        return JavaConversions.asScalaBuffer(list);
    }

    public static Properties props(String key, String value) {
        Properties properties = new Properties();
        properties.put(key, value);
        return properties;
    }
    public static Properties props(List<String> keyThenVal) {
        Properties properties = new Properties();
        Iterator<String> items = keyThenVal.iterator();
        while (items.hasNext()) {
            String key = items.next();
            if (!items.hasNext())
                throw new NoSuchElementException("No more elements after key: " + key + ", list: " + keyThenVal);
            properties.put(key, items.next());
        }
        return properties;
    }

    public static List<TopicPartition> filterPartitionsByTopic(Collection<TopicPartition> topicPartitions, String topic) {
        List<TopicPartition> result = new ArrayList<>();
        for (TopicPartition tp : topicPartitions) {
            if (tp.topic().equals(topic))
                result.add(tp);
        }
        return result;
    }

}
