// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.integration;

import io.confluent.kafka.databalancing.ConfluentRebalancerCommand;
import io.confluent.kafka.databalancing.RebalancerConfig;
import io.confluent.kafka.databalancing.exception.NoRebalanceInProgressException;
import io.confluent.kafka.databalancing.exception.ValidationException;
import io.confluent.kafka.databalancing.integration.utils.EmbeddedKafkaCluster;
import io.confluent.kafka.databalancing.license.DefaultLicenseValidator;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.ReplicaAssignment;
import io.confluent.kafka.databalancing.view.DefaultRebalancerFactory;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.serialization.StringSerializer;
import org.apache.kafka.common.utils.SystemTime;
import org.apache.kafka.test.TestCondition;
import org.apache.kafka.test.TestUtils;
import org.junit.After;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

public class ConfluentRebalancerIntegrationTest {
    private static final Logger LOG = LoggerFactory.getLogger(ConfluentRebalancerIntegrationTest.class);

    @Rule
    public TemporaryFolder tmpFolder = new TemporaryFolder();
    private static final int ORIG_NUM_BROKERS = 3;
    List<EmbeddedKafkaCluster> clusters;
    EmbeddedKafkaCluster metricsCluster;

    @Before
    public void setUp() throws IOException {
        clusters = new ArrayList<>();
        metricsCluster = new EmbeddedKafkaCluster(3);
        clusters.add(metricsCluster);

    }

    @After
    public void tearDown() {
        for (EmbeddedKafkaCluster cluster : clusters) {
            cluster.shutDown();
        }
    }

    /**
     * The basic test is as follows:
     *  1. Create three topics with 5 partitions and two replicas.
     *  2. Produce a decent volume of data to the cluster.
     *  3. Get the current cluster layout.
     *  4. Add 5 more brokers. Now there will be 5 empty brokers.
     *  5. Kick of the rebalance command.
     *  6. Get the new cluster layout from ZK.. The data should be evenly distributed amongst all 10 brokers now.
     */
    @Test
    public void basicExecutionTest() throws IOException, InterruptedException {
        EmbeddedKafkaCluster cluster = new EmbeddedKafkaCluster(ORIG_NUM_BROKERS, metricsCluster);
        clusters.add(cluster);
        metricsCluster.ensureTopicCreated(EmbeddedKafkaCluster.METRICS_TOPIC,
                EmbeddedKafkaCluster.METRICS_TOPIC_PARTITIONS);
        createAndSeedTopics(3, cluster);
        final int newNumBrokers = 6;
        cluster.growTo(newNumBrokers);

        ReplicaAssignment origAssignment = getPartitionAssignment(cluster);
        rebalanceCluster(cluster, metricsCluster, tmpFolder, 1000000, 60000);
        ReplicaAssignment newAssignment = getPartitionAssignment(cluster);

        assertEquals("Did not expect a change in the TopicPartitions in the cluster",
                origAssignment.topicPartitions(), newAssignment.topicPartitions());

        List<Integer> newBrokers = new ArrayList<>();
        for (int i = ORIG_NUM_BROKERS; i < newNumBrokers; ++i) {
            newBrokers.add(i);
        }

        Set<Integer> newReplicas = new HashSet<>();
        // Validate that replicas have moved onto the new brokers.
        for (TopicPartition tp : origAssignment.topicPartitions()) {
            boolean disjoint =  Collections.disjoint(origAssignment.replicas(tp), newBrokers);
            assertTrue("Expected no new broker to host replicas in the original assignment", disjoint);

            newReplicas.addAll(newAssignment.replicas(tp));
        }
        assertTrue("Expected all new brokers to have some replicas", newReplicas.containsAll(newBrokers));

        Map<Broker, Collection<TopicPartition>> origBrokerLeaderCounts = origAssignment.brokerToLeaders();
        Map<Broker, Collection<TopicPartition>> newBrokerLeaderCounts = newAssignment.brokerToLeaders();
        Map<Broker, Collection<TopicPartition>> origBrokerReplicaCounts = origAssignment.brokerToReplicas();
        Map<Broker, Collection<TopicPartition>> newBrokerReplicaCounts = newAssignment.brokerToReplicas();
        int origBrokerLeaderCountSum = 0;
        int newBrokerLeaderCountSum = 0;
        int origBrokerReplicaCountSum = 0;
        int newBrokerReplicaCountSum = 0;

        for (Broker broker : origAssignment.brokers()) {
            origBrokerLeaderCountSum += origBrokerLeaderCounts.get(broker).size();
            newBrokerLeaderCountSum += newBrokerLeaderCounts.get(broker).size();
            origBrokerReplicaCountSum += origBrokerReplicaCounts.get(broker).size();
            newBrokerReplicaCountSum += newBrokerReplicaCounts.get(broker).size();

        }
        assertTrue("Expected leader counts on original brokers to reduce",
                newBrokerLeaderCountSum < origBrokerLeaderCountSum );

        assertTrue("Expected replica counts on original brokers to reduce",
                newBrokerReplicaCountSum < origBrokerReplicaCountSum);
    }

    /**
     * In this test, we rebalance a cluster with a high throttle, and then rebalance the same configuration with 1/2
     * the original throttle. We expect the second run to take longer than the first run.
     */
    @Test
    public void throttlingTest() throws IOException, InterruptedException {
        EmbeddedKafkaCluster cluster = new EmbeddedKafkaCluster(ORIG_NUM_BROKERS, metricsCluster);
        clusters.add(cluster);
        metricsCluster.ensureTopicCreated(EmbeddedKafkaCluster.METRICS_TOPIC,
                EmbeddedKafkaCluster.METRICS_TOPIC_PARTITIONS);
        createAndSeedTopics(3, cluster);
        final int newNumBrokers = 6;
        cluster.growTo(newNumBrokers);

        long firstRunTime = rebalanceCluster(cluster, metricsCluster, tmpFolder, 1000000, 60000);

        // Shutdown this cluster, otherwise we will have 2 zk and 20 kafka brokers on one host.
        cluster.shutDown();
        // remove this cluster from the list of clusters tracked by the test, otherwise shutdown will be called twice.
        clusters.remove(cluster);

        EmbeddedKafkaCluster secondCluster = new EmbeddedKafkaCluster(ORIG_NUM_BROKERS, metricsCluster);
        clusters.add(secondCluster);
        createAndSeedTopics(3, secondCluster);
        secondCluster.growTo(newNumBrokers);

        long secondRunTime = rebalanceCluster(secondCluster, metricsCluster, tmpFolder, 500000, 120000);
        secondCluster.shutDown();
        LOG.debug("Run time with throttle of 1000000: {}; Run time with throttle of 500000: {}", firstRunTime, secondRunTime);
        assertTrue("Expected second run to take more time than the first run, since it has half the throttle. " +
                "First run time: " + firstRunTime + "; second run time: " + secondRunTime,
                secondRunTime > firstRunTime);
    }

    /**
     * Test shrinking the cluster from 5 brokers to 3, and making sure that the data moves off the 3 clusters
     * we are targeting.
     */
    @Test
    public void shrinkClusterTest() throws IOException, InterruptedException {
        final int origNumBrokers = 5;
        EmbeddedKafkaCluster cluster = new EmbeddedKafkaCluster(origNumBrokers, metricsCluster);
        clusters.add(cluster);
        metricsCluster.ensureTopicCreated(EmbeddedKafkaCluster.METRICS_TOPIC,
                EmbeddedKafkaCluster.METRICS_TOPIC_PARTITIONS);
        createAndSeedTopics(3, cluster);
        ReplicaAssignment origAssignment = getPartitionAssignment(cluster);
        LOG.info("original brokers: {}", origAssignment.brokers().toString());
        String rebalancerConfig = createRebalancerConfig(metricsCluster, tmpFolder);
        assertTrue(cluster.removeBrokerWithId(4));
        ConfluentRebalancerCommand.run(new DefaultRebalancerFactory(), new DefaultLicenseValidator(new SystemTime()),
                "execute", "--zookeeper", cluster.zkConnectString(), "--metrics-bootstrap-server",
                metricsCluster.brokerList(), "--config-file", rebalancerConfig, "--force",
                "--throttle", "1000000", "--remove-broker-ids", "3,4", "--verbose");

        TestUtils.waitForCondition(new RebalanceFinishedCondition(cluster), 120000,
                "Rebalance did not finish in requisite amount of time.");

        ReplicaAssignment newAssignment = getPartitionAssignment(cluster);
        final int newNumBrokers = 3;

        assertEquals("Did not expect a change in the TopicPartitions in the cluster",
                origAssignment.topicPartitions(), newAssignment.topicPartitions());

        List<Integer> oldBrokers = new ArrayList<>();
        for (int i = newNumBrokers; i < origNumBrokers; ++i) {
            oldBrokers.add(i);
        }

        Set<Integer> origReplicas = new HashSet<>();
        // Validate that replicas have moved away from the old brokers.
        for (TopicPartition tp : newAssignment.topicPartitions()) {
            boolean disjoint =  Collections.disjoint(newAssignment.replicas(tp), oldBrokers);
            assertTrue("Expected no original broker to host replicas in the new assignment", disjoint);
            origReplicas.addAll(origAssignment.replicas(tp));
        }
        assertTrue("Expected all original brokers to have some replicas in the original assignment",
                origReplicas.containsAll(oldBrokers));
    }

    @Test(expected = ValidationException.class)
    public void missingBrokerTest() throws IOException, InterruptedException {
        EmbeddedKafkaCluster cluster = new EmbeddedKafkaCluster(ORIG_NUM_BROKERS + 1, metricsCluster);
        clusters.add(cluster);
        metricsCluster.ensureTopicCreated(EmbeddedKafkaCluster.METRICS_TOPIC,
                EmbeddedKafkaCluster.METRICS_TOPIC_PARTITIONS);
        createAndSeedTopics(3, cluster);
        cluster.removeBrokerWithId(ORIG_NUM_BROKERS);
        rebalanceCluster(cluster, metricsCluster, tmpFolder, 1000000, 20000);
    }

    private static long rebalanceCluster(EmbeddedKafkaCluster cluster, EmbeddedKafkaCluster metricsCluster,
                                         TemporaryFolder tmpFolder, long throttle,
                                         long timeout) throws IOException, InterruptedException {
        String configFile = createRebalancerConfig(metricsCluster, tmpFolder);
        long startTime = System.currentTimeMillis();
        ConfluentRebalancerCommand.run(new DefaultRebalancerFactory(), new DefaultLicenseValidator(new SystemTime()),
                "execute", "--zookeeper", cluster.zkConnectString(),
                "--metrics-bootstrap-server", metricsCluster.brokerList(), "--config-file",
                configFile, "--verbose", "--force", "--throttle", Long.toString(throttle));
        LOG.info("Ran rebalance command..");
        TestUtils.waitForCondition(new RebalanceFinishedCondition(cluster), timeout,
                "Rebalance did not complete on time. Aborting.");
        return System.currentTimeMillis() - startTime;
    }

    private static List<String> createAndSeedTopics(int numTopics, EmbeddedKafkaCluster cluster) {
        List<String> topics = new ArrayList<>();
        for (int i = 0; i < numTopics; ++i) {
            String topicName = "arandomtopic" + i;
            cluster.createTopic(topicName, 5, 2);
            topics.add(topicName);
            LOG.info("topic: {} created..", topicName);
        }

        KafkaProducer<String, String> producer = createProducer(ORIG_NUM_BROKERS, cluster);
        for (int i = 0; i < 100000; ++i) {
            for (String topic : topics) {
                producer.send(new ProducerRecord<>(topic, Integer.toString(i), Integer.toString(i)));
            }
        }
        producer.close();
        return topics;
    }

    private static class RebalanceFinishedCondition implements TestCondition {

        private final EmbeddedKafkaCluster cluster;

        RebalanceFinishedCondition(EmbeddedKafkaCluster cluster) {
            this.cluster = cluster;
        }

        @Override
        public boolean conditionMet() {
            try {
                LOG.info("Checking rebalance status..");
                ConfluentRebalancerCommand.run(new DefaultRebalancerFactory(), new DefaultLicenseValidator(new SystemTime()),
                        "status", "--zookeeper", cluster.zkConnectString());
            } catch (NoRebalanceInProgressException e) {
                    return true;
            }
            return false;
        }
    }

    private static ReplicaAssignment getPartitionAssignment(EmbeddedKafkaCluster cluster) {
        return ReplicaAssignment.fromScala(cluster.zkUtils.getReplicaAssignmentForTopics(
                cluster.zkUtils.getAllTopics()));
    }


    private static String createRebalancerConfig(EmbeddedKafkaCluster metricsCluster, TemporaryFolder tmpFolder) throws IOException {
        Properties rebalanceProps = new Properties();
        rebalanceProps.put(RebalancerConfig.METRICS_BOOTSTRAP_SERVERS_CONFIG, metricsCluster.brokerList());
        File configFile = tmpFolder.newFile();
        try (FileOutputStream outputStream = new FileOutputStream(configFile)) {
            rebalanceProps.store(outputStream, "Configs for the confluent rebalancer");
        }
        return configFile.getAbsolutePath();
    }

    private static KafkaProducer<String, String> createProducer(int numBrokers, EmbeddedKafkaCluster cluster) {
        Properties props = new Properties();
        props.put(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG, cluster.brokerList(0, numBrokers));
        props.put(ProducerConfig.ACKS_CONFIG, "all");
        return new KafkaProducer<>(props, new StringSerializer(), new StringSerializer());
    }
}
