// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.license;

import io.confluent.license.License;
import org.apache.kafka.common.utils.MockTime;
import org.apache.kafka.common.utils.SystemTime;
import org.jose4j.lang.JoseException;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.util.concurrent.TimeUnit;

public class DefaultLicenseValidatorTest {
    private KeyPair keyPair;

    @Before
    public void setUp() {
        keyPair = generateKeyPair();
    }

    @Test
    public void testValidLicense() throws LicenseValidator.ValidationFailedException {
        DefaultLicenseValidator validator = new DefaultLicenseValidator(new SystemTime());
        String licenseStr = generateLicense();
        validator.validateLicense(licenseStr, keyPair.getPublic());
    }

    @Test
    public void testInvalidLicense() {
        DefaultLicenseValidator validator = new DefaultLicenseValidator(new SystemTime());
        String licenseStr = "foobarbazbadlicense";
        try {
            validator.validateLicense(licenseStr, keyPair.getPublic());
        } catch (LicenseValidator.ValidationFailedException e) {
            Assert.assertTrue(e.getMessage().startsWith("Invalid License."));
        }
    }

    @Test
    public void testExpiredLicense() {
        MockTime time = new MockTime(System.currentTimeMillis());
        DefaultLicenseValidator validator = new DefaultLicenseValidator(time);
        // Generate a license which expires immediately.
        String licenseStr = generateLicense(System.currentTimeMillis());
        time.sleep(10);
        try {
            validator.validateLicense(licenseStr, keyPair.getPublic());
        } catch (LicenseValidator.ValidationFailedException e) {
            Assert.assertTrue(e.getMessage().startsWith("License expired on"));
        }
    }


    private String generateLicense() {
        return generateLicense(System.currentTimeMillis() + TimeUnit.DAYS.toMillis(365));
    }

    private String generateLicense(long expiryMs) {
        try {
            return License.sign(keyPair.getPrivate(), "confluent-rebalancer:LicenseValidatorTest", expiryMs, true);
        } catch (JoseException e) {
            throw new RuntimeException(e);
        }
    }

    private KeyPair generateKeyPair() {
        final KeyPairGenerator keyGen;
        try {
            keyGen = KeyPairGenerator.getInstance("RSA");
            keyGen.initialize(2048, SecureRandom.getInstance("SHA1PRNG"));
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException(e);
        }
        return keyGen.generateKeyPair();
    }
}
