// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.kafka.databalancing.report;

import io.confluent.kafka.databalancing.metric.Metrics;
import io.confluent.kafka.databalancing.topology.Broker;
import io.confluent.kafka.databalancing.topology.ReplicaAssignment;
import kafka.admin.BrokerMetadata;
import org.apache.kafka.common.TopicPartition;
import org.junit.Test;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import static io.confluent.kafka.databalancing.TestUtils.concat;
import static io.confluent.kafka.databalancing.TestUtils.createBrokers;
import static io.confluent.kafka.databalancing.TestUtils.createMap;
import static io.confluent.kafka.databalancing.TestUtils.entry;
import static java.util.Arrays.asList;
import static org.junit.Assert.*;

public class RebalanceReportTest {

    private final List<BrokerMetadata> brokers = concat(createBrokers(100, 103, "rack1"),
            createBrokers(103, 106, "rack2"));
    private final Map<TopicPartition, List<Integer>> partitions = createMap(
            entry(new TopicPartition("t1", 0), asList(100, 101)),
            entry(new TopicPartition("t2", 0), asList(103, 104)),
            entry(new TopicPartition("t2", 1), asList(104, 105)),
            entry(new TopicPartition("t3", 0), asList(105, 103)),
            entry(new TopicPartition("t3", 1), asList(100)),
            entry(new TopicPartition("t4", 0), asList(103)));
    private final Map<String, Integer> reps = createMap(entry("t1", 2), entry("t2", 2), entry("t3", 2), entry("t4", 2));
    private final Map<TopicPartition, Long> partitionToSize = createMap(
            entry(new TopicPartition("t1", 0), 162334L),
            entry(new TopicPartition("t2", 0), 1513244L),
            entry(new TopicPartition("t2", 1), 1054666L),
            entry(new TopicPartition("t3", 0), 2123325L),
            entry(new TopicPartition("t3", 1), 1240099L),
            entry(new TopicPartition("t4", 0), 2024454L));

    private final Metrics metrics = new Metrics(partitionToSize);
    private final ReplicaAssignment currentAssignment = new ReplicaAssignment(partitions);

    private final Map<TopicPartition, List<Integer>> proposedPartitions = createMap(
            entry(new TopicPartition("t1", 0), asList(100, 104)),
            entry(new TopicPartition("t2", 0), asList(104, 100)),
            entry(new TopicPartition("t2", 1), asList(102, 105)),
            entry(new TopicPartition("t3", 0), asList(105, 102)),
            entry(new TopicPartition("t3", 1), asList(101, 103)),
            entry(new TopicPartition("t4", 0), asList(103, 101)));

    private final Map<TopicPartition, List<Integer>> proposedPartitionsWithRemovedBrokers = createMap(
            entry(new TopicPartition("t1", 0), asList(103, 102)),
            entry(new TopicPartition("t2", 0), asList(105, 100)),
            entry(new TopicPartition("t2", 1), asList(102, 103)),
            entry(new TopicPartition("t3", 0), asList(102, 103)),
            entry(new TopicPartition("t3", 1), asList(100, 105)),
            entry(new TopicPartition("t4", 0), asList(105, 100)));

    private final String expectedOut = "You are about to move 7 replica(s) for 6 partitions to 5 broker(s) with total size 9.4 MB.\n" +
            "The preferred leader for 3 partition(s) will be changed.\n" +
            "In total, the assignment for 6 partitions will be changed.\n" +
            "\n" +
            "The following brokers will require more disk space during the rebalance and, in some cases, after the rebalance:\n" +
            "\tBroker     Current (MB)    During Rebalance (MB)  After Rebalance (MB)\n" +
            "\t100        1.4             2.9                    1.7            \n" +
            "\t101        0.2             3.4                    3.3            \n" +
            "\t102        0               3.2                    3.2            \n" +
            "\t103        5.7             6.9                    3.3            \n" +
            "\t104        2.6             2.7                    1.7            \n" +
            "\n" +
            "Min/max stats for brokers (before -> after):\n" +
            "\tType  Leader Count                 Replica Count                Size (MB)                          \n" +
            "\tMin   0 (id: 101) -> 1 (id: 100)   0 (id: 102) -> 2 (id: 100)   0 (id: 102) -> 1.7 (id: 100)       \n" +
            "\tMax   2 (id: 100) -> 1 (id: 100)   3 (id: 103) -> 2 (id: 100)   5.7 (id: 103) -> 3.3 (id: 101)     \n" +
            "\n" +
            "Rack stats (before -> after):\n" +
            "\tRack       Leader Count    Replica Count   Size (MB)           \n" +
            "\track1      2 -> 3          3 -> 6          1.6 -> 8.1          \n" +
            "\track2      4 -> 3          7 -> 6          11.4 -> 8.1         \n" +
            "\n";

    private final String expectedOutVerbose = expectedOut + "Broker stats (before -> after):\n" +
            "\tBroker     Leader Count    Replica Count   Size (MB)           \n" +
            "\t100        2 -> 1          2 -> 2          1.4 -> 1.7          \n" +
            "\t101        0 -> 1          1 -> 2          0.2 -> 3.3          \n" +
            "\t102        0 -> 1          0 -> 2          0 -> 3.2            \n" +
            "\t103        2 -> 1          3 -> 2          5.7 -> 3.3          \n" +
            "\t104        1 -> 1          2 -> 2          2.6 -> 1.7          \n" +
            "\t105        1 -> 1          2 -> 2          3.2 -> 3.2          \n" +
            "\n";

    private final String expectedOutWithRemovedBrokers = "You are about to move 10 replica(s) for 6 partitions to 4 broker(s) with total size 12.9 MB.\n" +
            "The preferred leader for 5 partition(s) will be changed.\n" +
            "In total, the assignment for 6 partitions will be changed.\n" +
            "\n" +
            "You have requested all replicas to be moved out of 2 broker(s) with id(s): 101,104.\n" +
            "After the rebalance, these broker(s) will have no replicas.\n" +
            "\n" +
            "The following brokers will require more disk space during the rebalance and, in some cases, after the rebalance:\n" +
            "\tBroker     Current (MB)    During Rebalance (MB)  After Rebalance (MB)\n" +
            "\t100        1.4             4.9                    4.8            \n" +
            "\t102        0               3.3                    3.3            \n" +
            "\t103        5.7             6.9                    3.3            \n" +
            "\t105        3.2             8                      4.8            \n" +
            "\n" +
            "Min/max stats for brokers (before -> after):\n" +
            "\tType  Leader Count                 Replica Count                Size (MB)                          \n" +
            "\tMin   0 (id: 101) -> 0 (id: 101)   0 (id: 102) -> 0 (id: 101)   0 (id: 102) -> 0 (id: 101)         \n" +
            "\tMax   2 (id: 100) -> 2 (id: 102)   3 (id: 103) -> 3 (id: 100)   5.7 (id: 103) -> 4.8 (id: 100)     \n" +
            "\n" +
            "Rack stats (before -> after):\n" +
            "\tRack       Leader Count    Replica Count   Size (MB)           \n" +
            "\track1      2 -> 3          3 -> 6          1.6 -> 8.1          \n" +
            "\track2      4 -> 3          7 -> 6          11.4 -> 8.1         \n" +
            "\n";

    private final String expectedOutVerboseWithRemovedBrokers = expectedOutWithRemovedBrokers + "Broker stats (before -> after):\n" +
            "\tBroker     Leader Count    Replica Count   Size (MB)           \n" +
            "\t100        2 -> 1          2 -> 3          1.4 -> 4.8          \n" +
            "\t101        0 -> 0          1 -> 0          0.2 -> 0            \n" +
            "\t102        0 -> 2          0 -> 3          0 -> 3.3            \n" +
            "\t103        2 -> 1          3 -> 3          5.7 -> 3.3          \n" +
            "\t104        1 -> 0          2 -> 0          2.6 -> 0            \n" +
            "\t105        1 -> 2          2 -> 3          3.2 -> 4.8          \n" +
            "\n";

    @Test
    public void testOut() {
        List<Broker> brokersToRemove = Collections.emptyList();
        ReplicaAssignment proposedAssignment = new ReplicaAssignment(proposedPartitions);
        RebalanceReport report = new RebalanceReport(brokers, currentAssignment, proposedAssignment, metrics,
                brokersToRemove);
        ByteArrayOutputStream outBytes = new ByteArrayOutputStream();
        report.print(new PrintStream(outBytes), false);
        String out = new String(outBytes.toByteArray());
        assertEquals(expectedOut, out);
    }

    @Test
    public void testOutVerbose() {
        List<Broker> brokersToRemove = Collections.emptyList();
        ReplicaAssignment proposedAssignment = new ReplicaAssignment(proposedPartitions);
        RebalanceReport report = new RebalanceReport(brokers, currentAssignment, proposedAssignment, metrics,
                brokersToRemove);
        ByteArrayOutputStream outBytes = new ByteArrayOutputStream();
        report.print(new PrintStream(outBytes), true);
        String out = new String(outBytes.toByteArray());
        assertEquals(expectedOutVerbose, out);
    }

    @Test
    public void testOutWithRemovedBrokers() {
        List<Broker> brokersToRemove = asList(new Broker(101), new Broker(104));
        ReplicaAssignment proposedAssignment = new ReplicaAssignment(proposedPartitionsWithRemovedBrokers);
        RebalanceReport report = new RebalanceReport(brokers, currentAssignment, proposedAssignment, metrics,
                brokersToRemove);
        ByteArrayOutputStream outBytes = new ByteArrayOutputStream();
        report.print(new PrintStream(outBytes), false);
        String out = new String(outBytes.toByteArray());
        assertEquals(expectedOutWithRemovedBrokers, out);
    }

    @Test
    public void testOutVerboseWithRemovedBrokers() {
        List<Broker> brokersToRemove = asList(new Broker(101), new Broker(104));
        ReplicaAssignment proposedAssignment = new ReplicaAssignment(proposedPartitionsWithRemovedBrokers);
        RebalanceReport report = new RebalanceReport(brokers, currentAssignment, proposedAssignment, metrics,
                brokersToRemove);
        ByteArrayOutputStream outBytes = new ByteArrayOutputStream();
        report.print(new PrintStream(outBytes), true);
        String out = new String(outBytes.toByteArray());
        assertEquals(expectedOutVerboseWithRemovedBrokers, out);
    }

}
