// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter;

import com.github.zafarkhaja.semver.Version;
import com.google.common.base.Strings;
import com.google.common.collect.ImmutableSet;
import com.google.inject.AbstractModule;
import com.google.inject.BindingAnnotation;
import com.google.inject.Inject;
import com.google.inject.Provides;
import com.google.inject.Singleton;
import io.confluent.controlcenter.streams.SingleGroupPartitionGrouper;
import io.confluent.controlcenter.util.DefaultZKUtilsFactory;
import io.confluent.controlcenter.util.ZKUtilsFactory;
import org.apache.commons.mail.Email;
import org.apache.commons.mail.EmailException;
import org.apache.commons.mail.SimpleEmail;
import org.apache.kafka.common.protocol.SecurityProtocol;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.Target;
import java.util.Properties;

import static java.lang.annotation.RetentionPolicy.RUNTIME;

public class ControlCenterConfigModule extends AbstractModule {

  private static final Logger log = LoggerFactory.getLogger(ControlCenterConfigModule.class);

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface EnableSubTopologies {
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface SubTopologyInputTopics {
  }

  private final ControlCenterConfig controlCenterConfig;
  private final boolean enableSubTopologies;
  private final ImmutableSet<String> subTopologyTopics;

  public ControlCenterConfigModule(ControlCenterConfig controlCenterConfig) {
    this(controlCenterConfig, false, null);
  }

  public ControlCenterConfigModule(ControlCenterConfig controlCenterConfig, boolean enableSubTopologies, ImmutableSet<String> subTopologyTopics) {
    this.controlCenterConfig = controlCenterConfig;
    this.enableSubTopologies = enableSubTopologies;
    this.subTopologyTopics = subTopologyTopics;
  }

  @Override
  protected void configure() {
    if (enableSubTopologies) {
      //We have to use static injection because this class is instantiated by Kafka Streams
      requestStaticInjection(SingleGroupPartitionGrouper.class);
    }
  }

  @Provides
  protected ControlCenterConfig provideControlCenterConfig() {
    return controlCenterConfig;
  }

  @Provides
  @EnableSubTopologies
  protected boolean areSubTopologiesEnabled() {
    return enableSubTopologies;
  }

  @Provides
  @SubTopologyInputTopics
  protected ImmutableSet<String> provideSubTopologyTopics() {
    return subTopologyTopics;
  }

  @Provides
  @Inject
  protected SecurityProtocol provideSecurityProtocol(ControlCenterConfig config) {
    return SecurityProtocol.forName(config.getString(ControlCenterConfig.CONTROL_CENTER_CONSUMER_SECURITY_PROTOCOL));
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface InternalProperties {
  }

  @Provides
  @InternalProperties
  protected Properties provideInternalProperties() {
    Properties out = new Properties();
    try {
      out.load(getClass().getClassLoader().getResourceAsStream("controlcenter.internal.properties"));
    } catch (Exception e) {
      log.warn("unable to load internal properties file");
    }
    return out;
  }

  @Provides
  @Inject
  @Singleton
  protected Version provideControlCenterVersion(@InternalProperties Properties properties) throws IOException {
    return Version.valueOf(properties.getProperty("version"));
  }

  @Provides
  @Inject
  protected ZKUtilsFactory getZKUtilsFactory() {
    return new DefaultZKUtilsFactory();
  }

  @Provides
  @Inject
  protected Email provideBaseEmail(ControlCenterConfig config) throws EmailException {
    Email email = new SimpleEmail();
    email.setHostName(config.getString(ControlCenterConfig.CONTROL_CENTER_EMAIL_HOST_NAME));
    email.setSmtpPort(config.getInt(ControlCenterConfig.CONTROL_CENTER_EMAIL_PORT));
    email.setSSLOnConnect(config.getBoolean(ControlCenterConfig.CONTROL_CENTER_EMAIL_SSL_CHECKSERVERIDENTITY));
    email.setStartTLSRequired(config.getBoolean(ControlCenterConfig.CONTROL_CENTER_EMAIL_STARTTLS_REQUIRED));
    email.setFrom(config.getString(ControlCenterConfig.CONTROL_CENTER_EMAIL_FROM));
    String bounceAddress = config.getString(ControlCenterConfig.CONTROL_CENTER_EMAIL_BOUNCE_ADDRESS);
    if (!Strings.isNullOrEmpty(bounceAddress)) {
      email.setBounceAddress(bounceAddress);
    }

    String username = config.getString(ControlCenterConfig.CONTROL_CENTER_EMAIL_USERNAME);
    if (!Strings.isNullOrEmpty(username)) {
      email.setAuthentication(username, config.getString(ControlCenterConfig.CONTROL_CENTER_EMAIL_PASSWORD));
    }
    return email;
  }
}
