package io.confluent.controlcenter;

import com.google.common.base.Predicate;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Iterables;
import com.google.common.collect.Sets;
import com.google.inject.Inject;

import org.apache.kafka.common.protocol.Errors;
import org.apache.kafka.common.protocol.SecurityProtocol;
import org.apache.kafka.common.requests.MetadataResponse.PartitionMetadata;
import org.apache.kafka.common.requests.MetadataResponse.TopicMetadata;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Properties;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.TimeUnit;

import io.confluent.controlcenter.streams.TopicStoreMaster;
import io.confluent.controlcenter.util.ZKUtilsFactory;
import kafka.admin.AdminUtils;
import kafka.admin.RackAwareMode;
import kafka.admin.TopicCommand;
import kafka.log.LogConfig$;
import kafka.server.KafkaConfig$;
import kafka.utils.ZkUtils;

public class KafkaHelper {

  private static final String COMPACT = "compact";
  private static final String COMPACT_DELETE = "compact,delete";

  public static class KafkaAdminUtils {
    public Set<TopicMetadata> fetchTopicMetadataFromZk(Set<String> topics, ZkUtils zkUtils, SecurityProtocol protocol) {
      return scala.collection.JavaConversions.setAsJavaSet(
          AdminUtils.fetchTopicMetadataFromZk(scala.collection.JavaConversions.asScalaSet(topics), zkUtils, protocol)
      );
    }

    public TopicMetadata fetchTopicMetadataFromZk(String topic, ZkUtils zkUtils, SecurityProtocol protocol) {
      return fetchTopicMetadataFromZk(Sets.<String>newHashSet(topic), zkUtils, protocol).iterator().next();
    }


    public void createTopic(ZkUtils zkUtils, String topic, int partitions, int replication, Properties props, RackAwareMode.Enforced$ module$) {
      AdminUtils.createTopic(zkUtils, topic, partitions, replication, props, module$);
    }
  }

  private static final Logger log = LoggerFactory.getLogger(KafkaHelper.class);
  private final ZkUtils zkUtils;
  private final KafkaAdminUtils kafkaAdminUtils;
  private final SecurityProtocol protocol;

  @Inject
  public KafkaHelper(
      SecurityProtocol protocol,
      ControlCenterConfig controlCenterConfig,
      ZKUtilsFactory zkUtilsFactory,
      KafkaAdminUtils kafkaAdminUtils
  ) {
    log.warn("kafka securityProtocol={}", protocol);
    this.protocol = protocol;
    this.zkUtils = zkUtilsFactory.create(controlCenterConfig.getString(ControlCenterConfig.ZOOKEEPER_CONNECT_CONFIG));
    this.kafkaAdminUtils = kafkaAdminUtils;
  }

  protected static int minIsr(int replication) {
    return Math.min(3, (replication < 3 ? 1 : replication - 1));
  }

  public boolean checkPartitions(TopicMetadata topicMetadata, int partitions) {
    return topicMetadata.partitionMetadata().size() == partitions;
  }

  public boolean checkReplicas(PartitionMetadata partitionMetadata, int replicas) {
    return partitionMetadata.replicas().size() == replicas;
  }

  public boolean checkMinIsr(PartitionMetadata partitionMetadata, int minIsr) {
    return partitionMetadata.isr().size() >= minIsr;
  }

  private Properties getProperties(
      final long retentionMs,
      final long retentionBytes,
      final int minIsr,
      final long changeLogSegmentBytes,
      final String topic,
      final String cleanupPolicy
  ) {
    Properties props = new Properties();
    props.setProperty(KafkaConfig$.MODULE$.MinInSyncReplicasProp(), "" + minIsr);
    if (cleanupPolicy != null) {
      long topicRetentionMs = retentionMs;
      if (topic.contains(Rollup.FIFTEEN_SECONDS.name())) {
        topicRetentionMs = Rollup.FIFTEEN_SECONDS.getRetainMillis() + TimeUnit.DAYS.toMillis(1);
      } else if (topic.contains(Rollup.ONE_HOUR.name())) {
        topicRetentionMs = Rollup.ONE_HOUR.getRetainMillis() + TimeUnit.DAYS.toMillis(1);
      } else if (topic.contains(Rollup.ONE_WEEK.name())) {
        topicRetentionMs = Rollup.ONE_WEEK.getRetainMillis() + TimeUnit.DAYS.toMillis(1);
      } else {
        topicRetentionMs += TimeUnit.DAYS.toMillis(1);
      }
      props.setProperty(LogConfig$.MODULE$.CleanupPolicyProp(), cleanupPolicy);
      props.setProperty(LogConfig$.MODULE$.SegmentBytesProp(), "" + changeLogSegmentBytes);
      props.setProperty(LogConfig$.MODULE$.DeleteRetentionMsProp(), "" + topicRetentionMs);
    } else {
      props.setProperty(LogConfig$.MODULE$.MessageTimestampTypeProp(), "LogAppendTime");
      props.setProperty(LogConfig$.MODULE$.RetentionMsProp(), "" + retentionMs);
      props.setProperty(LogConfig$.MODULE$.RetentionBytesProp(), "" + retentionBytes);
    }
    return props;
  }

  public boolean checkCreateTopics(Set<String> topics, int partitions, int replication, long retentionMs, long retentionBytes, long changeLogSegmentBytes) {
    return checkCreateTopics(topics, partitions, replication, retentionMs, retentionBytes, changeLogSegmentBytes, null);
  }

  private boolean checkCreateTopics(Set<String> topics, int partitions, int replication, long retentionMs, long retentionBytes, long changeLogSegmentBytes, String cleanupPolicy) {
    Set<TopicMetadata> k = kafkaAdminUtils.fetchTopicMetadataFromZk(topics, zkUtils, protocol);

    if (k.size() != topics.size()) {
      log.error("expected count={} topics actual={}", topics.size(), k.size());
      return false;
    }

    Set<String> invalidTopics = Sets.newHashSet();
    int minIsr = minIsr(replication);
    boolean topicInError = false;
    for (TopicMetadata topicMetadata : k) {
      if (Errors.UNKNOWN_TOPIC_OR_PARTITION.equals(topicMetadata.error())) {
        invalidTopics.add(topicMetadata.topic());
        continue;
      } else if (topicMetadata.error() != null && !Errors.NONE.equals(topicMetadata.error())) {
        log.error("topic={} in error={}", topicMetadata.topic(), topicMetadata.error());
        topicInError = true;
        continue;
      }
      if (!checkPartitions(topicMetadata, partitions)) {
        log.error("found topic={} with partitions={} instead of expectedPartitions={}", topicMetadata.topic(), topicMetadata.partitionMetadata().size(), partitions);
        topicInError = true;
      }
      for (PartitionMetadata partitionMetadata : topicMetadata.partitionMetadata()) {
        if (partitionMetadata.error() != null && !Errors.NONE.equals(partitionMetadata.error())) {
          log.error("topic={} partition={} in error={}", topicMetadata.topic(), partitionMetadata.partition(), partitionMetadata.error());
          topicInError = true;
          continue;
        }
        if (!checkReplicas(partitionMetadata, replication)) {
          log.error("found topic={} with replication={} instead of expectedReplication={}", topicMetadata.topic(), partitionMetadata.replicas().size(), replication);
          topicInError = true;
        }
        if (!checkMinIsr(partitionMetadata, minIsr)) {
          log.error("found topic={} with isr={} below minIsr={}", topicMetadata.topic(), partitionMetadata.isr().size(), minIsr);
          topicInError = true;
        }
      }
    }
    if (topicInError) {
      return false;
    }
    for (String topic : invalidTopics) {
      try {
        String topicCleanupPolicy = cleanupPolicy;
        if (topicCleanupPolicy == null && topic.endsWith("-changelog")) {
          topicCleanupPolicy = COMPACT;
        }
        Properties props = getProperties(retentionMs, retentionBytes, minIsr, changeLogSegmentBytes, topic, topicCleanupPolicy);
        kafkaAdminUtils.createTopic(zkUtils, topic, partitions, replication, props, RackAwareMode.Enforced$.MODULE$);
        TopicMetadata topicMetadata = kafkaAdminUtils.fetchTopicMetadataFromZk(topic, zkUtils, protocol);
        assert checkPartitions(topicMetadata, partitions);
        assert topic.equals(topicMetadata.topic());
        assert (topicMetadata.error() == null || Errors.NONE.equals(topicMetadata.error()));
        log.info("created topic={} partitions={} replication={} minIsr={} retention={}", topic, partitions, replication, minIsr, retentionMs);
      } catch (Throwable e) {
        topicInError = true;
        log.error("failed to create topic={} partitions={} replication={} minIsr={} retention={}", topic, partitions, replication, minIsr, retentionMs, e);
      }
    }
    return !topicInError;
  }

  public boolean deleteTopic(final String topic) {
    final TopicCommand.TopicCommandOptions commandOptions = new TopicCommand.TopicCommandOptions(new String[]{
        "--zookeeper", zkUtils.toString(),
        "--delete", "--topic", topic});
    try {
      TopicCommand.deleteTopic(zkUtils, commandOptions);
    } catch (final IllegalArgumentException e) {
      //topic doesn't exist
      return false;
    }
    catch (final RuntimeException e) {
      log.error("ERROR: Deleting topic {} failed.", topic, e);
      return false;
    }
    return true;
  }

  public static class ControlCenterPreconditions implements Callable<Boolean> {

    private final ControlCenterConfig config;
    private final KafkaHelper kafkaHelper;
    private final TopicStoreMaster topicStoreMaster;

    @Inject
    public ControlCenterPreconditions(
        ControlCenterConfig config,
        KafkaHelper kafkaHelper,
        TopicStoreMaster topicStoreMaster
    ) {
      this.config = config;
      this.kafkaHelper = kafkaHelper;
      this.topicStoreMaster = topicStoreMaster;
    }

    @Override
    public Boolean call() throws Exception {
      Set<String> internalTopics = topicStoreMaster.getInternalTopicNames();
      boolean createdTopics = kafkaHelper.checkCreateTopics(
          ImmutableSet.of(config.getString(ControlCenterConfig.CONTROL_CENTER_COMMAND_TOPIC)),
          1,
          config.getInt(ControlCenterConfig.CONTROL_CENTER_COMMAND_TOPIC_REPLICATION_CONFIG),
          config.getLong(ControlCenterConfig.CONTROL_CENTER_COMMAND_TOPIC_RETENTION_MS_CONFIG),
          config.getLong(ControlCenterConfig.CONTROL_CENTER_COMMAND_TOPIC_RETENTION_BYTES_CONFIG),
          config.getLong(ControlCenterConfig.CONTROL_CENTER_INTERNAL_CHANGELOG_SEGMENT_BYTES_CONFIG),
          COMPACT
      );

      createdTopics &= kafkaHelper.checkCreateTopics(
          topicStoreMaster.getStoreTopicNames(),
          config.getInt(ControlCenterConfig.CONTROL_CENTER_INTERNAL_PARTITIONS_CONFIG),
          config.getInt(ControlCenterConfig.CONTROL_CENTER_INTERNAL_REPLICATION_CONFIG),
          config.getLong(ControlCenterConfig.CONTROL_CENTER_INTERNAL_RETENTION_MS_CONFIG),
          config.getLong(ControlCenterConfig.CONTROL_CENTER_INTERNAL_RETENTION_BYTES_CONFIG),
          config.getLong(ControlCenterConfig.CONTROL_CENTER_INTERNAL_CHANGELOG_SEGMENT_BYTES_CONFIG),
          COMPACT
      );

      createdTopics &= kafkaHelper.checkCreateTopics(
          topicStoreMaster.getWindowedStoreTopicNames(),
          config.getInt(ControlCenterConfig.CONTROL_CENTER_INTERNAL_PARTITIONS_CONFIG),
          config.getInt(ControlCenterConfig.CONTROL_CENTER_INTERNAL_REPLICATION_CONFIG),
          config.getLong(ControlCenterConfig.CONTROL_CENTER_INTERNAL_RETENTION_MS_CONFIG),
          config.getLong(ControlCenterConfig.CONTROL_CENTER_INTERNAL_RETENTION_BYTES_CONFIG),
          config.getLong(ControlCenterConfig.CONTROL_CENTER_INTERNAL_CHANGELOG_SEGMENT_BYTES_CONFIG),
          COMPACT_DELETE
      );

      createdTopics &= kafkaHelper.checkCreateTopics(
          topicStoreMaster.getIntermediateTopicNames(),
          config.getInt(ControlCenterConfig.CONTROL_CENTER_INTERNAL_PARTITIONS_CONFIG),
          config.getInt(ControlCenterConfig.CONTROL_CENTER_INTERNAL_REPLICATION_CONFIG),
          config.getLong(ControlCenterConfig.CONTROL_CENTER_INTERNAL_RETENTION_MS_CONFIG),
          config.getLong(ControlCenterConfig.CONTROL_CENTER_INTERNAL_RETENTION_BYTES_CONFIG),
          config.getLong(ControlCenterConfig.CONTROL_CENTER_INTERNAL_CHANGELOG_SEGMENT_BYTES_CONFIG)
      );


      createdTopics &= kafkaHelper.checkCreateTopics(
          ImmutableSet.of(config.getString(ControlCenterConfig.CONTROL_CENTER_TOPIC_CONFIG)),
          config.getInt(ControlCenterConfig.CONTROL_CENTER_MONITORING_TOPIC_PARTITIONS_CONFIG),
          config.getInt(ControlCenterConfig.CONTROL_CENTER_MONITORING_TOPIC_REPLICATION_CONFIG),
          config.getLong(ControlCenterConfig.CONTROL_CENTER_MONITORING_TOPIC_RETENTION_MS_CONFIG),
          config.getLong(ControlCenterConfig.CONTROL_CENTER_MONITORING_TOPIC_RETENTION_BYTES_CONFIG),
          config.getLong(ControlCenterConfig.CONTROL_CENTER_INTERNAL_CHANGELOG_SEGMENT_BYTES_CONFIG)
      );

      return createdTopics;
    }
  }
}
