package io.confluent.controlcenter.command;

import com.google.common.io.Files;
import com.google.inject.AbstractModule;
import com.google.inject.BindingAnnotation;
import com.google.inject.Inject;
import com.google.inject.Provides;
import com.google.inject.Singleton;

import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.streams.StreamsConfig;
import org.apache.kafka.streams.state.ReadOnlyKeyValueStore;
import org.slf4j.Logger;

import java.io.IOException;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.Target;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Properties;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import io.confluent.command.CommandStore;
import io.confluent.command.record.Command.CommandConfigType;
import io.confluent.command.record.Command.CommandKey;
import io.confluent.command.record.Command.CommandMessage;
import io.confluent.controlcenter.ControlCenterConfig;
import io.confluent.controlcenter.ControlCenterConfigModule;
import io.confluent.controlcenter.streams.SingleGroupPartitionGrouper;
import io.confluent.controlcenter.streams.TopicStoreMaster;

import static java.lang.annotation.RetentionPolicy.RUNTIME;

public class CommandModule extends AbstractModule {

  private static final Logger log = org.slf4j.LoggerFactory.getLogger(CommandModule.class);

  @Override
  protected void configure() {}


  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface CommandStateDir {

  }

  @Provides
  @Inject
  @CommandStateDir
  protected Path getCommandStateDir(ControlCenterConfig controlCenterConfig) throws IOException {
    Path commandStateDir = Paths.get(
        controlCenterConfig.getString(ControlCenterConfig.CONTROL_CENTER_DATA_DIR),
        controlCenterConfig.getString(ControlCenterConfig.CONTROL_CENTER_ID_CONFIG),
        "cp-command"
    );
    Files.createParentDirs(commandStateDir.toFile());
    return commandStateDir;
  }

  @Provides
  @Singleton
  protected CommandStore provideCommandStore(
      ControlCenterConfig controlCenterConfig,
      @ControlCenterConfigModule.EnableSubTopologies boolean enableSubTopologies,
      @CommandStateDir Path commandStateDir
  ) throws TimeoutException, InterruptedException {
    Properties props = new Properties();
    props.put(
        StreamsConfig.BOOTSTRAP_SERVERS_CONFIG,
        controlCenterConfig.getList(ControlCenterConfig.CONTROL_CENTER_BOOTSTRAP_SERVERS_CONFIG)
    );
    props.put(
        StreamsConfig.ZOOKEEPER_CONNECT_CONFIG,
        controlCenterConfig.getString(ControlCenterConfig.ZOOKEEPER_CONNECT_CONFIG)
    );
    props.put(
        StreamsConfig.STATE_DIR_CONFIG,
        commandStateDir.toString()
    );
    String streamsAppId = TopicStoreMaster.nameJoin(
        controlCenterConfig.getString(ControlCenterConfig.CONTROL_CENTER_NAME_CONFIG),
        controlCenterConfig.getString(ControlCenterConfig.CONTROL_CENTER_ID_CONFIG),
        "command"
    );
    props.put(StreamsConfig.APPLICATION_ID_CONFIG, streamsAppId);
    props.putAll(controlCenterConfig.valuesWithPrefix(ControlCenterConfig.CONFLUENT_CONTROLCENTER_KSTREAMS_PREFIX));
    props.putAll(controlCenterConfig.originalsWithPrefix(ControlCenterConfig.CONFLUENT_CONTROLCENTER_KSTREAMS_PREFIX));
    props.put(StreamsConfig.consumerPrefix(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG), "earliest");
    props.put(StreamsConfig.producerPrefix(ProducerConfig.ACKS_CONFIG), "all");
    props.put(StreamsConfig.NUM_STREAM_THREADS_CONFIG, "1");
    if (enableSubTopologies) {
      props.put(
          StreamsConfig.PARTITION_GROUPER_CLASS_CONFIG,
          SingleGroupPartitionGrouper.class
      );
    }

    StreamsConfig streamsConfig = new StreamsConfig(props);
    CommandStore commander = new CommandStore.Builder()
        .topic(controlCenterConfig.getString(ControlCenterConfig.CONTROL_CENTER_COMMAND_TOPIC))
        .store(CommandConfigType.MONITORING_TRIGGER_CONFIG)
        .store(CommandConfigType.ACTION_CONFIG)
        .store(CommandConfigType.LICENSE_INFO)
        .store(CommandConfigType.CLUSTER_METADATA)
        .streamsConfig(streamsConfig)
        .producerConfig(streamsConfig.getProducerConfigs("c3-command"))
        .build();
    return commander;
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface TriggerConfigs {}

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface ActionConfigs {}

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface ClusterMetadataStore {}

  @Provides
  @Inject
  @Singleton
  @TriggerConfigs
  protected ReadOnlyKeyValueStore<CommandKey, CommandMessage> provideTriggerConfigs(
      CommandStore commandStore
  ) {
    return commandStore.type(CommandConfigType.MONITORING_TRIGGER_CONFIG);
  }

  @Provides
  @Inject
  @Singleton
  @ActionConfigs
  protected ReadOnlyKeyValueStore<CommandKey, CommandMessage> provideActionConfigs(
      CommandStore commandStore
  ) {
    return commandStore.type(CommandConfigType.ACTION_CONFIG);
  }

  @Provides
  @Inject
  @Singleton
  @ClusterMetadataStore
  protected ReadOnlyKeyValueStore<CommandKey, CommandMessage> provideClusterMetadata(
      CommandStore commandStore
  ) {
    return commandStore.type(CommandConfigType.CLUSTER_METADATA);
  }
}
