// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.data;

import com.google.inject.Inject;
import com.google.inject.Provider;
import io.confluent.command.CommandStore;
import io.confluent.command.cluster_metadata.CommandClusterMetadata.ClusterMetadata;
import io.confluent.command.cluster_metadata.CommandClusterMetadata.KafkaClusterMetadata;
import io.confluent.command.cluster_metadata.CommandClusterMetadata.KafkaConnectClusterMetadata;
import io.confluent.command.record.Command.CommandConfigType;
import io.confluent.command.record.Command.CommandKey;
import io.confluent.command.record.Command.CommandMessage;
import io.confluent.controlcenter.ControlCenterConfig;
import io.confluent.controlcenter.command.CommandModule.ClusterMetadataStore;
import io.confluent.controlcenter.util.KafkaClusterMetadataClientFactory;
import org.apache.kafka.streams.state.ReadOnlyKeyValueStore;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.Callable;


public class ClusterMetadataInitializer implements Callable<Void> {
  private static final Logger log = LoggerFactory.getLogger(new Object() {}.getClass().getEnclosingClass());
  private final KafkaClusterMetadataClientFactory kafkaClusterMetadataClientFactory;
  private final ControlCenterConfig controlCenterConfig;
  private final Provider<CommandStore> commandStore;
  private final Provider<ReadOnlyKeyValueStore<CommandKey, CommandMessage>> clusterMetadataStore;

  @Inject
  public ClusterMetadataInitializer(
      ControlCenterConfig controlCenterConfig,
      KafkaClusterMetadataClientFactory kafkaClusterMetadataClientFactory,
      Provider<CommandStore> commandStore,
      @ClusterMetadataStore Provider<ReadOnlyKeyValueStore<CommandKey, CommandMessage>> clusterMetadataStore
  ) {
    this.controlCenterConfig = controlCenterConfig;
    this.kafkaClusterMetadataClientFactory = kafkaClusterMetadataClientFactory;
    this.commandStore = commandStore;
    this.clusterMetadataStore = clusterMetadataStore;
  }

  @Override
  public Void call() throws Exception {
    String c3ClusterId = updateC3KafkaClusterInfo();
    updateConnectClusterInfo(c3ClusterId);
    return null;
  }

  private void updateConnectClusterInfo(final String c3ClusterId) {
    final String connectZkConnect = controlCenterConfig.getConnectZookeeperConnect();
    final String realConnectKafkaClusterId = getClusterId(connectZkConnect);
    final String connectKafkaClusterId = realConnectKafkaClusterId == null? "kafka-connect" : realConnectKafkaClusterId;
    if (realConnectKafkaClusterId == null || c3ClusterId == null || !realConnectKafkaClusterId.equals(c3ClusterId)) {
      //Only upsert a Kafka entry for Connect if it's different from C3s
      updateKafkaClusterInfo(connectKafkaClusterId, connectZkConnect, null);
    }
    updateConnectClusterInfo(connectKafkaClusterId, controlCenterConfig.getList(ControlCenterConfig.CONTROL_CENTER_CONNECT_CLUSTER));
  }

  private void updateConnectClusterInfo(String connectKafkaClusterId, List<String> restUrls) {
    final CommandKey key = CommandKey.newBuilder()
        .setConfigType(CommandConfigType.CLUSTER_METADATA)
        .setGuid("connect-default")
        .build();
    final CommandMessage existingMsg = clusterMetadataStore.get().get(key);
    final CommandMessage.Builder upsertBldr = getInitialConnectClusterMetadata(connectKafkaClusterId, restUrls);
    if (existingMsg != null) {
      upsertBldr.getClusterMetadataBuilder().setDisplayName(existingMsg.getClusterMetadata().getDisplayName());
    }
    commandStore.get().sendCommand(key, upsertBldr.build());
  }

  private CommandMessage.Builder getInitialConnectClusterMetadata(String kafkaClusterId, List<String> restUrls) {
    return CommandMessage.newBuilder()
        .setClusterMetadata(
            ClusterMetadata.newBuilder()
                .setDisplayName("Kafka Connect")
                .setConnect(KafkaConnectClusterMetadata.newBuilder()
                    .setKafkaClusterId(kafkaClusterId)
                    .addAllRestUrls(restUrls)
                )
        );
  }

  private String updateC3KafkaClusterInfo() {
    final String c3zkConnect = controlCenterConfig.getString(ControlCenterConfig.ZOOKEEPER_CONNECT_CONFIG);
    final String c3ClusterId = getClusterId(c3zkConnect);
    if (c3ClusterId == null) {
      return null;
    }
    updateKafkaClusterInfo(c3ClusterId, c3zkConnect, controlCenterConfig.getList(ControlCenterConfig.CONTROL_CENTER_BOOTSTRAP_SERVERS_CONFIG));
    return c3ClusterId;
  }

  private void updateKafkaClusterInfo(final String clusterId, final String zkConnect, final List<String> bootstrapSvrs) {
    final CommandKey key = CommandKey.newBuilder()
        .setConfigType(CommandConfigType.CLUSTER_METADATA)
        .setGuid(clusterId)
        .build();
    final CommandMessage existingMsg = clusterMetadataStore.get().get(key);
    final CommandMessage.Builder upsertBldr = getInitialKafkaClusterMetadata(zkConnect, bootstrapSvrs);
    if (existingMsg != null) {
      upsertBldr.getClusterMetadataBuilder().setDisplayName(existingMsg.getClusterMetadata().getDisplayName());
    }
    commandStore.get().sendCommand(key, upsertBldr.build());
  }

  private CommandMessage.Builder getInitialKafkaClusterMetadata(final String c3zkConnect, List<String> bootstrapSvrs) {
    if (bootstrapSvrs == null) {
      bootstrapSvrs = new ArrayList<>();
    }
    return CommandMessage.newBuilder()
        .setClusterMetadata(
            ClusterMetadata.newBuilder()
                .setKafka(
                    KafkaClusterMetadata.newBuilder()
                        .addAllZookeeperConnect(Arrays.asList(c3zkConnect.split(",")))
                        .addAllBootstrapServers(bootstrapSvrs)
                )
        );
  }

  private String getClusterId(String zookeeperConnect) {
    try {
      return kafkaClusterMetadataClientFactory.create(zookeeperConnect).queryClusterId();
    } catch (Throwable t) {
      log.error("Error trying to get cluster id for cluster {}", zookeeperConnect, t);
    }
    return null;
  }

}
