// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.rest;

import com.google.common.base.Strings;
import com.google.common.collect.Lists;
import com.google.common.collect.Ordering;
import com.google.inject.Inject;
import com.google.inject.Provider;

import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.state.KeyValueIterator;
import org.apache.kafka.streams.state.ReadOnlyKeyValueStore;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.UUID;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeoutException;

import javax.ws.rs.ClientErrorException;
import javax.ws.rs.DELETE;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.NotFoundException;
import javax.ws.rs.POST;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;

import io.confluent.command.CommandStore;
import io.confluent.command.record.Command.CommandConfigType;
import io.confluent.command.record.Command.CommandKey;
import io.confluent.command.record.Command.CommandMessage;
import io.confluent.command.record.alert.CommandAlert.ActionConfig;
import io.confluent.command.record.alert.CommandAlert.MonitoringTriggerConfig;
import io.confluent.controlcenter.alert.record.Alert.AlertInfo;
import io.confluent.controlcenter.command.CommandModule.ActionConfigs;
import io.confluent.controlcenter.command.CommandModule.TriggerConfigs;
import io.confluent.controlcenter.data.ClusterMetadataDao;
import io.confluent.controlcenter.serialization.SerializationModule.TriggerEventsKeyPrefixSerde;
import io.confluent.controlcenter.streams.TopicStoreModule.AlertHistoryStore;
import io.confluent.controlcenter.util.KVQuery;
import io.confluent.serializers.OrderedKeyUberSerde;

@Path("/2.0/alerts")
@Produces(MediaType.APPLICATION_JSON)
public class AlertsResource {

  private static final Logger log = LoggerFactory.getLogger(AlertsResource.class);
  private static final Ordering<AlertInfo> ALERT_INFO_ORDERING = new Ordering<AlertInfo>() {
    @Override
    public int compare(AlertInfo left, AlertInfo right) {
      return Ordering.natural()
          .reverse()
          .compare(
              left.getTimestamp(),
              right.getTimestamp()
          );
    }
  };

  private final Provider<ReadOnlyKeyValueStore<CommandKey, CommandMessage>> monitoringTriggerConfigs;
  private final Provider<ReadOnlyKeyValueStore<CommandKey, CommandMessage>> actionConfigs;
  private final CommandStore commandStore;
  private final Provider<ReadOnlyKeyValueStore<Bytes, AlertInfo>> alertHistoryStore;
  private final OrderedKeyUberSerde<AlertInfo> alertRangeSerde;

  @Inject
  public AlertsResource(
      CommandStore commandStore,
      @TriggerConfigs Provider<ReadOnlyKeyValueStore<CommandKey, CommandMessage>> monitoringTriggerConfigs,
      @ActionConfigs Provider<ReadOnlyKeyValueStore<CommandKey, CommandMessage>> actionConfigs,
      @AlertHistoryStore Provider<ReadOnlyKeyValueStore<Bytes, AlertInfo>> alertHistoryStore,
      @TriggerEventsKeyPrefixSerde OrderedKeyUberSerde<AlertInfo> alertRangeSerde
  ) {
    this.commandStore = commandStore;
    this.monitoringTriggerConfigs = monitoringTriggerConfigs;
    this.actionConfigs = actionConfigs;
    this.alertHistoryStore = alertHistoryStore;
    this.alertRangeSerde = alertRangeSerde;
  }

  @GET
  @Path("/triggers/{guid}")
  public MonitoringTriggerConfig getTrigger(
      @PathParam("guid") String guid
  ) {
    CommandKey key = getCommandKey(guid, CommandConfigType.MONITORING_TRIGGER_CONFIG);
    CommandMessage baseMessage = monitoringTriggerConfigs.get().get(key);
    if (baseMessage == null) {
      throw new NotFoundException("unable to find config with guid=" + guid);
    }
    return externalizeClusterId(baseMessage.getMonitoringTriggerConfig());
  }

  @GET
  @Path("/triggers")
  public List<MonitoringTriggerConfig> getTriggers() {
    List<MonitoringTriggerConfig> out = Lists.newArrayList();
    try (KeyValueIterator<CommandKey, CommandMessage> i = monitoringTriggerConfigs.get().all()) {
      while (i.hasNext()) {
        out.add(
            externalizeClusterId(
                i.next().value.getMonitoringTriggerConfig()
            )
        );
      }
    }
    return out;
  }

  @POST
  @Path("/triggers")
  public MonitoringTriggerConfig createTrigger(
      MonitoringTriggerConfig config
  ) throws InterruptedException, ExecutionException, TimeoutException {
    if (!Strings.isNullOrEmpty(config.getGuid())) {
      return updateTrigger(config.getGuid(), config);
    }
    verifyMonitoringTriggerConfig(config);
    config = internalizeClusterIdBuilder(config)
        .setGuid(UUID.randomUUID().toString())
        .build();
    log.trace("config={}", config);
    commandStore.awaitCommand(
        getCommandKey(config.getGuid(), CommandConfigType.MONITORING_TRIGGER_CONFIG),
        CommandMessage.newBuilder()
            .setMonitoringTriggerConfig(config)
            .build()
    );
    return externalizeClusterId(config);
  }

  @PUT
  @Path("/triggers/{guid}")
  public MonitoringTriggerConfig updateTrigger(
      @PathParam("guid") String guid,
      MonitoringTriggerConfig config
  ) throws InterruptedException, ExecutionException, TimeoutException {
    if (!guid.equals(config.getGuid())) {
      throw new ClientErrorException("config guid does not match rest guid", Response.Status.CONFLICT);
    }
    verifyMonitoringTriggerConfig(config);
    config = internalizeClusterIdBuilder(config).build();
    CommandKey key = getCommandKey(guid, CommandConfigType.MONITORING_TRIGGER_CONFIG);
    CommandMessage baseMessage = monitoringTriggerConfigs.get().get(key);
    log.trace("config={}", config);
    log.trace("baseMessage={}", baseMessage);
    commandStore.awaitCommand(
        key,
        CommandMessage.newBuilder(baseMessage)
            .setMonitoringTriggerConfig(config)
            .build()
    );
    return externalizeClusterId(config);
  }

  @DELETE
  @Path("/triggers/{guid}")
  public void deleteTrigger(
      @PathParam("guid") String guid
  ) throws InterruptedException, ExecutionException, TimeoutException {
    CommandKey key = getCommandKey(guid, CommandConfigType.MONITORING_TRIGGER_CONFIG);
    log.info("deleting trigger={} and removing from associated actions", guid);
    commandStore.awaitCommand(key, null);
    try (KeyValueIterator<CommandKey, CommandMessage> i =  actionConfigs.get().all()) {
      while (i.hasNext()) {
        KeyValue<CommandKey, CommandMessage> keyValue = i.next();
        if (keyValue.value.getActionConfig().getTriggerGuidList().contains(guid)) {
          log.debug("removing trigger={} from action={}", guid, keyValue.key.getGuid());
          ActionConfig.Builder acBuild = ActionConfig.newBuilder(keyValue.value.getActionConfig());
          acBuild.clearTriggerGuid();
          for (String triggerGuid : keyValue.value.getActionConfig().getTriggerGuidList()) {
            if (!triggerGuid.equals(guid)) {
              acBuild.addTriggerGuid(triggerGuid);
            }
          }
          commandStore.sendCommand(
              keyValue.key,
              CommandMessage.newBuilder(keyValue.value)
                  .setActionConfig(acBuild)
                  .build()
          );
        }
      }
    }
  }

  @POST
  @Path("/actions")
  public ActionConfig createAction(
      ActionConfig config
  ) throws InterruptedException, ExecutionException, TimeoutException {
    if (!Strings.isNullOrEmpty(config.getGuid())) {
      return updateAction(config.getGuid(), config);
    }
    verifyActionConfig(config);
    config  = ActionConfig.newBuilder(config)
        .setGuid(UUID.randomUUID().toString())
        .build();
    log.trace("config={}", config);
    commandStore.awaitCommand(
        getCommandKey(config.getGuid(), CommandConfigType.ACTION_CONFIG),
        CommandMessage.newBuilder()
            .setActionConfig(config)
            .build()
    );
    return config;
  }

  @GET
  @Path("/actions/{guid}")
  public ActionConfig getAction(
      @PathParam("guid") String guid
  ) {
    CommandKey key = getCommandKey(guid, CommandConfigType.ACTION_CONFIG);
    CommandMessage baseMessage = actionConfigs.get().get(key);
    if (baseMessage == null) {
      throw new NotFoundException("unable to find config with guid=" + guid);
    }
    return baseMessage.getActionConfig();
  }

  @GET
  @Path("/actions")
  public List<ActionConfig> getActions(
  ) {
    List<ActionConfig> out = Lists.newArrayList();
    try (KeyValueIterator<CommandKey, CommandMessage> i =  actionConfigs.get().all()) {
      while (i.hasNext()) {
        out.add(i.next().value.getActionConfig());
      }
    }
    return out;
  }

  @PUT
  @Path("/actions/{guid}")
  public ActionConfig updateAction(
      @PathParam("guid") String guid,
      ActionConfig config
  ) throws InterruptedException, ExecutionException, TimeoutException {
    if (!guid.equals(config.getGuid())) {
      throw new ClientErrorException("config guid does not match rest guid", Response.Status.CONFLICT);
    }
    verifyActionConfig(config);
    CommandKey key = getCommandKey(guid, CommandConfigType.ACTION_CONFIG);
    CommandMessage baseMessage = actionConfigs.get().get(key);
    log.trace("config={}", config);
    log.trace("baseMessage={}", baseMessage);
    commandStore.awaitCommand(
        key,
        CommandMessage.newBuilder(baseMessage)
            .setActionConfig(config)
            .build()
    );
    return config;
  }

  @DELETE
  @Path("/actions/{guid}")
  public void deleteAction(
      @PathParam("guid") String guid
  ) throws InterruptedException, ExecutionException, TimeoutException {
    CommandKey key = getCommandKey(guid, CommandConfigType.ACTION_CONFIG);
    commandStore.awaitCommand(key, null);
  }

  @GET
  @Path("history")
  public List<AlertInfo> getHistory(
      @QueryParam("ts") Long timestamp,
      @QueryParam("guid") @DefaultValue("") String guid,
      @QueryParam("limit") @DefaultValue("50") Long limit
  ) {
    if (limit > 500) {
      throw new ClientErrorException("limit must be < 500", Response.Status.BAD_REQUEST);
    }
    List<AlertInfo> out = Lists.newArrayList();
    Bytes startKey = null;
    if (timestamp != null) {
       startKey = alertRangeSerde.key(
          AlertInfo.newBuilder()
              .setGuid(guid)
              .setTimestamp(timestamp)
              .build()
      );
    }
    try (KeyValueIterator<Bytes, AlertInfo> iter = KVQuery.rangeFrom(alertHistoryStore.get(), startKey)) {
      while (iter.hasNext() && out.size() < limit) {
        out.add(externalizeClusterId(iter.next().value));
      }
    }
    return ALERT_INFO_ORDERING.sortedCopy(out);
  }


  private static CommandKey getCommandKey(String guid, CommandConfigType type) {
    return CommandKey.newBuilder()
        .setGuid(guid)
        .setConfigType(type)
        .build();
  }

  private void verifyActionConfig(ActionConfig config) {
    if (config.getIntervalMs() < 0) {
      throw new ClientErrorException("intervalMs must be >=0", Response.Status.BAD_REQUEST);
    }
    for (String triggerGuid : config.getTriggerGuidList()) {
      CommandMessage message = monitoringTriggerConfigs.get().get(
          getCommandKey(triggerGuid, CommandConfigType.MONITORING_TRIGGER_CONFIG)
      );
      if (message == null) {
        throw new ClientErrorException("invalid triggerGuid=" + triggerGuid, Response.Status.BAD_REQUEST);
      }
    }
  }

  private static void verifyMonitoringTriggerConfig(MonitoringTriggerConfig config) {
    if (config.getLagMs() < 0) {
      throw new ClientErrorException("lagMs must be >=0", Response.Status.BAD_REQUEST);
    }
  }

  private static MonitoringTriggerConfig.Builder internalizeClusterIdBuilder(MonitoringTriggerConfig config) {
    return MonitoringTriggerConfig.newBuilder(config)
        .setClusterId(ClusterMetadataDao.getInternalKafkaId(config.getClusterId()));
  }

  private static MonitoringTriggerConfig externalizeClusterId(MonitoringTriggerConfig config) {
    return MonitoringTriggerConfig.newBuilder(config)
        .setClusterId(ClusterMetadataDao.getExternalKafkaId(config.getClusterId()))
        .build();
  }

  private static AlertInfo externalizeClusterId(AlertInfo alertInfo) {
    return AlertInfo.newBuilder(alertInfo)
        .setMonitoringTrigger(externalizeClusterId(alertInfo.getMonitoringTrigger()))
        .build();
  }
}
