// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.rest;

import com.google.inject.Inject;
import com.google.protobuf.GeneratedMessageV3;

import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonSerializer;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.module.SimpleModule;
import com.github.zafarkhaja.semver.Version;

import org.eclipse.jetty.servlet.FilterHolder;
import org.eclipse.jetty.servlet.ServletContextHandler;
import org.eclipse.jetty.util.resource.Resource;
import org.eclipse.jetty.util.resource.ResourceCollection;
import org.glassfish.jersey.servlet.ServletProperties;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.container.ContainerRequestContext;
import javax.ws.rs.container.ContainerResponseContext;
import javax.ws.rs.container.ContainerResponseFilter;
import javax.ws.rs.core.Configurable;
import javax.ws.rs.ext.ContextResolver;

import io.confluent.command.record.alert.CommandAlert.ActionConfig;
import io.confluent.command.record.alert.CommandAlert.MonitoringTriggerConfig;
import io.confluent.controlcenter.alert.record.Alert.AlertInfo;
import io.confluent.rest.Application;
import io.confluent.rest.RestConfig;
import io.confluent.serializers.UberSerde;

public class ControlCenterApplication extends Application<RestConfig> {

  private static final Logger log = LoggerFactory.getLogger(ControlCenterApplication.class);

  private final StaticFileResource staticFileResource;
  private final MessageDeliveryResource messageDeliveryResource;
  private final KafkaResource kafkaResource;
  private final ConnectResource connectResource;
  private final LicenseResource licenseResource;
  private final StatusResource statusResource;
  private final Version version;
  private final AuthResource authResource;
  private final AlertsResource alertsResource;
  private final CommandResource commandResource;
  private final ClusterResource clusterResource;
  private final UberMessageBodyRW<? extends GeneratedMessageV3> uberMessageBodyRW;
  private final UberSerde<AlertInfo> alertInfoUberSerde;
  private final UberSerde<MonitoringTriggerConfig> monitoringTriggerConfigUberSerde;
  private final UberSerde<ActionConfig> actionConfigUberSerde;

  @Inject
  public ControlCenterApplication(
      RestConfig config,
      AlertsResource alertsResource,
      StaticFileResource staticFileResource,
      KafkaResource kafkaResource,
      ClusterResource clusterResource,
      MessageDeliveryResource messageDeliveryResource,
      ConnectResource connectResource,
      LicenseResource licenseResource,
      StatusResource statusResource,
      AuthResource authResource,
      Version version,
      CommandResource commandResource,
      UberMessageBodyRW<? extends GeneratedMessageV3> uberMessageBodyRW,
      UberSerde<AlertInfo> alertInfoUberSerde,
      UberSerde<MonitoringTriggerConfig> monitoringTriggerConfigUberSerde,
      UberSerde<ActionConfig> actionConfigUberSerde
  ) {
    super(config);
    this.alertsResource = alertsResource;
    this.staticFileResource = staticFileResource;
    this.kafkaResource = kafkaResource;
    this.clusterResource = clusterResource;
    this.messageDeliveryResource = messageDeliveryResource;
    this.connectResource = connectResource;
    this.licenseResource = licenseResource;
    this.statusResource = statusResource;
    this.authResource = authResource;
    this.version = version;
    this.commandResource = commandResource;
    this.uberMessageBodyRW = uberMessageBodyRW;
    this.alertInfoUberSerde = alertInfoUberSerde;
    this.monitoringTriggerConfigUberSerde = monitoringTriggerConfigUberSerde;
    this.actionConfigUberSerde = actionConfigUberSerde;
  }


  @Override
  protected ResourceCollection getStaticResources() {
    return new ResourceCollection(Resource.newClassPathResource("io/confluent/controlcenter/rest/static"));
  }

  @Override
  protected void configurePostResourceHandling(ServletContextHandler context) {
    context.addFilter(new FilterHolder(new Filter() {
      @Override
      public void init(FilterConfig filterConfig) throws ServletException {

      }

      @Override
      public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain) throws IOException, ServletException {
        String path = ((HttpServletRequest) request).getPathInfo();
        // redirect /dist/* to static files, everything else gets mapped to /
        if(path.startsWith("/dist/")) {
          chain.doFilter(request, response);
        } else {
          request.getRequestDispatcher("/").forward(request, response);
        }
      }

      @Override
      public void destroy() {

      }
    }), "/*", null);
  }

  @Override
  public void setupResources(Configurable<?> config, RestConfig appConfig) {
    config.register(alertsResource);
    config.register(kafkaResource);
    config.register(clusterResource);
    config.register(connectResource);
    config.register(licenseResource);
    config.register(messageDeliveryResource);
    config.register(statusResource);
    config.register(authResource);
    config.register(staticFileResource);
    config.register(commandResource);
    config.register(new ContextResolver<ObjectMapper>() {
      @Override
      public ObjectMapper getContext(Class<?> type) {
        return getJsonMapper();
      }
    });
    config.register(uberMessageBodyRW);
    config.register(new ContainerResponseFilter() {
      @Override
      public void filter(ContainerRequestContext requestContext, ContainerResponseContext responseContext) throws IOException {
        responseContext.getHeaders().add("X-Confluent-Control-Center-Version", version.toString());
      }
    });
    // forward empty 404 from staticFileResource to static files
    config.property(ServletProperties.FILTER_FORWARD_ON_404, true);
  }

  @Override
  protected ObjectMapper getJsonMapper() {
    ObjectMapper om =  super.getJsonMapper();
    //TODO(norwood) lets inject this whole module
    SimpleModule module = new SimpleModule("ProtoJsonSerializerModule", new com.fasterxml.jackson.core.Version(0, 1, 0, "alpha"));
    module.addSerializer(AlertInfo.class, new JsonSerializer<AlertInfo>() {
      @Override
      public void serialize(AlertInfo value, JsonGenerator gen, SerializerProvider serializers) throws IOException, JsonProcessingException {
        gen.writeRawValue(alertInfoUberSerde.toJSON(value));
      }

      @Override
      public Class<AlertInfo> handledType() {
        return AlertInfo.class;
      }
    });
    module.addSerializer(MonitoringTriggerConfig.class, new JsonSerializer<MonitoringTriggerConfig>() {
      @Override
      public void serialize(MonitoringTriggerConfig value, JsonGenerator gen, SerializerProvider serializers) throws IOException, JsonProcessingException {
        gen.writeRawValue(monitoringTriggerConfigUberSerde.toJSON(value));
      }

      @Override
      public Class<MonitoringTriggerConfig> handledType() {
        return MonitoringTriggerConfig.class;
      }
    });
    module.addSerializer(ActionConfig.class, new JsonSerializer<ActionConfig>() {
      @Override
      public void serialize(ActionConfig value, JsonGenerator gen, SerializerProvider serializers) throws IOException, JsonProcessingException {
        gen.writeRawValue(actionConfigUberSerde.toJSON(value));
      }

      @Override
      public Class<ActionConfig> handledType() {
        return ActionConfig.class;
      }
    });

    om.registerModule(module);
    return om;
  }
}
