// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.rest.res;

import com.fasterxml.jackson.annotation.JsonInclude;
import com.google.common.base.Strings;
import com.google.common.collect.ComparisonChain;
import com.google.common.collect.Maps;
import io.confluent.controlcenter.Rollup;
import io.confluent.controlcenter.keys.Keys.KeyType;
import io.confluent.controlcenter.record.Controlcenter.MemberInfo;
import io.confluent.controlcenter.rest.Metric;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Objects;

@JsonStandard
public class DeliveryResponse {

  public enum DeliveryType {
    CONSUMER_GROUPS(
        KeyType.LOOKUP_CLIENTTYPE,
        KeyType.EXPECTED_PRODUCTION,
        KeyType.INFO_GROUPS_IN_CLIENTTYPE,
        KeyType.LOOKUP_CLIENTTYPE_GROUP,
        KeyType.EXPECTED_PRODUCTION_GROUP,
        KeyType.INFO_TOPICS_IN_CLIENTTYPE
    ),
    CONSUMER_GROUP(
        KeyType.LOOKUP_CLIENTTYPE_GROUP,
        KeyType.EXPECTED_PRODUCTION_GROUP,
        KeyType.INFO_CLIENTIDS_IN_GROUP,
        KeyType.LOOKUP_CLIENTTYPE_CLIENT,
        KeyType.EXPECTED_PRODUCTION_CLIENT,
        KeyType.INFO_TOPICPARTITIONS_IN_GROUP
        ),
    CONSUMER(
        KeyType.LOOKUP_CLIENTTYPE_CLIENT,
        KeyType.EXPECTED_PRODUCTION_CLIENT,
        KeyType.INFO_TOPICPARTITIONS_IN_CLIENT,
        KeyType.LOOKUP_CLIENTTYPE_CLIENT_TOPIC_PARTITION,
        KeyType.EXPECTED_PRODUCTION_IN_TOPICPARTITION_FOR_CLIENT,
        null
    ),
    CONSUMER_GROUPS_TOPICS(
        KeyType.LOOKUP_CLIENTTYPE,
        KeyType.EXPECTED_PRODUCTION,
        KeyType.INFO_TOPICS_IN_CLIENTTYPE,
        KeyType.LOOKUP_CLIENTTYPE_TOPIC,
        KeyType.EXPECTED_PRODUCTION_IN_TOPIC,
        KeyType.INFO_GROUPS_IN_CLIENTTYPE
    ),
    CONSUMER_GROUP_TOPICS(
        KeyType.LOOKUP_CLIENTTYPE_GROUP,
        KeyType.EXPECTED_PRODUCTION_GROUP,
        KeyType.INFO_TOPICPARTITIONS_IN_GROUP,
        KeyType.LOOKUP_CLIENTTYPE_GROUP_TOPICPARTITION,
        KeyType.EXPECTED_PRODUCTION_IN_TOPICPARTITION_FOR_GROUP,
        KeyType.INFO_CLIENTIDS_IN_GROUP
        );

    public final KeyType baseConsumptionKeyType;
    public final KeyType baseProductionKeyType;
    public final KeyType subGroupingKeyType;
    public final KeyType subConsumptionKeyType;
    public final KeyType subProductionKeyType;
    public final KeyType secondarySubGroupingKeyType;

    DeliveryType(
        KeyType baseConsumptionKeyType,
        KeyType baseProductionKeyType,
        KeyType subGroupingKeyType,
        KeyType subConsumptionKeyType,
        KeyType subProductionKeyType,
        KeyType secondarySubGroupingKeyType
    ) {
      this.baseConsumptionKeyType = baseConsumptionKeyType;
      this.baseProductionKeyType = baseProductionKeyType;
      this.subGroupingKeyType = subGroupingKeyType;
      this.subConsumptionKeyType = subConsumptionKeyType;
      this.subProductionKeyType = subProductionKeyType;
      this.secondarySubGroupingKeyType = secondarySubGroupingKeyType;
    }
  }

  @JsonStandard
  public static class MemberResponse implements Comparable<MemberResponse> {
    @JsonInclude(JsonInclude.Include.NON_EMPTY)
    public String clientId;
    @JsonInclude(JsonInclude.Include.NON_EMPTY)
    public String group;
    @JsonInclude(JsonInclude.Include.NON_EMPTY)
    public String topic;
    @JsonInclude(JsonInclude.Include.NON_EMPTY)
    public Integer partition;
    @JsonInclude(JsonInclude.Include.NON_EMPTY)
    public Map<Metric, ArrayList<Object>> produced = Maps.newTreeMap();
    @JsonInclude(JsonInclude.Include.NON_EMPTY)
    public Map<Metric, ArrayList<Object>> consumed = Maps.newTreeMap();

    public MemberResponse(MemberInfo memberInfo) {
      this.clientId = memberInfo.getClientId();
      this.group = memberInfo.getGroup();

      // first try to fill from topicpartition
      this.topic = Strings.emptyToNull(memberInfo.getTopicPartition().getTopic());
      if (this.topic != null) {
        this.partition = memberInfo.getTopicPartition().getPartition();
      } else {
        // fallback to just topic
        this.topic = Strings.emptyToNull(memberInfo.getTopic());
      }
    }

    @Override
    public boolean equals(Object o) {
      if (this == o) {
        return true;
      }
      if (o == null || getClass() != o.getClass()) {
        return false;
      }
      MemberResponse that = (MemberResponse) o;
      return Objects.equals(clientId, that.clientId) &&
             Objects.equals(group, that.group) &&
             Objects.equals(topic, that.topic) &&
             Objects.equals(partition, that.partition);
    }

    @Override
    public int hashCode() {
      return Objects.hash(clientId, group, topic, partition);
    }

    @Override
    public int compareTo(MemberResponse o) {
      return ComparisonChain.start()
          .compare(Strings.nullToEmpty(clientId), Strings.nullToEmpty(o.clientId))
          .compare(Strings.nullToEmpty(group), Strings.nullToEmpty(o.group))
          .compare(Strings.nullToEmpty(topic), Strings.nullToEmpty(o.topic))
          .compare(partition == null ? -1 : partition, o.partition == null ? -1 : o.partition)
          .result();
    }
  }

  public Rollup rollup;
  public DeliveryType type;
  public long startTimeMs;
  public long stepMs;
  @JsonInclude(JsonInclude.Include.NON_NULL)
  public String group;
  @JsonInclude(JsonInclude.Include.NON_NULL)
  public String clientId;
  @JsonInclude(JsonInclude.Include.NON_EMPTY)
  public Map<Metric, ArrayList<Object>> produced = null;
  @JsonInclude(JsonInclude.Include.NON_EMPTY)
  public Map<Metric, ArrayList<Object>> consumed = null;
  @JsonInclude(JsonInclude.Include.NON_NULL)
  public List<MemberResponse> members = null;
  @JsonInclude(JsonInclude.Include.NON_NULL)
  public List<MemberResponse> sources = null;
  @JsonInclude(JsonInclude.Include.NON_NULL)
  public List<MemberResponse> destinations = null;

  public DeliveryResponse(
      Rollup rollup,
      long startTimeMs,
      long stepMs,
      String group,
      String clientId,
      DeliveryType deliveryType
  ) {
    this.rollup = rollup;
    this.startTimeMs = startTimeMs;
    this.stepMs = stepMs;
    this.group = group;
    this.clientId = clientId;
    this.type = deliveryType;
  }

  public List<MemberResponse> getMembers() {
    if (members == null) {
      return null;
    }
    Collections.sort(members);
    return members;
  }

  public List<MemberResponse> getSources() {
    if (sources == null) {
      return null;
    }
    Collections.sort(sources);
    return sources;
  }

  public List<MemberResponse> getDestinations() {
    if (destinations == null) {
      return null;
    }
    Collections.sort(destinations);
    return destinations;
  }
}
