// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.serialization;

import com.google.inject.Guice;
import com.google.inject.Injector;

import io.confluent.controlcenter.command.CommandModule;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.common.errors.SerializationException;
import org.joda.time.DateTime;

import java.io.PrintStream;
import java.util.Properties;

import io.confluent.controlcenter.ControlCenterConfig;
import io.confluent.controlcenter.ControlCenterConfigModule;
import io.confluent.controlcenter.streams.StreamsConfigModule;
import io.confluent.controlcenter.streams.TopicStoreMaster;
import io.confluent.controlcenter.streams.TopicStoreModule;
import kafka.common.MessageFormatter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class UberFormatter implements MessageFormatter {
  private static final Logger log = LoggerFactory.getLogger(new Object() {}.getClass().getEnclosingClass());
  private TopicStoreMaster topicStoreMaster;
  private boolean allowErrors;
  private boolean printTopic;
  private boolean printPartition;
  private boolean printTs;
  private boolean printKey;
  private String fieldSeparator;
  private String lineSeparator;

  @Override
  public void init(Properties props) {
    String configFile = props.getProperty("config");
    try {
      final ControlCenterConfig config = new ControlCenterConfig(configFile);
      Injector injector = Guice.createInjector(
          new ControlCenterConfigModule(config),
          new StreamsConfigModule(),
          new SerializationModule(),
          new TopicStoreModule()
      );
      topicStoreMaster = injector.getInstance(TopicStoreMaster.class);
    } catch (Exception e) {
      log.error("Init failed: ", e);
      throw new RuntimeException("Failed to load config", e);
    }
    allowErrors = getBoolean(props, "allow.errors", true);
    printTopic = getBoolean(props, "print.topic", true);
    printPartition = getBoolean(props, "print.partition", true);
    printTs = getBoolean(props, "print.timestamp", true);
    printKey = getBoolean(props, "print.key", true);
    lineSeparator = props.getProperty("line.separator", "\n");
    fieldSeparator = props.getProperty("field.separator", "\t");
  }

  private boolean getBoolean(Properties props, String name, boolean defaultValue) {
    if (props.containsKey(name)) {
      return Boolean.parseBoolean(props.getProperty(name).trim());
    }
    return defaultValue;
  }

  @Override
  public void writeTo(ConsumerRecord<byte[], byte[]> consumerRecord, PrintStream output) {
    TopicStoreMaster.SerdePair serdes = topicStoreMaster.getSerdesForTopic(consumerRecord.topic());
    if (serdes == null && !allowErrors) {
      throw new SerializationException("No serde found for topic: " + consumerRecord.topic());
    }

    if (printTopic) {
      output.append(consumerRecord.topic());
      output.append(fieldSeparator);
    }
    if (printPartition) {
      output.append(Integer.toString(consumerRecord.partition()));
      output.append(fieldSeparator);
    }
    if (printTs) {
      output.append(new DateTime(consumerRecord.timestamp()).toString());
      output.append(fieldSeparator);
    }
    if (printKey) {
      output.append(getKeyString(consumerRecord.topic(), consumerRecord.key(), serdes));
      output.append(fieldSeparator);
    }
    output.append(getValueString(consumerRecord.topic(), consumerRecord.value(), serdes));
    output.append(lineSeparator);
  }

  private String getValueString(String topic, byte[] valueBytes, TopicStoreMaster.SerdePair serdes) {
    String valueStr = "null";
    if (valueBytes != null) {
      if (serdes == null || serdes.valueSerde == null) {
        if (allowErrors) {
          valueStr = "ERROR: No serde found";
        } else {
          throw new SerializationException("No value serde found for topic: " + topic);
        }
      } else {
        try {
          Object value = serdes.valueSerde.deserialize(valueBytes);
          valueStr = serdes.valueSerde.toJSON(value);
        } catch (SerializationException e) {
          if (allowErrors) {
            valueStr = "ERROR: " + e.getMessage();
          } else {
            throw e;
          }
        }
      }
    }
    return valueStr;
  }

  private String getKeyString(String topic, byte[] keyBytes, TopicStoreMaster.SerdePair serdes) {
    String keyStr = "null";
    if (keyBytes != null) {
      if (serdes == null || serdes.keySerde == null) {
        if (allowErrors) {
          keyStr = "ERROR: No serde found";
        } else {
          throw new SerializationException("No key serde found for topic: " + topic);
        }
      } else {
        try {
          if (serdes.windowedSerde != null) {
            WindowStoreSerde.WindowedKey windowedKey = serdes.windowedSerde.deserialize(keyBytes);
            keyStr = serdes.windowedSerde.toJSON(windowedKey);
          } else {
            Object key = serdes.keySerde.deserialize(keyBytes);
            keyStr = serdes.keySerde.toJSON(key);
          }
        } catch (SerializationException e) {
          if (allowErrors) {
            keyStr = "ERROR: " + e.getMessage();
          } else {
            throw e;
          }
        }
      }
    }
    return keyStr;
  }

  @Override
  public void close() {
  }

}
