// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.serialization;

import com.google.inject.Guice;
import com.google.inject.Injector;

import io.confluent.controlcenter.command.CommandModule;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.KafkaException;
import org.apache.kafka.common.errors.SerializationException;
import org.joda.time.DateTime;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Properties;

import io.confluent.controlcenter.ControlCenterConfig;
import io.confluent.controlcenter.ControlCenterConfigModule;
import io.confluent.controlcenter.streams.StreamsConfigModule;
import io.confluent.controlcenter.streams.TopicStoreMaster;
import io.confluent.controlcenter.streams.TopicStoreModule;
import kafka.common.MessageReader;

public class UberReader implements MessageReader {
  private String topic;
  private boolean parseTopic;
  private boolean parsePartition;
  private boolean parseTs;
  private boolean parseKey;
  private String fieldSeparator;
  private BufferedReader reader;
  private int lineNumber = 0;
  private TopicStoreMaster topicStoreMaster;

  @Override
  public void init(InputStream inputStream, Properties props) {
    topic = props.getProperty("topic");
    parseTopic = getBoolean(props, "parse.topic", topic == null);
    parsePartition = getBoolean(props, "parse.partition", true);
    parseTs = getBoolean(props, "parse.timestamp", true);
    parseKey = getBoolean(props, "parse.key", true);
    fieldSeparator = props.getProperty("field.separator", "\t");
    if (topic == null && !parseTopic) {
      throw new SerializationException("Either the topic or parse.topic properties must be set");
    }
    reader = new BufferedReader(new InputStreamReader(inputStream));
    String configFile = props.getProperty("config");
    try {
      final ControlCenterConfig config = new ControlCenterConfig(configFile);
      Injector injector = Guice.createInjector(
          new ControlCenterConfigModule(config),
          new StreamsConfigModule(),
          new SerializationModule(),
          new TopicStoreModule()
      );
      topicStoreMaster = injector.getInstance(TopicStoreMaster.class);
    } catch (Exception e) {
      throw new SerializationException("Failed to load config " + e.getMessage(), e);
    }
  }

  private boolean getBoolean(Properties props, String name, boolean defaultValue) {
    if (props.containsKey(name)) {
      return Boolean.parseBoolean(props.getProperty(name).trim());
    }
    return defaultValue;
  }

  private ProducerRecord<byte[], byte[]> parseLine(String line) {
    String topicName = topic;
    Integer partition = null;
    Long timestamp = null;
    byte[] keyBytes = null;
    int readIdx = 0;
    if (parseTopic) {
      int topicEndIdx = line.indexOf(fieldSeparator, readIdx);
      if (topicEndIdx < 0) {
        throw new KafkaException("No topic found in line " + line);
      }
      String topicField = nullableSubstring(line, readIdx, topicEndIdx);
      if (topicField != null) {
        topicName = topicField;
      }
      readIdx = topicEndIdx + fieldSeparator.length();
    }
    TopicStoreMaster.SerdePair serdes = topicStoreMaster.getSerdesForTopic(topicName);
    if (serdes == null) {
      throw new SerializationException("No serde found for topic: " + topicName);
    }
    if (parsePartition) {
      int partitionEndIdx = line.indexOf(fieldSeparator, readIdx);
      if (partitionEndIdx < 0) {
        throw new KafkaException("No partition found in line " + line);
      }
      String partitionField = nullableSubstring(line, readIdx, partitionEndIdx);
      if (partitionField != null) {
        partition = Integer.parseInt(partitionField);
      }
      readIdx = partitionEndIdx + fieldSeparator.length();
    }
    if (parseTs) {
      int tsEndIdx = line.indexOf(fieldSeparator, readIdx);
      if (tsEndIdx < 0) {
        throw new KafkaException("No timestamp found in line " + line);
      }
      String timestampField = nullableSubstring(line, readIdx, tsEndIdx);
      if (timestampField != null) {
        timestamp = DateTime.parse(timestampField).getMillis();
      }
      readIdx = tsEndIdx + fieldSeparator.length();
    }
    if (parseKey) {
      int keyEndIdx = line.indexOf(fieldSeparator, readIdx);
      if (keyEndIdx < 0) {
        throw new KafkaException("No key found in line " + line);
      }
      String keyField = nullableSubstring(line, readIdx, keyEndIdx);
      if (keyField != null) {
        keyBytes = serdes.keySerde.fromJSON(keyField);
      }
      readIdx = keyEndIdx + fieldSeparator.length();
    }
    byte[] valBytes = serdes.valueSerde.fromJSON(line.substring(readIdx));
    return new ProducerRecord<>(topicName, partition, timestamp, keyBytes, valBytes);
  }

  private String nullableSubstring(String string, int startIdx, int endIdx) {
    String result = string.substring(startIdx, endIdx);
    if (result.toLowerCase().equals("null")) {
      return null;
    }
    return result;
  }

  @Override
  public ProducerRecord<byte[], byte[]> readMessage() {
    try {
      while (true) {
        lineNumber += 1;
        String line = reader.readLine();
        if (line == null) {
          return null;
        }
        if (line.startsWith("#") || line.trim().length() == 0) {
          continue;
        }
        return parseLine(line);
      }
    } catch (Exception e) {
      throw new KafkaException(String.format("Error line %s: %s", lineNumber, e.getMessage()), e);
    }
  }

  @Override
  public void close() {
  }

}
