// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.serialization;

import com.google.gson.Gson;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonPrimitive;
import io.confluent.serializers.UberSerde;
import org.apache.kafka.streams.state.StateSerdes;
import org.apache.kafka.streams.state.internals.WindowStoreUtils;

import java.nio.ByteBuffer;

public class WindowStoreSerde<T> {
  private final Gson gson = new Gson();
  private final UberSerde<T> keySerde;
  private final StateSerdes<T, T> serdes;

  public WindowStoreSerde(UberSerde<T> keySerde) {
    this.keySerde = keySerde;
    //This object is needed by KStreams WindowStoreUtils API but only the keySerde is used
    serdes = new StateSerdes<>("dummy", keySerde, keySerde);
  }

  public static class WindowedKey<T> {
    public final T key;
    public final long timestamp;
    public final int seqNum;

    public WindowedKey(T key, long timestamp, int seqNum) {
      this.key = key;
      this.timestamp = timestamp;
      this.seqNum = seqNum;
    }
  }

  public byte[] serialize(WindowedKey<T> windowedKey) {
    return WindowStoreUtils.toBinaryKey(windowedKey.key, windowedKey.timestamp, windowedKey.seqNum, serdes);
  }

  public WindowedKey<T> deserialize(byte[] bytes) {
    if (bytes == null) {
      return null;
    }
    T key = WindowStoreUtils.keyFromBinaryKey(bytes, serdes);
    long timestamp = WindowStoreUtils.timestampFromBinaryKey(bytes);
    int seqNum = sequenceNumberFromBinaryKey(bytes);
    return new WindowedKey<>(key, timestamp, seqNum);
  }

  //TODO(rhoover): call WindowStoreUtils in KStreams once seq number extraction is available
  private int sequenceNumberFromBinaryKey(byte[] bytes) {
    return ByteBuffer.wrap(bytes).getInt(bytes.length - 8);
  }

  public byte[] fromJSON(String json) {
    WindowedKey windowedKey = gson.fromJson(json, WindowedKey.class);
    //TODO(rhoover): use TypeAdapter?
    String objJson = gson.toJson(windowedKey.key);
    T key = keySerde.deserialize(keySerde.fromJSON(objJson));
    return WindowStoreUtils.toBinaryKey(key, windowedKey.timestamp, windowedKey.seqNum, serdes);
  }

  public String toJSON(WindowedKey<T> windowedKey) {
    Gson gson = new Gson();
    JsonObject doc = new JsonObject();
    //TODO(rhoover): use TypeAdapter?
    doc.add("key", gson.fromJson(keySerde.toJSON(windowedKey.key), JsonElement.class));
    doc.add("timestamp", new JsonPrimitive(windowedKey.timestamp));
    doc.add("seq_num", new JsonPrimitive(windowedKey.seqNum));
    return gson.toJson(doc);
  }
}
