// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;

import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.kstream.KStream;
import org.apache.kafka.streams.kstream.KTable;
import org.apache.kafka.streams.kstream.KeyValueMapper;
import org.apache.kafka.streams.kstream.Predicate;
import org.apache.kafka.streams.kstream.TimeWindows;
import org.apache.kafka.streams.kstream.Window;
import org.apache.kafka.streams.kstream.Windowed;
import org.apache.kafka.streams.kstream.Windows;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collection;
import java.util.Map;
import java.util.Set;

import io.confluent.controlcenter.Rollup;
import io.confluent.monitoring.record.Monitoring;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;

/**
 *  This is used to take a stream and split it with the mappers from
 *  getMappers(), create a number of rollups and the pass them to the
 *  implementing class in filterAndAggegateRollup() to do what they will.
 *
 * @param <O>
 */
public abstract class AbstractMonitoringMessageRollupStreamExtension<O>
    implements StreamExtension<O, byte[], MonitoringMessage> {

  private static final Logger log = LoggerFactory.getLogger(AbstractMonitoringMessageRollupStreamExtension.class);
  private static final Set<Monitoring.ClientType> DEFAULT_VALID_CLIENT_TYPES = ImmutableSet.of(Monitoring.ClientType.CONSUMER);

  public Map<Rollup, KTable<Windowed<Bytes>, O>> extend(KStream<byte[], MonitoringMessage> stream) {
    // Create a record for each "group by" key
    KStream<Bytes, O> explodedStream = stream.filter(new Predicate<byte[], MonitoringMessage>() {
      @Override
      public boolean test(byte[] key, MonitoringMessage value) {
        return validClientTypes().contains(value.getClientType());
      }
    }).flatMap(new KeyValueMapper<byte[], MonitoringMessage, Iterable<KeyValue<Bytes, O>>>() {
      @Override
      public Iterable<KeyValue<Bytes, O>> apply(byte[] key, MonitoringMessage value) {
        Collection<KeyValue<Bytes, O>> out = Lists.newArrayList();
        for (TypeMapper<O> mapper : getMappers()) {
          out.add(mapper.apply(key, value));
        }
        return out;
      }
    });
    // Re-partition so that all records in each group are in the same partition
    KStream<Bytes, O> groupedStream = shuffle(explodedStream);
    // Filter and aggregate per group per rollup
    Map<Rollup, KTable<Windowed<Bytes>, O>> rollupAggregates = Maps.newHashMap();
    for (Rollup rollup : Rollup.values()) {
      final Windows windows = TimeWindows
          .of(rollup.getMillis())
          .until(rollup.getRetainMillis());
      rollupAggregates.put(rollup,
                           filterAndAggregateRollup(
                               groupedStream,
                               windows,
                               TopicStoreMaster.nameJoin(name(), rollup.toString())
                           )
      );
    }
    return rollupAggregates;
  }

  /**
   * name to refer to this extension. used to generate windows, so needs to be unique within
   * a kstreams dag
   * @return
   */
  protected abstract String name();

  protected Set<Monitoring.ClientType> validClientTypes() {
    return DEFAULT_VALID_CLIENT_TYPES;
  }

  /**
   * mappers that will be used to create N forks for the stream, one for each mapper
   * @return
   */
  protected abstract Iterable<? extends TypeMapper<O>> getMappers();

  /**
   * Filter and aggregate per group per rollup
   */
  protected abstract KTable<Windowed<Bytes>, O> filterAndAggregateRollup(
      KStream<Bytes, O> stream,
      final Windows<Window> windows,
      final String name
  );

  /**
   * Re-partition the stream so that all records in each group are in the same partition
   * @return the re-partitioned stream
   */
  protected abstract KStream<Bytes, O> shuffle(KStream<Bytes, O> stream);
}
