// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams;

import com.google.inject.Inject;
import com.google.inject.Provider;
import com.google.protobuf.Descriptors;

import org.apache.commons.mail.Email;
import org.apache.commons.mail.EmailException;
import org.apache.kafka.common.serialization.Serdes;
import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.KafkaStreams;
import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.StreamsConfig;
import org.apache.kafka.streams.kstream.JoinWindows;
import org.apache.kafka.streams.kstream.KStream;
import org.apache.kafka.streams.kstream.KStreamBuilder;
import org.apache.kafka.streams.kstream.KTable;
import org.apache.kafka.streams.kstream.KeyValueMapper;
import org.apache.kafka.streams.kstream.ValueMapper;
import org.apache.kafka.streams.kstream.Windowed;
import org.apache.kafka.streams.processor.StateStoreSupplier;
import org.apache.kafka.streams.processor.StreamPartitioner;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Map;
import java.util.Set;

import io.confluent.command.record.alert.CommandAlert;
import io.confluent.controlcenter.Rollup;
import io.confluent.controlcenter.alert.record.Alert.AlertInfo;
import io.confluent.controlcenter.keys.Keys.KeyType;
import io.confluent.controlcenter.record.Controlcenter.TriggerActions;
import io.confluent.controlcenter.record.Controlcenter.TriggerEvent;
import io.confluent.controlcenter.record.Controlcenter.TriggerMeasurement;
import io.confluent.controlcenter.record.Controlcenter.VerifiableTriggerInfo;
import io.confluent.controlcenter.record.Controlcenter.WindowedClusterGroup;
import io.confluent.controlcenter.serialization.OrderedKeyPrefixedSerdeSupplier;
import io.confluent.controlcenter.streams.TopicStoreMaster.Store;
import io.confluent.controlcenter.streams.TopicStoreMaster.Topic;
import io.confluent.controlcenter.streams.TopicStoreModule.ActualGroupConsumptionAggregateTopic;
import io.confluent.controlcenter.streams.TopicStoreModule.AlertHistoryStore;
import io.confluent.controlcenter.streams.TopicStoreModule.ExpectedGroupConsumptionAggregateTopic;
import io.confluent.controlcenter.streams.TopicStoreModule.MonitoringErrorTopic;
import io.confluent.controlcenter.streams.TopicStoreModule.MonitoringTopic;
import io.confluent.controlcenter.streams.TopicStoreModule.MonitoringTriggerStore;
import io.confluent.controlcenter.streams.TopicStoreModule.ReKeyedMonitoringTopic;
import io.confluent.controlcenter.streams.TopicStoreModule.TriggerActionsStore;
import io.confluent.controlcenter.streams.TopicStoreModule.TriggerEventsStore;
import io.confluent.controlcenter.streams.TopicStoreModule.VerifierStore;
import io.confluent.controlcenter.streams.TopicStoreModule.ReKeyedTriggerEventTopic;
import io.confluent.controlcenter.streams.alert.AlertHistoryProcessorSupplier;
import io.confluent.controlcenter.streams.alert.AlertTransformerSupplier;
import io.confluent.controlcenter.streams.alert.GroupMeasurement;
import io.confluent.controlcenter.streams.alert.MonitoringTriggerMeasurementJoiner;
import io.confluent.controlcenter.streams.alert.MonitoringTriggerTransformerSupplier;
import io.confluent.controlcenter.streams.alert.SendEmail;
import io.confluent.controlcenter.streams.alert.TriggerEventUtil;
import io.confluent.controlcenter.streams.cluster.ClusterStreamExtension;
import io.confluent.controlcenter.streams.consumergroups.ConsumerGroupStreamExtension;
import io.confluent.controlcenter.streams.group.GroupStreamExtension;
import io.confluent.controlcenter.streams.monitoring.MonitoringMessageHasWindow;
import io.confluent.controlcenter.streams.monitoring.MonitoringMessageType;
import io.confluent.controlcenter.streams.monitoring.MonitoringStreamExtension;
import io.confluent.controlcenter.streams.verify.MonitoringVerifier;
import io.confluent.controlcenter.streams.verify.VerifyTransformerSupplier;
import io.confluent.monitoring.common.MonitoringMessageUtil;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.serializers.OrderedKeyPrefixedSerde;

public class StreamProvider implements Provider<KafkaStreams> {

  private static final Logger log = LoggerFactory.getLogger(StreamProvider.class);

  private final StreamsConfig streamsConfig;
  private final MonitoringStreamExtension monitoringStreamExtension;
  private final GroupStreamExtension groupStreamExtension;
  private final ClusterStreamExtension clusterStreamExtension;
  private final ConsumerGroupStreamExtension consumerGroupStreamExtension;
  private final Set<StateStoreSupplier> storeSuppliers;
  private final VerifyTransformerSupplier verifyTransformerSupplier;
  private final Store<Bytes, MonitoringVerifier, MonitoringVerifier> verifierStore;
  private final Topic<Void, MonitoringMessage, Void, Void> monitoringTopic;
  private final Topic<Void, MonitoringMessage, Void, Void> rekeyedMonitoringTopic;
  private final Topic<Void, MonitoringMessage, Void, Void> errorTopic;
  private final Topic<WindowedClusterGroup, MonitoringMessage, Void, Void> rekeyedExpectedGroupConsumptionTopic;
  private final Topic<WindowedClusterGroup, MonitoringMessage, Void, Void> rekeyedActualGroupConsumptionTopic;
  private final Topic<Bytes, TriggerEvent, Void, Void> rekeyedTriggerEventTopic;
  private final SourcePartitionTrackingTransformerSupplier
      sourcePartitionTrackingTransformerSupplier;
  private final MonitoringTriggerTransformerSupplier monitoringTriggerTransformerSupplier;
  private final Store<Bytes, TriggerMeasurement, TriggerMeasurement> monitoringTriggerStore;
  private final Store<Bytes, TriggerActions, TriggerActions> triggerActionsStore;
  private final Store<Bytes, VerifiableTriggerInfo, VerifiableTriggerInfo> triggerEventsStore;
  private final Store<Bytes, AlertInfo, AlertInfo> alertHistoryStore;
  private final OrderedKeyPrefixedSerde<KeyType, MonitoringMessage> keySerde;
  private final GroupMeasurement groupMeasurement;
  private final AlertTransformerSupplier alertTransformerSupplier;
  private final AlertHistoryProcessorSupplier alertHistoryProcessorSupplier;
  private final SendEmail emailSender;

  @Inject
  public StreamProvider(
      StreamsConfig streamsConfig,
      MonitoringStreamExtension monitoringStreamExtension,
      GroupStreamExtension groupStreamExtension,
      ClusterStreamExtension clusterStreamExtension,
      ConsumerGroupStreamExtension consumerGroupStreamExtension,
      Set<StateStoreSupplier> storeSuppliers,
      VerifyTransformerSupplier verifyTransformerSupplier,
      @VerifierStore Store<Bytes, MonitoringVerifier, MonitoringVerifier> verifierStore,
      @MonitoringTopic Topic<Void, MonitoringMessage, Void, Void> monitoringTopic,
      @ReKeyedMonitoringTopic Topic<Void, MonitoringMessage, Void, Void> rekeyedMonitoringTopic,
      @MonitoringErrorTopic Topic<Void, MonitoringMessage, Void, Void> errorTopic,
      @ExpectedGroupConsumptionAggregateTopic Topic<WindowedClusterGroup, MonitoringMessage, Void, Void> rekeyedExpectedGroupConsumptionTopic,
      @ActualGroupConsumptionAggregateTopic Topic<WindowedClusterGroup, MonitoringMessage, Void, Void> rekeyedActualGroupConsumptionTopic,
      SourcePartitionTrackingTransformerSupplier sourcePartitionTrackingTransformerSupplier,
      MonitoringTriggerTransformerSupplier monitoringTriggerTransformerSupplier,
      @MonitoringTriggerStore Store<Bytes, TriggerMeasurement, TriggerMeasurement> monitoringTriggerStore,
      @TriggerActionsStore Store<Bytes, TriggerActions, TriggerActions> triggerActionsStore,
      @TriggerEventsStore Store<Bytes, VerifiableTriggerInfo, VerifiableTriggerInfo> triggerEventsStore,
      @AlertHistoryStore Store<Bytes, AlertInfo, AlertInfo> alertHistoryStore,
      AlertTransformerSupplier alertTransformerSupplier,
      AlertHistoryProcessorSupplier alertHistoryProcessorSupplier,
      OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> keySerdeSupplier,
      GroupMeasurement groupMeasurement,
      SendEmail emailSender,
      @ReKeyedTriggerEventTopic Topic<Bytes, TriggerEvent, Void, Void> rekeyedTriggerEventTopic

  ) {
    this.streamsConfig = streamsConfig;
    this.monitoringStreamExtension = monitoringStreamExtension;
    this.groupStreamExtension = groupStreamExtension;
    this.clusterStreamExtension = clusterStreamExtension;
    this.consumerGroupStreamExtension = consumerGroupStreamExtension;
    this.storeSuppliers = storeSuppliers;
    this.verifyTransformerSupplier = verifyTransformerSupplier;
    this.sourcePartitionTrackingTransformerSupplier = sourcePartitionTrackingTransformerSupplier;
    this.monitoringTopic = monitoringTopic;
    this.rekeyedMonitoringTopic = rekeyedMonitoringTopic;
    this.verifierStore = verifierStore;
    this.errorTopic = errorTopic;
    this.rekeyedActualGroupConsumptionTopic = rekeyedActualGroupConsumptionTopic;
    this.rekeyedExpectedGroupConsumptionTopic = rekeyedExpectedGroupConsumptionTopic;
    this.monitoringTriggerTransformerSupplier = monitoringTriggerTransformerSupplier;
    this.monitoringTriggerStore = monitoringTriggerStore;
    this.alertTransformerSupplier = alertTransformerSupplier;
    this.alertHistoryProcessorSupplier = alertHistoryProcessorSupplier;
    this.triggerActionsStore = triggerActionsStore;
    this.triggerEventsStore = triggerEventsStore;
    this.alertHistoryStore = alertHistoryStore;
    this.keySerde = keySerdeSupplier.get();
    this.groupMeasurement = groupMeasurement;
    this.emailSender = emailSender;
    this.rekeyedTriggerEventTopic = rekeyedTriggerEventTopic;
  }

  @Override
  public KafkaStreams get() {
    try {
      KStreamBuilder builder = new KStreamBuilder();

      for (StateStoreSupplier storeSupplier : storeSuppliers) {
        builder.addStateStore(storeSupplier);
      }

      KStream<byte[], MonitoringMessage> input = builder.stream(Serdes.ByteArray(), monitoringTopic.valueSerde, monitoringTopic.name);

      final StreamPartitioner<byte[], MonitoringMessage> partitioner = new StreamPartitioner<byte[], MonitoringMessage>() {
        @Override
        public Integer partition(byte[] key, MonitoringMessage value, int numPartitions) {
          return MonitoringMessageUtil.partitionFor(value, numPartitions);
        }
      };
      KStream<byte[], MonitoringMessage>[] branches = input.transform(
          sourcePartitionTrackingTransformerSupplier
      ).flatMap(
          new Identity<byte[], Iterable<KeyValue<byte[], MonitoringMessage>>>()
      ).through(
          Serdes.ByteArray(),
          rekeyedMonitoringTopic.valueSerde,
          partitioner,
          rekeyedMonitoringTopic.name
      ).transform(
          verifyTransformerSupplier,
          verifierStore.name
      ).flatMap(
          new Identity<byte[], Iterable<KeyValue<byte[], MonitoringMessage>>>()
      ).branch(
          MonitoringMessageType.IS_NORMAL,
          MonitoringMessageType.IS_ERROR
      );

      // we need to filter out messages with no window on an non-error branch -- those are
      // heartbeat messages from clients. We don't need to do that on error branch, because all
      // error messages must have a window set.
      input = branches[0].filter(new MonitoringMessageHasWindow());

      // We need to write to "error-topic" to force timestamp re-extraction; since KStreams
      // extract timestamps only in the beginning of the topology and we set timestamps to error
      // messages during transform, we need to re-extract timestamps from windows in error messages.
      //
      // we are ok with duplicates from error-topic
      // TODO(apovzner) need to deal with lost messages from error-topic
      KStream<byte[], MonitoringMessage> errorInput = branches[1]
          .through(
              Serdes.ByteArray(),
              errorTopic.valueSerde,
              partitioner,
              errorTopic.name
          );

      KStream<byte[], MonitoringMessage> merged = builder.merge(input, errorInput);
      Map<Rollup, KTable<Windowed<Bytes>, MonitoringMessage>> expectedConsumptionAggr =
          this.consumerGroupStreamExtension.extend(merged);
      Map<Rollup, KTable<Windowed<Bytes>, MonitoringMessage>> actualConsumptionAggr =
          this.monitoringStreamExtension.extend(merged);
      this.groupStreamExtension.extend(merged);
      this.clusterStreamExtension.extend(merged);

      // temporary re-key until we change key types
      KStream<WindowedClusterGroup, MonitoringMessage> rekeyedCgExpectedConsumption =
          expectedConsumptionAggr.get(Rollup.FIFTEEN_SECONDS)
              .filter(groupMeasurement)
              .toStream(
                new KeyValueMapper<Windowed<Bytes>, MonitoringMessage, WindowedClusterGroup>() {
                @Override
                public WindowedClusterGroup apply(Windowed<Bytes> key,
                                                  MonitoringMessage value) {
                  // filtered out messages will come with null value
                  if (value != null) {
                    MonitoringMessage keyMM = keySerde.toProto(key.key());
                    return WindowedClusterGroup.newBuilder()
                        .setClusterId(keyMM.getClusterId())
                        .setGroup(keyMM.getGroup())
                        .setWindow(key.window().start()).build();
                  }

                  return WindowedClusterGroup.newBuilder()
                      .setWindow(key.window().start()).build();
                }
              }
          ).filter(new ValueNotNull<WindowedClusterGroup, MonitoringMessage>()
                   // Not doing verifiable through here, because in the worst case we will create
                   // two same measurements, which could create two same trigger events; and this
                   // is only if we have 0 or very small lag that we actually create two same
                   // trigger events.
          ).through(
              this.rekeyedExpectedGroupConsumptionTopic.keySerde,
              this.rekeyedExpectedGroupConsumptionTopic.valueSerde,
              this.rekeyedExpectedGroupConsumptionTopic.name
          );

      KStream<WindowedClusterGroup, MonitoringMessage> rekeyedCgActualConsumption =
          actualConsumptionAggr.get(Rollup.FIFTEEN_SECONDS)
              .filter(groupMeasurement)
              .toStream(
              new KeyValueMapper<Windowed<Bytes>, MonitoringMessage, WindowedClusterGroup>() {
                @Override
                public WindowedClusterGroup apply(Windowed<Bytes> key,
                                              MonitoringMessage value) {
                  // filtered out messages will come with null value
                  if (value != null) {
                    MonitoringMessage keyMM = keySerde.toProto(key.key());
                    return WindowedClusterGroup.newBuilder()
                        .setClusterId(keyMM.getClusterId())
                        .setGroup(keyMM.getGroup())
                        .setWindow(key.window().start()).build();
                  }
                  return WindowedClusterGroup.newBuilder()
                      .setWindow(key.window().start()).build();
                }
              }

          ).filter(new ValueNotNull<WindowedClusterGroup, MonitoringMessage>()
                   // Not doing verifiable through here, because in the worst case we will create
                   // two same measurements, which could create two same trigger events; and this
                   // is only if we have 0 or very small lag that we actually create two same
                   // trigger events.
          ).through(
              this.rekeyedActualGroupConsumptionTopic.keySerde,
              this.rekeyedActualGroupConsumptionTopic.valueSerde,
              this.rekeyedActualGroupConsumptionTopic.name
          );

      // TODO(apovzner) We are going to do join on ktables, so will not need to define JoinWindows
      final StreamPartitioner<Bytes, TriggerEvent> triggerTopicPartitioner = new StreamPartitioner<Bytes, TriggerEvent>() {
        @Override
        public Integer partition(Bytes key, TriggerEvent value, int numPartitions) {
          return TriggerEventUtil.partitionFor(key, value, numPartitions);
        }
      };
      KStream<Bytes, AlertInfo> alertInfoKStream = rekeyedCgActualConsumption.outerJoin(
          rekeyedCgExpectedConsumption,
          new MonitoringTriggerMeasurementJoiner(),
          JoinWindows.of(1000),
          this.rekeyedActualGroupConsumptionTopic.keySerde,
          this.rekeyedExpectedGroupConsumptionTopic.valueSerde,
          this.rekeyedActualGroupConsumptionTopic.valueSerde
      ).transform(
          monitoringTriggerTransformerSupplier,
          monitoringTriggerStore.name
      ).flatMap(
          new Identity<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>>()
          // Not doing verifiable through because in the worst case we send two same alerts
      ).through(
          rekeyedTriggerEventTopic.keySerde,
          rekeyedTriggerEventTopic.valueSerde,
          triggerTopicPartitioner,
          rekeyedTriggerEventTopic.name
      ).transform(
          alertTransformerSupplier,
          triggerActionsStore.name,
          triggerEventsStore.name
      ).flatMap(
          new Identity<Bytes, Iterable<KeyValue<Bytes, AlertInfo>>>()
      );
      alertInfoKStream.foreach(emailSender);
      alertInfoKStream.process(
          alertHistoryProcessorSupplier,
          alertHistoryStore.name
      );

      return new KafkaStreams(builder, streamsConfig);
    } catch (Throwable t) {
      log.error("creating streams", t);
    }

    return null;
  }
}
