package io.confluent.controlcenter.streams;

import com.google.common.collect.ImmutableSet;
import com.google.common.io.Files;
import com.google.inject.AbstractModule;
import com.google.inject.BindingAnnotation;
import com.google.inject.Inject;
import com.google.inject.Provides;
import com.google.inject.Singleton;
import io.confluent.controlcenter.ControlCenterConfig;
import io.confluent.controlcenter.ControlCenterConfigModule;
import io.confluent.controlcenter.serialization.MonitoringMessageSerde;
import io.confluent.controlcenter.serialization.WindowStoreSerde;
import io.confluent.monitoring.record.Monitoring;
import io.confluent.serializers.UberSerde;

import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.common.serialization.Serdes;
import org.apache.kafka.streams.StreamsConfig;

import java.io.IOException;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.Target;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Properties;

import static java.lang.annotation.RetentionPolicy.RUNTIME;

public class StreamsConfigModule extends AbstractModule {

  @Override
  protected void configure() {
  }

  @Provides
  @Inject
  @Singleton
  protected StreamsConfig provideMonitoringStreamsConfig(
      ControlCenterConfig controlCenterConfig,
      @ControlCenterConfigModule.EnableSubTopologies boolean enableSubTopologies,
      @StreamsStateDir Path streamsStateDir
      ) {
    Properties props = new Properties();
    props.put(
        StreamsConfig.BOOTSTRAP_SERVERS_CONFIG,
        controlCenterConfig.getList(ControlCenterConfig.CONTROL_CENTER_BOOTSTRAP_SERVERS_CONFIG)
    );
    props.put(
        StreamsConfig.ZOOKEEPER_CONNECT_CONFIG,
        controlCenterConfig.getString(ControlCenterConfig.ZOOKEEPER_CONNECT_CONFIG)
    );
    props.put(
        StreamsConfig.STATE_DIR_CONFIG,
        streamsStateDir.toString()
    );
    String streamsAppId = TopicStoreMaster.nameJoin(
        controlCenterConfig.getString(ControlCenterConfig.CONTROL_CENTER_NAME_CONFIG),
        controlCenterConfig.getString(ControlCenterConfig.CONTROL_CENTER_ID_CONFIG)
    );
    props.put(StreamsConfig.APPLICATION_ID_CONFIG, streamsAppId);
    props.put(StreamsConfig.TIMESTAMP_EXTRACTOR_CLASS_CONFIG, WindowExtractor.class);
    props.put(StreamsConfig.consumerPrefix(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG), "earliest");
    props.put(StreamsConfig.CACHE_MAX_BYTES_BUFFERING_CONFIG, 0);
    props.put(
        StreamsConfig.REPLICATION_FACTOR_CONFIG,
        controlCenterConfig.getInt(ControlCenterConfig.CONTROL_CENTER_INTERNAL_REPLICATION_CONFIG)
    );
    props.put(StreamsConfig.producerPrefix(ProducerConfig.ACKS_CONFIG), "all");
    props.putAll(controlCenterConfig.valuesWithPrefix(ControlCenterConfig.CONFLUENT_CONTROLCENTER_KSTREAMS_PREFIX));
    props.putAll(controlCenterConfig.originalsWithPrefix(ControlCenterConfig.CONFLUENT_CONTROLCENTER_KSTREAMS_PREFIX));
    if (enableSubTopologies) {
      props.put(
          StreamsConfig.PARTITION_GROUPER_CLASS_CONFIG,
          SingleGroupPartitionGrouper.class
      );
    }

    return new StreamsConfig(props);
  }

  @BindingAnnotation
  @Retention(RUNTIME)
  @Target({ElementType.FIELD, ElementType.PARAMETER, ElementType.METHOD})
  public @interface StreamsStateDir {

  }

  @Provides
  @Inject
  @StreamsStateDir
  protected Path getStreamsStateDir(ControlCenterConfig controlCenterConfig) throws IOException {
    Path streamsStateDir = Paths.get(
        controlCenterConfig.getString(ControlCenterConfig.CONTROL_CENTER_DATA_DIR),
        controlCenterConfig.getString(ControlCenterConfig.CONTROL_CENTER_ID_CONFIG),
        "kafka-streams"
    );
    Files.createParentDirs(streamsStateDir.toFile());
    return streamsStateDir;
  }
}
