// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.inject.AbstractModule;
import com.google.inject.Inject;
import com.google.inject.Provider;
import com.google.inject.Provides;
import com.google.inject.Singleton;

import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.KafkaStreams;
import org.apache.kafka.streams.StreamsConfig;
import org.apache.kafka.streams.processor.StateStoreSupplier;
import org.apache.kafka.streams.state.QueryableStoreTypes;
import org.apache.kafka.streams.state.ReadOnlyKeyValueStore;
import org.apache.kafka.streams.state.ReadOnlyWindowStore;
import org.apache.kafka.streams.state.Stores;

import java.util.HashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import io.confluent.command.record.Command.CommandKey;
import io.confluent.command.record.Command.CommandMessage;
import io.confluent.controlcenter.ControlCenterConfig;
import io.confluent.controlcenter.Rollup;
import io.confluent.controlcenter.alert.record.Alert.AlertInfo;
import io.confluent.controlcenter.command.CommandModule.ActionConfigs;
import io.confluent.controlcenter.command.CommandModule.TriggerConfigs;
import io.confluent.controlcenter.record.Controlcenter.WindowedClusterGroup;
import io.confluent.controlcenter.record.Controlcenter.Cluster;
import io.confluent.controlcenter.record.Controlcenter.TriggerActions;
import io.confluent.controlcenter.record.Controlcenter.TriggerEventKey;
import io.confluent.controlcenter.record.Controlcenter.TriggerMeasurement;
import io.confluent.controlcenter.record.Controlcenter.VerifiableMonitoringMessage;
import io.confluent.controlcenter.record.Controlcenter.VerifiableTriggerInfo;
import io.confluent.controlcenter.record.Controlcenter.WindowedGrouping;
import io.confluent.controlcenter.serialization.SerializationModule.TriggerEventsKeyPrefixSerde;
import io.confluent.controlcenter.serialization.SerializationModule.MonitoringTriggerKeySerde;
import io.confluent.controlcenter.serialization.SerializationModule.TriggerActionsKeySerde;
import io.confluent.controlcenter.serialization.SerializationModule.TriggerEventsKeySerde;
import io.confluent.controlcenter.serialization.SerializationModule.VerifierKeySerde;
import io.confluent.controlcenter.streams.TopicStoreMaster.Store;
import io.confluent.controlcenter.streams.TopicStoreMaster.Topic;
import io.confluent.controlcenter.streams.TopicStoreModule.ClusterStore;
import io.confluent.controlcenter.streams.TopicStoreModule.GroupStore;
import io.confluent.controlcenter.streams.TopicStoreModule.MonitoringMessageAggregatorWindowsStore;
import io.confluent.controlcenter.streams.TopicStoreModule.MonitoringStreamStore;
import io.confluent.controlcenter.streams.TopicStoreModule.MonitoringTriggerStore;
import io.confluent.controlcenter.streams.TopicStoreModule.TriggerActionsStore;
import io.confluent.controlcenter.streams.TopicStoreModule.TriggerEventsStore;
import io.confluent.controlcenter.streams.TopicStoreModule.VerifierStore;
import io.confluent.controlcenter.streams.TopicStoreModule.AlertHistoryStore;
import io.confluent.controlcenter.streams.TopicStoreModule.ExpectedGroupConsumptionAggregateTopic;
import io.confluent.controlcenter.streams.TopicStoreModule.ActualGroupConsumptionAggregateTopic;
import io.confluent.controlcenter.streams.alert.AlertTransformerSupplier;
import io.confluent.controlcenter.streams.alert.MonitoringTriggerTransformerSupplier;
import io.confluent.controlcenter.streams.alert.AlertHistoryProcessorSupplier;
import io.confluent.controlcenter.streams.verify.MonitoringHeartbeatSender;
import io.confluent.controlcenter.streams.verify.MonitoringVerifier;
import io.confluent.controlcenter.streams.verify.VerifyTransformerSupplier;
import io.confluent.monitoring.common.Clock;
import io.confluent.monitoring.common.MonitoringMessageUtil;
import io.confluent.monitoring.common.MonitoringProducerDefaults;
import io.confluent.monitoring.common.SystemClock;
import io.confluent.monitoring.common.TimeBucket;
import io.confluent.monitoring.record.Monitoring.ClientType;
import io.confluent.monitoring.record.Monitoring.MessageType;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.serializers.OrderedKeyUberSerde;
import io.confluent.serializers.UberSerde;

public class StreamsModule extends AbstractModule {

  @Override
  protected void configure() {
    bind(KafkaStreams.class).toProvider(StreamProvider.class).asEagerSingleton();
  }

  @Provides
  @GroupStore
  @Inject
  protected Map<Rollup, ReadOnlyWindowStore<Bytes, WindowedGrouping>> getGroupStore(
      @GroupStore Store<Bytes, WindowedGrouping, WindowedGrouping> groupStore,
      KafkaStreams kstream
  ) {
    return getRollupStores(groupStore, kstream);
  }

  @Provides
  @ClusterStore
  @Inject
  protected ReadOnlyKeyValueStore<Bytes, Cluster> getClusterStore(
      @ClusterStore Store<Bytes, Cluster, Cluster> clusterStore,
      KafkaStreams kstream
  ) {
    return kstream.store(clusterStore.name, QueryableStoreTypes.<Bytes, Cluster>keyValueStore());
  }

  @Provides
  @MonitoringStreamStore
  @Inject
  protected Map<Rollup, ReadOnlyWindowStore<Bytes, MonitoringMessage>> getMonitoringStreamStore(
      @MonitoringStreamStore Store<Bytes, MonitoringMessage, MonitoringMessage> monitoringStreamStore,
      KafkaStreams kstream
  ) {
    return getRollupStores(monitoringStreamStore, kstream);
  }

  @Provides
  @MonitoringMessageAggregatorWindowsStore
  @Inject
  protected Map<Rollup, ReadOnlyWindowStore<Bytes, MonitoringMessage>> getMonitoringMessageAggregatorWindowsStore(
      @MonitoringMessageAggregatorWindowsStore Store<Bytes, MonitoringMessage, MonitoringMessage> monitoringMessageAggregatorWindowsStore,
      KafkaStreams kstream
  ) {
    return getRollupStores(monitoringMessageAggregatorWindowsStore, kstream);
  }


  static <K, V> Map<Rollup, ReadOnlyWindowStore<K, V>> getRollupStores(
      Store<K, ?, V> groupStore,
      KafkaStreams kstream
  ) {
    ImmutableMap.Builder<Rollup, ReadOnlyWindowStore<K, V>> builder = ImmutableMap.builder();
    for (Rollup rollup : Rollup.values()) {
      builder.put(
          rollup,
          kstream.store(
              TopicStoreMaster.nameJoin(groupStore.name, rollup.name()),
              QueryableStoreTypes.<K, V>windowStore()
          )
      );
    }
    return builder.build();
  }

  @Provides
  @AlertHistoryStore
  @Inject
  protected ReadOnlyKeyValueStore<Bytes, AlertInfo> getAlertHistoryStore(
      @AlertHistoryStore Store<Bytes, AlertInfo, AlertInfo> alertHistoryStore,
      KafkaStreams kstream
  ) {
    return kstream.store(alertHistoryStore.name, QueryableStoreTypes.<Bytes, AlertInfo>keyValueStore());
  }

  @Provides
  @Inject
  @Singleton
  protected Set<StateStoreSupplier> provideStoreSuppliers(
      @VerifierStore Store<Bytes, MonitoringVerifier, MonitoringVerifier> verifierStore,
      @MonitoringTriggerStore Store<Bytes, TriggerMeasurement, TriggerMeasurement> monitoringTriggerStore,
      @TriggerActionsStore Store<Bytes, TriggerActions, TriggerActions> triggerActionsStore,
      @TriggerEventsStore Store<Bytes, VerifiableTriggerInfo, VerifiableTriggerInfo> triggerEventsStore,
      @AlertHistoryStore Store<Bytes, AlertInfo, AlertInfo> alertHistoryStore,
      Set<Topic<?, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage>> verifiableTopics
  ) {
    Set<StateStoreSupplier> suppliers = new HashSet<>();
    ImmutableList<Store<Bytes, ?, ?>> transformerStores = ImmutableList.of(verifierStore,
                                                                           monitoringTriggerStore,
                                                                           triggerActionsStore,
                                                                           triggerEventsStore,
                                                                           alertHistoryStore);
    for (Store<Bytes, ?, ?> transformerStore: transformerStores) {
      suppliers.add(Stores.create(transformerStore.name)
                        .withKeys(transformerStore.keySerde)
                        .withValues(transformerStore.valueSerde)
                        .persistent()
                        .build());
    }

    for (Topic<?, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage> topic : verifiableTopics) {
      Set<String> storeNames = new HashSet<>();
      if (topic.rollup) {
        for (Rollup rollup : Rollup.values()) {
          storeNames.add(TopicStoreMaster.nameJoin(topic.store.name, rollup.name()));
        }
      } else {
        storeNames.add(topic.store.name);
      }
      for (String storeName : storeNames) {
        suppliers.add(
            Stores.create(storeName)
                .withKeys(topic.store.keySerde)
                .withValues(topic.store.valueSerde)
                .inMemory()
                .maxEntries(2 << 13)
                .build());
      }
    }
    return suppliers;
  }

  @Provides
  @Inject
  @Singleton
  protected VerifyTransformerSupplier provideVerifyTransformerSupplier(
      ControlCenterConfig config,
      @VerifierStore Store<Bytes, MonitoringVerifier, MonitoringVerifier> verifierStore,
      @VerifierKeySerde OrderedKeyUberSerde<MonitoringMessage> keySerde,
      Clock clock
  ) {
    return new VerifyTransformerSupplier(
        verifierStore.name,
        config.getInt(ControlCenterConfig.CONTROL_CENTER_MONITORING_TOPIC_PARTITIONS_CONFIG),
        clock,
        keySerde
    );
  }

  @Provides
  @Inject
  @Singleton
  protected MonitoringTriggerTransformerSupplier provideMonitoringTriggerTransformerSupplier(
      ControlCenterConfig config,
      @MonitoringTriggerStore Store<Bytes, TriggerMeasurement, TriggerMeasurement> monitoringTriggerStore,
      @MonitoringTriggerKeySerde OrderedKeyUberSerde<WindowedClusterGroup> keySerde,
      @TriggerConfigs Provider<ReadOnlyKeyValueStore<CommandKey, CommandMessage>> triggerConfigsProvider,
      Clock clock
  ) {
    return new MonitoringTriggerTransformerSupplier(
        monitoringTriggerStore.name,
        clock,
        keySerde,
        triggerConfigsProvider,
        config.getInt(ControlCenterConfig.CONTROL_CENTER_INTERNAL_PARTITIONS_CONFIG)
    );
  }

  @Provides
  @Inject
  @Singleton
  protected AlertTransformerSupplier provideAlertTransformerSupplier(
      ControlCenterConfig config,
      @TriggerActionsStore Store<Bytes, TriggerActions, TriggerActions> triggerActionsStore,
      @TriggerEventsStore Store<Bytes, VerifiableTriggerInfo, VerifiableTriggerInfo> triggerEventsStore,
      @TriggerActionsKeySerde OrderedKeyUberSerde<TriggerEventKey> actionsKeySerde,
      @TriggerEventsKeySerde OrderedKeyUberSerde<AlertInfo> eventsKeySerde,
      @TriggerConfigs Provider<ReadOnlyKeyValueStore<CommandKey, CommandMessage>> triggerConfigsProvider,
      @ActionConfigs Provider<ReadOnlyKeyValueStore<CommandKey, CommandMessage>> actionConfigsProvider,
      Clock clock
  ) {
    return new AlertTransformerSupplier(
        triggerActionsStore.name,
        triggerEventsStore.name,
        config.getInt(ControlCenterConfig.CONTROL_CENTER_MAX_TRIGGER_EVENTS_PER_ALERT_CONFIG),
        clock,
        actionsKeySerde,
        eventsKeySerde,
        triggerConfigsProvider,
        actionConfigsProvider
    );
  }

  @Provides
  @Inject
  @Singleton
  protected AlertHistoryProcessorSupplier provideAlertHistoryProcessorSupplier(
      @AlertHistoryStore Store<Bytes, AlertInfo, AlertInfo> alertHistoryStore,
      @TriggerEventsKeySerde OrderedKeyUberSerde<AlertInfo> alertKeySerde,
      @TriggerEventsKeyPrefixSerde OrderedKeyUberSerde<AlertInfo> alertRangeSerde
  ) {
    return new AlertHistoryProcessorSupplier(
        alertHistoryStore.name,
        alertKeySerde,
        alertRangeSerde
    );
  }

  @Provides
  @Inject
  @Singleton
  protected Clock provideClock() {
    return new SystemClock();
  }

  @Provides
  @Singleton
  @Inject
  protected MonitoringHeartbeatSender provideMonitoringHeartbeatSender(
      ControlCenterConfig cfg,
      StreamsConfig streamsConfig,
      UberSerde<MonitoringMessage> monitoringMessageUberSerde,
      @ExpectedGroupConsumptionAggregateTopic Topic<WindowedClusterGroup, MonitoringMessage, Void, Void> rekeyedExpectedGroupConsumptionTopic,
      @ActualGroupConsumptionAggregateTopic Topic<WindowedClusterGroup, MonitoringMessage, Void, Void> rekeyedActualGroupConsumptionTopic,
      Clock clock
      ) {
    MonitoringMessage baseMonitoringMessage = provideBaseHeartbeatMonitoringMessage(
        cfg.getString(ControlCenterConfig.CONTROL_CENTER_ID_CONFIG)
    );

    Map<String, Object> producerProperties = streamsConfig.getProducerConfigs(
        "confluent-control-center-heartbeat-sender-" +
        baseMonitoringMessage.getClientId()
    );
    // this will get us the defaults for kstreams
    producerProperties.put(
        ProducerConfig.ACKS_CONFIG,
        MonitoringProducerDefaults.ACKS_CONFIG
    );
    producerProperties.put(
        ProducerConfig.KEY_SERIALIZER_CLASS_CONFIG,
        MonitoringProducerDefaults.KEY_SERIALIZER_CLASS_CONFIG
    );
    producerProperties.put(
        ProducerConfig.VALUE_SERIALIZER_CLASS_CONFIG,
        MonitoringProducerDefaults.VALUE_SERIALIZER_CLASS_CONFIG
    );
    producerProperties.put(
        ProducerConfig.LINGER_MS_CONFIG,
        MonitoringProducerDefaults.LINGER_MS_CONFIG
    );
    producerProperties.put(
        ProducerConfig.BOOTSTRAP_SERVERS_CONFIG,
        cfg.getList(ControlCenterConfig.CONTROL_CENTER_BOOTSTRAP_SERVERS_CONFIG)
    );
    KafkaProducer kp = new KafkaProducer<>(producerProperties);
    Set<Topic<WindowedClusterGroup, MonitoringMessage, Void, Void>>
        internalPublishTopics = new HashSet<>();
    internalPublishTopics.add(rekeyedExpectedGroupConsumptionTopic);
    internalPublishTopics.add(rekeyedActualGroupConsumptionTopic);
    return new MonitoringHeartbeatSender(
        cfg.getString(ControlCenterConfig.CONTROL_CENTER_TOPIC_CONFIG),
        internalPublishTopics,
        baseMonitoringMessage,
        kp,
        monitoringMessageUberSerde,
        TimeBucket.SIZE,
        clock
    );
  }

  public static MonitoringMessage provideBaseHeartbeatMonitoringMessage(String clientId) {
    return MonitoringMessage
        .newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setClientType(ClientType.CONTROLCENTER)
        .setClientId(clientId)
        .setType(MessageType.HEARTBEAT)
        .setWindow(0)
        .setTopic("")
        .buildPartial();
  }
}
