// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams;

import com.google.common.base.Joiner;
import com.google.common.collect.ImmutableSet;
import com.google.inject.Inject;

import io.confluent.command.record.Command;
import io.confluent.command.record.Command.CommandKey;
import io.confluent.command.record.Command.CommandMessage;
import io.confluent.controlcenter.streams.TopicStoreModule.CommandTopic;
import org.apache.kafka.streams.StreamsConfig;
import org.apache.kafka.streams.state.QueryableStoreType;
import org.apache.kafka.streams.state.QueryableStoreTypes;
import org.apache.kafka.streams.state.ReadOnlyKeyValueStore;
import org.apache.kafka.streams.state.ReadOnlyWindowStore;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import io.confluent.controlcenter.Rollup;
import io.confluent.controlcenter.serialization.WindowStoreSerde;
import io.confluent.controlcenter.streams.TopicStoreModule.AllStores;
import io.confluent.controlcenter.streams.TopicStoreModule.AllTopics;
import io.confluent.controlcenter.streams.TopicStoreModule.MonitoringTopic;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.serializers.UberSerde;

/**
 * A central location for topics, stores, and serdes
 */
public class TopicStoreMaster {

  public static class SerdePair {
    public UberSerde keySerde;
    public UberSerde valueSerde;
    public WindowStoreSerde windowedSerde;

    SerdePair(UberSerde keySerde, UberSerde valueSerde) {
      this(keySerde, valueSerde, null);
    }

    SerdePair(UberSerde keySerde, UberSerde valueSerde, WindowStoreSerde windowedSerde) {
      this.keySerde = keySerde;
      this.valueSerde = valueSerde;
      this.windowedSerde = windowedSerde;
    }
  }

  public static class Topic<K, V, SK, ST> {
    public String name;
    public Store<SK, V, ST> store;
    public UberSerde<K> keySerde;
    public UberSerde<V> valueSerde;
    public boolean rollup;

    Topic(String name, Store<SK, V, ST> store, UberSerde<K> keySerde, UberSerde<V> valueSerde, boolean rollup) {
      this.name = name;
      this.store = store;
      this.keySerde = keySerde;
      this.valueSerde = valueSerde;
      this.rollup = rollup;
    }
  }

  public static class Store<K, V, T> {
    public String name;
    public boolean rollup;
    public UberSerde<K> keySerde;
    public UberSerde<V> valueSerde;
    public UberSerde<T> aggregateSerde;
    public WindowStoreSerde windowedSerde;
    public QueryableStoreType<ReadOnlyKeyValueStore<K, V>> queryableKeyValueStoreType;
    public QueryableStoreType<ReadOnlyWindowStore<K, V>> queryableWindowStoreType;

    Store(String name, UberSerde<K> keySerde, UberSerde<V> valueSerde, UberSerde<T> aggregateSerde, boolean rollup) {
      this(name, keySerde, valueSerde, aggregateSerde, null, rollup);
    }

    Store(String name, UberSerde<K> keySerde, UberSerde<V> valueSerde, UberSerde<T> aggregateSerde, WindowStoreSerde windowedSerde, boolean rollup) {
      this.name = name;
      this.keySerde = keySerde;
      this.valueSerde = valueSerde;
      this.aggregateSerde = aggregateSerde;
      this.windowedSerde = windowedSerde;
      this.rollup = rollup;
      if (windowedSerde == null) {
        this.queryableKeyValueStoreType = QueryableStoreTypes.keyValueStore();;
      } else {
        this.queryableWindowStoreType = QueryableStoreTypes.windowStore();
      }
    }
  }

  private final String streamsAppId;
  private final Topic<Void, MonitoringMessage, Void, Void> monitoringTopic;
  private final Topic<CommandKey, CommandMessage, Void, Void> commandTopic;
  private final Map<String, Topic> topicMap = new HashMap<>();
  private final Map<String, Store> storeMap = new HashMap<>();
  private final Map<String, Store> windowedStoreMap = new HashMap<>();

  @Inject
  public TopicStoreMaster(
      StreamsConfig streamsConfig,
      @MonitoringTopic Topic<Void, MonitoringMessage, Void, Void> monitoringTopic,
      @CommandTopic Topic<CommandKey, CommandMessage, Void, Void> commandTopic,
      @AllTopics ImmutableSet<Topic> topics,
      @AllStores ImmutableSet<Store> stores
  ) {
    this.streamsAppId = streamsConfig.getString(StreamsConfig.APPLICATION_ID_CONFIG);
    this.monitoringTopic = monitoringTopic;
    this.commandTopic = commandTopic;

    for (Topic topic : topics) {
      registerTopic(topic);
    }
    for (Store store : stores) {
      registerStore(store);
    }
  }

  private void registerTopic(Topic topic) {
    if (topic.rollup) {
      for (Rollup rollup : Rollup.values()) {
        registerTopic(nameJoin(topic.name, rollup.name()), topic);
      }
    }
    else {
      registerTopic(topic.name, topic);
    }
    if (topic.store != null) {
      registerStore(topic.store);
    }
  }

  private void registerTopic(String name, Topic topic) {
    topicMap.put(name, topic);
  }

  private void registerStore(Store store) {
    String prefixedName = nameJoin(streamsAppId, store.name);
    if (store.rollup) {
      for (Rollup rollup : Rollup.values()) {
        registerStore(nameJoin(prefixedName, rollup.name()), store);
      }
    } else {
      registerStore(prefixedName, store);
    }
  }

  private void registerStore(String prefixedName, Store store) {
    if (store.windowedSerde != null) {
      windowedStoreMap.put(nameJoin(prefixedName, "changelog"), store);
    } else {
      storeMap.put(nameJoin(prefixedName, "changelog"), store);
    }
  }

  public static String nameJoin(String... parts) {
    return Joiner.on("-").join(parts);
  }

  public ImmutableSet<String> getInputTopicNames() {
    ImmutableSet.Builder<String> builder = ImmutableSet.builder();
    builder.add(monitoringTopic.name);
    builder.add(commandTopic.name);
    return builder.build();
  }

  public ImmutableSet<String> getIntermediateTopicNames() {
    Set<String> names = new HashSet<>();
    names.addAll(topicMap.keySet());
    names.removeAll(getInputTopicNames());
    return ImmutableSet.copyOf(names);
  }

  public ImmutableSet<String> getStoreTopicNames() {
    return ImmutableSet.<String>builder()
        .addAll(storeMap.keySet())
        .build();
  }

  public ImmutableSet<String> getWindowedStoreTopicNames() {
    return ImmutableSet.<String>builder()
        .addAll(windowedStoreMap.keySet())
        .build();
  }

  public ImmutableSet<String> getInternalTopicNames() {
    ImmutableSet.Builder<String> builder = ImmutableSet.builder();
    builder.addAll(getIntermediateTopicNames());
    builder.addAll(storeMap.keySet());
    builder.addAll(windowedStoreMap.keySet());
    return builder.build();
  }

  public SerdePair getSerdesForTopic(String topicName) {
    if (topicName.endsWith("-changelog")) {
      if (storeMap.containsKey(topicName)) {
        Store store = storeMap.get(topicName);
        return new SerdePair(store.keySerde, store.aggregateSerde, store.windowedSerde);
      }
      if (windowedStoreMap.containsKey(topicName)) {
        Store store = windowedStoreMap.get(topicName);
        return new SerdePair(store.keySerde, store.aggregateSerde, store.windowedSerde);
      }
    }
    if (topicMap.containsKey(topicName)) {
      Topic topic = topicMap.get(topicName);
      return new SerdePair(topic.keySerde, topic.valueSerde);
    }
    return null;
  }

  public static String getInternalTopicName(StreamsConfig streamsConfig, String topic) {
    String streamsAppId = streamsConfig.getString(StreamsConfig.APPLICATION_ID_CONFIG);
    return nameJoin(streamsAppId, topic);
  }
}
