package io.confluent.controlcenter.streams.alert;

import com.google.common.base.Joiner;
import com.google.common.collect.Iterables;
import com.google.inject.Provider;

import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.kstream.Transformer;
import org.apache.kafka.streams.kstream.TransformerSupplier;
import org.apache.kafka.streams.processor.ProcessorContext;
import org.apache.kafka.streams.state.KeyValueIterator;
import org.apache.kafka.streams.state.KeyValueStore;
import org.apache.kafka.streams.state.ReadOnlyKeyValueStore;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

import io.confluent.command.record.Command.CommandKey;
import io.confluent.command.record.Command.CommandMessage;
import io.confluent.command.record.alert.CommandAlert.MonitoringTriggerConfig;
import io.confluent.controlcenter.alert.record.Alert.TriggerInfo;
import io.confluent.controlcenter.record.Controlcenter.ClientTriggerMeasurement;
import io.confluent.controlcenter.record.Controlcenter.TriggerEvent;
import io.confluent.controlcenter.record.Controlcenter.TriggerMeasurement;
import io.confluent.controlcenter.record.Controlcenter.WindowedClusterGroup;
import io.confluent.controlcenter.util.KVQuery;
import io.confluent.monitoring.common.Clock;
import io.confluent.monitoring.common.TimeBucket;
import io.confluent.serializers.OrderedKeyUberSerde;


/**
 * This transformer takes trigger measurements taken from joining expected and actual
 * consumption streams and emits triggers if measurements are in triggerable state after and
 * a requested lag time.
 *
 * Input: key = Windowed clusterID:consumerGroup
 *        value = TriggerMeasurement
 * Output: key (Bytes) = clusterID:consumerGroup
 *        value = TriggerEvent
 */
public class MonitoringTriggerTransformerSupplier implements
   TransformerSupplier<WindowedClusterGroup, TriggerMeasurement, KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>>> {

  private static final Logger log = LoggerFactory.getLogger(MonitoringTriggerTransformerSupplier.class);

  private static final long PUNCTUATE_INTERVAL_MS = TimeBucket.SIZE;
  private static final long DEFAULT_MAX_PUNCTUATE_MS = TimeUnit.SECONDS.toMillis(1);
  private static final long MAX_WAIT_TO_MOVE_C3_CLOCK = 8 * TimeBucket.SIZE;

  private final String storeName;
  private final Clock clock;
  private final OrderedKeyUberSerde<WindowedClusterGroup> keySerde;
  private final Provider<? extends ReadOnlyKeyValueStore<CommandKey, CommandMessage>> triggerConfigsProvider;
  private final int numInternalPartitions;

  public MonitoringTriggerTransformerSupplier(
      String storeName,
      Clock clock,
      OrderedKeyUberSerde<WindowedClusterGroup> keySerde,
      Provider<? extends ReadOnlyKeyValueStore<CommandKey, CommandMessage>> triggerConfigsProvider,
      int numInternalPartitions
  ) {
    this.storeName = storeName;
    this.clock = clock;
    this.keySerde = keySerde;
    this.triggerConfigsProvider = triggerConfigsProvider;
    this.numInternalPartitions = numInternalPartitions;
  }

  @Override
  public Transformer<WindowedClusterGroup, TriggerMeasurement, KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>>> get() {
    return new MonitoringTriggerTransformer(
        PUNCTUATE_INTERVAL_MS,
        DEFAULT_MAX_PUNCTUATE_MS,
        storeName,
        triggerConfigsProvider
    );
  }

  private class MonitoringTriggerTransformer implements
      Transformer<WindowedClusterGroup, TriggerMeasurement, KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>>> {
    //TODO - remove this when we have rangeFrom() support in Kafka Streams
    private final Bytes LAST_KEY = Bytes.wrap(new byte[] { (byte)0xff });

    private final String storeName;
    private final long maxPunctuateMs;
    private final long punctuateIntervalMs;
    private final Provider<? extends ReadOnlyKeyValueStore<CommandKey, CommandMessage>> triggerConfigsProvider;

    private ProcessorContext context;
    // Stores measurements that are in triggerable state at the time of arrival to the transformer
    // key = ClusterID:ConsumerGroup:Window
    private KeyValueStore<Bytes, TriggerMeasurement> kvStore;
    private long lastPunctuateTimeMs = -1;
    /**
     * Approximated C3 time based on arrivalTime of messages; C3 time moves only when arrivalTime
     * of next message is greater than approxC3Time. Or, if MAX_WAIT_TO_MOVE_C3_CLOCK wallclock
     * time passed since we moved approxC3Time, then
     * approxC3Time += MAX_WAIT_TO_MOVE_C3_CLOCK
     */
    private long approxC3Time = 0;
    // when we move approxC3Time either because of arrived message or MAX_WAIT_TO_MOVE_C3_CLOCK
    // time passed since the last update, lastTimeMoveC3Clock = current wallclock time
    private long lastTimeMoveC3Clock = -1;
    private Bytes lastPunctuateKey = null;

    public MonitoringTriggerTransformer(
        long punctuateIntervalMs,
        long maxPunctuateMs,
        String storeName,
        Provider<? extends ReadOnlyKeyValueStore<CommandKey, CommandMessage>> triggerConfigsProvider
    ) {
      this.punctuateIntervalMs = punctuateIntervalMs;
      this.maxPunctuateMs = maxPunctuateMs;
      this.storeName = storeName;
      this.triggerConfigsProvider = triggerConfigsProvider;
    }

    @SuppressWarnings("unchecked")
    @Override
    public void init(ProcessorContext context) {
      this.context = context;
      kvStore = (KeyValueStore<Bytes, TriggerMeasurement>) this.context.getStateStore(storeName);
    }

    @Override
    public KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>> transform(
        WindowedClusterGroup key,
        TriggerMeasurement triggerMeasurement
    ) {
      if (triggerMeasurement == null ||
          (!triggerMeasurement.hasClientMeasurement() && !triggerMeasurement.getHeartbeat())) {
        log.trace("Ignoring null trigger measurement ");
        return null;
      }

      Bytes storeKey = keySerde.key(key);
      if (triggerMeasurement.getHeartbeat()) {
        // in case no messages arrive for long time, try to punctuate
        log.trace("Monitoring measurement C3 heartbeat");
        updateC3Clock(null, clock.currentTimeMillis());
        List<KeyValue<Bytes, TriggerEvent>> ret = new ArrayList<>();
        // pass heartbeat downstream to every partition of re-keyed trigger event topic
        for (int partition = 0; partition < numInternalPartitions; ++partition) {
          TriggerEvent msg = TriggerEvent.newBuilder()
              .setHeartbeat(true)
              .setInternalTopicPartition(partition)
              .setTimestamp(approxC3Time)
              .setWindow(triggerMeasurement.getWindow())
              .build();
          ret.add(KeyValue.pair(Bytes.wrap(new byte[]{}), msg));
        }
        // possibly pass alerts if time expired
        KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>> pair = maybePunctuate();
        if (pair != null) {
          Iterables.addAll(ret, pair.value);
        }

        return KeyValue.pair(
            Bytes.wrap(new byte[]{}),
            (Iterable<KeyValue<Bytes, TriggerEvent>>) ret
        );
      }

      if (log.isTraceEnabled()) {
        log.trace("trigger transform storeKey={} measurement={}", storeKey,
                  logMeasurement(triggerMeasurement));
      }

      // update c3 clock based on arrival time of a trigger measurement
      updateC3Clock(triggerMeasurement.getClientMeasurement().getArrivalTime(), clock.currentTimeMillis());

      try {
        // each measurement is an update since it comes from KTable, so either delete if value is
        // null or not in triggerable state, or update the store
        TriggerMeasurement storedMeasurement = kvStore.get(storeKey);
        // make sure that we store arrival time of the first measurement
        long arrivalTime = Math.min(triggerMeasurement.getClientMeasurement().getArrivalTime(),
                                    storedMeasurement == null ? Long.MAX_VALUE :
                                    storedMeasurement.getClientMeasurement().getArrivalTime());

        triggerMeasurement = TriggerMeasurement.newBuilder(triggerMeasurement)
            .setGroup(key.getGroup())
            .setClusterId(key.getClusterId())
            .setWindow(key.getWindow())
            .setClientMeasurement(
                ClientTriggerMeasurement.newBuilder(triggerMeasurement.getClientMeasurement()).setArrivalTime(arrivalTime))
            .build();

        if (isTriggerable(triggerMeasurement)) {
          if (log.isTraceEnabled()) {
            log.trace("Measurement for key={} is in triggerable state: {}",
                      storeKey, logMeasurement(triggerMeasurement));
          }
          kvStore.put(storeKey, triggerMeasurement);
        } else {
          kvStore.delete(storeKey);
        }
      } catch (Exception e) {
        log.warn("Failed to update trigger store. May not fire a trigger when needed.", e);
      }

      return maybePunctuate();
    }

    @Override
    public KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>> punctuate(long timestamp) {
      // we don't want KStreams to call punctuate
      return null;
    }

    private KeyValue<Bytes, Iterable<KeyValue<Bytes, TriggerEvent>>> maybePunctuate() {
      // timestamp is either last known message arrival time or last known message arrival time +
      // wait time (approximated arrival time when we don't move C3 clock for too long
      if (lastPunctuateTimeMs < 0) {
        // we will do the first punctuate after 'punctuate interval'
        lastPunctuateTimeMs = approxC3Time;
      }
      if (approxC3Time - lastPunctuateTimeMs < punctuateIntervalMs) {
        // we don't want to punctuate more often than 'punctuate interval'
        return null;
      }

      lastPunctuateTimeMs = approxC3Time;
      long punctuateTimeout = clock.currentTimeMillis() + maxPunctuateMs;

      List<KeyValue<Bytes, TriggerEvent>> triggerList = new ArrayList<>();
      try(KeyValueIterator<Bytes, TriggerMeasurement> iter = KVQuery.rangeFrom(kvStore, lastPunctuateKey)) {
        long count = 0;
        while (iter.hasNext()) {
          KeyValue<Bytes, TriggerMeasurement> entry = iter.next();
          if (entry == null) {
            continue;
          }
          if (entry.value == null) {
            log.debug("removing null value key={} from store={}", entry.key, storeName);
            kvStore.delete(entry.key);
          } else {
            if (log.isTraceEnabled()) {
              log.trace("checking key={} measurement={} store={}", entry.key,
                        logMeasurement(entry.value), storeName);
            }

            // fire triggers based on current time
            if (createTriggerEventsIfTriggerable(entry.key, entry.value, triggerList, approxC3Time)) {
              kvStore.delete(entry.key);
            } else {
              kvStore.put(entry.key, TriggerMeasurement.newBuilder(entry.value).
                  setLastTriggerEventTimestamp(approxC3Time).build());
            }
          }
          lastPunctuateKey = entry.key;
          count++;
          if (clock.currentTimeMillis() > punctuateTimeout) {
            log.debug("stopping punctuate at key={} store={}", lastPunctuateKey, storeName);
            break;
          }
          if (!iter.hasNext()) {
            log.debug("finished all punctuation for store={}", storeName);
            lastPunctuateKey = null;
          }
          log.debug("punctuated on count={} records", count);
        }
      } catch (Exception e) {
        log.error("failure during punctuate for store={} e={}", storeName, e.getMessage());
      }

      if (triggerList.isEmpty()) {
        return null;
      }

      // the key of KeyValue does not matter, we are going to flat map and use key of trigger event
      return KeyValue.pair(
          Bytes.wrap(new byte[]{}),
          (Iterable<KeyValue<Bytes, TriggerEvent>>) triggerList
      );
    }

    @Override
    public void close() {
      log.info("closing store={}", storeName);
    }

    /**
     * Updates C3 clock (approximation) based on message arrival time (if method is called on
     * arrival of message) and current time.
     * @param arrivalTime arrivalTime field of a message if called on message arrival, or null if
     *                    called on C3 heartbeat.
     * @param currentTime current wallclock time
     */
    private void updateC3Clock(Long arrivalTime, Long currentTime) {
      if (arrivalTime != null) {
        if (arrivalTime > approxC3Time) {
          approxC3Time = arrivalTime;
          lastTimeMoveC3Clock = currentTime;
          log.trace("Moved C3 clock based on arrival time: {}", approxC3Time);
        }
      }
      // wait for some time until we see the first message with an arrival time set
      if (lastTimeMoveC3Clock < 0) {
        lastTimeMoveC3Clock = currentTime;
      }
      // check if too much real time passed since we moved C3 clock
      if (lastTimeMoveC3Clock + MAX_WAIT_TO_MOVE_C3_CLOCK < currentTime) {
        approxC3Time += (currentTime - lastTimeMoveC3Clock);
        lastTimeMoveC3Clock = currentTime;
        log.trace("Moved C3 clock based on wallclock time: {}", approxC3Time);
      }
    }

    /**
     * Returns true if measurement is triggerable on at least one metric.
     */
    private boolean isTriggerable(TriggerMeasurement triggerMeasurement) {
      try (KeyValueIterator<CommandKey, CommandMessage> iter = triggerConfigsProvider.get().all()) {
        while (iter.hasNext()) {
          MonitoringTriggerConfig triggerConfig = iter.next().value.getMonitoringTriggerConfig();
          if (isMeasurementTriggerable(triggerMeasurement, triggerConfig)) {
            return true;
          }
        }
      } catch (Exception e) {
        log.error("Failure while checking if event is triggerable.", e);
      }
      return false;
    }

    private boolean isMeasurementTriggerable(TriggerMeasurement triggerMeasurement,
                                             MonitoringTriggerConfig triggerConfig) {
      if (triggerConfig != null &&
          triggerConfig.getClusterId().equals(triggerMeasurement.getClusterId()) &&
          triggerConfig.getGroup().equals(triggerMeasurement.getGroup())) {
        Long measurementValue;
        switch (triggerConfig.getMetric()) {
          case CONSUMPTION_DIFF:
            measurementValue = triggerMeasurement.getClientMeasurement().getConsumptionDiff();
            break;
          case AVG_LATENCY:
            measurementValue = triggerMeasurement.getClientMeasurement().getAvgLatency();
            break;
          case MAX_LATENCY:
            measurementValue = triggerMeasurement.getClientMeasurement().getMaxLatency();
            break;
          default:
            return false;
        }
        switch (triggerConfig.getCondition()) {
          case GREATER_THAN:
            return (measurementValue > triggerConfig.getLongValue());
          case LESS_THAN:
            return (measurementValue < triggerConfig.getLongValue());
          case EQUAL:
            return (measurementValue == triggerConfig.getLongValue());
          case NOT_EQUAL:
            return (measurementValue != triggerConfig.getLongValue());
        }
      }
      return false;
    }

    /**
     * Checks if measurement is triggerable, and creates a trigger event for each metric that is
     * triggerable. Created trigger events are added to triggerList.
     * @return true there will be no trigger events created for this measurement either because
     * trigger events were already created or measurement not triggerable anymore
     */
    private boolean createTriggerEventsIfTriggerable(Bytes measurementKey,
                                                     TriggerMeasurement triggerMeasurement,
                                                     List<KeyValue<Bytes, TriggerEvent>> triggerList,
                                                     long now) {
      boolean waitForAtLeastOneTrigger = false;
      try (KeyValueIterator<CommandKey, CommandMessage> iter = triggerConfigsProvider.get().all()) {
        while (iter.hasNext()) {
          MonitoringTriggerConfig triggerConfig = iter.next().value.getMonitoringTriggerConfig();
          if (isMeasurementTriggerable(triggerMeasurement, triggerConfig)) {
            Long deadline = triggerMeasurement.getClientMeasurement().getArrivalTime() + triggerConfig.getLagMs();
            Long lastTriggerTimestamp = triggerMeasurement.getLastTriggerEventTimestamp();
            if (deadline <= now && deadline > lastTriggerTimestamp) {
              // still triggerable after the lag and we did not send this trigger yet
              TriggerInfo.Builder info = TriggerInfo.newBuilder()
                  .setWindow(triggerMeasurement.getWindow());

              switch (triggerConfig.getMetric()) {
                case CONSUMPTION_DIFF:
                  info.setLongValue(triggerMeasurement.getClientMeasurement().getConsumptionDiff());
                  break;
                case AVG_LATENCY:
                  info.setLongValue(triggerMeasurement.getClientMeasurement().getAvgLatency());
                  break;
                case MAX_LATENCY:
                  info.setLongValue(triggerMeasurement.getClientMeasurement().getMaxLatency());
                  break;
                default:
                  log.warn("Unexpected trigger metric, not sending trigger event: {}", triggerConfig);
                  continue;
              }

              TriggerEvent triggerEvent = TriggerEvent.newBuilder()
                  .setTriggerGuid(triggerConfig.getGuid())
                  .setTimestamp(approxC3Time)
                  .setInfo(info)
                  .build();
              log.trace("Fire trigger={}", triggerEvent);

              // Set window 0, so that all events for the same clusterId, consumer Group goes
              // into the same internal topic partition
              WindowedClusterGroup triggerKey =
                  WindowedClusterGroup.newBuilder(keySerde.toProto(measurementKey)).setWindow(0).build();
              triggerList.add(KeyValue.pair(keySerde.key(triggerKey), triggerEvent)
              );
            } else if (deadline > now) {
              // possibly triggerable in the future
              waitForAtLeastOneTrigger = true;
            }
          }
        }
      } catch (Exception e) {
        log.error("Failure while checking if event is triggerable.", e);
      }
      return !waitForAtLeastOneTrigger;
    }

  }

  private static String logMeasurement(TriggerMeasurement triggerMeasurement) {
    return Joiner.on(' ').join(
        "arrivalTime=", triggerMeasurement.getClientMeasurement().getArrivalTime(),
        "consumptionDiff=", triggerMeasurement.getClientMeasurement().getConsumptionDiff(),
        "avgLatency=", triggerMeasurement.getClientMeasurement().getAvgLatency(),
        "maxLatency=", triggerMeasurement.getClientMeasurement().getMaxLatency()
    );
  }
}
