// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams.consumergroups;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.google.inject.Inject;
import io.confluent.controlcenter.Rollup;
import io.confluent.controlcenter.keys.Keys.KeyType;
import io.confluent.controlcenter.record.Controlcenter.ClientGroup;
import io.confluent.controlcenter.record.Controlcenter.ClientGroupProductionAggregate;
import io.confluent.controlcenter.record.Controlcenter.TopicPartition;
import io.confluent.controlcenter.record.Controlcenter.TopicPartitionClientGroup;
import io.confluent.controlcenter.record.Controlcenter.VerifiableMonitoringMessage;
import io.confluent.controlcenter.serialization.OrderedKeyPrefixedSerdeSupplier;
import io.confluent.controlcenter.streams.StreamExtension;
import io.confluent.controlcenter.streams.TopicStoreMaster;
import io.confluent.controlcenter.streams.TopicStoreMaster.Store;
import io.confluent.controlcenter.streams.TopicStoreMaster.Topic;
import io.confluent.controlcenter.streams.TopicStoreModule.AggregatedTopicPartitionTableWindowsStore;
import io.confluent.controlcenter.streams.TopicStoreModule.ConsumerGroupAggTPTopic;
import io.confluent.controlcenter.streams.TopicStoreModule.ConsumerGroupGroupAggTopic;
import io.confluent.controlcenter.streams.TopicStoreModule.MonitoringMessageAggregatorWindowsStore;
import io.confluent.controlcenter.streams.monitoring.MonitoringMessageAggregator;
import io.confluent.controlcenter.streams.monitoring.MonitoringMessageInitializer;
import io.confluent.controlcenter.streams.verify.Verifiable;
import io.confluent.monitoring.common.MonitoringMessageUtil;
import io.confluent.monitoring.record.Monitoring;
import io.confluent.monitoring.record.Monitoring.MessageType;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.serializers.OrderedKeyPrefixedSerde;
import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.kstream.Aggregator;
import org.apache.kafka.streams.kstream.Initializer;
import org.apache.kafka.streams.kstream.KStream;
import org.apache.kafka.streams.kstream.KTable;
import org.apache.kafka.streams.kstream.KeyValueMapper;
import org.apache.kafka.streams.kstream.Predicate;
import org.apache.kafka.streams.kstream.TimeWindows;
import org.apache.kafka.streams.kstream.Windowed;
import org.apache.kafka.streams.kstream.Windows;
import org.apache.kafka.streams.kstream.internals.TimeWindow;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import static io.confluent.controlcenter.ControlCenterProtoHelper.clientGroupBuilder;
import static io.confluent.controlcenter.ControlCenterProtoHelper.clientGroupProductionAggregateBuilder;
import static io.confluent.controlcenter.ControlCenterProtoHelper.topicPartition;

public class ConsumerGroupStreamExtension implements StreamExtension<MonitoringMessage, byte[], MonitoringMessage> {

  private static final Logger log = LoggerFactory.getLogger(ConsumerGroupStreamExtension.class);

  private final Verifiable verifiable;
  private final Topic<TopicPartition, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage> consumerGroupAggTpTopic;
  private final Topic<Bytes, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage> consumerGroupGroupAggTopic;
  private final Store<TopicPartition, MonitoringMessage, ClientGroupProductionAggregate> aggregatedTopicPartitionTableWindowsStore;
  private final Store<Bytes, MonitoringMessage, MonitoringMessage> monitoringMessageAggregatorWindowsStore;
  private final OrderedKeyPrefixedSerde<KeyType, MonitoringMessage> expectedProductionSerde;
  private final OrderedKeyPrefixedSerde<KeyType, MonitoringMessage> expectedProductionGroupSerde;
  private final OrderedKeyPrefixedSerde<KeyType, MonitoringMessage> expectedProductionGroupClientTopicPartitionSerde;
  private final OrderedKeyPrefixedSerde<KeyType, MonitoringMessage> expectedProductionClientSerde;
  private final OrderedKeyPrefixedSerde<KeyType, MonitoringMessage> expectedProductionInTopicSerde;
  private final OrderedKeyPrefixedSerde<KeyType, MonitoringMessage> expectedProductionInTopicPartitionForGroupSerde;
  private final OrderedKeyPrefixedSerde<KeyType, MonitoringMessage> expectedProductionInTopicPartitionForClientSerde;

  private static boolean clientGroupHasValues(ClientGroup clientGroup, String group, String clientId) {
    return Objects.equals(clientGroup.getGroup(), group) &&
           Objects.equals(clientGroup.getClient(), clientId);
  }

  @Inject
  public ConsumerGroupStreamExtension(
      Verifiable verifiable,
      @ConsumerGroupAggTPTopic Topic<TopicPartition, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage> consumerGroupAggTpTopic,
      @ConsumerGroupGroupAggTopic Topic<Bytes, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage> consumerGroupGroupAggTopic,
      @AggregatedTopicPartitionTableWindowsStore Store<TopicPartition, MonitoringMessage, ClientGroupProductionAggregate> aggregatedTopicPartitionTableWindowsStore,
      @MonitoringMessageAggregatorWindowsStore Store<Bytes, MonitoringMessage, MonitoringMessage> monitoringMessageAggregatorWindowsStore,
      OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> keySerdeSupplier
    ) {
    this.verifiable = verifiable;
    this.consumerGroupAggTpTopic = consumerGroupAggTpTopic;
    this.consumerGroupGroupAggTopic = consumerGroupGroupAggTopic;
    this.aggregatedTopicPartitionTableWindowsStore = aggregatedTopicPartitionTableWindowsStore;
    this.monitoringMessageAggregatorWindowsStore = monitoringMessageAggregatorWindowsStore;
    this.expectedProductionSerde = keySerdeSupplier.get(KeyType.EXPECTED_PRODUCTION);
    this.expectedProductionGroupSerde = keySerdeSupplier.get(KeyType.EXPECTED_PRODUCTION_GROUP);
    this.expectedProductionGroupClientTopicPartitionSerde = keySerdeSupplier.get(KeyType.EXPECTED_PRODUCTION_GROUP_CLIENT_TOPIC_PARTITION);
    this.expectedProductionClientSerde  = keySerdeSupplier.get(KeyType.EXPECTED_PRODUCTION_CLIENT);
    this.expectedProductionInTopicSerde  = keySerdeSupplier.get(KeyType.EXPECTED_PRODUCTION_IN_TOPIC);
    this.expectedProductionInTopicPartitionForGroupSerde  = keySerdeSupplier.get(KeyType.EXPECTED_PRODUCTION_IN_TOPICPARTITION_FOR_GROUP);
    this.expectedProductionInTopicPartitionForClientSerde = keySerdeSupplier.get(KeyType.EXPECTED_PRODUCTION_IN_TOPICPARTITION_FOR_CLIENT);
  }

  Windows<TimeWindow> windows(Rollup rollup) {
    return TimeWindows
        .of(rollup.getMillis())
        .until(rollup.getRetainMillis());
  }

  String stageName(Rollup rollup, String stage) {
    return TopicStoreMaster.nameJoin(stage, rollup.name());
  }

  private MonitoringMessage getKeyMM(ClientGroup cg, TopicPartition tp) {
    return MonitoringMessage.newBuilder()
        .setClientType(Monitoring.ClientType.PRODUCER)
        .setClusterId(tp.getClusterId())
        .setGroup(cg.getGroup())
        .setClientId(cg.getClient())
        .setTopic(tp.getTopic())
        .setPartition(tp.getPartition())
        .build();
  }

  private Collection<KeyValue<Bytes, MonitoringMessage>> consumerGroupData(
      ClientGroup cg,
      TopicPartition tp,
      MonitoringMessage value
  ) {
    MonitoringMessage keyMM = getKeyMM(cg, tp);
    return ImmutableList.of(
        KeyValue.pair(
            expectedProductionSerde.key(keyMM),
            value
        ),
        KeyValue.pair(
            expectedProductionGroupSerde.key(keyMM),
            value
        )
    );
  }

  private Collection<KeyValue<Bytes, MonitoringMessage>> consumerData(
      ClientGroup cg,
      TopicPartition tp,
      MonitoringMessage value
  ) {
    MonitoringMessage keyMM = getKeyMM(cg, tp);
    return ImmutableList.of(
        KeyValue.pair(
            expectedProductionGroupClientTopicPartitionSerde.key(keyMM),
            value
        ),
        KeyValue.pair(
            expectedProductionClientSerde.key(keyMM),
            value
        )
    );
  }

  private Collection<KeyValue<Bytes, MonitoringMessage>> aggregateTopicData(
      ClientGroup cg,
      TopicPartition tp,
      MonitoringMessage value
  ) {
    MonitoringMessage keyMM = getKeyMM(cg, tp);
    return ImmutableList.of(
        KeyValue.pair(
            expectedProductionInTopicSerde.key(keyMM),
            value
        ),
        KeyValue.pair(
            expectedProductionInTopicPartitionForGroupSerde.key(keyMM),
            value
        )
    );
  }

  private Collection<KeyValue<Bytes, MonitoringMessage>> consumerTopicData(
      ClientGroup cg,
      TopicPartition tp,
      MonitoringMessage value
  ) {
    MonitoringMessage keyMM = getKeyMM(cg, tp);
    return ImmutableList.of(
        KeyValue.pair(
            expectedProductionInTopicPartitionForClientSerde.key(keyMM),
            value
        )
    );
  }


  @Override
  public Map<Rollup, KTable<Windowed<Bytes>, MonitoringMessage>> extend(KStream<byte[], MonitoringMessage> stream) {
    KStream<TopicPartition, MonitoringMessage> monitoringMessagesByTopicPartition =
        rekeyMessagesByTopicPartition(stream);
    Map<Rollup, KTable<Windowed<Bytes>, MonitoringMessage>> rollupAggregates = Maps.newHashMap();
    for (Rollup rollup : Rollup.values()) {
      KStream<Windowed<TopicPartition>, ClientGroupProductionAggregate> clientGroupProductionAggregates =
          clientGroupProductionAggregateKStream(rollup, monitoringMessagesByTopicPartition);
      KStream<Bytes, MonitoringMessage> messageDiffsByTopicPartititionClientGroup =
          extractDiffsAndRekeyByTopicPartitionClientGroup(rollup, clientGroupProductionAggregates);
      rollupAggregates.put(rollup,
                           aggregateAndStoreFinalData(rollup,
                                                      messageDiffsByTopicPartititionClientGroup));
    }
    return rollupAggregates;
  }

  /**
   * @param input stream of monitoring messages
   * @return stream re-keyed with by TopicPartition
   */
  private KStream<TopicPartition, MonitoringMessage> rekeyMessagesByTopicPartition(
      KStream<byte[], MonitoringMessage> input
  ) {
    KStream<TopicPartition, MonitoringMessage> rekeyed = input.map(
        new KeyValueMapper<byte[], MonitoringMessage, KeyValue<TopicPartition, MonitoringMessage>>() {
          @Override
          public KeyValue<TopicPartition, MonitoringMessage> apply(
              byte[] key,
              MonitoringMessage value
          ) {
            return KeyValue.pair(topicPartition(value), value);
          }
        }
    );
    return verifiable.through(
        rekeyed,
        consumerGroupAggTpTopic
    );
  }

  /**
   * generates clientgroup and message diffs/aggregates
   * {{topic: t, partition: 0}} => {clientType: P,  count: 11}
   * {{topic: t, partition: 0}} => {clientType: C, client: c, group: g}
   * {{topic: t, partition: 0}} => {clientType: P, count: 4}
   * {{topic: t, partition: 0}} => {clientType: C, client: c, group: g}
   * {{topic: t, partition: 0}} => {clientType: P, count: 6}
   * {{topic: t, partition: 0}} => {clientType: C, client: c2, group: g2}
   * <p>
   * would output
   * <p>
   * {topic: t, partition: 0} => {
   * agg: {topic: t, partition: 0, count: 11}
   * diff: {topic: t, partition: 0, count: 11},
   * clientGroups: [],
   * newClientGroup: null
   * }
   * {topic: t, partition: 0} => {
   * agg: {topic: t, partition: 0, count: 11}
   * diff: null,
   * clientGroups: [{client: c, group: g}],
   * newClientGroup: {client:c, group: g}
   * }
   * {topic: t, partition: 0} => {
   * agg: {topic: t, partition: 0, count: 15}
   * diff: {topic: t, partition: 0, count: 4},
   * clientGroups: [{client: c, group: g}],
   * newClientGroup: null
   * }
   * {topic: t, partition: 0} => {
   * agg: {topic: t, partition: 0, count: 15}
   * diff: null,
   * clientGroups: [{client: c, group: g}],
   * newClientGroup: null
   * }
   * {topic: t, partition: 0} => {
   * agg: {topic: t, partition: 0, count: 21}
   * diff: {topic: t, partition: 0, count: 6},
   * clientGroups: [{client: c, group: g}],
   * newClientGroup: null
   * }
   * {topic: t, partition: 0} => {
   * agg: {topic: t, partition: 0, count: 21}
   * diff: null,
   * clientGroups: [{client: c, group: g}, {client: c2, group: g2}],
   * newClientGroup: {client:c2, group: g2}
   * }
   */
  private KStream<Windowed<TopicPartition>, ClientGroupProductionAggregate> clientGroupProductionAggregateKStream(
      Rollup rollup,
      KStream<TopicPartition, MonitoringMessage> input
  ) {
    final Windows<TimeWindow> windows = windows(rollup);
    final String stageName = stageName(rollup, aggregatedTopicPartitionTableWindowsStore.name);
    return input.groupByKey(
        aggregatedTopicPartitionTableWindowsStore.keySerde,
        aggregatedTopicPartitionTableWindowsStore.valueSerde
    ).aggregate(
        new Initializer<ClientGroupProductionAggregate>() {
          @Override
          public ClientGroupProductionAggregate apply() {
            return ClientGroupProductionAggregate.getDefaultInstance();
          }
        },
        new Aggregator<TopicPartition, MonitoringMessage, ClientGroupProductionAggregate>() {
          private MonitoringMessageAggregator<Void> aggregator = new MonitoringMessageAggregator<>();

          @Override
          public ClientGroupProductionAggregate apply(
              TopicPartition aggKey,
              MonitoringMessage value,
              ClientGroupProductionAggregate aggregate
          ) {
            ClientGroupProductionAggregate.Builder builder = clientGroupProductionAggregateBuilder(aggregate);
            switch (value.getClientType()) {
              case CONSUMER:
                ClientGroup.Builder newClientGroup = clientGroupBuilder(value);
                for (ClientGroup cg : aggregate.getClientGroupList()) {
                  if (clientGroupHasValues(cg, value.getGroup(), value.getClientId())) {
                    newClientGroup.setWindow(-1);
                    break;
                  }
                }
                builder = builder.setNewClientGroup(newClientGroup);
                if (newClientGroup.getWindow() > -1) {
                  builder.addClientGroup(newClientGroup);
                  // having a new client group means that we will be using this aggregate downstream
                  // we need to ensure that the new aggregate has some uniqueness to it, and since
                  // we know that this cg is new we can us its name.
                  builder.getAggBuilder()
                      .setClientId(newClientGroup.getClient())
                      .setGroup(newClientGroup.getGroup());
                }
                return builder.build();
              case PRODUCER:
                if (!aggregate.hasAgg() || aggregate.getAgg().getWindow() < 0) {
                  builder.setDiff(value);
                  builder.setAgg(value);
                } else {
                  builder.setAgg(aggregator.aggregate(value, aggregate.getAgg()));
                  if (builder.getAgg().getCount() > aggregate.getAgg().getCount() ||
                      builder.getAgg().getType() == MessageType.ERROR) {
                    builder.setDiff(value);
                  } else {
                    builder.setDiff(MonitoringMessageUtil.baseMonitoringMessage());
                  }
                }
                return builder.build();
              case UNRECOGNIZED:
                log.warn("unrecognized client type");
                break;
            }
            return builder.build();
          }
        },
        windows,
        aggregatedTopicPartitionTableWindowsStore.aggregateSerde,
        stageName
    ).toStream();
  }

  /**
   * extracts diffs from a stream that looks like
   *
   * {{topic: t, partition: 0}} => {
   * agg: {topic: t, partition: 0, count: 11}
   * diff: {topic: t, partition: 0, count: 4},
   * clientGroups: []
   * newClientGroup: null
   * }
   * {{topic: t, partition: 0}} => {
   * agg: {topic: t, partition: 0, count: 11}
   * diff: null,
   * clientGroups: [{client: c, group: g}]
   * newClientGroup: {client:c, group: g}
   * }
   * {{topic: t, partition: 0}} => {
   * agg: {topic: t, partition: 0, count: 15}
   * diff: {topic: t, partition: 0, count: 4},
   * clientGroups: [{client: c, group: g}]
   * newClientGroup: null
   * }
   * {{topic: t, partition: 0}} => {
   * agg: {topic: t, partition: 0, count: 21}
   * diff: {topic: t, partition: 0, count: 6},
   * clientGroups: [{client: c, group: g}, {client: c2, group: g2}]
   * newClientGroup: {client:c2, group: g2}
   * }
   *
   * transforms to
   *
   * {{topic: t, partition: 0}, {client: c, group: g}} => {topic: t, partition: 0, count: 11}
   * {{topic: t, partition: 0}, {client: c, group: g}} => {topic: t, partition: 0, count: 4}
   * {{topic: t, partition: 0}, {client: c, group: g}} => {topic: t, partition: 0, count: 6}
   * {{topic: t, partition: 0}, {client: c2, group: g2}} => {topic: t, partition: 0, count: 21}
   * {{topic: t, partition: 0}, {client: c, group: g}} => {topic: t, partition: 0, count: 11}
   * {{topic: t, partition: 0}, {group: g}} => {topic: t, partition: 0, count: 11}
   * {{topic: t, partition: 1}, {client: c, group: g}} => {topic: t, partition: 1, count: 4}
   * {{topic: t, partition: 1}, {group: g}} => {topic: t, partition: 1, count: 4}
   * <p>
   * and outputs a repartitioned stream that looks like
   * lookupkey1({{topic: t, partition: 0}, {client: c, group: g}}) => {topic: t, partition: 0, count: 11}
   * lookupkey2({{topic: t, partition: 0}, {client: c, group: g}}) => {topic: t, partition: 0, count: 11}
   * lookupkey3({{topic: t, partition: 0}, {client: c, group: g}}) => {topic: t, partition: 0, count: 11}
   * lookupkey4({{topic: t, partition: 0}, {client: c, group: g}}) => {topic: t, partition: 0, count: 11}
   * ....
   *
   */
  private KStream<Bytes, MonitoringMessage> extractDiffsAndRekeyByTopicPartitionClientGroup(
      Rollup rollup,
      KStream<Windowed<TopicPartition>, ClientGroupProductionAggregate> aggregates
  ) {
    KStream<Bytes, MonitoringMessage> diffs = aggregates.filter(
        new Predicate<Windowed<TopicPartition>, ClientGroupProductionAggregate>() {
          @Override
          public boolean test(Windowed<TopicPartition> key, ClientGroupProductionAggregate value) {
            return (
                       value.hasNewClientGroup() &&
                       value.getNewClientGroup().getWindow() > -1
                   )
                   ||
                   (
                       value.hasDiff() &&
                       value.getDiff().getWindow() > -1
                   )
                ;
          }
        }
    ).flatMap(
        new KeyValueMapper<
            Windowed<TopicPartition>, ClientGroupProductionAggregate,
            Iterable<KeyValue<Bytes, MonitoringMessage>>
            >() {

          @Override
          public Iterable<KeyValue<Bytes, MonitoringMessage>> apply(
              Windowed<TopicPartition> key,
              ClientGroupProductionAggregate value
          ) {
            Set<String> usedGroups = Sets.newHashSet();

            // create tpcgs with diffs for already extant cgs and
            // the full value for any new clientgroups
            List<KeyValue<TopicPartitionClientGroup, MonitoringMessage>> tpcgs = Lists.newArrayList();
            if (
                value.hasDiff() && value.getDiff().getWindow() > -1 &&
                value.getClientGroupList() != null && value.getClientGroupCount() > 0
                ) {
              ClientGroup newClientGroup = value.getNewClientGroup();
              for (ClientGroup cg : value.getClientGroupList()) {
                if (!clientGroupHasValues(newClientGroup, cg.getGroup(), cg.getClient())) {
                  tpcgs.add(KeyValue.pair(topicPartition(key.key(), cg), value.getDiff()));
                }
              }
            }

            // rekey and aggregate over all tpcgs
            Map<Bytes, MonitoringMessage> localAggregate = Maps.newHashMap();
            List<KeyValue<Bytes, MonitoringMessage>> monitoringMessages = Lists.newArrayList();
            for (KeyValue<TopicPartitionClientGroup, MonitoringMessage> kv : tpcgs) {
              monitoringMessages.addAll(consumerData(kv.key.getClientGroup(), kv.key.getTopicPartition(), kv.value));
              monitoringMessages.addAll(consumerTopicData(kv.key.getClientGroup(), kv.key.getTopicPartition(), kv.value));
              if (usedGroups.add(kv.key.getClientGroup().getGroup())) {
                monitoringMessages.addAll(consumerGroupData(kv.key.getClientGroup(), kv.key.getTopicPartition(), kv.value));
                monitoringMessages.addAll(aggregateTopicData(kv.key.getClientGroup(), kv.key.getTopicPartition(), kv.value));
              }
            }

            if (value.hasNewClientGroup() && value.getNewClientGroup().getWindow() >= 0) {
              ClientGroup newClientGroup = value.getNewClientGroup();
              monitoringMessages.addAll(consumerData(newClientGroup, key.key(), value.getAgg()));
              monitoringMessages.addAll(consumerTopicData(newClientGroup, key.key(), value.getAgg()));
              for (ClientGroup cg : value.getClientGroupList()) {
                if (!cg.equals(newClientGroup)) {
                  usedGroups.add(cg.getGroup());
                }
              }
              if (usedGroups.add(newClientGroup.getGroup())) {
                monitoringMessages.addAll(consumerGroupData(newClientGroup, key.key(), value.getAgg()));
                monitoringMessages.addAll(aggregateTopicData(newClientGroup, key.key(), value.getAgg()));
              }
            }

            MonitoringMessageAggregator<Bytes> mma = new MonitoringMessageAggregator<>();
            for (KeyValue<Bytes, MonitoringMessage> kv : monitoringMessages) {
              if (!localAggregate.containsKey(kv.key)) {
                localAggregate.put(kv.key, kv.value);
              } else {
                localAggregate.put(kv.key, mma.aggregate(kv.value, localAggregate.get(kv.key)));
              }
            }

            List<KeyValue<Bytes, MonitoringMessage>> aggregatedOutput = Lists.newArrayList();
            for (Map.Entry<Bytes, MonitoringMessage> entry : localAggregate.entrySet()) {
              if (entry.getValue().getCount() > 0 || entry.getValue().getType() == MessageType.ERROR) {
                aggregatedOutput.add(KeyValue.pair(
                    entry.getKey(),
                    entry.getValue()
                ));
              }
            }
            return aggregatedOutput;
          }
        }
    );
    return verifiable.through(
        diffs, consumerGroupGroupAggTopic, rollup
    );
  }

  private KTable<Windowed<Bytes>, MonitoringMessage> aggregateAndStoreFinalData(final Rollup rollup, KStream<Bytes, MonitoringMessage> out) {
    return out.groupByKey(
        monitoringMessageAggregatorWindowsStore.keySerde,
        monitoringMessageAggregatorWindowsStore.valueSerde
    ).aggregate(
        new MonitoringMessageInitializer(),
        MonitoringMessageAggregator.<Bytes>stripped(),
        windows(rollup),
        monitoringMessageAggregatorWindowsStore.aggregateSerde,
        stageName(rollup, monitoringMessageAggregatorWindowsStore.name)
    );
  }
}
