// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams.group;

import com.google.common.cache.Cache;
import com.google.common.cache.CacheBuilder;
import com.google.common.collect.ImmutableSet;

import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.kstream.Aggregator;

import java.util.Collection;

import io.confluent.controlcenter.record.Controlcenter.MemberInfo;
import io.confluent.controlcenter.record.Controlcenter.WindowedGrouping;

public class WindowedGroupingAggregator
    implements Aggregator<Bytes, WindowedGrouping, WindowedGrouping> {

  public static final int DEFAULT_MAX_MEMBERS_DIFF = 2 << 5;
  public static final int DEFAULT_CACHE_SIZE = 2 << 12;

  private final Cache<Bytes, Integer> memberListSize;
  private final int maxMembersDiff;

  public WindowedGroupingAggregator() {
    this(DEFAULT_MAX_MEMBERS_DIFF, DEFAULT_CACHE_SIZE);
  }

  protected WindowedGroupingAggregator(int maxMembersDiff, int cacheSize) {
    this.maxMembersDiff = maxMembersDiff;
    memberListSize = CacheBuilder.newBuilder()
        .maximumSize(cacheSize)
        .build();
  }

  @Override
  public WindowedGrouping apply(
      Bytes aggKey,
      WindowedGrouping value,
      WindowedGrouping aggregate
  ) {
    WindowedGrouping.Builder builder = WindowedGrouping.newBuilder(aggregate).setWindow(value.getWindow());
    Integer lastMemberListSize = memberListSize.getIfPresent(aggKey);
    lastMemberListSize = lastMemberListSize == null ? 1 : lastMemberListSize;
    if (
        value.getMembersCount() > 0 &&
        (
            value.getMembersCount() > maxMembersDiff ||
            aggregate.getMembersCount() >= lastMemberListSize + maxMembersDiff
        )
        ) {
      builder.clearMembers();
      builder.addAllMembers(uniqueMemberInfos(value.getMembersList(), aggregate.getMembersList()));
      memberListSize.put(aggKey, builder.getMembersCount());
    } else {
      for (MemberInfo mi : value.getMembersList()) {
        builder.addMembers(mi);
      }
    }
    return builder.build();
  }

  public static Collection<? extends MemberInfo> uniqueMemberInfos(Iterable<? extends MemberInfo>... iters) {
    ImmutableSet.Builder<MemberInfo> memberInfos = ImmutableSet.builder();
    for (Iterable<? extends MemberInfo> mis : iters) {
      if (mis != null) {
        memberInfos.addAll(mis);
      }
    }
    return memberInfos.build();
  }
}
