// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams.monitoring;

import com.google.common.collect.ImmutableList;
import com.google.inject.Inject;

import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.kstream.KStream;
import org.apache.kafka.streams.kstream.KTable;
import org.apache.kafka.streams.kstream.Window;
import org.apache.kafka.streams.kstream.Windowed;
import org.apache.kafka.streams.kstream.Windows;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import io.confluent.controlcenter.keys.Keys.KeyType;
import io.confluent.controlcenter.record.Controlcenter.VerifiableMonitoringMessage;
import io.confluent.controlcenter.serialization.OrderedKeyPrefixedSerdeSupplier;
import io.confluent.controlcenter.streams.AbstractMonitoringMessageRollupStreamExtension;
import io.confluent.controlcenter.streams.TopicStoreMaster.Store;
import io.confluent.controlcenter.streams.TopicStoreMaster.Topic;
import io.confluent.controlcenter.streams.TopicStoreModule.MonitoringMsgExtensionTopic;
import io.confluent.controlcenter.streams.TopicStoreModule.MonitoringStreamStore;
import io.confluent.controlcenter.streams.TypeMapper;
import io.confluent.controlcenter.streams.verify.Verifiable;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;

public class MonitoringStreamExtension extends
    AbstractMonitoringMessageRollupStreamExtension<MonitoringMessage> {

  private static final Logger log = LoggerFactory.getLogger(MonitoringStreamExtension.class);

  private final Verifiable verifiable;
  private final Topic<Bytes, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage> monitoringMsgExtensionTopic;
  private final Store<Bytes, MonitoringMessage, MonitoringMessage> monitoringStreamStore;
  private final OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> orderedKeyPrefixedSerdeSupplier;

  @Inject
  public MonitoringStreamExtension(
      Verifiable verifiable,
      @MonitoringMsgExtensionTopic Topic<Bytes, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage> monitoringMsgExtensionTopic,
      @MonitoringStreamStore Store<Bytes, MonitoringMessage, MonitoringMessage> monitoringStreamStore,
      OrderedKeyPrefixedSerdeSupplier<KeyType, MonitoringMessage> orderedKeyPrefixedSerdeSupplier
  ) {
    this.verifiable = verifiable;
    this.monitoringMsgExtensionTopic = monitoringMsgExtensionTopic;
    this.monitoringStreamStore = monitoringStreamStore;
    this.orderedKeyPrefixedSerdeSupplier = orderedKeyPrefixedSerdeSupplier;
  }

  @Override
  protected String name() {
    return monitoringStreamStore.name;
  }

  @Override
  protected Iterable<? extends TypeMapper<MonitoringMessage>> getMappers() {
    return ImmutableList.of(
        new MonitoringMessageTypeMapper(orderedKeyPrefixedSerdeSupplier.get(KeyType.LOOKUP_CLIENTTYPE)),
        new MonitoringMessageTypeMapper(orderedKeyPrefixedSerdeSupplier.get(KeyType.LOOKUP_CLIENTTYPE_GROUP)),
        new MonitoringMessageTypeMapper(orderedKeyPrefixedSerdeSupplier.get(KeyType.LOOKUP_CLIENTTYPE_CLIENT)),
        new MonitoringMessageTypeMapper(orderedKeyPrefixedSerdeSupplier.get(KeyType.LOOKUP_CLIENTTYPE_TOPIC)),
        new MonitoringMessageTypeMapper(orderedKeyPrefixedSerdeSupplier.get(KeyType.LOOKUP_CLIENTTYPE_TOPICPARTITION)),
        new MonitoringMessageTypeMapper(orderedKeyPrefixedSerdeSupplier.get(KeyType.LOOKUP_CLIENTTYPE_CLIENT_TOPIC_PARTITION)),
        new MonitoringMessageTypeMapper(orderedKeyPrefixedSerdeSupplier.get(KeyType.LOOKUP_CLIENTTYPE_GROUP_TOPICPARTITION))
    );
  }

  /**
   * this will filter out any window that is outside of window's maintainms then aggregate the
   * remaining messages then write to monitoringdb with rollup based on window's name
   *
   * @param stream  base stream to filter and aggregate on
   * @param windows named windows that we want to roll up
   */
  protected KTable<Windowed<Bytes>, MonitoringMessage> filterAndAggregateRollup(
      KStream<Bytes, MonitoringMessage> stream,
      final Windows<Window> windows,
      final String name
  ) {
    return stream.groupByKey(
        monitoringStreamStore.keySerde,
        monitoringStreamStore.valueSerde
    ).aggregate(
        new MonitoringMessageInitializer(),
        MonitoringMessageAggregator.<Bytes>stripped(),
        windows,
        monitoringStreamStore.aggregateSerde,
        name
    );
  }

  @Override
  protected KStream<Bytes, MonitoringMessage> shuffle(KStream<Bytes, MonitoringMessage> stream) {
    return verifiable.through(stream, monitoringMsgExtensionTopic);
  }
}
