// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams.verify;

import com.google.common.collect.Lists;

import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.clients.producer.RecordMetadata;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import io.confluent.controlcenter.record.Controlcenter.WindowedClusterGroup;
import io.confluent.controlcenter.streams.TopicStoreMaster.Topic;
import io.confluent.monitoring.common.Clock;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.serializers.UberSerde;

/**
 * This class periodically sends heartbeats to monitoring topic, so that KStreams call punctuate
 * even when clients are not running (not publishing metrics to monitoring topic). This is a
 * temporary solution until KStreams supports wall-clock punctuate.
 */
public class MonitoringHeartbeatSender implements Runnable {

  private static final Logger log = LoggerFactory.getLogger(MonitoringHeartbeatSender.class);

  private final long timeout;
  private final MonitoringMessage baseMonitoringMessage;
  private final String publishTopic;
  private final Set<Topic<WindowedClusterGroup, MonitoringMessage, Void, Void>> internalPublishTopics;
  private final UberSerde<MonitoringMessage> monitoringMessageSerde;
  private final Clock clock;

  private KafkaProducer<byte[], byte[]> heartbeatProducer;

  public MonitoringHeartbeatSender(
      String publishTopic,
      Set<Topic<WindowedClusterGroup, MonitoringMessage, Void, Void>> internalPublishTopics,
      MonitoringMessage baseMonitoringMessage,
      KafkaProducer<byte[], byte[]> heartbeatProducer,
      UberSerde<MonitoringMessage> monitoringMessageUberSerde,
      long timeout,
      Clock clock
  ) {
    this.timeout = timeout;
    this.publishTopic = publishTopic;
    this.internalPublishTopics = internalPublishTopics;
    this.baseMonitoringMessage = baseMonitoringMessage;
    this.heartbeatProducer = heartbeatProducer;
    this.monitoringMessageSerde = monitoringMessageUberSerde;
    this.clock = clock;
  }

  @Override
  public void run() {
    try {
      sendHeartbeat();
    } catch (Throwable t) {
      log.error("Terminating heartbeat thread", t);
    }
  }

  public void close() {
    try {
      heartbeatProducer.close();
    } catch (Exception e) {
      log.error("Failed to close monitoring heartbeat producer", e);
    }
    log.info("Closed monitoring heartbeat producer");
  }

  private void sendHeartbeat() {
    long now = clock.currentTimeMillis();
    MonitoringMessage heartbeatMsg = MonitoringMessage.newBuilder(this.baseMonitoringMessage)
        .setTimestamp(now)
        .setWindow(now)
        .build();

    try {
      List<Future<RecordMetadata>> futures = Lists.newArrayList();
      sendHeartbeatToAllPartitions(this.publishTopic, null, heartbeatMsg, futures);
      for (Topic<WindowedClusterGroup, MonitoringMessage, Void, Void> internalPublishTopic : internalPublishTopics) {
        sendHeartbeatToAllPartitions(internalPublishTopic.name,
                                     internalPublishTopic.keySerde.serialize(
                                         WindowedClusterGroup.newBuilder().setWindow(0).build()),
                                     heartbeatMsg,
                                     futures);
      }
      long endTime = now + timeout;
      Iterator<Future<RecordMetadata>> i = futures.iterator();
      while (endTime > System.currentTimeMillis() && i.hasNext()) {
        Future<RecordMetadata> f = i.next();
        f.get(endTime - System.currentTimeMillis(), TimeUnit.MILLISECONDS);
      }
    } catch (Throwable t) {
      log.warn("failed to publish monitoring heartbeat message", t);
    }
  }

  private void sendHeartbeatToAllPartitions(String topic,
                                            byte[] key,
                                            MonitoringMessage heartbeatMsg,
                                            List<Future<RecordMetadata>> futures) {
    int partitions = this.heartbeatProducer.partitionsFor(topic).size();
    for (int partition = 0; partition < partitions; ++partition) {
      futures.add(this.heartbeatProducer.send(
          new ProducerRecord<>(
              topic,
              partition,
              key,
              this.monitoringMessageSerde.serialize(heartbeatMsg)
          )
      ));
    }
  }
}
