// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams.verify;

import java.util.Objects;

import io.confluent.controlcenter.record.Controlcenter;
import io.confluent.monitoring.record.Monitoring;

/**
 * Represents monitoring message sequence number and corresponding metadata: timestamp when sequence
 * number was created, min and max time window seen in the last set of monitoring messages.
 */
public class SequenceWithMetadata implements Comparable<SequenceWithMetadata> {

  private final long sequence;
  /**
   * Timestamp is the current C3 time (based on heartbeats) of the message arrival
   */
  private final long timestamp;
  private final long minWindow;
  private final long maxWindow;

  public SequenceWithMetadata(long sequence, long timestamp, long minWindow, long maxWindow) {
    this.sequence = sequence;
    this.timestamp = timestamp;
    this.minWindow = minWindow;
    this.maxWindow = maxWindow;
  }

  /**
   * Constructs the object using appropriate fields in the monitoring message
   */
  public SequenceWithMetadata(Monitoring.MonitoringMessage monitoringMessage, long timestamp) {
    this(monitoringMessage.getSequence(), timestamp,
         monitoringMessage.getMinWindow(), monitoringMessage.getMaxWindow());
  }

  /**
   * Constructs the object from SequenceInfo
   */
  public SequenceWithMetadata(Controlcenter.SequenceInfo info) {
    this(info.getSequence(), info.getTimestamp(), info.getMinWindow(), info.getMaxWindow());
  }

  public SequenceWithMetadata(long sequence) {
    this(sequence, -1, -1, -1);
  }

  public SequenceWithMetadata(long sequence, Long timestamp) {
    this(sequence, timestamp, -1, -1);
  }

  /**
   * Creates a new object with next sequence number after the sequence number of 'seq' and all other
   * fields (metadata) are null.
   */
  public SequenceWithMetadata nextWithTimestamp(Long timestamp) {
    return new SequenceWithMetadata(nextSequence(), timestamp);
  }

  /**
   * Creates a new object with the previous sequence number before the sequence number of 'seq',
   * with given timestmap and other fields set to null.
   *
   * @return A new object or null if the sequence number of 'seq' is 0.
   */
  public SequenceWithMetadata previousWithTimestamp(Long timestamp) {
    return this.sequence <= 0 ? null : new SequenceWithMetadata(this.sequence - 1, timestamp);
  }

  /**
   * Creates a new object with the same sequence number and timestamp, and min and max windows set
   * to 'window'.
   */
  public SequenceWithMetadata withMinMaxWindow(Long window) {
    return new SequenceWithMetadata(sequence, timestamp, window, window);
  }

  public Controlcenter.SequenceInfo toSequenceInfo() {
    return Controlcenter.SequenceInfo.newBuilder().setSequence(sequence)
        .setTimestamp(timestamp)
        .setMinWindow(minWindow)
        .setMaxWindow(maxWindow)
        .build();
  }

  public long sequence() {
    return this.sequence;
  }

  public long nextSequence() {
    return this.sequence + 1;
  }

  /**
   * Returns message arrival time, in terms of C3 heartbeat time
   */
  public long timestamp() {
    return this.timestamp;
  }

  public long minWindow() {
    return this.minWindow;
  }

  public long maxWindow() {
    return this.maxWindow;
  }

  @Override
  public int compareTo(SequenceWithMetadata sequenceWithMetadata) {
    if (this == sequenceWithMetadata) {
      return 0;
    }

    return Long.compare(sequence, sequenceWithMetadata.sequence);
  }

  @Override
  public boolean equals(Object o) {
    if (this == o) {
      return true;
    }
    if (o == null || getClass() != o.getClass()) {
      return false;
    }

    SequenceWithMetadata other = (SequenceWithMetadata) o;
    return Objects.equals(sequence, other.sequence) &&
           Objects.equals(timestamp, other.timestamp) &&
           Objects.equals(minWindow, other.minWindow) &&
           Objects.equals(maxWindow, other.maxWindow);
  }

  @Override
  public int hashCode() {
    return Objects.hash(sequence, timestamp, minWindow, maxWindow);
  }

  @Override
  public String toString() {
    return "{Sequence=" + this.sequence + ", Timestamp=" + this.timestamp +
           ", minWindow=" + this.minWindow + ", maxWindow=" + this.maxWindow + "}";
  }
}
