package io.confluent.controlcenter.streams.verify;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.collect.Maps;
import com.google.common.util.concurrent.RateLimiter;

import org.apache.kafka.streams.kstream.ValueTransformer;
import org.apache.kafka.streams.kstream.ValueTransformerSupplier;
import org.apache.kafka.streams.processor.ProcessorContext;
import org.apache.kafka.streams.state.KeyValueStore;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Map;
import java.util.concurrent.TimeUnit;

import io.confluent.controlcenter.record.Controlcenter.VerifiableMonitoringMessage;
import io.confluent.monitoring.common.Clock;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;

/**
 * This transformer checks and filters out duplicates after KStreams.through() is called
 */
public class ThroughTopicVerifierTransformerSupplier
    implements ValueTransformerSupplier<VerifiableMonitoringMessage, MonitoringMessage> {

  private final String storeName;
  private final Clock clock;

  public ThroughTopicVerifierTransformerSupplier(
      String storeName,
      Clock clock
  ) {
    this.storeName = storeName;
    this.clock = clock;
  }

  @Override
  public ValueTransformer<VerifiableMonitoringMessage, MonitoringMessage> get() {
    return new ThroughTopicVerifierTransformer(storeName);
  }

  private class ThroughTopicVerifierTransformer
      implements ValueTransformer<VerifiableMonitoringMessage, MonitoringMessage> {

    private final Logger log =
        LoggerFactory.getLogger(ThroughTopicVerifierTransformer.class);


    private final String storeName;
    private ProcessorContext context;
    private KeyValueStore<String, VerifiableMonitoringMessage> kvStore;

    ThroughTopicVerifierTransformer(String storeName) {
      this.storeName = storeName;
    }

    @SuppressWarnings("unchecked")
    @Override
    public void init(ProcessorContext context) {
      this.context = context;
      kvStore = (KeyValueStore<String, VerifiableMonitoringMessage>) this.context.getStateStore(storeName);
    }

    @Override
    public MonitoringMessage transform(VerifiableMonitoringMessage verifiableMonitoringMessage) {
      if (verifiableMonitoringMessage == null) {
        return null;
      }
      String key = verifiableMonitoringMessage.getGuid();
      MonitoringMessage out = null;

      try {
        VerifiableMonitoringMessage previous = kvStore.get(key);
        kvStore.put(
            key,
            VerifiableMonitoringMessage.newBuilder(verifiableMonitoringMessage)
                .clearMonitoringMessage()
                .setLastSeen(clock.currentTimeMillis())
                .build()
        );
        if (previous == null) {
          out = verifiableMonitoringMessage.getMonitoringMessage();
        } else {
          log.warn("dropping duplicate message key={} lastSeen={} store={}", key, previous.getLastSeen(), storeName);
        }
      } catch (Exception e) {
        // either rocksdb exception or unexpected exception from the verifier
        log.warn("failed to update key={} store={} {}", key, storeName, e);
      }

      return out;
    }

    @Override
    public MonitoringMessage punctuate(long l) {
      return null;
    }

    @Override
    public void close() {
      log.info("closing store={}", storeName);
    }
  }
}
