package io.confluent.controlcenter.streams.verify;

import com.google.common.hash.HashCode;
import com.google.common.hash.Hasher;
import com.google.common.hash.Hashing;
import com.google.inject.Inject;

import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.kstream.KStream;
import org.apache.kafka.streams.kstream.KeyValueMapper;

import java.nio.charset.StandardCharsets;
import java.util.concurrent.TimeUnit;

import io.confluent.controlcenter.Rollup;
import io.confluent.controlcenter.record.Controlcenter.VerifiableMonitoringMessage;
import io.confluent.controlcenter.streams.TopicStoreMaster;
import io.confluent.controlcenter.streams.TopicStoreMaster.Topic;
import io.confluent.controlcenter.streams.ValueNotNull;
import io.confluent.monitoring.common.Clock;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;

public class Verifiable {

  public static final long STORE_PUNCTUATE_INTERVAL_MS = TimeUnit.MINUTES.toMillis(2);
  public static final long STORE_RETAIN_MS = TimeUnit.HOURS.toMillis(12);

  // this needs to be small enough to allow all our stores to hit this value and not time us out
  public static final long STORE_MAX_PUNCTUATE_MS = TimeUnit.SECONDS.toMillis(1);

  @Inject
  private Clock clock;

  public <K> KStream<K, MonitoringMessage> through(
      final KStream<K, MonitoringMessage> inputStream,
      final Topic<K, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage> topic
  ) {
    assert !topic.rollup;
    return throughImpl(inputStream, topic, null);
  }

  public <K> KStream<K, MonitoringMessage> through(
      final KStream<K, MonitoringMessage> inputStream,
      final Topic<K, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage> topic,
      final Rollup rollup
  ) {
    assert topic.rollup;
    return throughImpl(inputStream, topic, rollup);
  }

  /**
   * rollup can be null here
   * <p>
   * TODO: make rollup Optional<KeyMaster.Rollup> when we upgrade to Java 8
   */
  protected <K> KStream<K, MonitoringMessage> throughImpl(
      final KStream<K, MonitoringMessage> inputStream,
      final Topic<K, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage> topic,
      final Rollup rollup
  ) {
    final String topicName = rollup == null ? topic.name : TopicStoreMaster.nameJoin(topic.name, rollup.name());
    final String storeName = rollup == null ? topic.store.name : TopicStoreMaster.nameJoin(topic.store.name, rollup.name());
    return inputStream.map(
        new KeyValueMapper<K, MonitoringMessage, KeyValue<K, VerifiableMonitoringMessage>>() {
          @Override
          public KeyValue<K, VerifiableMonitoringMessage> apply(K key, MonitoringMessage value) {
            String hash = getHash(topic, key, value, rollup).toString();
            return KeyValue.pair(
                key,
                VerifiableMonitoringMessage.newBuilder()
                    .setMonitoringMessage(value)
                    .setGuid(hash)
                    .build()
            );
          }
        }
    ).through(
        topic.keySerde,
        topic.valueSerde,
        topicName
    ).transformValues(
        new ThroughTopicVerifierTransformerSupplier(
            storeName,
            clock
        ),
        storeName
    ).filter(
        new ValueNotNull<K, MonitoringMessage>()
    );
  }

  /**
   * rollup can be null
   * <p>
   * TODO: make rollup Optional<KeyMaster.Rollup> when we upgrade to Java 8
   */
  protected <K> HashCode getHash(
      final Topic<K, VerifiableMonitoringMessage, String, VerifiableMonitoringMessage> topic,
      final K key,
      final MonitoringMessage value,
      final Rollup rollup) {
    Hasher hasher = Hashing.murmur3_128().newHasher()
        .putBytes(value.toByteArray())
        .putBytes(topic.keySerde.serialize(key));
    if (rollup != null) {
      hasher.putString(rollup.name(), StandardCharsets.UTF_8);
    }
    return hasher.hash();
  }
}
