// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.tools;

import com.google.common.base.Joiner;
import com.google.inject.Guice;
import com.google.inject.Inject;
import com.google.inject.Injector;
import io.confluent.controlcenter.ControlCenterConfig;
import io.confluent.controlcenter.ControlCenterConfigModule;
import io.confluent.controlcenter.KafkaHelper;
import io.confluent.controlcenter.serialization.SerializationModule;
import io.confluent.controlcenter.streams.StreamsConfigModule;
import io.confluent.controlcenter.streams.TopicStoreMaster;
import io.confluent.controlcenter.streams.TopicStoreModule;
import kafka.tools.StreamsResetter;
import org.apache.kafka.common.utils.SystemTime;
import org.apache.kafka.streams.StreamsConfig;

import java.io.IOException;
import java.nio.file.*;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.ArrayList;
import java.util.List;

public class Resetter {
  protected final ControlCenterConfig config;
  protected final TopicStoreMaster topicStoreMaster;
  protected final StreamsConfig streamsConfig;
  protected final StreamsResetter streamsResetter;
  protected final KafkaHelper kafkaHelper;

  @Inject
  public Resetter(ControlCenterConfig config,
                  TopicStoreMaster topicStoreMaster,
                  StreamsConfig streamsConfig,
                  StreamsResetter streamsResetter,
                  KafkaHelper kafkaHelper) {
    this.config = config;
    this.streamsConfig = streamsConfig;
    this.topicStoreMaster = topicStoreMaster;
    this.streamsResetter = streamsResetter;
    this.kafkaHelper = kafkaHelper;
  }

  public int run() {
    Joiner joiner = Joiner.on(",");
    List<String> argsList = new ArrayList<>();
    argsList.add("--zookeeper");
    argsList.add(config.getString(ControlCenterConfig.ZOOKEEPER_CONNECT_CONFIG));
    argsList.add("--bootstrap-servers");
    argsList.add(joiner.join(config.getList(ControlCenterConfig.CONTROL_CENTER_BOOTSTRAP_SERVERS_CONFIG)));
    argsList.add("--application-id");
    argsList.add(streamsConfig.getString(StreamsConfig.APPLICATION_ID_CONFIG));
    argsList.add("--input-topics");
    argsList.add(joiner.join(topicStoreMaster.getInputTopicNames()));
    argsList.add("--intermediate-topics");
    argsList.add(joiner.join(topicStoreMaster.getIntermediateTopicNames()));
    int exit = streamsResetter.run(argsList.toArray(new String[0]));

    System.out.println("Deleting intermediate topics");
    for (String topic : topicStoreMaster.getIntermediateTopicNames()) {
      kafkaHelper.deleteTopic(topic);
    }
    System.out.println("Done");

    deleteLocalData();

    return exit;
  }

  private void deleteLocalData() {
    Path dataDir = Paths.get(config.getString(
        ControlCenterConfig.CONTROL_CENTER_DATA_DIR),
        config.getString(ControlCenterConfig.CONTROL_CENTER_ID_CONFIG));
    System.out.println("Deleting local RocksDB data in " + dataDir.toAbsolutePath());
    try {
      Files.walkFileTree(dataDir, new SimpleFileVisitor<Path>() {
        @Override
        public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
          Files.delete(file);
          return FileVisitResult.CONTINUE;
        }

        @Override
        public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException {
          Files.delete(dir);
          return FileVisitResult.CONTINUE;
        }
      });
      System.out.println("Done");
    } catch (NoSuchFileException e) {
      System.out.println("Does not exist");
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  public static void main(String[] args) throws Exception {
    if (args.length == 0) {
      System.err.println("You must provide a path to the config file");
      System.exit(1);
    }

    final ControlCenterConfig config = new ControlCenterConfig(args[0]);
    Injector injector = Guice.createInjector(
        new ControlCenterConfigModule(config),
        new StreamsConfigModule(),
        new SerializationModule(),
        new TopicStoreModule()
    );
    Resetter resetter = injector.getInstance(Resetter.class);
    System.exit(resetter.run());
  }

}
