package io.confluent.controlcenter;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;

import io.confluent.controlcenter.util.ZKUtilsFactory;
import org.apache.kafka.common.Node;
import org.apache.kafka.common.protocol.Errors;
import org.apache.kafka.common.protocol.SecurityProtocol;
import org.apache.kafka.common.requests.MetadataResponse;
import org.easymock.EasyMock;
import org.easymock.IArgumentMatcher;
import org.junit.Before;
import org.junit.Test;

import java.util.Collections;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.TimeUnit;

import kafka.admin.RackAwareMode;
import kafka.log.LogConfig$;
import kafka.server.KafkaConfig$;
import kafka.utils.ZkUtils;

import static org.easymock.EasyMock.and;
import static org.easymock.EasyMock.anyObject;
import static org.easymock.EasyMock.createMock;
import static org.easymock.EasyMock.eq;
import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.replay;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

public class KafkaHelperTest {

  public static final String TOPIC1 = "topic1-changelog";
  public static final String TOPIC2 = "topic2";
  public static final Set<String> TOPICS = ImmutableSet.of(TOPIC1, TOPIC2);
  KafkaHelper.KafkaAdminUtils mockAdminUtils;
  KafkaHelper kafkaHelper;

  ZKUtilsFactory zkUtilsFactory = new ZKUtilsFactory() {
    @Override
    public ZkUtils create(String zookeeperConnect) {
      return createMock(ZkUtils.class);
    }
  };

  private MetadataResponse.PartitionMetadata buildPartitionMetadata(int partition, int replicas, int isr) {
    return new MetadataResponse.PartitionMetadata(
        null,
        partition,
        null,
        Collections.nCopies(replicas, new Node(0, "", 0)),
        Collections.nCopies(isr, new Node(0, "", 0))
    );
  }

  private MetadataResponse.TopicMetadata buildTopicMetadata(String topic, Errors errors, int partitions, int replicas, int isr) {
    ImmutableList.Builder<MetadataResponse.PartitionMetadata> partitionMetadata = ImmutableList.builder();
    for (int i = 0; i < partitions; i++) {
      partitionMetadata.add(
          buildPartitionMetadata(i, replicas, isr)
      );
    }
    return new MetadataResponse.TopicMetadata(errors, topic, false, partitionMetadata.build());
  }

  private Set<MetadataResponse.TopicMetadata> topicMetadatas(Set<String> topics, Errors errors, int partitions, int replicas, int isr) {
    ImmutableSet.Builder<MetadataResponse.TopicMetadata> out = ImmutableSet.builder();
    for (String topic : topics) {
      out.add(buildTopicMetadata(topic, errors, partitions, replicas, isr));
    }
    return out.build();
  }

  private Set<MetadataResponse.TopicMetadata> topicMetadatas(Set<String> topics, int partitions, int replicas, int isr) {
    return topicMetadatas(topics, null, partitions, replicas, isr);
  }

  @Before
  public void before() {
    mockAdminUtils = createMock(KafkaHelper.KafkaAdminUtils.class);
    kafkaHelper = new KafkaHelper(SecurityProtocol.PLAINTEXT, new ControlCenterConfig(), zkUtilsFactory, mockAdminUtils);
  }

  @Test
  public void testVerifiesExtantTopic() {
    expect(mockAdminUtils.fetchTopicMetadataFromZk(eq(TOPICS), anyObject(ZkUtils.class), eq(SecurityProtocol.PLAINTEXT))).andReturn(
        topicMetadatas(TOPICS, 1, 1, 1)
    );
    replay(mockAdminUtils);
    assertTrue(kafkaHelper.checkCreateTopics(TOPICS, 1, 1, 100, -1, 24));
  }

  @Test
  public void testVerifiesExtantTopicWithMultiplePartitions() {
    expect(mockAdminUtils.fetchTopicMetadataFromZk(eq(TOPICS), anyObject(ZkUtils.class), eq(SecurityProtocol.PLAINTEXT))).andReturn(
        topicMetadatas(TOPICS, 35, 35, 34)
    );
    replay(mockAdminUtils);
    assertTrue(kafkaHelper.checkCreateTopics(TOPICS, 35, 35, 100, -1, 24));
  }

  @Test
  public void testFailsWithLowIsr() {
    expect(mockAdminUtils.fetchTopicMetadataFromZk(eq(TOPICS), anyObject(ZkUtils.class), eq(SecurityProtocol.PLAINTEXT))).andReturn(
        topicMetadatas(TOPICS, 2, 5, 2)
    );
    replay(mockAdminUtils);
    assertFalse(kafkaHelper.checkCreateTopics(TOPICS, 2, 5, 100, -1, 24));
  }

  @Test
  public void testFailsImproperReturnSetSize() {
    expect(mockAdminUtils.fetchTopicMetadataFromZk(eq(TOPICS), anyObject(ZkUtils.class), eq(SecurityProtocol.PLAINTEXT))).andReturn(
        topicMetadatas(Sets.<String>newHashSet(TOPIC1), 2, 5, 2)
    );
    replay(mockAdminUtils);
    assertFalse(kafkaHelper.checkCreateTopics(TOPICS, 2, 5, 100, -1, 24));
  }

  @Test
  public void testCreatesMissingTopic() {
    expect(mockAdminUtils.fetchTopicMetadataFromZk(eq(TOPICS), anyObject(ZkUtils.class), eq(SecurityProtocol.PLAINTEXT))).andReturn(
        ImmutableSet.<MetadataResponse.TopicMetadata>builder()
            .addAll(topicMetadatas(Sets.<String>newHashSet(TOPIC1), 2, 5, 5))
            .addAll(topicMetadatas(Sets.<String>newHashSet(TOPIC2), Errors.UNKNOWN_TOPIC_OR_PARTITION, 0, 0, 0))
            .build()
    );
    mockAdminUtils.createTopic(anyObject(ZkUtils.class), eq(TOPIC2), eq(2), eq(5), basicProperties(3, 100), anyObject(RackAwareMode.Enforced$.class));
    expect(mockAdminUtils.fetchTopicMetadataFromZk(eq(TOPIC2), anyObject(ZkUtils.class), eq(SecurityProtocol.PLAINTEXT))).andReturn(
        buildTopicMetadata(TOPIC2, null, 2, 5, 3)
    );

    replay(mockAdminUtils);
    assertTrue(kafkaHelper.checkCreateTopics(TOPICS, 2, 5, 100, -1, 24));
  }

  @Test
  public void testCreatesMultipleMissingTopic() {
    expect(mockAdminUtils.fetchTopicMetadataFromZk(eq(TOPICS), anyObject(ZkUtils.class), eq(SecurityProtocol.PLAINTEXT))).andReturn(
        ImmutableSet.<MetadataResponse.TopicMetadata>builder()
            .addAll(topicMetadatas(TOPICS, Errors.UNKNOWN_TOPIC_OR_PARTITION, 0, 0, 0))
            .build()
    );
    mockAdminUtils.createTopic(anyObject(ZkUtils.class), eq(TOPIC1), eq(2), eq(5), changelogProperties(3, 100), anyObject(RackAwareMode.Enforced$.class));
    expect(mockAdminUtils.fetchTopicMetadataFromZk(eq(TOPIC1), anyObject(ZkUtils.class), eq(SecurityProtocol.PLAINTEXT))).andReturn(
        buildTopicMetadata(TOPIC1, null, 2, 5, 3)
    );
    mockAdminUtils.createTopic(anyObject(ZkUtils.class), eq(TOPIC2), eq(2), eq(5), basicProperties(3, 100), anyObject(RackAwareMode.Enforced$.class));
    expect(mockAdminUtils.fetchTopicMetadataFromZk(eq(TOPIC2), anyObject(ZkUtils.class), eq(SecurityProtocol.PLAINTEXT))).andReturn(
        buildTopicMetadata(TOPIC2, Errors.NONE, 2, 5, 3)
    );

    replay(mockAdminUtils);
    assertTrue(kafkaHelper.checkCreateTopics(TOPICS, 2, 5, 100, -1, 24));
  }

  @Test
  public void testErrorCreatingTopic() {
    expect(mockAdminUtils.fetchTopicMetadataFromZk(eq(TOPICS), anyObject(ZkUtils.class), eq(SecurityProtocol.PLAINTEXT))).andReturn(
        ImmutableSet.<MetadataResponse.TopicMetadata>builder()
            .addAll(topicMetadatas(Sets.<String>newHashSet(TOPIC1), 2, 5, 5))
            .addAll(topicMetadatas(Sets.<String>newHashSet(TOPIC2), Errors.UNKNOWN_TOPIC_OR_PARTITION, 0, 0, 0))
            .build()
    );
    mockAdminUtils.createTopic(anyObject(ZkUtils.class), eq(TOPIC2), eq(2), eq(5), basicProperties(4, 100), anyObject(RackAwareMode.Enforced$.class));
    expect(mockAdminUtils.fetchTopicMetadataFromZk(eq(TOPIC2), anyObject(ZkUtils.class), eq(SecurityProtocol.PLAINTEXT))).andReturn(
        buildTopicMetadata(TOPIC2, Errors.CORRUPT_MESSAGE, 2, 5, 5)
    );

    replay(mockAdminUtils);
    assertFalse(kafkaHelper.checkCreateTopics(TOPICS, 2, 5, 100, -1, 24));
  }

  @Test
  public void testIncorrectlyCreatedTopic() {
    expect(mockAdminUtils.fetchTopicMetadataFromZk(eq(TOPICS), anyObject(ZkUtils.class), eq(SecurityProtocol.PLAINTEXT))).andReturn(
        ImmutableSet.<MetadataResponse.TopicMetadata>builder()
            .addAll(topicMetadatas(Sets.<String>newHashSet(TOPIC1), 2, 5, 5))
            .addAll(topicMetadatas(Sets.<String>newHashSet(TOPIC2), Errors.UNKNOWN_TOPIC_OR_PARTITION, 0, 0, 0))
            .build()
    );
    mockAdminUtils.createTopic(anyObject(ZkUtils.class), eq(TOPIC2), eq(2), eq(5), basicProperties(4, 100), anyObject(RackAwareMode.Enforced$.class));
    expect(mockAdminUtils.fetchTopicMetadataFromZk(eq(TOPIC2), anyObject(ZkUtils.class), eq(SecurityProtocol.PLAINTEXT))).andReturn(
        buildTopicMetadata(TOPIC2, null, 1, 3, 3)
    );

    replay(mockAdminUtils);
    assertFalse(kafkaHelper.checkCreateTopics(TOPICS, 2, 5, 100, -1, 24));
  }

  @Test
  public void testChangelogTopicsAreCompacted() {
    expect(mockAdminUtils.fetchTopicMetadataFromZk(eq(TOPICS), anyObject(ZkUtils.class), eq(SecurityProtocol.PLAINTEXT))).andReturn(
        ImmutableSet.<MetadataResponse.TopicMetadata>builder()
            .addAll(topicMetadatas(Sets.<String>newHashSet(TOPIC2), 2, 5, 5))
            .addAll(topicMetadatas(Sets.<String>newHashSet(TOPIC1), Errors.UNKNOWN_TOPIC_OR_PARTITION, 0, 0, 0))
            .build()
    );
    mockAdminUtils.createTopic(anyObject(ZkUtils.class), eq(TOPIC1), eq(2), eq(5), changelogProperties(3, 100), anyObject(RackAwareMode.Enforced$.class));
    expect(mockAdminUtils.fetchTopicMetadataFromZk(eq(TOPIC1), anyObject(ZkUtils.class), eq(SecurityProtocol.PLAINTEXT))).andReturn(
        buildTopicMetadata(TOPIC1, Errors.NONE, 2, 5, 3)
    );

    replay(mockAdminUtils);
    assertTrue(kafkaHelper.checkCreateTopics(TOPICS, 2, 5, 100, -1, 24));
  }

  public static Properties propertiesMinIsrCheck(int minIsr) {
    EasyMock.reportMatcher(new PropertyIsSet(KafkaConfig$.MODULE$.MinInSyncReplicasProp(), "" + minIsr));
    return null;
  }

  public static Properties propertiesIsCompacted() {
    EasyMock.reportMatcher(new PropertyIsSet(LogConfig$.MODULE$.CleanupPolicyProp(), "compact"));
    return null;
  }

  public static Properties propertiesIsNotCompacted() {
    EasyMock.reportMatcher(new PropertyIsSet(LogConfig$.MODULE$.CleanupPolicyProp(), null));
    return null;
  }

  public static Properties propertiesDeleteRetentionCheck(long retention) {
    EasyMock.reportMatcher(new PropertyIsSet(LogConfig$.MODULE$.DeleteRetentionMsProp(), "" + (retention + TimeUnit.DAYS.toMillis(1))));
    return null;
  }

  public static Properties propertiesRetentionMsCheck(long retention) {
    EasyMock.reportMatcher(new PropertyIsSet(LogConfig$.MODULE$.RetentionMsProp(), "" + retention));
    return null;
  }

  private Properties basicProperties(int minIsr, long retention) {
    return and(
        and(
            propertiesRetentionMsCheck(retention),
            propertiesIsNotCompacted()
        ),
        propertiesMinIsrCheck(minIsr)
    );
  }

  private Properties changelogProperties(int minIsr, long retention) {
    return and(
        and(
            propertiesDeleteRetentionCheck(retention),
            propertiesIsCompacted()
        ),
        propertiesMinIsrCheck(minIsr)
    );
  }

  private static class PropertyIsSet implements IArgumentMatcher {

    private final String property;
    private final String value;

    public PropertyIsSet(String property, String value) {
      this.property = property;
      this.value = value;
    }

    @Override
    public boolean matches(Object actual) {
      if (actual == null || !(actual instanceof Properties)) {
        return false;
      }

      String actualValue = ((Properties) actual).getProperty(property);

      if (value == null) {
        if (actualValue == null) {
          return true;
        }
        return false;
      }

      return value.equals(actualValue);
    }

    @Override
    public void appendTo(StringBuffer buffer) {
      buffer.append("PropertyIsSet(expects Properties with ");
      buffer.append(property);
      buffer.append("=");
      buffer.append(value);
    }
  }

  @Test
  public void minIsrHandlesBadInput() {
    assertEquals(1, KafkaHelper.minIsr(0));
    assertEquals(1, KafkaHelper.minIsr(-1));
    assertEquals(1, KafkaHelper.minIsr(Integer.MIN_VALUE));
  }

  @Test
  public void minIsrHandlesGoodInput() {
    assertEquals(1, KafkaHelper.minIsr(1));
    assertEquals(1, KafkaHelper.minIsr(2));
    assertEquals(2, KafkaHelper.minIsr(3));
    assertEquals(3, KafkaHelper.minIsr(4));
    assertEquals(3, KafkaHelper.minIsr(4<<3));
    assertEquals(3, KafkaHelper.minIsr(Integer.MAX_VALUE));
  }

}
