// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.serialization;

import io.confluent.command.CommandStore;
import io.confluent.command.record.Command;
import io.confluent.command.record.Command.CommandKey;
import io.confluent.command.record.Command.CommandMessage;
import io.confluent.monitoring.record.Monitoring;
import io.confluent.serializers.ProtoSerde;
import io.confluent.serializers.UberSerde;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.junit.Test;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.nio.file.Paths;
import java.util.Properties;

import static org.junit.Assert.*;

public class UberReaderTest {
  @Test
  public void readMessage() throws Exception {
    Properties props = new Properties();
    UberReader reader = new UberReader();
    String monitorMsgJson = "_confluent-monitoring\t3\t2016-06-07T16:56:23.286-07:00\tnull\t{\"clientType\":\"PRODUCER\",\"clientId\":\"myproducer\",\"session\":\"c1c45a6c-3878-44c2-af00-6d81e16c5c41\",\"sequence\":\"13\",\"window\":\"1465068255000\",\"timestamp\":\"1465068278749\",\"topic\":\"test-topic\",\"count\":\"61\",\"aggregateBytes\":\"488\",\"aggregateCrc\":1.008406505E9,\"samplePeriod\":\"15000\",\"minWindow\":\"1465068075000\",\"maxWindow\":\"1465068255000\"}";
    InputStream inputStream = new ByteArrayInputStream(monitorMsgJson.getBytes(StandardCharsets.UTF_8));
    reader.init(inputStream, props);
    ProducerRecord<byte[], byte[]> msg = reader.readMessage();
    assertEquals("_confluent-monitoring", msg.topic());
    assertEquals(3, msg.partition().intValue());
    assertEquals(1465343783286L, msg.timestamp().longValue());
    assertNull(msg.key());
    UberSerde<Monitoring.MonitoringMessage> serde = new ProtoSerde<>(Monitoring.MonitoringMessage.getDefaultInstance());
    Monitoring.MonitoringMessage monMsg = serde.deserialize(msg.value());
    assertEquals("myproducer", monMsg.getClientId());
  }

  @Test
  public void readMessageWithNulls() throws Exception {
    Properties props = new Properties();
    props.setProperty("topic", "_confluent-monitoring");
    props.setProperty("parse.topic", "true");
    props.setProperty("parse.key", "true");
    props.setProperty("parse.partition", "true");
    props.setProperty("parse.timestamp", "true");
    UberReader reader = new UberReader();
    String json = "null\tNULL\tNull\tnuLl\t{\"topic\":\"test-topic\"}";
    InputStream inputStream = new ByteArrayInputStream(json.getBytes(StandardCharsets.UTF_8));
    reader.init(inputStream, props);
    ProducerRecord<byte[], byte[]> msg = reader.readMessage();
    assertEquals("_confluent-monitoring", msg.topic());
    assertNull(msg.partition());
    assertNull(msg.timestamp());
    assertNull(msg.key());
    UberSerde<Monitoring.MonitoringMessage> serde = new ProtoSerde<>(Monitoring.MonitoringMessage.getDefaultInstance());
    Monitoring.MonitoringMessage monMsg = serde.deserialize(msg.value());
    assertEquals("test-topic", monMsg.getTopic());
  }

  @Test
  public void readMessageWithDefaultTopicAndSkipKeyAndTimestamp() throws Exception {
    Properties props = new Properties();
    props.setProperty("topic", "_confluent-monitoring");
    props.setProperty("parse.key", "false");
    props.setProperty("parse.timestamp", "false");
    UberReader reader = new UberReader();
    String json = "3\t{\"topic\":\"test-topic\"}";
    InputStream inputStream = new ByteArrayInputStream(json.getBytes(StandardCharsets.UTF_8));
    reader.init(inputStream, props);
    ProducerRecord<byte[], byte[]> msg = reader.readMessage();
    assertEquals("_confluent-monitoring", msg.topic());
    assertEquals(3, msg.partition().intValue());
    assertNull(msg.timestamp());
    assertNull(msg.key());
    UberSerde<Monitoring.MonitoringMessage> serde = new ProtoSerde<>(Monitoring.MonitoringMessage.getDefaultInstance());
    Monitoring.MonitoringMessage monMsg = serde.deserialize(msg.value());
    assertEquals("test-topic", monMsg.getTopic());
  }

  @Test
  public void readMessageWithDefaultTopicAndSkipKey() throws Exception {
    Properties props = new Properties();
    props.setProperty("topic", "_confluent-monitoring");
    props.setProperty("parse.key", "false");
    props.setProperty("parse.timestamp", "true");
    UberReader reader = new UberReader();
    String json = "3\t2016-06-07T16:56:23.286-07:00\t{\"topic\":\"test-topic\"}";
    InputStream inputStream = new ByteArrayInputStream(json.getBytes(StandardCharsets.UTF_8));
    reader.init(inputStream, props);
    ProducerRecord<byte[], byte[]> msg = reader.readMessage();
    assertEquals("_confluent-monitoring", msg.topic());
    assertEquals(3, msg.partition().intValue());
    assertEquals(1465343783286L, msg.timestamp().longValue());
    assertNull(msg.key());
    UberSerde<Monitoring.MonitoringMessage> serde = new ProtoSerde<>(Monitoring.MonitoringMessage.getDefaultInstance());
    Monitoring.MonitoringMessage monMsg = serde.deserialize(msg.value());
    assertEquals("test-topic", monMsg.getTopic());
  }

  @Test
  public void readMessageWithDefaultTopicAndSkipTimestamp() throws Exception {
    Properties props = new Properties();
    props.setProperty("topic", "_confluent-monitoring");
    props.setProperty("parse.key", "true");
    props.setProperty("parse.timestamp", "false");
    UberReader reader = new UberReader();
    String json = "3\tnull\t{\"topic\":\"test-topic\"}";
    InputStream inputStream = new ByteArrayInputStream(json.getBytes(StandardCharsets.UTF_8));
    reader.init(inputStream, props);
    ProducerRecord<byte[], byte[]> msg = reader.readMessage();
    assertEquals("_confluent-monitoring", msg.topic());
    assertEquals(3, msg.partition().intValue());
    assertNull(msg.timestamp());
    assertNull(msg.key());
    UberSerde<Monitoring.MonitoringMessage> serde = new ProtoSerde<>(Monitoring.MonitoringMessage.getDefaultInstance());
    Monitoring.MonitoringMessage monMsg = serde.deserialize(msg.value());
    assertEquals("test-topic", monMsg.getTopic());
  }

  @Test
  public void readMessageWithDefaultTopicAndSkipPartition() throws Exception {
    Properties props = new Properties();
    props.setProperty("topic", "_confluent-monitoring");
    props.setProperty("parse.partition", "false");
    UberReader reader = new UberReader();
    String json = "2016-06-07T16:56:23.286-07:00\tnull\t{\"topic\":\"test-topic\"}";
    InputStream inputStream = new ByteArrayInputStream(json.getBytes(StandardCharsets.UTF_8));
    reader.init(inputStream, props);
    ProducerRecord<byte[], byte[]> msg = reader.readMessage();
    assertEquals("_confluent-monitoring", msg.topic());
    assertNull(msg.partition());
    assertEquals(1465343783286L, msg.timestamp().longValue());
    assertNull(msg.key());
    UberSerde<Monitoring.MonitoringMessage> serde = new ProtoSerde<>(Monitoring.MonitoringMessage.getDefaultInstance());
    Monitoring.MonitoringMessage monMsg = serde.deserialize(msg.value());
    assertEquals("test-topic", monMsg.getTopic());
  }

  @Test
  public void readCmdTopic() throws Exception {
    Properties props = new Properties();
    props.setProperty("topic", "_confluent-command");
    props.setProperty("parse.partition", "false");
    props.setProperty("parse.timestamp", "false");
    UberReader reader = new UberReader();
    String json = "{\"configType\":\"CLUSTER_METADATA\",\"guid\":\"xxxxx\"}\t{\"clusterMetadata\":{\"displayName\":\"Funky Cluster\"}}";
    InputStream inputStream = new ByteArrayInputStream(json.getBytes(StandardCharsets.UTF_8));
    reader.init(inputStream, props);
    ProducerRecord<byte[], byte[]> msg = reader.readMessage();
    assertEquals("_confluent-command", msg.topic());
    UberSerde<CommandKey> keySerde = CommandStore.getCommandKeySerde();
    UberSerde<CommandMessage> valSerde = CommandStore.getCommandMessageSerde();
    CommandKey key = keySerde.deserialize(msg.key());
    assertEquals("xxxxx", key.getGuid());
    CommandMessage cmdMsg = valSerde.deserialize(msg.value());
    assertEquals("Funky Cluster", cmdMsg.getClusterMetadata().getDisplayName());
  }
}