package io.confluent.controlcenter.streams.alert;


import org.junit.Test;

import io.confluent.controlcenter.record.Controlcenter.TriggerMeasurement;
import io.confluent.monitoring.common.MonitoringMessageUtil;
import io.confluent.monitoring.record.Monitoring;
import io.confluent.monitoring.record.Monitoring.ClientType;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

public class MonitoringTriggerMeasurementJoinerTest {

  private static final String DEFAULT_CONSUMER_GROUP = "cg-0";

  private static final MonitoringMessage makeMonitoringMessage(ClientType clientType,
                                                               String group,
                                                               Monitoring.MessageType msgType,
                                                               Long arrivalTime,
                                                               Long count, Long totalLatency,
                                                               Long minLatency, Long maxLatency) {
    return MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setClientType(clientType)
        .setGroup(group)
        .setCount(count)
        .setAggregateBytes(count)
        .setAggregateCrc(0)
        .setTotalLatency(totalLatency)
        .setMinLatency(minLatency)
        .setMaxLatency(maxLatency)
        .setType(msgType)
        .setArrivalTime(arrivalTime)
        .build();
  }

  @Test
  public void testJoiningValues() {
    Long expectedCount = 10L;
    Long actualCount = 10L;
    Long totalLatency = 89L;
    Long minLatency = 100L;
    Long maxLatency = 200L;
    Long expectedMsgArrivalTime = 12345L;
    Long actualMsgArrivalTime = 12348L;

    MonitoringTriggerMeasurementJoiner joiner = new MonitoringTriggerMeasurementJoiner();
    MonitoringMessage expectedConsumptionMsg = makeMonitoringMessage(ClientType.PRODUCER,
                                                                     DEFAULT_CONSUMER_GROUP,
                                                                     Monitoring.MessageType.NORMAL,
                                                                     expectedMsgArrivalTime,
                                                                     expectedCount, 0L, 0L, 0L);
    MonitoringMessage actualConsumptionMsg = makeMonitoringMessage(ClientType.CONSUMER,
                                                                   DEFAULT_CONSUMER_GROUP,
                                                                   Monitoring.MessageType.NORMAL,
                                                                   actualMsgArrivalTime,
                                                                   actualCount, totalLatency,
                                                                   minLatency, maxLatency);
    TriggerMeasurement measurement = joiner.apply(actualConsumptionMsg, expectedConsumptionMsg);
    assertEquals(measurement.getGroup(), DEFAULT_CONSUMER_GROUP);
    assertEquals(0l, measurement.getClientMeasurement().getConsumptionDiff());
    assertEquals(maxLatency.longValue(), measurement.getClientMeasurement().getMaxLatency());
    assertEquals(totalLatency / actualCount, measurement.getClientMeasurement().getAvgLatency());
    assertFalse(measurement.getClientMeasurement().getHasError());
    assertEquals(12348, measurement.getClientMeasurement().getArrivalTime());
  }

  @Test
  public void testExpectedActualCountIsJoined() {
    Long expectedCount = 10L;
    Long actualCount = 9L;
    Long expectedMsgArrivalTime = 12345L;
    Long actualMsgArrivalTime = 12348L;

    MonitoringTriggerMeasurementJoiner joiner = new MonitoringTriggerMeasurementJoiner();
    MonitoringMessage expectedConsumptionMsg = makeMonitoringMessage(ClientType.PRODUCER,
                                                                     DEFAULT_CONSUMER_GROUP,
                                                                     Monitoring.MessageType.NORMAL,
                                                                     expectedMsgArrivalTime,
                                                                     expectedCount, 0L, 0L, 0L);
    MonitoringMessage actualConsumptionMsg = makeMonitoringMessage(ClientType.CONSUMER,
                                                                   DEFAULT_CONSUMER_GROUP,
                                                                   Monitoring.MessageType.NORMAL,
                                                                   actualMsgArrivalTime,
                                                                   actualCount, 0L, 0L, 0L);
    TriggerMeasurement measurement = joiner.apply(actualConsumptionMsg, expectedConsumptionMsg);
    assertEquals(1l, measurement.getClientMeasurement().getConsumptionDiff());
  }

  @Test
  public void testAverageLatencyIsJoined() {
    Long expectedCount = 10L;
    Long actualCount = 10L;
    Long totalLatency = 90l;
    Long expectedMsgArrivalTime = 12345L;
    Long actualMsgArrivalTime = 12348L;

    MonitoringTriggerMeasurementJoiner joiner = new MonitoringTriggerMeasurementJoiner();
    MonitoringMessage expectedConsumptionMsg = makeMonitoringMessage(ClientType.PRODUCER,
                                                                     DEFAULT_CONSUMER_GROUP,
                                                                     Monitoring.MessageType.NORMAL,
                                                                     expectedMsgArrivalTime,
                                                                     expectedCount, 0L, 0L, 0L);
    MonitoringMessage actualConsumptionMsg = makeMonitoringMessage(ClientType.CONSUMER,
                                                                   DEFAULT_CONSUMER_GROUP,
                                                                   Monitoring.MessageType.NORMAL,
                                                                   actualMsgArrivalTime,
                                                                   actualCount, totalLatency, 0L, 0L);
    TriggerMeasurement measurement = joiner.apply(actualConsumptionMsg, expectedConsumptionMsg);
    assertEquals(9, measurement.getClientMeasurement().getAvgLatency());
  }

  @Test
  public void testMaxLatencyIsJoined() {
    Long expectedCount = 10L;
    Long actualCount = 10L;
    Long maxLatency = 900L;
    Long expectedMsgArrivalTime = 12345L;
    Long actualMsgArrivalTime = 12348L;

    MonitoringTriggerMeasurementJoiner joiner = new MonitoringTriggerMeasurementJoiner();
    MonitoringMessage expectedConsumptionMsg = makeMonitoringMessage(ClientType.PRODUCER,
                                                                     DEFAULT_CONSUMER_GROUP,
                                                                     Monitoring.MessageType.NORMAL,
                                                                     expectedMsgArrivalTime,
                                                                     expectedCount, 0L, 0L, 0L);
    MonitoringMessage actualConsumptionMsg = makeMonitoringMessage(ClientType.CONSUMER,
                                                                   DEFAULT_CONSUMER_GROUP,
                                                                   Monitoring.MessageType.NORMAL,
                                                                   actualMsgArrivalTime,
                                                                   actualCount, 0L, 0L, maxLatency);
    TriggerMeasurement measurement = joiner.apply(actualConsumptionMsg, expectedConsumptionMsg);
    assertEquals(maxLatency.longValue(), measurement.getClientMeasurement().getMaxLatency());
  }

  @Test
  public void testErrorIsJoined() {
    Long expectedCount = 10L;
    Long actualCount = 10L;
    Long expectedMsgArrivalTime = 12345L;
    Long actualMsgArrivalTime = 12348L;

    MonitoringTriggerMeasurementJoiner joiner = new MonitoringTriggerMeasurementJoiner();
    MonitoringMessage expectedConsumptionMsg = makeMonitoringMessage(ClientType.PRODUCER,
                                                                     DEFAULT_CONSUMER_GROUP,
                                                                     Monitoring.MessageType.ERROR,
                                                                     expectedMsgArrivalTime,
                                                                     expectedCount, 0L, 0L, 0L);
    MonitoringMessage actualConsumptionMsg = makeMonitoringMessage(ClientType.CONSUMER,
                                                                   DEFAULT_CONSUMER_GROUP,
                                                                   Monitoring.MessageType.NORMAL,
                                                                   actualMsgArrivalTime,
                                                                   actualCount, 0L, 0L, 0L);
    TriggerMeasurement measurement = joiner.apply(actualConsumptionMsg, expectedConsumptionMsg);
    assertTrue(measurement.getClientMeasurement().getHasError());
  }

  @Test
  public void testErrorWithNoCountsNotJoined() {
    Long expectedMsgArrivalTime = 12345L;
    Long actualMsgArrivalTime = 12348L;

    MonitoringTriggerMeasurementJoiner joiner = new MonitoringTriggerMeasurementJoiner();
    MonitoringMessage expectedConsumptionMsg = makeMonitoringMessage(ClientType.PRODUCER,
                                                                     DEFAULT_CONSUMER_GROUP,
                                                                     Monitoring.MessageType.NORMAL,
                                                                     expectedMsgArrivalTime,
                                                                     0L, 0L, 0L, 0L);
    MonitoringMessage actualConsumptionMsg = makeMonitoringMessage(ClientType.CONSUMER,
                                                                   DEFAULT_CONSUMER_GROUP,
                                                                   Monitoring.MessageType.ERROR,
                                                                   actualMsgArrivalTime,
                                                                   0L, 0L, 0L, 0L);
    assertNull(joiner.apply(actualConsumptionMsg, expectedConsumptionMsg));
  }

  @Test
  public void testNullExpected() {
    Long actualCount = 10L;
    Long actualMsgArrivalTime = 12348L;

    MonitoringTriggerMeasurementJoiner joiner = new MonitoringTriggerMeasurementJoiner();
    MonitoringMessage actualConsumptionMsg = makeMonitoringMessage(ClientType.CONSUMER,
                                                                   DEFAULT_CONSUMER_GROUP,
                                                                   Monitoring.MessageType.NORMAL,
                                                                   actualMsgArrivalTime,
                                                                   actualCount, 0L, 0L, 0L);
    TriggerMeasurement measurement = joiner.apply(actualConsumptionMsg, null);
    assertEquals(-actualCount.longValue(), measurement.getClientMeasurement().getConsumptionDiff());
  }

  @Test
  public void testNullActual() {
    Long expectedCount = 10L;
    Long expectedMsgArrivalTime = 12345L;

    MonitoringTriggerMeasurementJoiner joiner = new MonitoringTriggerMeasurementJoiner();
    MonitoringMessage expectedConsumptionMsg = makeMonitoringMessage(ClientType.PRODUCER,
                                                                     DEFAULT_CONSUMER_GROUP,
                                                                     Monitoring.MessageType.ERROR,
                                                                     expectedMsgArrivalTime,
                                                                     expectedCount, 0L, 0L, 0L);
    TriggerMeasurement measurement = joiner.apply(null, expectedConsumptionMsg);
    assertNull(measurement);
  }
}
