package io.confluent.controlcenter.streams.group;

import com.google.common.collect.Sets;

import org.apache.kafka.common.utils.Bytes;
import org.junit.Test;

import java.util.Collection;

import io.confluent.controlcenter.record.Controlcenter.MemberInfo;
import io.confluent.controlcenter.record.Controlcenter.WindowedGrouping;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

public class WindowedGroupingAggregatorTest {

  public static final Bytes EMPTY_KEY = Bytes.wrap(new byte[]{});

  public static MemberInfo memberInGroup(String group) {
    return MemberInfo.newBuilder()
        .setGroup(group)
        .build();
  }

  @Test
  public void testBasicAggregatingMembers() {
    WindowedGroupingAggregator wga = new WindowedGroupingAggregator();
    WindowedGrouping aggregate = WindowedGrouping.getDefaultInstance();
    WindowedGrouping value = WindowedGrouping.newBuilder()
        .setWindow(100L)
        .addMembers(memberInGroup("group"))
        .build();

    WindowedGrouping output = wga.apply(EMPTY_KEY, value, aggregate);
    assertEquals(100L, output.getWindow());
    assertEquals(1, output.getMembersCount());
    assertEquals(output.getMembers(0), memberInGroup("group"));
  }

  @Test
  public void testAggregatingMembersCompressesMemberList() {
    WindowedGroupingAggregator wga = new WindowedGroupingAggregator(0, 0);
    WindowedGrouping aggregate = WindowedGrouping.getDefaultInstance();
    WindowedGrouping value = WindowedGrouping.newBuilder()
        .setWindow(100L)
        .addMembers(memberInGroup("group"))
        .addMembers(memberInGroup("group"))
        .build();

    WindowedGrouping output = wga.apply(EMPTY_KEY, aggregate, aggregate);
    output = wga.apply(EMPTY_KEY, value, output);
    assertEquals(100L, output.getWindow());
    assertEquals(1, output.getMembersCount());
    assertEquals(output.getMembers(0), memberInGroup("group"));
  }

  @Test
  public void testAggregatingMembersDoesntCompressesMemberList() {
    WindowedGroupingAggregator wga = new WindowedGroupingAggregator(5, 1);
    WindowedGrouping aggregate = WindowedGrouping.getDefaultInstance();
    WindowedGrouping value = WindowedGrouping.newBuilder()
        .setWindow(100L)
        .addMembers(memberInGroup("group"))
        .addMembers(memberInGroup("group"))
        .build();

    WindowedGrouping output = wga.apply(EMPTY_KEY, value, aggregate);
    output = wga.apply(EMPTY_KEY, value, output);
    assertEquals(100L, output.getWindow());
    assertEquals(4, output.getMembersCount());
    assertEquals(output.getMembers(0), memberInGroup("group"));
    assertEquals(output.getMembers(1), memberInGroup("group"));
  }

  @Test
  public void testAggregatingMembersCompressesOversizedMemberList() {
    WindowedGroupingAggregator wga = new WindowedGroupingAggregator(5, 1);
    WindowedGrouping aggregate = WindowedGrouping.getDefaultInstance();
    WindowedGrouping value = WindowedGrouping.newBuilder()
        .setWindow(100L)
        .addMembers(memberInGroup("group"))
        .addMembers(memberInGroup("group"))
        .addMembers(memberInGroup("group"))
        .addMembers(memberInGroup("group"))
        .addMembers(memberInGroup("group"))
        .addMembers(memberInGroup("group"))
        .build();

    WindowedGrouping output = wga.apply(EMPTY_KEY, value, aggregate);
    assertEquals(100L, output.getWindow());
    assertEquals(1, output.getMembersCount());
    assertEquals(memberInGroup("group"), output.getMembers(0));

    value = WindowedGrouping.newBuilder()
        .setWindow(100L)
        .addMembers(memberInGroup("group"))
        .addMembers(memberInGroup("group"))
        .addMembers(memberInGroup("group"))
        .addMembers(memberInGroup("group"))
        .addMembers(memberInGroup("group"))
        .addMembers(memberInGroup("group"))
        .addMembers(memberInGroup("group"))
        .addMembers(memberInGroup("group"))
        .build();
    output = wga.apply(EMPTY_KEY, value, output);
    assertEquals(100L, output.getWindow());
    assertEquals(1, output.getMembersCount());
    assertEquals(memberInGroup("group"), output.getMembers(0));
  }

  @Test
  public void testAddingOneMemberAtATime() {
    WindowedGroupingAggregator wga = new WindowedGroupingAggregator(5, 1);
    WindowedGrouping output = WindowedGrouping.getDefaultInstance();
    WindowedGrouping value = WindowedGrouping.newBuilder()
        .setWindow(100L)
        .addMembers(memberInGroup("group"))
        .build();

    for (int i = 0; i < 33; i++) {
      output = wga.apply(EMPTY_KEY, value, output);
      assertEquals(100L, output.getWindow());
      assertEquals((i % 6) + 1, output.getMembersCount());
      assertEquals(memberInGroup("group"), output.getMembers(0));
    }
  }

  @Test
  public void testUniqueMemberInfos() {
    Collection<? extends MemberInfo> out = WindowedGroupingAggregator.uniqueMemberInfos(
        Sets.newHashSet(memberInGroup("group")),
        Sets.newHashSet(memberInGroup("group"))
    );
    assertEquals(out.size(), 1);
    assertTrue(out.contains(memberInGroup("group")));

    out = WindowedGroupingAggregator.uniqueMemberInfos(
        out,
        Sets.newHashSet(memberInGroup("group2"))
    );
    assertEquals(out.size(), 2);
    assertTrue(out.contains(memberInGroup("group")));
    assertTrue(out.contains(memberInGroup("group2")));
  }

  @Test
  public void testUniqueMemberInfosHandlesNulls() {
    Collection<? extends MemberInfo> out = WindowedGroupingAggregator.uniqueMemberInfos(
        Sets.newHashSet(memberInGroup("group")),
        Sets.newHashSet(memberInGroup("group")),
        null
    );
    assertEquals(out.size(), 1);
    assertTrue(out.contains(memberInGroup("group")));
  }

  @Test
  public void testUniqueMemberInfosCanBeEmpty() {
    Collection<? extends MemberInfo> out = WindowedGroupingAggregator.uniqueMemberInfos();
    assertEquals(out.size(), 0);
    assertFalse(out.contains(memberInGroup("group")));
  }
}
