// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams.monitoring;

import io.confluent.monitoring.common.MonitoringMessageUtil;
import org.junit.Test;

import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.monitoring.record.Monitoring.ClientType;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotSame;

public class MonitoringMessageAggregatorTest {

  private static final String KEY = "this doesnt matter";

  private static final MonitoringMessage makeMonitoringMessage(Long count, Long bytes, Integer crc, Long
      totalLatency, Long minLatency, Long maxLatency) {
    return MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setCount(count)
        .setAggregateBytes(bytes)
        .setAggregateCrc(crc)
        .setTotalLatency(totalLatency)
        .setMinLatency(minLatency)
        .setMaxLatency(maxLatency)
        .build();
  }

  @Test
  public void testAggregatingValues() {
    long valueCount = 10l;
    long valueBytes = 1000l;
    int valueCrc = 213423438;
    long valueTotalLatency = 200l;
    long valueMinLatency = 94l;
    long valueMaxLatency = 94938l;

    long aggregateCount = 77l;
    long aggregateBytes = 45l;
    int aggregateCrc = 86445678;
    long aggregateTotalLatency = 847l;
    long aggregateMinLatency = 67l;
    long aggregateMaxLatency = 983l;

    MonitoringMessageAggregator ama = new MonitoringMessageAggregator();
    MonitoringMessage aggregate = makeMonitoringMessage(aggregateCount, aggregateBytes, aggregateCrc,
                                                        aggregateTotalLatency, aggregateMinLatency,
                                                        aggregateMaxLatency);
    MonitoringMessage value = makeMonitoringMessage(valueCount, valueBytes, valueCrc, valueTotalLatency,
                                                    valueMinLatency, valueMaxLatency);

    assertEquals(aggregate.getCount(), aggregateCount);
    assertEquals(aggregate.getAggregateBytes(), aggregateBytes);
    assertEquals(aggregate.getAggregateCrc(), aggregateCrc);
    assertEquals(aggregate.getTotalLatency(), aggregateTotalLatency);
    assertEquals(aggregate.getMinLatency(), aggregateMinLatency);
    assertEquals(aggregate.getMaxLatency(), aggregateMaxLatency);
    MonitoringMessage out = ama.apply(KEY, value, aggregate);
    assertEquals(out.getCount(), valueCount + aggregateCount);
    assertEquals(out.getAggregateBytes(), valueBytes + aggregateBytes);
    assertEquals(out.getAggregateCrc(), valueCrc ^ aggregateCrc);
    assertEquals(out.getTotalLatency(),
                 valueTotalLatency + aggregateTotalLatency);
    assertEquals(out.getMinLatency(),
                 Math.min(valueMinLatency, aggregateMinLatency));
    assertEquals(out.getMaxLatency(), Math.max(valueMaxLatency, aggregateMaxLatency));
  }

  @Test
  public void testNullValue() {
    long aggregateCount = 77l;
    long aggregateBytes = 45l;
    int aggregateCrc = 86445678;
    long aggregateTotalLatency = 847l;
    long aggregateMinLatency = 67l;
    long aggregateMaxLatency = 983l;

    MonitoringMessageAggregator ama = new MonitoringMessageAggregator();
    MonitoringMessage aggregate = makeMonitoringMessage(aggregateCount, aggregateBytes, aggregateCrc,
                                                        aggregateTotalLatency, aggregateMinLatency,
                                                        aggregateMaxLatency);
    MonitoringMessage out = ama.apply(KEY, null, aggregate);
    assertEquals(out, aggregate);
    assertEquals(aggregate.getCount(), aggregateCount);
    assertEquals(aggregate.getAggregateBytes(), aggregateBytes);
    assertEquals(aggregate.getAggregateCrc(), aggregateCrc);
    assertEquals(aggregate.getTotalLatency(), aggregateTotalLatency);
    assertEquals(aggregate.getMinLatency(), aggregateMinLatency);
    assertEquals(aggregate.getMaxLatency(), aggregateMaxLatency);
  }

  @Test
  public void testNullAggregate() {
    MonitoringMessageAggregator ama = new MonitoringMessageAggregator();
    MonitoringMessage value = MonitoringMessageUtil.baseMonitoringMessage();
    MonitoringMessage out = ama.apply(KEY, value, null);
    assertEquals(out, null);
  }

  @Test
  public void testClientTypeIsAggregated() {
    MonitoringMessageAggregator ama = new MonitoringMessageAggregator();
    MonitoringMessage aggregate = MonitoringMessageUtil.baseMonitoringMessage();
    MonitoringMessage value = MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setClientType(ClientType.PRODUCER)
        .build();
    assertEquals(aggregate.getClientType(), ClientType.CONSUMER);
    MonitoringMessage out = ama.apply(KEY, value, aggregate);
    assertEquals(out.getClientType(), ClientType.PRODUCER);
  }

  @Test
  public void testTopicIsAggregated() {
    String topic = "blah blah blah";
    MonitoringMessageAggregator ama = new MonitoringMessageAggregator();
    MonitoringMessage aggregate = MonitoringMessageUtil.baseMonitoringMessage();
    MonitoringMessage value = MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setTopic(topic)
        .build();
    assertNotSame(aggregate.getTopic(), topic);
    MonitoringMessage out = ama.apply(KEY, value, aggregate);
    assertEquals(out.getTopic(), topic);
  }

  @Test
  public void testClientIdIsAggregated() {
    String clientId = "blah blah blah";
    MonitoringMessageAggregator ama = new MonitoringMessageAggregator();
    MonitoringMessage aggregate = MonitoringMessageUtil.baseMonitoringMessage();
    MonitoringMessage value = MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setClientId(clientId)
        .build();
    assertNotSame(aggregate.getClientId(), clientId);
    MonitoringMessage out = ama.apply(KEY, value, aggregate);
    assertEquals(out.getClientId(), clientId);
  }

  @Test
  public void testGroupIsAggregated() {
    String group = "blah blah blah";
    MonitoringMessageAggregator ama = new MonitoringMessageAggregator();
    MonitoringMessage aggregate = MonitoringMessageUtil.baseMonitoringMessage();
    MonitoringMessage value = MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setGroup(group)
        .build();
    assertNotSame(aggregate.getGroup(), group);
    MonitoringMessage out = ama.apply(KEY, value, aggregate);
    assertEquals(out.getGroup(), group);
  }

  @Test
  public void testPartitionIsAggregated() {
    int partition = 28980;
    MonitoringMessageAggregator ama = new MonitoringMessageAggregator();
    MonitoringMessage aggregate = MonitoringMessageUtil.baseMonitoringMessage();
    MonitoringMessage value = MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setPartition(partition)
        .build();
    assertNotSame(aggregate.getPartition(), partition);
    MonitoringMessage out = ama.apply(KEY, value, aggregate);
    assertEquals(out.getPartition(), partition);
  }

  @Test
  public void testTopicIsNotAggregatedWhenStripped() {
    String topic = "blah blah blah";
    MonitoringMessageAggregator ama = MonitoringMessageAggregator.stripped();
    MonitoringMessage aggregate = MonitoringMessageUtil.baseMonitoringMessage();
    MonitoringMessage value = MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setTopic(topic)
        .build();
    assertNotSame(aggregate.getTopic(), topic);
    MonitoringMessage out = ama.apply(KEY, value, aggregate);
    assertEquals("", out.getTopic());
  }

  @Test
  public void testClientIdIsNotAggregatedWhenStripped() {
    String clientId = "blah blah blah";
    MonitoringMessageAggregator ama = MonitoringMessageAggregator.stripped();
    MonitoringMessage aggregate = MonitoringMessageUtil.baseMonitoringMessage();
    MonitoringMessage value = MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setClientId(clientId)
        .build();
    assertNotSame(aggregate.getClientId(), clientId);
    MonitoringMessage out = ama.apply(KEY, value, aggregate);
    assertEquals("", out.getClientId());
  }

  @Test
  public void testPartitionIsNotAggregatedWhenStripped() {
    int partition = 28980;
    MonitoringMessageAggregator ama = MonitoringMessageAggregator.stripped();
    MonitoringMessage aggregate = MonitoringMessageUtil.baseMonitoringMessage();
    MonitoringMessage value = MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setPartition(partition)
        .build();
    assertNotSame(aggregate.getPartition(), partition);
    MonitoringMessage out = ama.apply(KEY, value, aggregate);
    assertEquals(0, out.getPartition());
  }
}
