package io.confluent.controlcenter.streams.verify;


import com.google.common.collect.ImmutableSet;
import com.google.common.util.concurrent.ThreadFactoryBuilder;

import junit.framework.AssertionFailedError;

import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.clients.producer.RecordMetadata;
import org.apache.kafka.clients.producer.internals.FutureRecordMetadata;
import org.apache.kafka.clients.producer.internals.ProduceRequestResult;
import org.apache.kafka.common.PartitionInfo;
import org.apache.kafka.common.TopicPartition;
import org.easymock.EasyMock;
import org.easymock.IAnswer;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.powermock.api.mockito.PowerMockito;
import org.powermock.core.classloader.annotations.PrepareForTest;
import org.powermock.modules.junit4.PowerMockRunner;

import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import io.confluent.controlcenter.ControlCenterConfig;
import io.confluent.controlcenter.streams.MockClock;
import io.confluent.controlcenter.streams.StreamsModule;
import io.confluent.monitoring.common.Clock;
import io.confluent.monitoring.common.SystemClock;
import io.confluent.monitoring.common.TimeBucket;
import io.confluent.monitoring.record.Monitoring;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.serializers.ProtoSerde;
import io.confluent.serializers.UberSerde;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertTrue;

@RunWith(PowerMockRunner.class)
@PrepareForTest(MonitoringHeartbeatSender.class)
public class MonitoringHeartbeatSenderTest {

  private static final long PUBLISH_PERIOD_MS = 1;
  public static final String TOPIC = "monitoring_topic";
  public static final String ID = "0";
  public static final String BOOTSTRAP = "bootstrap";

  private ControlCenterConfig config;
  private KafkaProducer<byte[], byte[]> monitoringProducer;
  private MonitoringHeartbeatSender heartbeatSender;

  private List<ProducerRecord<byte[], byte[]>> sent = new CopyOnWriteArrayList<>();
  private UberSerde<Monitoring.MonitoringMessage> monitoringMessageUberSerde =
      new ProtoSerde<>(Monitoring.MonitoringMessage.getDefaultInstance());
  final ExecutorService es = Executors.newSingleThreadScheduledExecutor(
      new ThreadFactoryBuilder().build()
  );

  @Before
  public void setUp() throws Exception {
    monitoringProducer = EasyMock.createMock(KafkaProducer.class);
  }

  private void shutdownTasksAndAwaitTermination() {
    es.shutdown();
    try {
      if (!es.awaitTermination(10, TimeUnit.SECONDS)) {
        es.shutdownNow();
        if (!es.awaitTermination(60, TimeUnit.SECONDS)) {
          throw new AssertionFailedError("Failed to shutdown heartbeat sender");
        }
      }
    } catch (InterruptedException ie) {
      es.shutdownNow();
      throw new AssertionFailedError("Failed to cleanly shutdown heartbeat sender");
    }
  }

  @Test
  public void testCloseClosesProducer() {
    monitoringProducer.close();
    EasyMock.expectLastCall().once();
    EasyMock.replay(monitoringProducer);
    MonitoringHeartbeatSender mhs = new MonitoringHeartbeatSender(
        TOPIC,
        null,
        StreamsModule.provideBaseHeartbeatMonitoringMessage(ID),
        monitoringProducer,
        monitoringMessageUberSerde,
        TimeBucket.SIZE,
        new MockClock(100)
    );
    mhs.close();
    EasyMock.verify(monitoringProducer);
    assertEquals(0, sent.size());
  }

  @Test
  public void testSendHeartbeats() {
    expectPartitions(1);
    expectSend();
    EasyMock.replay(monitoringProducer);
    MonitoringHeartbeatSender mhs = new MonitoringHeartbeatSender(
        TOPIC,
        null,
        StreamsModule.provideBaseHeartbeatMonitoringMessage(ID),
        monitoringProducer,
        monitoringMessageUberSerde,
        PUBLISH_PERIOD_MS,
        new MockClock(500)
    );

    es.submit(mhs);
    try {
      // wait so that interceptor publishes heartbeat messages
      Thread.sleep(PUBLISH_PERIOD_MS + 1000);
    } catch (InterruptedException ie) {
      throw new AssertionFailedError("Did not finish waiting for publish");
    }

    shutdownTasksAndAwaitTermination();
    EasyMock.verify(monitoringProducer);
    if (sent.size() > 0) {
      ProducerRecord<byte[], byte[]> record = sent.get(0);
      MonitoringMessage msg = monitoringMessageUberSerde.deserialize(record.value());
      assertEquals(Monitoring.ClientType.CONTROLCENTER, msg.getClientType());
      assertEquals(Monitoring.MessageType.HEARTBEAT, msg.getType());
      assertNotEquals(msg.getWindow(), 0);
      assertEquals(msg.getWindow(), 500);
    }
  }

  private void expectPartitions(final int numPartitions) {
    EasyMock.expect(monitoringProducer.partitionsFor(EasyMock.anyObject(String.class))).andAnswer(
        new IAnswer<List<PartitionInfo>>() {
          @Override
          public List<PartitionInfo> answer() throws Throwable {
            ArrayList<PartitionInfo> partionsInfoList = new ArrayList<>();
            for (int i = 0; i < numPartitions; i++) {
              partionsInfoList.add(EasyMock.createMock(PartitionInfo.class));
            }
            return partionsInfoList;
          }
        }
    ).anyTimes();
  }

  private void expectSend() {
    EasyMock.expect(monitoringProducer.send(EasyMock.anyObject(ProducerRecord.class))).andAnswer(
        new IAnswer<Future<RecordMetadata>>() {
          @Override
          public Future<RecordMetadata> answer() throws Throwable {
            ProducerRecord<byte[], byte[]>
                record =
                (ProducerRecord<byte[], byte[]>) EasyMock.getCurrentArguments()[0];
            ProduceRequestResult result = new ProduceRequestResult();
            FutureRecordMetadata future = new FutureRecordMetadata(result, 0, 0, 0, 0, 0);
            TopicPartition topicPartition = new TopicPartition(record.topic(), 0);
            sent.add(record);
            result.done(topicPartition, 0, null);
            return future;
          }
        }).anyTimes();
  }
}
