// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams.verify;

import org.junit.Test;

import io.confluent.controlcenter.record.Controlcenter.SequenceInfo;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

public class SequenceWithMetadataTest {

  @Test
  public void testCreateWithMetadata() {
    final long seq = 4l;
    final long timestamp = 10002l;
    final long minWindow = 100l;
    final long maxWindow = 200l;

    SequenceWithMetadata seqWithMeta = new SequenceWithMetadata(seq, timestamp, minWindow,
                                                                maxWindow);
    assertEquals(seq, seqWithMeta.sequence());
    assertEquals(seq + 1, seqWithMeta.nextSequence());
    assertEquals(timestamp, seqWithMeta.timestamp());
    assertEquals(minWindow, seqWithMeta.minWindow());
    assertEquals(maxWindow, seqWithMeta.maxWindow());
  }

  @Test
  public void testCreateFromSequenceInfo() {
    final long seq = 6l;
    final long timestamp = 102202l;
    final long minWindow = 120l;
    final long maxWindow = 240l;

    SequenceInfo info = SequenceInfo.newBuilder()
        .setSequence(seq)
        .setTimestamp(timestamp)
        .setMinWindow(minWindow)
        .setMaxWindow(maxWindow)
        .build();

    SequenceWithMetadata seqWithMeta = new SequenceWithMetadata(info);
    assertEquals(seq, seqWithMeta.sequence());
    assertEquals(timestamp, seqWithMeta.timestamp());
    assertEquals(minWindow, seqWithMeta.minWindow());
    assertEquals(maxWindow, seqWithMeta.maxWindow());
  }

  @Test
  public void testCreateWithoutMetadata() {
    final long seq = 4l;

    SequenceWithMetadata seqWithMeta = new SequenceWithMetadata(seq);
    assertEquals(seq, seqWithMeta.sequence());
    assertEquals(seq + 1, seqWithMeta.nextSequence());
    assertEquals(-1, seqWithMeta.timestamp());
    assertEquals(-1, seqWithMeta.minWindow());
    assertEquals(-1, seqWithMeta.maxWindow());
  }

  @Test
  public void testCreateWithTimestampOnly() {
    final long seq = 4l;
    final long timestamp = 10002l;

    SequenceWithMetadata seqWithMeta = new SequenceWithMetadata(seq, timestamp);
    assertEquals(seq, seqWithMeta.sequence());
    assertEquals(seq + 1, seqWithMeta.nextSequence());
    assertEquals(timestamp, seqWithMeta.timestamp());
    assertEquals(-1, seqWithMeta.minWindow());
    assertEquals(-1, seqWithMeta.maxWindow());
  }

  @Test
  public void testCreateNextWithTimestamp() {
    final long seq = 4l;
    final long timestamp = 10002l;

    SequenceWithMetadata sequence = new SequenceWithMetadata(seq);
    SequenceWithMetadata nextSequence = sequence.nextWithTimestamp(timestamp);

    assertEquals(seq + 1, nextSequence.sequence());
    assertEquals(timestamp, nextSequence.timestamp());
    assertEquals(-1, nextSequence.minWindow());
    assertEquals(-1, nextSequence.maxWindow());
  }

  @Test
  public void testCreatePreviousWithTimestamp() {
    final long seq = 13l;
    final long timestamp = 10002l;

    SequenceWithMetadata sequence = new SequenceWithMetadata(seq);
    SequenceWithMetadata prevSequence = sequence.previousWithTimestamp(timestamp);
    assertEquals(seq - 1, prevSequence.sequence());
    assertEquals(timestamp, prevSequence.timestamp());
    assertEquals(-1, prevSequence.minWindow());
    assertEquals(-1, prevSequence.maxWindow());
  }

  @Test
  public void testCreatePreviousOfFirstWithTimestamp() {
    final long seq = 0l;

    SequenceWithMetadata sequence = new SequenceWithMetadata(seq);
    SequenceWithMetadata prevSequence = sequence.previousWithTimestamp(12034l);
    assertNull(prevSequence);
  }

  @Test
  public void testToSequenceInfo() {
    final long seq = 43l;
    final long timestamp = 10302l;
    final long minWindow = 103l;
    final long maxWindow = 203l;

    SequenceWithMetadata sequence = new SequenceWithMetadata(seq, timestamp, minWindow, maxWindow);
    SequenceInfo info = sequence.toSequenceInfo();
    assertEquals(seq, info.getSequence());
    assertEquals(timestamp, info.getTimestamp());
    assertEquals(minWindow, info.getMinWindow());
    assertEquals(maxWindow, info.getMaxWindow());
  }

  @Test
  public void testEquals() {
    final long seq = 4l;
    final long timestamp = 10002l;
    final long minWindow = 100l;
    final long maxWindow = 200l;

    SequenceWithMetadata sequence1 = new SequenceWithMetadata(seq, timestamp, minWindow, maxWindow);
    SequenceWithMetadata sequence2 = new SequenceWithMetadata(seq, timestamp, minWindow, maxWindow);
    assertTrue(sequence1.equals(sequence2));
    assertTrue(sequence2.equals(sequence1));
  }

  @Test
  public void testNotEquals() {
    final long seq = 4l;
    final long timestamp = 10002l;
    final long minWindow = 100l;
    final long maxWindow = 200l;

    SequenceWithMetadata sequence1 = new SequenceWithMetadata(seq, timestamp, minWindow, maxWindow);
    SequenceWithMetadata sequence2 = new SequenceWithMetadata(seq + 2,
                                                              timestamp + 12344,
                                                              minWindow + 1,
                                                              maxWindow + 1);
    assertFalse(sequence1.equals(sequence2));
    assertFalse(sequence2.equals(sequence1));
  }

  @Test
  public void testNotEqualsSameSequence() {
    final long seq = 4l;
    final long timestamp = 10002l;
    final long minWindow = 100l;
    final long maxWindow = 200l;

    SequenceWithMetadata sequence1 = new SequenceWithMetadata(seq, timestamp, minWindow, maxWindow);
    SequenceWithMetadata sequence2 = new SequenceWithMetadata(seq,
                                                              timestamp + 12344,
                                                              minWindow + 1,
                                                              maxWindow + 1);
    assertFalse(sequence1.equals(sequence2));
    assertFalse(sequence2.equals(sequence1));
  }

  @Test
  public void testNotEqualsSameSequenceAndTimestamp() {
    final long seq = 4l;
    final long timestamp = 10002l;
    final long minWindow = 100l;
    final long maxWindow = 200l;

    SequenceWithMetadata sequence1 = new SequenceWithMetadata(seq, timestamp, minWindow, maxWindow);
    SequenceWithMetadata sequence2 = new SequenceWithMetadata(seq, timestamp,
                                                              minWindow + 1, maxWindow + 1);
    assertFalse(sequence1.equals(sequence2));
    assertFalse(sequence2.equals(sequence1));
  }

  @Test
  public void testNotEqualsSameSequenceAndTimestampAndMinWindow() {
    final long seq = 4l;
    final long timestamp = 10002l;
    final long minWindow = 100l;
    final long maxWindow = 200l;

    SequenceWithMetadata sequence1 = new SequenceWithMetadata(seq, timestamp, minWindow, maxWindow);
    SequenceWithMetadata sequence2 = new SequenceWithMetadata(seq, timestamp,
                                                              minWindow, maxWindow + 1);
    assertFalse(sequence1.equals(sequence2));
    assertFalse(sequence2.equals(sequence1));
  }

  @Test
  public void testEqualsNullMetadata() {
    final long seq = 4l;

    SequenceWithMetadata sequence1 = new SequenceWithMetadata(seq);
    SequenceWithMetadata sequence2 = new SequenceWithMetadata(seq);
    assertTrue(sequence1.equals(sequence2));
    assertTrue(sequence2.equals(sequence1));
  }

  @Test
  public void testEqualsNullWindows() {
    final long seq = 4l;
    final long timestamp = 10002l;

    SequenceWithMetadata sequence1 = new SequenceWithMetadata(seq, timestamp);
    SequenceWithMetadata sequence2 = new SequenceWithMetadata(seq, timestamp);
    assertTrue(sequence1.equals(sequence2));
    assertTrue(sequence2.equals(sequence1));
  }

  @Test
  public void testNotEqualsNullMetadata() {
    final long seq = 4l;
    final long timestamp = 10002l;

    SequenceWithMetadata sequence1 = new SequenceWithMetadata(seq);
    SequenceWithMetadata sequence2 = new SequenceWithMetadata(seq, timestamp);
    assertFalse(sequence1.equals(sequence2));
    assertFalse(sequence2.equals(sequence1));
  }

  @Test
  public void testNotEqualsNullWindows() {
    final long seq = 4l;
    final long timestamp = 10002l;
    final long minWindow = 100l;
    final long maxWindow = 200l;

    SequenceWithMetadata sequence1 = new SequenceWithMetadata(seq, timestamp);
    SequenceWithMetadata sequence2 = new SequenceWithMetadata(seq, timestamp, minWindow, maxWindow);
    assertFalse(sequence1.equals(sequence2));
    assertFalse(sequence2.equals(sequence1));
  }

  @Test
  public void testNotEqualsNullObject() {
    SequenceWithMetadata sequence1 = new SequenceWithMetadata(5l);
    assertFalse(sequence1.equals(null));
  }

  @Test
  public void testEqualsSameObject() {
    SequenceWithMetadata sequence1 = new SequenceWithMetadata(5l);
    assertTrue(sequence1.equals(sequence1));
  }

  @Test
  public void testCompareToEquals() {
    final long seq = 4l;
    final long timestamp = 10002l;
    final long minWindow = 100l;
    final long maxWindow = 200l;

    SequenceWithMetadata sequence1 = new SequenceWithMetadata(seq, timestamp, minWindow, maxWindow);
    SequenceWithMetadata sequence2 = new SequenceWithMetadata(seq, timestamp + 1, minWindow + 1,
                                                              maxWindow + 1);

    assertEquals(0, sequence1.compareTo(sequence2));
    assertEquals(0, sequence2.compareTo(sequence1));
  }

  @Test
  public void testCompareToNotEqual() {
    final long seq = 4l;
    final long timestamp = 10002l;
    final long minWindow = 100l;
    final long maxWindow = 200l;

    SequenceWithMetadata sequence1 = new SequenceWithMetadata(seq, timestamp, minWindow, maxWindow);
    SequenceWithMetadata sequence2 = new SequenceWithMetadata(seq + 1, timestamp,
                                                              minWindow, maxWindow);

    assertTrue(sequence1.compareTo(sequence2) < 0);
    assertTrue(sequence2.compareTo(sequence1) > 0);
  }

  @Test
  public void testCompareSameObjects() {
    SequenceWithMetadata sequence1 = new SequenceWithMetadata(5l, 12423l, 33333l, 55555555l);
    assertEquals(0, sequence1.compareTo(sequence1));
  }

  @Test
  public void testHashCode() {
    final long seq = 4l;
    final long timestamp = 10002l;
    final long minWindow = 100l;
    final long maxWindow = 200l;

    SequenceWithMetadata sequence1 = new SequenceWithMetadata(seq, timestamp, minWindow, maxWindow);
    SequenceWithMetadata sequence2 = new SequenceWithMetadata(seq, timestamp, minWindow, maxWindow);
    assertEquals(sequence1.hashCode(), sequence2.hashCode());
  }

  @Test
  public void testHashCodeDifferent() {
    final long seq = 4l;
    final long timestamp = 10002l;
    final long minWindow = 100l;
    final long maxWindow = 200l;

    SequenceWithMetadata sequence1 = new SequenceWithMetadata(seq, timestamp, minWindow, maxWindow);
    SequenceWithMetadata sequence2 = new SequenceWithMetadata(seq + 1, timestamp + 1,
                                                              minWindow + 1, maxWindow + 1);
    SequenceWithMetadata sequence3 = new SequenceWithMetadata(seq + 1);
    SequenceWithMetadata sequence4 = new SequenceWithMetadata(seq + 1, timestamp);
    assertNotEquals(sequence1.hashCode(), sequence2.hashCode());
    assertNotEquals(sequence1.hashCode(), sequence3.hashCode());
    assertNotEquals(sequence1.hashCode(), sequence4.hashCode());
  }
}
