package io.confluent.controlcenter.streams.verify;

import org.apache.kafka.clients.producer.Producer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.serialization.Serializer;
import org.apache.kafka.streams.kstream.ValueTransformer;
import org.apache.kafka.streams.processor.StateStoreSupplier;
import org.apache.kafka.streams.processor.internals.RecordCollector;
import org.apache.kafka.streams.processor.internals.RecordCollectorImpl;
import org.apache.kafka.streams.state.KeyValueStore;
import org.apache.kafka.streams.state.Stores;
import org.easymock.EasyMock;
import org.junit.Before;
import org.junit.Test;

import java.io.File;
import java.nio.file.Files;
import java.util.UUID;
import java.util.concurrent.TimeUnit;

import io.confluent.controlcenter.Rollup;
import io.confluent.controlcenter.record.Controlcenter.VerifiableMonitoringMessage;
import io.confluent.controlcenter.streams.MockClock;
import io.confluent.controlcenter.streams.MockProcessorContext;
import io.confluent.controlcenter.streams.TopicStoreMaster;
import io.confluent.monitoring.common.MonitoringMessageUtil;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.serializers.ProtoSerde;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

public class ThroughTopicVerifierTransformerSupplierTest {

  private static final long START_TIME_MS = 1;
  private static final Rollup rollup = Rollup.FIFTEEN_SECONDS;
  public static final int ENTRIES = 2 << 4;
  private final String streamTopic = "stream-topic";
  private final String storeName = TopicStoreMaster.nameJoin(streamTopic, rollup.name());
  private final String guid = UUID.randomUUID().toString();
  private final String firstSession = "Session-one";
  private final long retainMs = TimeUnit.SECONDS.toMillis(4);
  private final MockClock mockClock = new MockClock();

  private MockProcessorContext context;
  private Producer<byte[], byte[]> mockProducer;
  private KeyValueStore<String, VerifiableMonitoringMessage> kvStore;
  private File stateDir;

  private ThroughTopicVerifierTransformerSupplier supplier;
  private ValueTransformer<VerifiableMonitoringMessage, MonitoringMessage> transformer;

  @Before
  public void setUp() throws Exception {
    StateStoreSupplier storeSupplier = Stores.create(storeName)
        .withStringKeys()
        .withValues(new ProtoSerde<>(VerifiableMonitoringMessage.getDefaultInstance()))
        .inMemory()
        .maxEntries(ENTRIES)
        .build();

    // the code block below just to be able to test persistent kv store backed by rocksdb
    stateDir = Files.createTempDirectory("test").toFile();
    stateDir.deleteOnExit();
    mockProducer = EasyMock.createMock(Producer.class);
    RecordCollector recordCollector = new RecordCollectorImpl(mockProducer, "fakeTaskId") {
      @Override
      public <K1, V1> void send(ProducerRecord<K1, V1> record, Serializer<K1> keySerializer,
                                Serializer<V1> valueSerializer) {
        // nothing
      }
    };
    context = new MockProcessorContext(stateDir, recordCollector);

    kvStore = (KeyValueStore<String, VerifiableMonitoringMessage>) storeSupplier.get();
    kvStore.init(context, kvStore);
    context.register(kvStore, false, null);

    mockClock.setCurrentTimeMillis(START_TIME_MS);
    supplier = new ThroughTopicVerifierTransformerSupplier(
        storeName,
        mockClock
    );
    transformer = supplier.get();
    transformer.init(context);
  }

  @Test
  public void testVerifyPassThrough() {
    VerifiableMonitoringMessage msg = createMessage(guid);
    MonitoringMessage ret = transformer.transform(msg);
    assertEquals(msg.getMonitoringMessage(), ret);

    // verify the key was written to kv store
    assertNotNull(kvStore.get(msg.getGuid()));
  }

  @Test
  public void testVerifyFilterDuplicate() {
    VerifiableMonitoringMessage msg = createMessage(guid);
    MonitoringMessage ret = transformer.transform(msg);
    assertEquals(msg.getMonitoringMessage(), ret);
    assertNull(transformer.transform(msg));
  }

  @Test
  public void testMultipleGuidsNoDuplicates() {
    final int sessionCount = 5;
    for (int i = 0; i < sessionCount; i++) {
      VerifiableMonitoringMessage msg = createMessage(String.format("%d-%s", i, firstSession));
      MonitoringMessage ret = transformer.transform(msg);
      assertEquals(msg.getMonitoringMessage(), ret);
    }

    // verify we have 'sessionCount' number of entries in the db
    for (int i = 0; i < sessionCount; i++) {
      String session = String.format("%d-%s", i, firstSession);
      assertNotNull(kvStore.get(session));
    }
  }

  @Test
  public void testMultipleGuidsHasDuplicatesOverLimit() {
    final int sessionCount = ENTRIES * 2;
    for (int i = 0; i < sessionCount; i++) {
      VerifiableMonitoringMessage msg = createMessage(String.format("%d-%s", i, firstSession));
      MonitoringMessage ret = transformer.transform(msg);
      assertEquals(msg.getMonitoringMessage(), ret);
    }

    // verify we have 'sessionCount' number of entries in the db
    for (int i = 0; i < ENTRIES; i++) {
      String session = String.format("%d-%s", i, firstSession);
      assertNull(kvStore.get(session));
    }

    for (int i = ENTRIES; i < ENTRIES * 2; i++) {
      String session = String.format("%d-%s", i, firstSession);
      assertNotNull(kvStore.get(session));
    }
  }


  @Test
  public void testPunctuateDoesNotRemoveRecentMessage() {
    VerifiableMonitoringMessage msg = createMessage(guid);
    MonitoringMessage ret = transformer.transform(msg);
    assertEquals(msg.getMonitoringMessage(), ret);
    // make sure it's a new object (so we can compare later)
    VerifiableMonitoringMessage msgBefore = kvStore.get(guid);
    assertNotNull(msgBefore);
    assertEquals(msgBefore.getMonitoringMessage(), MonitoringMessage.getDefaultInstance());

    mockClock.setCurrentTimeMillis(START_TIME_MS + retainMs / 2);

    assertNull(transformer.punctuate(0));

    // state should not change
    VerifiableMonitoringMessage verifierAfter = kvStore.get(guid);
    assertNotNull(verifierAfter);
    assertEquals(msgBefore, verifierAfter);
    assertEquals(msgBefore.getMonitoringMessage(), verifierAfter.getMonitoringMessage());
  }

  /**
   * This transformer cares only about session and sequence number
   */
  private VerifiableMonitoringMessage createMessage(String guid) {
    MonitoringMessage monitoringMessage = MonitoringMessageUtil.baseMonitoringMessage();
    return VerifiableMonitoringMessage.newBuilder()
        .setMonitoringMessage(monitoringMessage)
        .setGuid(guid)
        .build();
  }
}
