// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.streams.verify;

import com.google.inject.Guice;
import com.google.inject.Injector;
import com.google.inject.Key;

import org.apache.kafka.clients.producer.Producer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.serialization.Serializer;
import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.kstream.Transformer;
import org.apache.kafka.streams.processor.StateStoreSupplier;
import org.apache.kafka.streams.processor.internals.RecordCollector;
import org.apache.kafka.streams.processor.internals.RecordCollectorImpl;
import org.apache.kafka.streams.state.KeyValueStore;
import org.apache.kafka.streams.state.Stores;
import org.easymock.EasyMock;
import org.junit.Before;
import org.junit.Test;

import java.io.File;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.TimeUnit;

import io.confluent.controlcenter.serialization.MonitoringVerifierSerde;
import io.confluent.controlcenter.serialization.SerializationModule;
import io.confluent.controlcenter.serialization.SerializationModule.VerifierKeySerde;
import io.confluent.controlcenter.streams.MockClock;
import io.confluent.controlcenter.streams.MockProcessorContext;
import io.confluent.monitoring.common.MonitoringMessageUtil;
import io.confluent.monitoring.common.TimeBucket;
import io.confluent.monitoring.record.Monitoring;
import io.confluent.monitoring.record.Monitoring.MessageType;
import io.confluent.monitoring.record.Monitoring.MonitoringMessage;
import io.confluent.serializers.OrderedKeyUberSerde;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

public class VerifyTransformerSupplierTest {

  private static final String STORE_NAME = "MonitoringVerifierTestStore";
  private static final long RETAIN_MS = TimeUnit.SECONDS.toMillis(2);
  private static final long MAX_PUNCTUATE_MS = TimeUnit.SECONDS.toMillis(2);
  private static final long PUNCTUATE_INTERVAL_MS = 3 * TimeBucket.SIZE;
  private static final long START_TIME_MS = 1;
  private final String clusterId = "fakeCluster";
  private final String clientId = "test-client";
  private final String topic = "test-topic";
  private final Integer partition = 0;
  private final String firstSession = "Session-one";
  private final long samplePeriod = TimeBucket.SIZE;
  private final long timeout = samplePeriod;
  private final long startSequence = 0;
  private final byte[] key = new byte[]{};
  private final MockClock mockClock = new MockClock();
  private final Injector injector = Guice.createInjector(new SerializationModule());
  private final OrderedKeyUberSerde<MonitoringMessage> keySerde =
      injector.getInstance(new Key<OrderedKeyUberSerde<MonitoringMessage>>(VerifierKeySerde.class) {});

  private MockProcessorContext context;
  private Producer<byte[], byte[]> mockProducer;
  private KeyValueStore<Bytes, MonitoringVerifier> kvStore;
  private File stateDir;

  private VerifyTransformerSupplier supplier;
  private Transformer<byte[],
      MonitoringMessage,
      KeyValue<byte[], Iterable<KeyValue<byte[], MonitoringMessage>>>> transformer;

  @Before
  public void setUp() throws Exception {
    StateStoreSupplier storeSupplier = Stores.create(STORE_NAME)
        .withKeys(keySerde)
        .withValues(new MonitoringVerifierSerde())
        .persistent()
        .build();

    // aahh all that code block below just to be able to test persistent kv store backed by rocksdb
    stateDir = Files.createTempDirectory("test").toFile();
    stateDir.deleteOnExit();
    mockProducer = EasyMock.createMock(Producer.class);
    RecordCollector recordCollector = new RecordCollectorImpl(mockProducer, "fakeTaskId") {
      @Override
      public <K1, V1> void send(ProducerRecord<K1, V1> record, Serializer<K1> keySerializer,
                                Serializer<V1> valueSerializer) {
        // nothing
      }
    };
    context = new MockProcessorContext(stateDir, recordCollector);

    kvStore = (KeyValueStore<Bytes, MonitoringVerifier>) storeSupplier.get();
    kvStore.init(context, kvStore);
    context.register(kvStore, false, null);

    mockClock.setCurrentTimeMillis(START_TIME_MS);
    supplier = new VerifyTransformerSupplier(timeout, STORE_NAME, RETAIN_MS, MAX_PUNCTUATE_MS,
                                             PUNCTUATE_INTERVAL_MS, 2, mockClock, keySerde);
    transformer = supplier.get();
    transformer.init(context);
  }

  @Test
  public void testVerifyPassThrough() {
    final long startWindow = 1000;
    final long startTime = 1100;

    MonitoringMessage msg = createMessage(startSequence, firstSession, startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret = verifyTransformReturnsOneKV(key, msg);
    assertEquals(msg, ret.value);

    // verify the key was written to kv store
    verifyMessageKeyStored(kvStore, msg);
  }

  @Test
  public void testVerifySetsArrivalTime() {
    final long startWindow = 1000;
    final long startTime = 1100;
    int msgCount = 0;

    // C3 heartbeat messages to set partition clocks
    msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence, 0L, startTime + 1, 0));
    msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence, 0L, startTime + 2, 1));
    assertEquals(0, msgCount);

    MonitoringMessage msg = createMessage(MessageType.NORMAL, Monitoring.ClientType.PRODUCER, partition, 1,
                                          startSequence, firstSession, startWindow, startTime, 1, false);
    KeyValue<byte[], MonitoringMessage> ret = verifyTransformReturnsOneKV(key, msg);
    assertEquals(startTime + 2, ret.value.getArrivalTime());

    MonitoringMessage msg2 = createMessage(MessageType.NORMAL, Monitoring.ClientType.PRODUCER,
                                           partition, 0, startSequence + 1, firstSession,
                                           startWindow, startTime, 1, false);
    KeyValue<byte[], MonitoringMessage> ret2 = verifyTransformReturnsOneKV(key, msg2);
    assertEquals(startTime + 1, ret2.value.getArrivalTime());
  }

  @Test
  public void testVerifyFilterDuplicate() {
    final long startWindow = 1000;
    final long startTime = 1100;

    MonitoringMessage msg = createMessage(startSequence, firstSession, startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret1 = verifyTransformReturnsOneKV(key, msg);
    assertEquals(msg, ret1.value);
    assertNull(transformer.transform(key, msg));
  }

  @Test
  public void testVerifyFilterMessageAfterDone() {
    final long startWindow = 1000;
    final long startTime = 1100;

    MonitoringMessage msg1 = createMessageWithShutdown(startSequence, firstSession,
                                                       startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret1 = verifyTransformReturnsOneKV(key, msg1);
    assertEquals(msg1, ret1.value);
    MonitoringMessage msg2 = createMessage(startSequence + 1, firstSession,
                                           startWindow + 10, startTime + 10);
    assertNull(transformer.transform(key, msg2));
  }

  @Test
  public void testVerifyFilterSecondShutdown() {
    final long startWindow = 1000;
    final long startTime = 1100;

    MonitoringMessage msg = createMessageWithShutdown(startSequence, firstSession,
                                                      startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret1 = verifyTransformReturnsOneKV(key, msg);
    assertEquals(msg, ret1.value);
    assertNull(transformer.transform(key, msg));
  }

  @Test
  public void testVerifyFilterHeartbeatMessages() {
    final long startWindow = 1000;
    final long startTime = 1100;

    // heartbeat message
    MonitoringMessage hMsg = createHeartbeatMessageNoShutdown(startSequence, firstSession,
                                                              startTime);
    assertNull(transformer.transform(key, hMsg));

    MonitoringMessage msg = createMessage(startSequence + 1, firstSession, startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret1 = verifyTransformReturnsOneKV(key, msg);
    assertEquals(msg, ret1.value);

    // one more heartbeat message
    MonitoringMessage hMsg2 = createHeartbeatMessageNoShutdown(startSequence + 2, firstSession,
                                                               startTime + 20);
    assertNull(transformer.transform(key, hMsg2));
  }

  @Test
  public void testVerifyFilterMessageHigherSequenceAfterShutdown() {
    final long startWindow = 1000;
    final long startTime = 1100;

    MonitoringMessage msg1 = createMessageWithShutdown(startSequence + 1, firstSession,
                                                       startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret1 = verifyTransformReturnsOneKV(key, msg1);
    assertEquals(msg1, ret1.value);
    MonitoringMessage msg2 = createMessage(startSequence + 2, firstSession,
                                           startWindow + 10, startTime + 10);
    assertNull(transformer.transform(key, msg2));
  }

  @Test
  public void testVerifyFilterMessageOutOfOrderShutdown() {
    final long startWindow = 1000;
    final long startTime = 1100;

    MonitoringMessage msg1 = createMessage(startSequence + 1, firstSession,
                                           startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret1 = verifyTransformReturnsOneKV(key, msg1);
    assertEquals(msg1, ret1.value);
    MonitoringMessage msg2 = createMessageWithShutdown(startSequence, firstSession,
                                                       startWindow + 10, startTime + 10);
    assertNull(transformer.transform(key, msg2));
  }

  @Test
  public void testTwoSessions() {
    final long startWindow = 1000;
    final long startTime = 1100;

    MonitoringMessage msg1 = createMessage(startSequence, firstSession,
                                           startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret1 = verifyTransformReturnsOneKV(key, msg1);
    assertEquals(msg1, ret1.value);
    // make sure it's a new object (so we can compare later)
    MonitoringVerifier verifierBeforeSessionChange =
        new MonitoringVerifier(getVerifier(kvStore, msg1).createVerifierInfo());
    assertNotNull(verifierBeforeSessionChange);

    MonitoringMessage msg2 = createMessage(startSequence, "session-2",
                                           startWindow + 10, startTime + 10);
    KeyValue<byte[], MonitoringMessage> ret2 = verifyTransformReturnsOneKV(key, msg2);
    assertEquals(msg2, ret2.value);
    verifyMessageKeyStored(kvStore, msg2);

    // the last transform must change the state of verifier for the first session
    MonitoringVerifier verifierAfterSessionChange = getVerifier(kvStore, msg1);
    assertNotNull(verifierAfterSessionChange);
    assertNotEquals(verifierAfterSessionChange, verifierBeforeSessionChange);
  }

  @Test
  public void testSessionEnd() {
    final long startWindow = 1000;
    final long startTime = 1100;
    final String baseSessionId = "session";
    List<MonitoringMessage> msgList = new ArrayList<>();

    final int count = 10;
    for (int i = 0; i < count; i++) {
      MonitoringMessage msg = createMessage(startSequence,
                                            String.format("FF%d-%s", i, baseSessionId),
                                            startWindow, startTime);
      KeyValue<byte[], MonitoringMessage> ret = verifyTransformReturnsOneKV(key, msg);
      assertEquals(msg, ret.value);
      msgList.add(msg);
    }

    // all session except last should be marked as ended
    for (MonitoringMessage msg : msgList) {
      MonitoringVerifier verifier = getVerifier(kvStore, msg);
      if (!verifier.getSession().equals(String.format("FF%d-%s", count - 1, baseSessionId))) {
        assertTrue(verifier.isSessionEnded());
      } else {
        assertFalse(verifier.isSessionEnded());
      }
    }
  }

  @Test
  public void testTwoSessionsDifferentPartition() {
    final long startWindow = 1000;
    final long startTime = 1100;

    MonitoringMessage msg1 = createMessage(startSequence, firstSession,
                                           startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret1 = verifyTransformReturnsOneKV(key, msg1);
    assertEquals(msg1, ret1.value);
    // make sure it's a new object (so we can compare later)
    MonitoringVerifier verifierBeforeSessionChange =
        new MonitoringVerifier(getVerifier(kvStore, msg1).createVerifierInfo());
    assertNotNull(verifierBeforeSessionChange);

    MonitoringMessage msg2 = MonitoringMessage.newBuilder(createMessage(startSequence, "session-2", startWindow + 10, startTime + 10))
        .setPartition(partition + 1)
        .build();
    KeyValue<byte[], MonitoringMessage> ret2 = verifyTransformReturnsOneKV(key, msg2);
    assertEquals(msg2, ret2.value);
    verifyMessageKeyStored(kvStore, msg2);

    // must not change the state of verifier for the first session
    MonitoringVerifier verifierAfterSessionChange = getVerifier(kvStore, msg1);
    assertNotNull(verifierAfterSessionChange);
    assertEquals(verifierAfterSessionChange, verifierBeforeSessionChange);
  }

  @Test
  public void testPunctuate() {
    final long startWindow = 1000;
    final long startTime = 1100;

    MonitoringMessage msg1 = createMessage(startSequence, firstSession,
                                           startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret1 = verifyTransformReturnsOneKV(key, msg1);
    assertEquals(msg1, ret1.value);
    // make sure it's a new object (so we can compare later)
    MonitoringVerifier verifierBefore =
        new MonitoringVerifier(getVerifier(kvStore, msg1).createVerifierInfo());
    assertNotNull(verifierBefore);

    assertNull(transformer.punctuate(0));

    // since we did punctuate right away, state should not change
    MonitoringVerifier verifierAfter = getVerifier(kvStore, msg1);
    assertNotNull(verifierAfter);
    assertEquals(verifierBefore, verifierAfter);
  }

  @Test
  public void testPunctuateRemovesOldVerifiers() {
    final String baseSessionId = "session";
    final long startWindow = 1000;
    final long startTime = 1100;
    List<MonitoringMessage> msgListDone = new ArrayList<>();
    List<MonitoringMessage> msgListInProg = new ArrayList<>();

    // simulate sessions with only one message that could be reclaimed
    assertNull(transformer.punctuate(startTime));

    final int sessionCount = 5;
    for (int i = 0; i < sessionCount; i++) {
      MonitoringMessage msg = createMessageWithShutdown(startSequence,
                                                        String.format("%d-%s", i, baseSessionId),
                                                        startWindow, startTime);
      KeyValue<byte[], MonitoringMessage> ret = verifyTransformReturnsOneKV(key, msg);
      assertEquals(msg, ret.value);
      MonitoringVerifier verifier = getVerifier(kvStore, msg);
      assertTrue(verifier.isDone());
      msgListDone.add(msg);
    }

    // simulate some sessions that are still in progress
    final int numPartitions = 4;
    for (int i = 0; i < numPartitions; i++) {
      MonitoringMessage msg = MonitoringMessage.newBuilder(createMessage(startSequence, "other-session", startWindow, startTime))
          .setPartition(partition + i)
          .build();
      KeyValue<byte[], MonitoringMessage> ret = verifyTransformReturnsOneKV(key, msg);
      assertEquals(msg, ret.value);
      MonitoringVerifier verifier = getVerifier(kvStore, msg);
      assertFalse(verifier.isDone());
      msgListInProg.add(msg);
    }

    mockClock.setCurrentTimeMillis(START_TIME_MS + PUNCTUATE_INTERVAL_MS + 100);

    // not enough time to expire sequence numbers (after highest)
    assertNull(transformer.punctuate(startTime + PUNCTUATE_INTERVAL_MS));

    // sessions that are done are expected to be deleted
    for (MonitoringMessage msg : msgListDone) {
      assertNull(getVerifier(kvStore, msg));
    }

    // but not ended sessions should still be there
    for (MonitoringMessage msg : msgListInProg) {
      MonitoringVerifier verifier = getVerifier(kvStore, msg);
      assertNotNull(verifier);
      assertFalse(verifier.isDone());
    }
  }

  @Test
  public void testTransformCallsPunctuateOnC3HeartbeatOnly() {
    final long startWindow = 1000;
    final long startTime = 1100;
    int msgCount = 0;

    // first heartbeat to set time
    msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence, 0L, START_TIME_MS, 0));
    msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence, 0L, START_TIME_MS, 1));
    assertEquals(0, msgCount);

    // simulate one session with a missing sequence
    MonitoringMessage msg = createMessage(startSequence, "s0", startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret = verifyTransformReturnsOneKV(key, msg);
    assertEquals(MonitoringMessage.newBuilder(msg).setArrivalTime(START_TIME_MS).build(), ret.value);
    MonitoringMessage msg1 = createMessage(startSequence + 2, "s0", startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret1 = verifyTransformReturnsOneKV(key, msg1);
    assertEquals(MonitoringMessage.newBuilder(msg1).setArrivalTime(START_TIME_MS).build(), ret1.value);

    // start session for a different client/topic/partition
    MonitoringMessage newMsg = MonitoringMessage.newBuilder(createMessage(startSequence, "s2",
                                                                          startWindow + TimeBucket.SIZE,
                                                                          startTime + TimeBucket.SIZE))
        .setPartition(2)
        .build();
    KeyValue<byte[], MonitoringMessage> newRet = verifyTransformReturnsOneKV(key, newMsg);
    assertEquals(MonitoringMessage.newBuilder(newMsg).setArrivalTime(START_TIME_MS).build(), newRet.value);

    // wait until we can expire sequence of the first session
    mockClock.setCurrentTimeMillis(START_TIME_MS + timeout + 100);

    // should not punctuate on normal message
    MonitoringMessage msg2 = MonitoringMessage.newBuilder(createMessage(startSequence + 1, "s2",
                                                                          startWindow + 2 * TimeBucket.SIZE,
                                                                          startTime + 2 * TimeBucket.SIZE))
        .setPartition(2)
        .build();
    // verify that transform can now expire sequences of different client/topic/partition,
    // because it also calls punctuate
    KeyValue<byte[], MonitoringMessage> ret2 = verifyTransformReturnsOneKV(key, msg2);
    assertEquals(MonitoringMessage.newBuilder(msg2).setArrivalTime(START_TIME_MS).build(), ret2.value);

    // should punctuate on second C3 heartbeat (after punctuate interval)
    msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence + 1, 0L,
                                                               START_TIME_MS + PUNCTUATE_INTERVAL_MS + 100, 0));
    assertEquals(0, msgCount);
    msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence + 1, 0L,
                                                               START_TIME_MS + PUNCTUATE_INTERVAL_MS + 100, 1));
    assertTrue(msgCount > 0);
  }

  @Test
  public void testFrequentPunctuate() {
    final long startWindow = 1000;
    final long startTime = 1100;
    int msgCount = 0;

    // first heartbeat to set time
    msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence, 0L, START_TIME_MS, 0));
    msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence, 0L, START_TIME_MS, 1));
    assertEquals(0, msgCount);

    MonitoringMessage msg = createMessage(startSequence, firstSession, startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret = verifyTransformReturnsOneKV(key, msg);
    assertEquals(MonitoringMessage.newBuilder(msg).setArrivalTime(START_TIME_MS).build(), ret.value);

    // there must be at least one window with missing monitoring data
    // first punctuate does not punctuate
    msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence + 1, 0L,
                                                               START_TIME_MS + 7 * samplePeriod, 0));
    msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence + 1, 0L,
                                                               START_TIME_MS + 7 * samplePeriod, 1));
    assertTrue(msgCount > 0);

    // if we call punctuate right away, there should be no missing windows
    assertNull(transformer.punctuate(START_TIME_MS + 7 * samplePeriod + 1));
    assertNull(transformer.punctuate(START_TIME_MS + 7 * samplePeriod + 2));
    assertNull(transformer.punctuate(START_TIME_MS + 7 * samplePeriod + 3));
  }

  @Test
  public void testPunctuateDoesNotDependOnWallClock() {
    MonitoringMessage msg = createMessage(startSequence, firstSession, 1000L, START_TIME_MS);
    KeyValue<byte[], MonitoringMessage> ret = verifyTransformReturnsOneKV(key, msg);
    assertEquals(msg, ret.value);

    // since the message arrived before the first heartbeat, we need 3 heartbeats to verify that
    // the next sequences is missing
    int msgCount = 0;
    for (int i = 0; i < 3; i++) {
      msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence + 1, 0L,
                                                                 START_TIME_MS + i * 7 * samplePeriod + 10, 0));
      msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence + 1, 0L,
                                                                 START_TIME_MS + i * 7 * samplePeriod + 10, 1));

      if (i < 2) {
        assertEquals(0, msgCount);
      }
    }
    assertTrue(msgCount > 0);
  }

  @Test
  public void testPunctuateTimestampBehindWallClock() {
    final long startWindow = 1000;
    final long startTime = 1100;

    MonitoringMessage msg = createMessage(startSequence, firstSession, startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret = verifyTransformReturnsOneKV(key, msg);
    assertEquals(msg, ret.value);

    // advance wall clock enough to normally expire sequence
    mockClock.setCurrentTimeMillis(START_TIME_MS + 7 * samplePeriod);

    // first punctuate does not expire anything
    transformer.punctuate(START_TIME_MS);

    // punctuate on timestamp < punctuate interval after the first punctuate should not expire
    // sequences even though wallclock time advanced further
    KeyValue<byte[], Iterable<KeyValue<byte[], MonitoringMessage>>> lst =
        transformer.punctuate(START_TIME_MS + samplePeriod);
    assertNull(lst);
  }

  @Test
  public void testSlowPartitionDoesNotMoveClock() {
    final long startWindow = 1000;
    final long startTime = 1100;

    MonitoringMessage msg = createMessage(startSequence, firstSession, startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret = verifyTransformReturnsOneKV(key, msg);
    assertEquals(msg, ret.value);

    // heartbeats from one partition only do not move clock (no sequences get expired)
    int count = transformAndReturnCount(key,
                                        createC3Heartbeat(startSequence, 0L, START_TIME_MS + 10, 0));
    count += transformAndReturnCount(key,
                                     createC3Heartbeat(startSequence + 1, 0L,
                                                       START_TIME_MS + 7 * samplePeriod + 10, 0));
    count += transformAndReturnCount(key,
                                     createC3Heartbeat(startSequence + 1, 0L,
                                                       START_TIME_MS + 14 * samplePeriod + 10, 0));
    assertTrue(count == 0);


    // first punctuate on the second partition should call punctuate but not verify
    count += transformAndReturnCount(key,
                                     createC3Heartbeat(startSequence + 1, 0L,
                                                       START_TIME_MS + 10, 1));
    assertTrue(count == 0);
    // third punctuate should expire at least one sequence
    count += transformAndReturnCount(key,
                                     createC3Heartbeat(startSequence + 1, 0L,
                                                       START_TIME_MS + 7 * samplePeriod + 10, 1));
    count += transformAndReturnCount(key,
                                     createC3Heartbeat(startSequence + 1, 0L,
                                                       START_TIME_MS + 14 * samplePeriod + 10, 1));
    assertTrue(count > 0);
  }

  @Test
  public void testTransformReturnsMissingWindows() {
    final long startWindow = 1000;
    final long startTime = 1100;
    int msgCount = 0;

    // first message did not arrive, but heartbeat arrived
    msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence, 0L, startTime, 0));
    msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence, 0L, startTime, 1));
    assertEquals(0, msgCount);

    MonitoringMessage msg = createMessage(startSequence + 1, firstSession, startWindow, startTime);
    KeyValue<byte[], MonitoringMessage> ret = verifyTransformReturnsOneKV(key, msg);
    assertEquals(MonitoringMessage.newBuilder(msg).setArrivalTime(startTime).build(), ret.value);

    // another heartbeat
    msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence + 1, 0L,
                                                               startTime + 2 * samplePeriod,
                                                               0));
    msgCount += transformAndReturnCount(key, createC3Heartbeat(startSequence + 1, 0L,
                                                               startTime + 2 * samplePeriod,
                                                               1));
    assertEquals(0, msgCount);

    MonitoringMessage msg2 = createMessage(startSequence + 3, firstSession,
                                           startWindow + 3 * samplePeriod,
                                           startTime + 3 * samplePeriod);
    KeyValue<byte[], Iterable<KeyValue<byte[], MonitoringMessage>>> lst =
        transformer.transform(key, msg2);
    assertNotNull(lst);

    // the returned list must contain msg2 and at least one more message
    Iterator<KeyValue<byte[], MonitoringMessage>> it = lst.value.iterator();
    boolean found = false;
    int count = 0;
    while (it.hasNext()) {
      KeyValue<byte[], MonitoringMessage> transformMsg = it.next();
      if (transformMsg.value.equals(MonitoringMessage.newBuilder(msg2).setArrivalTime(startTime + 2 * samplePeriod).build())) {
        found = true;
      }
      count++;
    }
    assertTrue(found);
    assertTrue(count > 1);
  }

  @Test
  public void testMonitoringDataFarBehind() {
    runTestMonitoringDataDifferentSpeed(timeout / 4, timeout / 10);
  }

  @Test
  public void testMonitoringDataAhead() {
    runTestMonitoringDataDifferentSpeed(0,  0);
  }

  public void runTestMonitoringDataDifferentSpeed(long delay, long delayIncrement) {
    final long startWindow = 1000;
    final long startTime = 1100;

    final int count = 50;
    long currentTime = START_TIME_MS;
    int transformReturnMsgCount = 0;
    for (int i = 0; i < count; ++i) {
      long window = startWindow + i * TimeBucket.SIZE;
      long ts = startTime + i * TimeBucket.SIZE;
      long sendTime = START_TIME_MS + i * TimeBucket.SIZE;

      // one metrics message
      MonitoringMessage msg = createMessage(startSequence + i, firstSession, window, ts);
      transformReturnMsgCount += transformAndReturnCount(key, msg);

      currentTime += 10 * delay;
      mockClock.setCurrentTimeMillis(currentTime);

      // another metrics message for another partition
      MonitoringMessage msg2 = createMessage(MessageType.NORMAL,
                                             Monitoring.ClientType.PRODUCER,
                                             1, 0, startSequence + i, "session2",
                                             window, ts, 1, false);
      transformReturnMsgCount += transformAndReturnCount(key, msg2);

      currentTime += 10 * delay;
      mockClock.setCurrentTimeMillis(currentTime);

      // one C3 heartbeat
      MonitoringMessage hb1 = createC3Heartbeat(startSequence + i, 0L, sendTime, 0);
      MonitoringMessage hb2 = createC3Heartbeat(startSequence + i, 0L, sendTime, 1);
      transformReturnMsgCount += transformAndReturnCount(key, hb1);
      transformReturnMsgCount += transformAndReturnCount(key, hb2);

      currentTime += delay;
      mockClock.setCurrentTimeMillis(currentTime);

      delay += delayIncrement;
    }

    assertEquals(count * 2, transformReturnMsgCount);
  }

  private KeyValue<byte[], MonitoringMessage> verifyTransformReturnsOneKV(
      byte[] key,
      MonitoringMessage value
  ) {
    KeyValue<byte[], Iterable<KeyValue<byte[], MonitoringMessage>>> ret =
        transformer.transform(key, value);
    assertNotNull(ret);
    Iterator<KeyValue<byte[], MonitoringMessage>> it = ret.value.iterator();
    assertTrue(it.hasNext());
    KeyValue<byte[], MonitoringMessage> msg = it.next();
    assertFalse(msg.value.getType() == MessageType.HEARTBEAT);
    assertFalse(it.hasNext());
    return msg;
  }

  private KeyValue<byte[], MonitoringMessage> verifyTransformReturnsMultipleKV(
      byte[] key,
      MonitoringMessage value
  ) {
    KeyValue<byte[], Iterable<KeyValue<byte[], MonitoringMessage>>> ret =
        transformer.transform(key, value);
    assertNotNull(ret);
    Iterator<KeyValue<byte[], MonitoringMessage>> it = ret.value.iterator();
    assertTrue(it.hasNext());
    KeyValue<byte[], MonitoringMessage> msg = it.next();
    assertTrue(it.hasNext());
    return msg;
  }

  private int transformAndReturnCount(byte[] key, MonitoringMessage value) {
    int count = 0;
    KeyValue<byte[], Iterable<KeyValue<byte[], MonitoringMessage>>> ret =
        transformer.transform(key, value);
    if (ret == null) {
      return count;
    }
    Iterator<KeyValue<byte[], MonitoringMessage>> it = ret.value.iterator();
    while (it.hasNext()) {
      ++count;
      it.next();
    }
    return count;
  }

  private MonitoringVerifier getVerifier(
      KeyValueStore<Bytes, MonitoringVerifier> store,
      MonitoringMessage msg
  ) {
    Bytes key = keySerde.key(msg);
    MonitoringMessage keyMM = keySerde.toProto(key);
    assertEquals(clusterId, keyMM.getClusterId());
    return store.get(key);
  }

  private void verifyMessageKeyStored(KeyValueStore<Bytes, MonitoringVerifier> store,
                                             MonitoringMessage msg) {
    MonitoringVerifier verifier = getVerifier(store, msg);
    MonitoringMessage baseMsg = verifier.getCopyOfBaseMonitoringMessage().build();
    assertEquals(clusterId, baseMsg.getClusterId());
    assertNotNull(verifier);
  }

  private MonitoringMessage createMessage(
      MessageType type,
      Monitoring.ClientType clientType,
      Integer part,
      Integer monitoringPartition,
      long sequence,
      String session,
      long window,
      Long timestamp,
      long count,
      boolean shutdown
  ) {
    return MonitoringMessage.newBuilder(MonitoringMessageUtil.baseMonitoringMessage())
        .setClusterId(clusterId)
        .setClientId(clientId)
        .setClientType(clientType)
        .setTopic(topic)
        .setPartition(part)
        .setMonitoringTopicPartition(monitoringPartition)
        .setType(type)
        .setSequence(sequence)
        .setSession(session)
        .setWindow(window)
        .setMinWindow(window)
        .setMaxWindow(window)
        .setTimestamp(timestamp)
        .setSamplePeriod(samplePeriod)
        .setCount(count)
        .setShutdown(shutdown)
        .build();
  }

  private MonitoringMessage createC3Heartbeat(long sequence, Long window, Long timestamp,
                                              int monitoringPartition) {
    return createMessage(MessageType.HEARTBEAT, Monitoring.ClientType.CONTROLCENTER, 0,
                         monitoringPartition, sequence, "", window, timestamp, 1, false);
  }

  private MonitoringMessage createMessageWithShutdown(
      long sequence, String session, Long window, Long timestamp) {
    return createMessage(MessageType.NORMAL, Monitoring.ClientType.PRODUCER, partition, 0,
                         sequence, session, window, timestamp, 1, true);
  }

  private MonitoringMessage createMessage(
      long sequence, String session, Long window, Long timestamp
  ) {
    return createMessage(MessageType.NORMAL, Monitoring.ClientType.PRODUCER, partition, 0,
                         sequence, session, window, timestamp, 1, false);
  }

  private MonitoringMessage createHeartbeatMessageNoShutdown(long sequence, String session,
                                                             Long timestamp) {
    return createMessage(MessageType.HEARTBEAT, Monitoring.ClientType.PRODUCER, partition, 0,
                         sequence, session, 0, timestamp, 0, false);
  }

}
