// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.controlcenter.util;

import com.google.common.collect.Iterators;

import org.apache.kafka.clients.producer.Producer;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.serialization.Serializer;
import org.apache.kafka.common.utils.Bytes;
import org.apache.kafka.streams.KeyValue;
import org.apache.kafka.streams.processor.StateStoreSupplier;
import org.apache.kafka.streams.processor.internals.RecordCollector;
import org.apache.kafka.streams.processor.internals.RecordCollectorImpl;
import org.apache.kafka.streams.state.KeyValueIterator;
import org.apache.kafka.streams.state.KeyValueStore;
import org.apache.kafka.streams.state.Stores;
import org.easymock.EasyMock;
import org.junit.Before;
import org.junit.Test;

import java.util.Arrays;
import java.util.List;

import io.confluent.controlcenter.record.Controlcenter.Cluster;
import io.confluent.controlcenter.record.Controlcenter.ClusterType;
import io.confluent.controlcenter.streams.MockProcessorContext;
import io.confluent.serializers.OrderedKeyProtoSerde;
import io.confluent.serializers.OrderedKeyUberSerde;
import io.confluent.serializers.ProtoSerde;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;


public class KVQueryTest {
  private final OrderedKeyUberSerde<Cluster> keySerde =  OrderedKeyProtoSerde.create(Cluster.getDefaultInstance(),
      Cluster.CLUSTERTYPE_FIELD_NUMBER,
      Cluster.CLUSTERID_FIELD_NUMBER
  );
  private final OrderedKeyUberSerde<Cluster> prefixKeySerde = keySerde.prefixKeySerde(1);
  private final ProtoSerde<Cluster> valueSerde = new ProtoSerde<>(Cluster.getDefaultInstance());
  private KeyValueStore<Bytes, Cluster> store;
  private final List<String> kafkaIds = Arrays.asList("1111", "2222", "3333", "4444", "5555");
  private final List<String> connectIds = Arrays.asList("aaaa", "bbbb", "cccc", "dddd", "eeee");

  @Before
  public void setUp() {
    StateStoreSupplier storeSupplier = Stores.create("test")
        .withKeys(keySerde)
        .withValues(valueSerde)
        .inMemory()
        .build();

    Producer mockProducer = EasyMock.createMock(Producer.class);
    RecordCollector recordCollector = new RecordCollectorImpl(mockProducer, "fakeTaskId") {
      @Override
      public <K1, V1> void send(ProducerRecord<K1, V1> record, Serializer<K1> keySerializer,
                                Serializer<V1> valueSerializer) {
        // nothing
      }
    };
    MockProcessorContext context = new MockProcessorContext(null, recordCollector);

    store = (KeyValueStore<Bytes, Cluster>) storeSupplier.get();
    store.init(context, store);
    context.register(store, false, null);

    for (String id : kafkaIds) {
      Cluster cluster = getCluster(ClusterType.KAFKA, id);
      store.put(keySerde.key(cluster), cluster);
    }

    for (String id : connectIds) {
      Cluster cluster = getCluster(ClusterType.KAFKA_CONNECT, id);
      store.put(keySerde.key(cluster), cluster);
    }
  }

  private Cluster getCluster(ClusterType type, String clusterId) {
    return Cluster.newBuilder()
          .setClusterType(type)
          .setClusterId(clusterId)
          .build();
  }

  @Test
  public void rangeFrom() throws Exception {
    Bytes startKey = keySerde.key(getCluster(ClusterType.KAFKA, "3333"));
    try (KeyValueIterator<Bytes, Cluster> iter = KVQuery.rangeFrom(store, startKey)) {
      assertEquals(8, Iterators.size(iter));
    }
    try (KeyValueIterator<Bytes, Cluster> iter = KVQuery.rangeFrom(store, startKey)) {
      KeyValue<Bytes, Cluster> first = iter.next();
      assertEquals("3333", first.value.getClusterId());
    }
    try (KeyValueIterator<Bytes, Cluster> iter = KVQuery.rangeFrom(store, startKey)) {
      KeyValue<Bytes, Cluster> last = Iterators.getLast(iter);
      assertEquals("eeee", last.value.getClusterId());
    }
  }

  @Test
  public void rangeFromNull() throws Exception {
    try (KeyValueIterator<Bytes, Cluster> iter = KVQuery.rangeFrom(store, null)) {
      assertEquals(10, Iterators.size(iter));
    }
    try (KeyValueIterator<Bytes, Cluster> iter = KVQuery.rangeFrom(store, null)) {
      KeyValue<Bytes, Cluster> first = iter.next();
      assertTrue(iter.hasNext());
      assertEquals("1111", first.value.getClusterId());
    }
    try (KeyValueIterator<Bytes, Cluster> iter = KVQuery.rangeFrom(store, null)) {
      KeyValue<Bytes, Cluster> last = Iterators.getLast(iter);
      assertEquals("eeee", last.value.getClusterId());
      assertFalse(iter.hasNext());
    }
  }

  @Test
  public void prefixCallOrder() throws Exception {
    Bytes prefixKey = prefixKeySerde.key(getCluster(ClusterType.KAFKA, ""));
    try (KeyValueIterator<Bytes, Cluster> iter = KVQuery.prefixScan(store, prefixKey)) {
      //Calls hasNext() first
      assertTrue(iter.hasNext());
      assertTrue(iter.hasNext());
      KeyValue<Bytes, Cluster> first = iter.next();
      assertEquals("1111", first.value.getClusterId());

      //Does not call has next
      KeyValue<Bytes, Cluster> second = iter.next();
      assertEquals("2222", second.value.getClusterId());
    }
  }

  @Test
  public void prefixScan() throws Exception {
    Bytes prefixKey = prefixKeySerde.key(getCluster(ClusterType.KAFKA, ""));
    try (KeyValueIterator<Bytes, Cluster> iter = KVQuery.prefixScan(store, prefixKey)) {
      assertEquals(5, Iterators.size(iter));
    }
    try (KeyValueIterator<Bytes, Cluster> iter = KVQuery.prefixScan(store, prefixKey)) {
      assertTrue(iter.hasNext());
      KeyValue<Bytes, Cluster> first = iter.next();
      assertEquals("1111", first.value.getClusterId());
    }
    try (KeyValueIterator<Bytes, Cluster> iter = KVQuery.prefixScan(store, prefixKey)) {
      KeyValue<Bytes, Cluster> last = Iterators.getLast(iter);
      assertEquals("5555", last.value.getClusterId());
      assertFalse(iter.hasNext());
    }
  }

  @Test
  public void prefixScanWithStart() throws Exception {
    Bytes prefixKey = prefixKeySerde.key(getCluster(ClusterType.KAFKA, ""));
    Bytes startKey = keySerde.key(getCluster(ClusterType.KAFKA, "3333"));
    try (KeyValueIterator<Bytes, Cluster> iter = KVQuery.prefixScan(store, prefixKey, startKey)) {
      assertEquals(3, Iterators.size(iter));
    }
    try (KeyValueIterator<Bytes, Cluster> iter = KVQuery.prefixScan(store, prefixKey, startKey)) {
      assertTrue(iter.hasNext());
      KeyValue<Bytes, Cluster> first = iter.next();
      assertEquals("3333", first.value.getClusterId());
    }
    try (KeyValueIterator<Bytes, Cluster> iter = KVQuery.prefixScan(store, prefixKey, startKey)) {
      KeyValue<Bytes, Cluster> last = Iterators.getLast(iter);
      assertEquals("5555", last.value.getClusterId());
      assertFalse(iter.hasNext());
    }
  }

  @Test(expected=UnsupportedOperationException.class)
  public void testRemove() throws Exception {
    Bytes prefixKey = prefixKeySerde.key(getCluster(ClusterType.KAFKA, ""));
    try (KeyValueIterator<Bytes, Cluster> iter = KVQuery.prefixScan(store, prefixKey)) {
      iter.remove();
    }
  }

}
