// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import moment from 'moment';
import { browserHistory } from 'react-router';

import { RangeTypes } from 'constants/monitoring/timeframe';
import { locationBeforeTransitionsSelector } from 'selectors/routing';
import { makeAction } from '../util';

import type { TimeframeSpec } from 'types/monitoring/timeframe';

import type {
  Action,
  Dispatchable,
  Dispatcher
} from 'types';

export const UPDATE_FETCH_EXTENDED_END_TIME = 'UPDATE_FETCH_EXTENDED_END_TIME';
export type UpdateFetchExtendedEndTimeAction = Action<'UPDATE_FETCH_EXTENDED_END_TIME', { endTimeMs: number | null }>
export const updateFetchExtendedEndTime = (endTimeMs: number | null): UpdateFetchExtendedEndTimeAction => makeAction(UPDATE_FETCH_EXTENDED_END_TIME, { endTimeMs });

export const UPDATE_END_TIME = 'UPDATE_END_TIME';
export type UpdateEndTimeAction = Action<'UPDATE_END_TIME', { endTimeMs: number }>
export const updateEndTime = (endTimeMs: number): UpdateEndTimeAction => makeAction(UPDATE_END_TIME, { endTimeMs });

export const SET_TIME_RANGE = 'SET_TIME_RANGE';
export type SetTimeRangeAction = Action<'SET_TIME_RANGE', TimeframeSpec>
export const setTimeRange = (timeRangeSpec: TimeframeSpec): SetTimeRangeAction => makeAction(SET_TIME_RANGE, timeRangeSpec);


export function updateTimeRange(timeRangeSpec: Object): Dispatchable {
  return (dispatch: Dispatcher) => {
    const { rangeType } = timeRangeSpec;

    let timeframeSpec = null;
    switch (rangeType) {
      case RangeTypes.FIXED:
        timeframeSpec = {
          rangeType,
          startTimeMs: timeRangeSpec.startTimeMs,
          stopTimeMs: timeRangeSpec.stopTimeMs
        };
        break;
      case RangeTypes.ROLLING:
        timeframeSpec = {
          rangeType,
          startTimeMs: moment().unix()*1000 - timeRangeSpec.lengthMs,
          stopTimeMs: moment().unix()*1000
        };
        break;
      case RangeTypes.GROWING_DAY:
        timeframeSpec = {
          rangeType,
          startTimeMs: moment().startOf('day').unix()*1000,
          stopTimeMs: moment().unix()*1000
        };
        break;
      case RangeTypes.GROWING_WEEK:
        timeframeSpec = {
          rangeType,
          startTimeMs: moment().startOf('week').unix()*1000,
          stopTimeMs: moment().unix()*1000
        };
        break;
      case RangeTypes.GROWING_MONTH:
        timeframeSpec = {
          rangeType,
          startTimeMs: moment().startOf('month').unix()*1000,
          stopTimeMs: moment().unix()*1000
        };
        break;
    }

    if (timeframeSpec != null) {
      dispatch(setTimeRange(timeframeSpec));
    }
  };
}

const ignorePreviousQueryParamsState = {
  state: {
    existingQueryParamBlacklist: ['fixed-end', 'fixed-start', 'growing', 'rolling']
  }
};

export function updateTimeRangeQueryParams(
  newTimeframeSpec: TimeframeSpec
): Dispatchable {
  return (dispatch: Dispatcher, getState: Function) => {
    const { pathname } = locationBeforeTransitionsSelector(getState());


    switch (newTimeframeSpec.rangeType) {
      case RangeTypes.GROWING_DAY:
        browserHistory.push({
          ...ignorePreviousQueryParamsState,
          pathname,
          query: {
            growing: 'day'
          }
        });
        break;
      case RangeTypes.GROWING_WEEK:
        browserHistory.push({
          ...ignorePreviousQueryParamsState,
          pathname,
          query: {
            growing: 'week'
          }
        });
        break;
      case RangeTypes.GROWING_MONTH:
        browserHistory.push({
          ...ignorePreviousQueryParamsState,
          pathname,
          query: {
            growing: 'month'
          }
        });
        break;
      case RangeTypes.ROLLING:
        const difference = moment.unix(newTimeframeSpec.stopTimeMs/1000).diff(moment.unix(newTimeframeSpec.startTimeMs/1000), 'minutes');
        browserHistory.push({
          ...ignorePreviousQueryParamsState,
          pathname,
          query: {
            rolling: difference
          }
        });
        break;
      default:
        browserHistory.push({
          ...ignorePreviousQueryParamsState,
          pathname,
          query: {
            'fixed-start': newTimeframeSpec.startTimeMs,
            'fixed-end': newTimeframeSpec.stopTimeMs
          }
        });
        break;
    }

    // TODO(DERRICK): Would be better to change things around so that this action is not calling a subsequent action below
    // Discussed changing Monitoring so that all clicks that result in a new timerange ONLY ever change the URL
    // Another process running on the component would respond to URL changes and call setTimeRange when necessary
    // The only other place that setTimeRange would need be called is for "growing" ranges where the endTime needs to up
    dispatch(setTimeRange(newTimeframeSpec));
  };
}
