// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.

import type { Action } from 'types';

const ReadyStates = {
  UNSET: 0,
  OPENED: 1,
  HEADERS_RECEIVED: 2,
  LOADING: 3,
  DONE: 4
};

const DataFormats = {
  JSON: 'JSON',
  MERGE_PATCH_JSON: 'MERGE_PATCH_JSON',
  PARAMS: 'PARAMS'
};

type DataFormatType = $Enum<typeof DataFormats>;

const Verbs = {
  DELETE: 'DELETE',
  GET: 'GET',
  PATCH: 'PATCH',
  POST: 'POST',
  PUT: 'PUT'
};

type VerbType = $Enum<typeof Verbs>;

const DEFAULT_FETCH_TIMEOUT_MS = 30000;

export const makeAction = (type: *, payload: *): Action<*, *> => ({ type, payload });

export const encodeQueryData = (data: Object, nonEmptyOnly: boolean = false): string =>
  Object
    .keys(data)
    .filter(key => !nonEmptyOnly || data[key] != null)
    .map(kv => `${encodeURIComponent(kv)}=${encodeURIComponent(data[kv])}`)
    .join('&');

function request(verb: VerbType, dataFormat: DataFormatType, path: string, data?: any, timeoutMs?: number): Promise<*> {

  const timeout = timeoutMs
    ? timeoutMs
    : DEFAULT_FETCH_TIMEOUT_MS;

  return new Promise((resolve: (data: ?Object) => void, reject: (error: Object) => void) => {

    function stateChangeListener() {
      if (this.readyState === ReadyStates.DONE) {
        let data;
        try {
          data = JSON.parse(this.responseText);
        } catch (e) {
          data = null;
        }
        if (this.status >= 200 && this.status <= 299) {
          resolve(data);
        } else {
          reject({
            data,
            status: this.status,
            message: this.status === 0
              ? 'Request Timed Out'
              : `HTTP Status: ${this.status}`
          });
        }
      }
    }

    const oReq = new XMLHttpRequest();
    oReq.timeout = timeout;
    oReq.onreadystatechange = stateChangeListener;
    oReq.open(verb, path, true);
    if (data) {
      if (dataFormat === DataFormats.MERGE_PATCH_JSON) {
        oReq.setRequestHeader('Content-Type', 'application/merge-patch+json');
        oReq.send(JSON.stringify(data));
        return;
      } else if (dataFormat === DataFormats.JSON) {
        oReq.setRequestHeader('Content-Type', 'application/json');
        const message = JSON.stringify(data);
        oReq.send(message);
        return;
      } else {
        oReq.setRequestHeader('Content-Type', 'application/x-www-form-urlencoded');
        const message = encodeQueryData(data);
        oReq.send(message);
        return;
      }
    }

    oReq.send(undefined);
  });
}

export const deleteJson= request.bind(null, Verbs.DELETE, DataFormats.JSON);
export const getJson = request.bind(null, Verbs.GET, DataFormats.JSON);
export const patchJson = request.bind(null, Verbs.PATCH, DataFormats.MERGE_PATCH_JSON);
export const postJson = request.bind(null, Verbs.POST, DataFormats.JSON);
export const postParams = request.bind(null, Verbs.POST, DataFormats.PARAMS);
export const putJson = request.bind(null, Verbs.PUT, DataFormats.JSON);
