// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React, { Component } from 'react';
import { connect } from 'react-redux';
import { browserHistory } from 'react-router';
import moment from 'moment';

import { MonitoringChart } from 'components/monitoring';
import { MonitoringDataTypes } from 'types/monitoring';
import { Modal } from 'components/modal';
import { HISTORY_INDEX_PATHNAME } from 'constants/alerts';
import { MetricTypes } from 'types/alerts';
import { alertsMonitoringData } from 'actions/alerts';
import { calcTimeSlotQuantizedMs } from 'selectors/monitoring';
import { historyItemSelector } from 'selectors/alerts';
import styles from './HistoryItemModal.css';
import { DateTimeFormats } from 'constants/timePicker';

import type { AlertSummary } from 'types/alerts';
import type { ClusterId } from 'types/clusters';

import {
  noTextLineSlotDataFormatter,
  noTextAverageLatencyFormatter,
  noTextMaxLatencyFormatter
} from 'components/charting/charts/hoverUtil';

const ONE_DAY_MS = 24*60*60*1000;
const THIRTY_MINS_MS = 30*60*1000;
const CHART_HEIGHT = 150;
const CHART_WIDTH = 460;

function constructAlertText(historyItem: AlertSummary): ?string {

  // TODO(mhowlett): this mapping would probably be better defined alongside the type definition.
  function metricText(historyItem: AlertSummary): string {
    switch (historyItem.monitoringTrigger.metric) {
      case MetricTypes.CONSUMPTION_DIFF:
        return 'consumption difference';
      case MetricTypes.MAX_LATENCY:
        return 'maximum latency';
      case MetricTypes.AVG_LATENCY:
        return 'average latency';
      case MetricTypes.CONNECTOR_STATE:
        return 'connector state';
      default:
        // TODO(mhowlett): change to 'N/A' when default metric value is explicitly specified.
        return 'consumption difference';
    }
  }

  function relativeValueText(historyItem: AlertSummary): string {
    if (historyItem.triggers.length > 0) {
      if (historyItem.triggers[0].longValue === historyItem.monitoringTrigger.longValue) {
        return 'equal to';
      }
      if (historyItem.triggers[0].longValue < historyItem.monitoringTrigger.longValue) {
        return 'less than';
      }
      return 'greater than';
    }
    return 'N/A';
  }

  function equalNotEqualText(historyItem: AlertSummary): string {
    if (historyItem.triggers.length > 0) {
      if (historyItem.triggers[0].longValue === historyItem.monitoringTrigger.longValue) {
        return 'equal to';
      }
      return 'not equal to';
    }
    return 'N/A';
  }

  function actualValueText(historyItem: AlertSummary): string {
    if (historyItem.triggers.length > 0) {
      if (historyItem.triggers[0].longValue != null) {
        return historyItem.triggers[0].longValue;
      }
    }
    return 'N/A';
  }

  switch (historyItem.monitoringTrigger.metric) {
    case MetricTypes.CONNECTOR_STATE:
      const connector = historyItem.monitoringTrigger.connector;
      if (connector == null) {
        return 'N/A';
      }
      return `The actual connector state on connector ${connector} was ${actualValueText(historyItem)} which is ${equalNotEqualText(historyItem)} the trigger value ${historyItem.monitoringTrigger.longValue}`;
    case MetricTypes.CONSUMPTION_DIFF:
    case MetricTypes.MAX_LATENCY:
    case MetricTypes.AVG_LATENCY:
    default:
      const consumerGroupName = historyItem.monitoringTrigger.group;
      if (consumerGroupName == null) {
        return 'N/A';
      }
      return `The actual ${metricText(historyItem)} on consumer group ${consumerGroupName} was ${actualValueText(historyItem)} which is ${relativeValueText(historyItem)} the threshold ${historyItem.monitoringTrigger.longValue}`;
    // TODO(mhowlett): separate out the default case when default metric value is explicitly specified.
  }
}

type HistoryItemModalProps = {
  itemId: string,
  startTimeSlotQuantizedMs: number,
  stopTimeSlotQuantizedMs: number,
  clusterId: ClusterId,
  monitoringData: Object,
  historyItem: AlertSummary,
  requestMonitoringData: Function,
};

export class HistoryItemModal extends Component {

  props: HistoryItemModalProps;

  handleClose() {
    browserHistory.push(HISTORY_INDEX_PATHNAME);
  }

  loadMonitoringData(props: HistoryItemModalProps) {
    const spec = {
      startTimeMs: props.startTimeSlotQuantizedMs,
      stopTimeMs: props.stopTimeSlotQuantizedMs,
      consumerGroup: props.historyItem.monitoringTrigger.group,
      clusterId: props.clusterId
    };
    this.props.requestMonitoringData(spec);
  }

  componentDidMount() {
    if (this.props.historyItem != null) {
      this.loadMonitoringData(this.props);
    }
  }

  componentWillReceiveProps(nextProps: Object) {
    if (this.props.historyItem == null && nextProps.historyItem != null) {
      this.loadMonitoringData(nextProps);
    }
  }

  render() {
    if (!this.props.historyItem) {
      return null;
    }
    const { historyItem, monitoringData } = this.props;
    const todayMs = new Date().getTime();

    let hoverFormatter = null;
    switch (historyItem.monitoringTrigger.metric) {
      case MetricTypes.AVG_LATENCY:
        hoverFormatter = noTextAverageLatencyFormatter;
        break;
      case MetricTypes.MAX_LATENCY:
        hoverFormatter = noTextMaxLatencyFormatter;
        break;
      case MetricTypes.CONSUMPTION_DIFF:
        hoverFormatter = noTextLineSlotDataFormatter;
        break;
    }

    return (
      <Modal onClose={this.handleClose.bind(this)}>
        <div className={styles.root}>
          <div className={styles.header}>{historyItem.monitoringTrigger.name}</div>
          <div className={styles.timestamp}>
            {todayMs - historyItem.timestamp > ONE_DAY_MS ?
              moment(historyItem.timestamp).format(`${DateTimeFormats.DATE} ${DateTimeFormats.TIME}`) :
              `Today @ ${moment(historyItem.timestamp).format('h:mm a')}`}
          </div>
          <div className={styles.chart}>
            {monitoringData
              ? <MonitoringChart
                  data={monitoringData}
                  width={CHART_WIDTH}
                  height={CHART_HEIGHT}
                  startTimeSlotQuantizedMs={this.props.startTimeSlotQuantizedMs}
                  stopTimeSlotQuantizedMs={this.props.stopTimeSlotQuantizedMs}
                  mouseOverTime={new Date(historyItem.timestamp)}
                  hoverFormatter={hoverFormatter}
                  type={
                    historyItem.monitoringTrigger.metric === MetricTypes.CONSUMPTION_DIFF ||
                    // TODO(mhowlett): remove null check when default metric value is explicitly specified.
                    historyItem.monitoringTrigger.metric == null
                      ? MonitoringDataTypes.DELIVERY
                      : MonitoringDataTypes.LATENCY
                  }
                />
              : null}
          </div>
          <div className={styles.description}>
            {constructAlertText(historyItem)}
          </div>
          <div className={styles.hl} />
          <div>
            <div className={styles['list-left']}>
              Actions:
            </div>
            <div className={styles['list-right']}>
              {
                // TODO(mhowlett): change index to action guid when available.
                historyItem.actions.map((action, i) => <div key={i}>{action.name}</div>)
              }
            </div>
          </div>
        </div>
      </Modal>
    );
  }
}

export default connect(
  (state, props) => {
    const historyItem = historyItemSelector(state, props.itemId);
    if (historyItem == null) {
      return {
        historyItem: null,
        startTimeSlotQuantizedMs: 0,
        stopTimeSlotQuantizedMs: 0,
        clusterId: 0,
        monitoringData: null
      };
    }

    const startTimeMs = historyItem.timestamp - THIRTY_MINS_MS;
    const stopTimeMs = historyItem.timestamp + THIRTY_MINS_MS;

    const startTimeSlotQuantizedMs = calcTimeSlotQuantizedMs(startTimeMs, CHART_WIDTH, 15000, startTimeMs, stopTimeMs);
    const stopTimeSlotQuantizedMs = calcTimeSlotQuantizedMs(stopTimeMs, CHART_WIDTH, 15000, startTimeMs, stopTimeMs);

    return {
      historyItem,
      startTimeSlotQuantizedMs,
      stopTimeSlotQuantizedMs,
      clusterId: historyItem.monitoringTrigger.clusterId,
      monitoringData: state.alerts.history.item.monitoringData
    };
  },
  {
    requestMonitoringData: alertsMonitoringData.request
  }
)(HistoryItemModal);
