// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React, { Component } from 'react';
import { connect } from 'react-redux';

import { assert } from 'utils';
import { setMouseOverTime } from 'actions/monitoring/layout';

import ChartHover from './ChartHover';
import GridLine from './GridLine';
import styles from './ChartArea.css';
import type { HoverFormatter } from './hoverUtil';

import type { MultiTimeSeries, TimeSeries, LatencyData } from 'types/monitoring/data';
import type { TimeScale, NumberScale } from 'types/monitoring';

import { aComponentIsFocusedSelector } from 'selectors/layout';


export class ChartArea extends Component {
  props: {
    children: any,
    xScale: TimeScale,
    yScale: NumberScale,
    tickTextFormatterCreator: Function,
    componentHeight: number,
    legend: any,
    setMouseOverTime: Function,

    multiSlotData?: MultiTimeSeries<LatencyData>,
    lineSlotData?: ?TimeSeries,
    areaSlotData?: ?TimeSeries,

    hoverFormatter: HoverFormatter,
    mouseOverTime: ?Date,

    solidLeadingGuide?: boolean,

    onFocusComponent?: Function
  };

  chartRoot: HTMLElement;
  handleMouseOut: Function;
  handleMouseMove: Function;

  constructor() {
    super();
    this.handleMouseMove = this.handleMouseMove.bind(this);
    this.handleMouseOut = this.handleMouseOut.bind(this);
  }

  handleMouseOut() {
    if (this.props.mouseMoveEnabled) {
      this.props.setMouseOverTime(null);
    }
  }

  calcMouseOverTime(clientX: number): ?Date {
    const rect = this.chartRoot.getBoundingClientRect();
    const offsetX = clientX - rect.left;
    const domainX = this.props.xScale.invert(offsetX);

    if (domainX < this.props.xScale.domain()[0] ||
        domainX > this.props.xScale.domain()[1]) {
      return null;
    }

    return domainX;
  }

  handleMouseMove(e: SyntheticMouseEvent) {
    if (this.props.mouseMoveEnabled) {
      this.props.setMouseOverTime(this.calcMouseOverTime(e.clientX));
    }
  }

  handleMouseDown(e: SyntheticMouseEvent) {
    const onFocusComponent = this.props.onFocusComponent;
    if (onFocusComponent) {
      const rect = this.chartRoot.getBoundingClientRect();
      const pageMouseX = e.pageX;
      const pageMouseY = e.pageY;
      const componentMouseX = pageMouseX - rect.left;
      const componentMouseY = pageMouseY - (rect.top + window.scrollY);
      onFocusComponent(pageMouseX, pageMouseY, componentMouseX, componentMouseY);
    }
  }

  render() {
    const {
      children,
      xScale,
      yScale,
      tickTextFormatterCreator,
      componentHeight,
      legend,

      multiSlotData,
      lineSlotData,
      areaSlotData,

      hoverFormatter,
      mouseOverTime,
      solidLeadingGuide
    } = this.props;

    assert(children && !Array.isArray(children), 'ChartArea must have exactly one child');

    const gridLines = yScale == null
      ? []
      : yScale.ticks(2).map(y => ({ y, screenY: Math.round(yScale(y)) }));

    const tickTextFormatter = tickTextFormatterCreator(gridLines);
    const yAxisTotalWidth = xScale.range()[0];
    const width = xScale.range()[1];

    return (
      <div
        ref={el => this.chartRoot = el}
        className={styles['chart-root']}
        style={{
          height: componentHeight,
          width: width
        }}
      >
        <svg
          height={componentHeight}
          width={width}
          onMouseMove={this.handleMouseMove}
          onMouseOut={this.handleMouseOut}
          onMouseDown={this.handleMouseDown.bind(this)}
        >
          {
            yScale && children
          }
          {
            gridLines.map(
              (gridLine, i) => (
                <GridLine
                  key={i}
                  yAxisTotalWidth={yAxisTotalWidth}
                  width={width}
                  screenY={gridLine.screenY}
                  domainY={gridLine.y}
                  tickTextFormatter={tickTextFormatter}
                  solid={solidLeadingGuide && i === gridLines.length - 1}
                />
              )
            )
          }
          {
            <ChartHover
              multiSlotData={multiSlotData}
              lineSlotData={lineSlotData}
              areaSlotData={areaSlotData}
              xScale={xScale}
              yScale={yScale}
              mouseOverTime={mouseOverTime}
              hoverFormatter={hoverFormatter}
            />
          }
        </svg>

        {legend}

      </div>
    );
  }
}

export default connect(
  (state) => ({
    mouseMoveEnabled: !aComponentIsFocusedSelector(state)
  }),
  {
    setMouseOverTime
  }
)(ChartArea);
