// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { format } from 'd3-format';

import type {
  TimeSeries,
  MultiTimeSeries,
  LatencyData
} from 'types/monitoring/data';

import type {
  TimeScale,
  NumberScale
} from 'types/monitoring';


export const LineExtendDirections = {
  UP: 'UP',
  DOWN: 'DOWN',
  UPDOWN: 'UPDOWN'
};

export type LineExtendDirection = $Enum<typeof LineExtendDirections>;

export function calcHoverInfo(
  mouseOverTime: Date,
  startTimeMs: number,
  stepMs: number,
  valuesLength: number,
  xScale: TimeScale
): ?{ ptX: number, idx: number } {

  const mouseOverMs = mouseOverTime.getTime();
  const quantizedBin = Math.floor(mouseOverMs / stepMs);
  const dataStartBin = Math.floor(startTimeMs / stepMs);
  const idx = quantizedBin - dataStartBin;
  if (idx < 0 || idx > valuesLength) {
    return null;
  }

  const hoverTimeMs = (quantizedBin + 0.5) * stepMs;

  const ptX = xScale(hoverTimeMs);

  return { ptX, idx };
}

const percFormatter = format('2d');
const siFormatter = format('.3s');
const formatter = format('3d');

export type HoverLabelInfo = {
  ptX: number,
  ptY: number,
  label: string,
  lineExtendDirection: LineExtendDirection,
  maxTextLength: number
}

export type HoverFormatter = (
  lineSlotData: ?TimeSeries,
  areaSlotData: ?TimeSeries,
  multiSlotData: ?MultiTimeSeries<LatencyData>,
  mouseOverTime: Date,
  xScale: TimeScale,
  yScale: NumberScale
) => ?HoverLabelInfo;


export function noTextMultiSlotDataFormatter(
  fieldName: string,
  lineSlotData: ?TimeSeries,
  areaSlotData: ?TimeSeries,
  multiSlotData: ?MultiTimeSeries<LatencyData>,
  mouseOverTime: Date,
  xScale: TimeScale,
  yScale: NumberScale
): ?HoverLabelInfo {
  if (multiSlotData == null) { return null; }

  const hoverLabelInfo = calcHoverInfo(
    mouseOverTime,
    multiSlotData.startTimeMs,
    multiSlotData.stepMs,
    multiSlotData.values[fieldName].length,
    xScale
  );
  if (hoverLabelInfo == null) { return null; }

  const val = multiSlotData.values[fieldName][hoverLabelInfo.idx];
  if (val == null) { return null; }

  return {
    ptX: hoverLabelInfo.ptX,
    ptY: yScale(val),
    label: '',
    lineExtendDirection: LineExtendDirections.DOWN,
    maxTextLength: 0
  };
}
export const noTextAverageLatencyFormatter = noTextMultiSlotDataFormatter.bind(null, 'averageLatency');
export const noTextMaxLatencyFormatter = noTextMultiSlotDataFormatter.bind(null, 'maxLatency');

export function noTextLineSlotDataFormatter(
  lineSlotData: ?TimeSeries,
  areaSlotData: ?TimeSeries,
  multiSlotData: ?MultiTimeSeries<LatencyData>,
  mouseOverTime: Date,
  xScale: TimeScale,
  yScale: NumberScale
): ?HoverLabelInfo {
  if (lineSlotData == null) { return null; }

  const hoverLabelInfo = calcHoverInfo(
    mouseOverTime,
    lineSlotData.startTimeMs,
    lineSlotData.stepMs,
    lineSlotData.values.length,
    xScale
  );
  if (hoverLabelInfo == null) { return null; }

  const val = lineSlotData.values[hoverLabelInfo.idx];
  if (val == null) { return null; }

  return {
    ptX: hoverLabelInfo.ptX,
    ptY: yScale(val),
    label: '',
    lineExtendDirection: LineExtendDirections.DOWN,
    maxTextLength: 0
  };
}

export function latencyFormatter(
  lineSlotData: ?TimeSeries,
  areaSlotData: ?TimeSeries,
  multiSlotData: ?MultiTimeSeries<LatencyData>,
  mouseOverTime: Date,
  xScale: TimeScale,
  yScale: NumberScale
): ?HoverLabelInfo {

  // formatter is only used in scenarios where there is latency data.
  if (!multiSlotData || !multiSlotData.stepMs) {
    return null;
  }

  const hoverLabelInfo = calcHoverInfo(
    mouseOverTime,
    multiSlotData.startTimeMs,
    multiSlotData.stepMs,
    multiSlotData.values.averageLatency.length,
    xScale
  );

  if (hoverLabelInfo == null) {
    return null;
  }

  const val = multiSlotData.values.averageLatency[hoverLabelInfo.idx];

  if (val == null) {
    return null;
  }

  return {
    ptX: hoverLabelInfo.ptX,
    ptY: yScale(val),
    label: `Avg latency: ${formatter(val)}ms`,
    lineExtendDirection: LineExtendDirections.UP,
    maxTextLength: 150
  };
}


export function produceConsumeFormatter(
  lineSlotData: ?TimeSeries,
  areaSlotData: ?TimeSeries,
  multiSlotData: ?MultiTimeSeries<LatencyData>,
  mouseOverTime: Date,
  xScale: TimeScale,
  yScale: NumberScale
): ?HoverLabelInfo {

  // this formatter is only for use in scenarios where there is both
  // line and area data.
  if (!lineSlotData || !areaSlotData) {
    return null;
  }

  const hoverLabelInfo = calcHoverInfo(
    mouseOverTime,
    lineSlotData.startTimeMs,
    lineSlotData.stepMs,
    lineSlotData.values.length,
    xScale
  );

  if (!hoverLabelInfo) {
    return null;
  }

  const { ptX, idx } = hoverLabelInfo;

  const lineVal = lineSlotData.values[idx];
  const areaVal = areaSlotData.values[idx];

  // if there is neither producer or consumer data, it's safe not to show hover info.
  if (lineVal == null || areaVal == null) {
    return null;
  }

  // only show % if it is meaningful to do so.
  const percText = lineVal && areaVal
    ? `${percFormatter((areaVal / lineVal) * 100)}% `
    : '';

  const ptY = !lineVal ? yScale.range()[0] : yScale(lineVal);

  // apply SI formatter conditionally, don't want "9.00 out of 16.0 messages"
  const areaValText = areaVal < 1000 ? formatter(areaVal) : siFormatter(areaVal);
  const lineValText = lineVal < 1000 ? formatter(lineVal) : siFormatter(lineVal);
  const label = `${percText}(${areaValText} of ${lineValText} messages)`;

  return {
    ptX,
    ptY,
    label,
    lineExtendDirection: LineExtendDirections.DOWN,
    maxTextLength: 175
  };
}
