// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
/* eslint-disable react/jsx-handler-names */
import React, { Component } from 'react';
import ReactSelect, { Creatable } from 'react-select';
import classNames from 'classnames';
import { uniqBy } from 'lodash';

import Label from 'components/form/Label';
import styles from './Select.less';
import { Icon } from 'components/icon';
import { constructPlaceholderText } from 'components/FieldInput/FieldInput';

import type { FormField, FieldSectionField, Option } from 'types/form';

const convertValueToOption = (val: string | Option): Option => {
  if (typeof val === 'object') {
    return val;
  }
  return {
    label: val,
    value: val
  };
};

export class Select extends Component {
  props: {
    active: boolean,
    canCreate: boolean,
    clearable: boolean,
    reduxFormInputProps: FormField,
    sectionField: FieldSectionField,
    searchable: boolean,
    showErrorIndicator: boolean,
    showInfoIcon: boolean
  };

  static defaultProps = {
    canCreate: false,
    clearable: false,
    searchable: true,
    showErrorIndicator: false,
    showInfoIcon: false
  };

  render() {
    const {
      active,
      canCreate,
      clearable,
      reduxFormInputProps,
      searchable,
      showErrorIndicator,
      showInfoIcon,
      sectionField
    } = this.props;

    const { multi, locked, id } = sectionField;

    let options = Array.isArray(sectionField.options) ? sectionField.options : [];

    const hasValue = reduxFormInputProps.value && reduxFormInputProps.value.length > 0;

    const showLabel = active || (multi ? hasValue : Boolean(reduxFormInputProps.value));

    // the value has to exist in the options (which exist in state) so we merge any new values
    // into the options here
    if (hasValue) {
      options = [
        ...options,
        ...(multi ? reduxFormInputProps.value.map(convertValueToOption) : [convertValueToOption(reduxFormInputProps.value)])
      ];
    }

    options = uniqBy(options.map(convertValueToOption), (option: Option) => option.value);

    const disabled = locked || reduxFormInputProps.disabled;
    const placeholder = constructPlaceholderText(sectionField);

    const selectProps = {
      clearable,
      disabled,
      inputProps: { id },
      multi,
      options,
      placeholder,
      searchable,
      value: reduxFormInputProps.value,
      onChange: selected => {
        reduxFormInputProps.onChange(multi ? selected.map(item => item.value) : selected.value);
      },
      onBlur: () => {
        reduxFormInputProps.onBlur(reduxFormInputProps.value);
      },
      onFocus: reduxFormInputProps.onFocus
    };

    return (
      <div
        className={classNames({
          [styles['select-root']]: !showErrorIndicator,
          [styles['select-root-error']]: showErrorIndicator,
          [styles['select-no-placeholder']]: placeholder == null
        })}
      >
        <Label
          disabled={disabled}
          id={id}
          showLabel={showLabel}
          text={placeholder}
          showInfoIcon={showInfoIcon}
        >
          {canCreate ? (
            <Creatable {...selectProps} />
          ) : (
            <ReactSelect {...selectProps} />
          )}

          {disabled &&
            <div className={styles['lock']}>
              <Icon name="lock" size={16} />
            </div>
          }
        </Label>
      </div>
    );
  }
}

export default Select;
