// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.

import React, { Component } from 'react';
import { browserHistory, withRouter } from 'react-router';
import { reduxForm } from 'redux-form';
import { connect } from 'react-redux';

import ManagementFormActionBar from './ManagementFormActionBar';
import { FormTypes } from 'types/form';
import { ContentPanel } from 'components/layout';
import { Form, FormSummary } from 'components/form';
import { StepLoader } from 'components/progress-indicator';
import { errorsAreVisibleSelector, optionsByFieldSelector } from 'selectors/management';
import { showErrors as showErrorsActionCreator } from 'actions/connectorForm';

import type { Props as ManagementFormActionBarProps } from './ManagementFormActionBar';

import type {
  FieldSections,
  ReduxFormContainerProps
} from 'types/form';

export type Props = {
  completed: boolean,
  completionRedirectPathname: string,
  errorsAreVisible: boolean,
  fieldSections: FieldSections,
  formActionBarProps: ManagementFormActionBarProps,
  handleSubmit: Function,
  isEdit: boolean,
  isLastStep: boolean,
  optionsByField: any,
  reduxFormProps: ReduxFormContainerProps,
  route: Object,
  router: Object,
  showErrorsActionCreator: () => any,
  showSummary: boolean,
  statefulFieldSections: any,
  statefulFieldSectionsBlacklist: any
};

export class ManagementForm extends Component {
  props: Props;

  unsetRouteLeaveHook: () => any;

  componentDidMount() {
    const { route, router } = this.props;

    this.unsetRouteLeaveHook = router.setRouteLeaveHook(route, () => {
      if (this.props.reduxFormProps.dirty && !this.props.completed) {
        return 'You have unsaved information, are you sure you want to leave this page?';
      }
    });
  }

  componentDidUpdate() {
    const { completed, completionRedirectPathname } = this.props;

    if (completed) {
      browserHistory.push(completionRedirectPathname);
    }
  }

  componentWillUnmount() {
    this.props.reduxFormProps.destroy();
    this.unsetRouteLeaveHook();
  }

  render() {
    const {
      completed,
      errorsAreVisible,
      fieldSections,
      formActionBarProps,
      isLastStep,
      optionsByField,
      reduxFormProps,
      showErrorsActionCreator,
      showSummary,
      statefulFieldSections,
      statefulFieldSectionsBlacklist
    } = this.props;

    const {
      asyncValidating,
      error,
      handleSubmit,
      invalid,
      submitting
    } = reduxFormProps;

    const asyncValidatingEntireForm = asyncValidating === true;

    // TODO(@dylanharrington): figure out how to kick off the spinner when the request is taking a long
    // time - possibly by checking for asyncValidatingEntireForm?
    if (completed || submitting) {
      return (
        <StepLoader
          step={isLastStep ? (submitting ? 2 : 1) : 1}
          stepTotal={isLastStep ? 2 : 1}
          text={!completed && (asyncValidatingEntireForm || !isLastStep) ? 'Validating' : 'Saving'}
        />
      );
    }

    const filteredStatefulFieldSections = statefulFieldSections.map(section => ({
      ...section,
      fields: section.fields.filter(field => !statefulFieldSectionsBlacklist.includes(field.id))
    }));

    return (
      <form onSubmit={handleSubmit}>
        <ContentPanel hasError={Boolean(error)}>
          {error && (
            <pre>Error: '{error}'</pre>
          )}

          {showSummary ? (
            <FormSummary
              formType={FormTypes.CONNECTOR_FORM}
            />
          ) : (
            <Form
              errorsAreVisible={errorsAreVisible}
              fieldSections={fieldSections}
              optionsByField={optionsByField}
              showTitles={true}
            />
          )}

          {!showSummary && filteredStatefulFieldSections.length > 0 && (
            <Form
              errorsAreVisible={errorsAreVisible}
              fieldSections={filteredStatefulFieldSections}
              optionsByField={optionsByField}
            />
          )}
        </ContentPanel>
        <ManagementFormActionBar
          {...formActionBarProps}
          invalid={invalid}
          showErrors={() => showErrorsActionCreator(formActionBarProps.currentStep)}
        />
      </form>
    );
  }
}

export default withRouter(reduxForm({
  destroyOnUnmount: false,
  form: FormTypes.CONNECTOR_FORM,
  propNamespace: 'reduxFormProps'
})(connect(
  state => ({
    errorsAreVisible: errorsAreVisibleSelector(state),
    optionsByField: optionsByFieldSelector(state)
  }),
  {
    showErrorsActionCreator
  }
)(ManagementForm)));
