// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React, { Component } from 'react';
import styles from './Clock.css';
import moment from 'moment';

import { DateTimeFormats } from 'constants/timePicker';

const CONTROLS = {
  hours: 0,
  minutes: 1,
  amPm: 2
};

// Given the desired time offset in hours and currTime,
// Will output a timestamp that represents the currTime shifted the supplied offset
// Will always loop within the current AM/PM range
export const nextHour: Function = (
  offset: number,
  currTime: number
): number => {
  // Handle changes to the hour
  // Keep hour between the 1-12th hour of the current AM/PM range
  const validVals = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12];
  const currHour = parseInt(moment.unix(currTime).format('h'), 10);
  const newHour = currHour + offset;
  const currAmPm = moment.unix(currTime).format('A');

  if (newHour > Math.max(...validVals)) {
    // Moving up to 13
    // AM range 1am, PM range 1pm
    return moment.unix(currTime).hour(currAmPm === 'AM' ? 1 : 13).seconds(0).unix();
  }

  if (newHour < Math.min(...validVals)) {
    // Moving down to 0
    // AM range 12am, PM range 12pm
    return moment.unix(currTime).hour(currAmPm === 'AM' ? 0 : 12).seconds(0).unix();
  }

  if (newHour === 11 && offset < 0) {
    // Moving down to 11
    // 12am to 11am, 12pm to 11pm
    return moment.unix(currTime).hour(currAmPm === 'AM' ? 11 : 23).seconds(0).unix();
  }

  if (newHour === 12 && offset > 0) {
    // Moving up to 12
    // 11am to 12am, 11pm to 12pm
    return moment.unix(currTime).hour(currAmPm === 'AM' ? 0 : 12).seconds(0).unix();
  }

  // Handle normal hour intervals
  return moment.unix(currTime).add(offset, 'hours').seconds(0).unix();
};

// Given the desired time offset in minutes and currTime,
// Will output a timestamp that represents the currTime shifted the supplied offset
// Will always loop within the current day range
export const nextMinute: Function = (
  offset: number,
  currTime: number
): number => {
  // Handle changes to the minute
  // Keep minutes between the 0 - 59 minute of the current hour
  const currMinute = moment.unix(currTime).minute();
  let newMinute = currMinute + offset;

  if (newMinute > 59) {
    newMinute = 0;
  } else if (newMinute < 0) {
    newMinute = 59;
  }

  return moment.unix(currTime).minute(newMinute).seconds(0).unix();
};

// Given the currTime,
// Will output current AM time as PM, and PM time as AM
// Example: 2:30AM -> 2:30PM
export const nextAmPm: Function = (currTime: number): number => {
  // Toggle between AM PM
  const currHour = moment.unix(currTime).hour();

  if (currHour >= 12) {
    // PM range
    return moment.unix(currTime).hour(currHour - 12).seconds(0).unix();
  } else {
    // AM range
    return moment.unix(currTime).hour(currHour + 12).seconds(0).unix();
  }
};

export class Clock extends Component {

  props: {
    onDateChange: Function,
    currTime: number
  };

  state: {
    hoverUp: boolean,
    hoverDown: boolean,
    hoverControl: ?number,
    currTime: number
  };

  constructor(props: any) {
    super();
    this.state = {
      hoverUp: false,
      hoverDown: false,
      hoverControl: null,
      currTime: props.currTime
    };
  }

  handleClick: Function = (control: number): void => {
    let newCurrTime = moment.unix(this.state.currTime);

    if (control === CONTROLS.hours) {
      newCurrTime = nextHour(
        this.state.hoverUp ? 1 : -1,
        this.state.currTime
      );
    } else if (control === CONTROLS.minutes) {
      newCurrTime = nextMinute(
        this.state.hoverUp ? 1 : -1,
        this.state.currTime
      );
    } else if (control === CONTROLS.amPm) {
      newCurrTime = nextAmPm(
        this.state.currTime
      );
    }

    this.setState({ currTime: newCurrTime });
    this.props.onDateChange(newCurrTime);
  };

  handleHover: Function = (
    hoverUp: boolean,
    hoverControl: number
  ): void => {
    this.setState({
      hoverUp: hoverUp,
      hoverDown: !hoverUp,
      hoverControl: hoverControl
    });
  };

  reset: Function = (): void => {
    this.setState({
      hoverUp: false,
      hoverDown: false,
      hoverControl: null
    });
  };

  componentWillReceiveProps(props: any) {
    this.setState({
      currTime: props.currTime
    });
  }

  render() {
    const optionsStyleName = (control) => {
      if (this.state.hoverControl === control) {
        return this.state.hoverUp ? 'options-shift-up' : 'options-shift-down';
      } else {
        return 'options';
      }
    };

    const controls = [
      {
        control: CONTROLS.hours,
        offset: 1,
        unit: 'hour',
        format: 'h'
      }, {
        control: CONTROLS.minutes,
        offset: 1,
        unit: 'minute',
        format: 'mm'
      }, {
        control: CONTROLS.amPm,
        offset: 12,
        unit: 'hours',
        format: 'A'
      }
    ];

    return (
      <time dateTime={moment.unix(this.props.currTime).format(DateTimeFormats.DATE_TIME)} className={styles['clock-root']}>
        {
          controls.map(data => (
            <div className={styles['control']} key={data.control}>
              <div className={styles[optionsStyleName(data.control)]}>
                <span className={styles['option']}>{moment.unix(this.props.currTime).add(data.offset, data.unit).format(data.format)}</span>
                <span className={styles['option']}>{moment.unix(this.props.currTime).format(data.format)}</span>
                <span className={styles['option']}>{moment.unix(this.props.currTime).subtract(data.offset, data.unit).format(data.format)}</span>
              </div>
              <div
                className={styles['up']}
                onClick={() => this.handleClick(data.control)}
                onMouseEnter={() => this.handleHover(true, data.control)}
                onMouseLeave={() => this.reset()}
              />
              <div
                className={styles['down']}
                onClick={() => this.handleClick(data.control)}
                onMouseEnter={() => this.handleHover(false, data.control)}
                onMouseLeave={() => this.reset()}
              />
            </div>
          ))
        }
      </time>
    );
  }
}

export default Clock;
