// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { get, omit } from 'lodash';

import * as actions from './selectionListActions';

import type { Action } from 'types';
import type {
  DeselectItemAction,
  EndRangeSelectAction,
  InitializeAction,
  SelectAllAction,
  SelectItemAction,
  StartRangeSelectAction
} from './selectionListActions';

export type SelectionListState = {
  startRangeIndex: number,
  selectedIdMap: {}
};

export type SelectionListsState = {
  [listName: string]: SelectionListState
};

const listInitialState = {
  startRangeIndex: -1,
  selectedIdMap: {}
};

const initialState = {};

function handleSelectAll(state: SelectionListState, action: SelectAllAction): SelectionListState {
  const { list, dataKey } = action.payload;

  return {
    startRangeIndex: -1,
    selectedIdMap: {
      ...(list.reduce((obj, row) => {
        obj[row[dataKey]] = true;
        return obj;
      }, {}))
    }
  };
}

function handleSelectItem(state: SelectionListState, action: SelectItemAction): SelectionListState {
  return {
    startRangeIndex: -1,
    selectedIdMap: {
      ...state.selectedIdMap,
      [action.payload.itemId]: true
    }
  };
}

function handleDeselectItem(state: SelectionListState, action: DeselectItemAction): SelectionListState {
  return {
    startRangeIndex: -1,
    selectedIdMap: {
      ...omit(state.selectedIdMap, action.payload.itemId)
    }
  };
}

function handleStartRangeSelect(state: SelectionListState, action: StartRangeSelectAction): SelectionListState {
  return {
    ...state,
    startRangeIndex: action.payload.startIndex
  };
}

function handleEndRangeSelect(state: SelectionListState, action: EndRangeSelectAction): SelectionListState {
  const { list, endIndex, dataKey } = action.payload;

  const isSelection = list[state.startRangeIndex][dataKey] in state.selectedIdMap;
  const [start, end] = state.startRangeIndex > endIndex ? [endIndex, state.startRangeIndex] : [state.startRangeIndex, endIndex];

  const ids = list.slice(start, end + 1).map(item => item[dataKey]);

  return {
    startRangeIndex: -1,
    selectedIdMap: {
      ...(isSelection ? ({
        ...state.selectedIdMap,
        ...(ids.reduce((obj, id) => {
          obj[id] = true;
          return obj;
        }, {}))
      }): omit(state.selectedIdMap, ids))
    }
  };
}

function listReducer(state: SelectionListState = listInitialState, action: Action<*, *>): SelectionListState {
  switch (action.type) {
    case actions.SELECTION_LIST_SELECT_ALL:
      return handleSelectAll(state, action);
    case actions.SELECTION_LIST_DESELECT_ALL:
      return listInitialState;
    case actions.SELECTION_LIST_SELECT_ITEM:
      return handleSelectItem(state, action);
    case actions.SELECTION_LIST_DESELECT_ITEM:
      return handleDeselectItem(state, action);
    case actions.SELECTION_LIST_START_RANGE_SELECT:
      return handleStartRangeSelect(state, action);
    case actions.SELECTION_LIST_END_RANGE_SELECT:
      return handleEndRangeSelect(state, action);
    default:
      return state;
  }
}

function handleInitialize(state: SelectionListsState, action: InitializeAction): SelectionListsState {
  const { listName } = action.payload;

  if (listName in state) {
    return state;
  }

  return {
    ...state,
    [listName]: {
      ...listInitialState
    }
  };
}

function reducer(state: SelectionListsState = initialState, action: Action<*, *>): SelectionListsState {
  switch (action.type) {
    case actions.SELECTION_LIST_INITIALIZE:
      return handleInitialize(state, action);
    default:
      const listName = get(action, 'payload.listName');
      if (listName && listName in state) {
        return {
          ...state,
          [listName]: listReducer(state[listName], action)
        };
      }
      return state;
  }
}

export default reducer;
