// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { browserHistory } from 'react-router';

import { formValueSelector } from 'redux-form';
import { InputTypes } from 'constants/management';
import { FormTypes } from 'types/form';

import { ACTIONS_INDEX_PATHNAME } from './index';

import {
  alertsCreateAction,
  alertsTriggers
} from 'actions/alerts';

import { ActionTypes } from 'types/alerts';
import type { Dispatcher } from 'types';

export const RateUnits = {
  PER_MINUTE: 'PER_MINUTE',
  PER_HOUR: 'PER_HOUR',
  PER_4_HOURS: 'PER_4_HOURS',
  PER_8_HOURS: 'PER_8_HOURS',
  PER_DAY: 'PER_DAY'
};

export type RateUnitType = $Enum<typeof RateUnits>;


export const ActionFieldIds = {
  ACTION_NAME: 'ACTION_NAME',
  ENABLED: 'ENABLED',
  TRIGGERS: 'TRIGGERS',
  ACTION_TYPE: 'ACTION_TYPE',
  EMAIL: 'EMAIL',
  SUBJECT: 'SUBJECT',
  MAX_SEND_RATE: 'MAX_SEND_RATE',
  RATE_UNITS: 'RATE_UNITS'
};


export function canSaveSelector(state: Object) {
  const getField = (fieldName) => formValueSelector(FormTypes.ALERTS_ACTION_FORM)(state, fieldName);

  const requiredFields = [
    ActionFieldIds.ACTION_NAME,
    ActionFieldIds.ENABLED,
    ActionFieldIds.TRIGGERS,
    ActionFieldIds.ACTION_TYPE,
    ActionFieldIds.EMAIL,
    ActionFieldIds.SUBJECT,
    ActionFieldIds.MAX_SEND_RATE,
    ActionFieldIds.RATE_UNITS
  ];

  if (requiredFields.filter(field => getField(field) == null).length > 0) {
    return false;
  }

  if (getField(ActionFieldIds.TRIGGERS).length === 0) {
    return false;
  }

  // TODO(mhowlett): check actionName for uniqueness.

  return true;
}

export function fieldSectionsSelector(alertsState: Object) {
  return [
    {
      title: 'General',
      fields: [
        {
          colSpan: 10,
          id: ActionFieldIds.ACTION_NAME,
          inputType: InputTypes.TEXT,
          instantErrorFeedback: true,
          placeholder: 'Action Name',
          required: true
        },
        {
          colSpan: 2,
          id: ActionFieldIds.ENABLED,
          inputType: InputTypes.SWITCH,
          instantErrorFeedback: true,
          placeholder: '',
          required: false,
          offText: 'disabled',
          onText: 'enabled'
        }
      ]
    },

    {
      title: 'Triggers',
      fields: [
        {
          colSpan: 12,
          id: ActionFieldIds.TRIGGERS,
          inputType: InputTypes.SELECT,
          multi: true,
          instantErrorFeedback: true,
          placeholder: 'Triggers',
          required: true,
          showInfoIcon: true,
          options: alertsState.triggers == null ? [] : alertsState.triggers.map((triggerSummary) => (
            {
              label: triggerSummary.name,
              value: triggerSummary.guid
            }
          ))
        }
      ]
    },

    {
      title: 'Actions',
      fields: [
        {
          colSpan: 12,
          id: ActionFieldIds.ACTION_TYPE,
          inputType: InputTypes.SELECT,
          multi: false,
          locked: true,
          instantErrorFeedback: true,
          placeholder: 'Action',
          required: true,
          showInfoIcon: true,
          options: [
            {
              label: 'Send email',
              value: ActionTypes.SEND_EMAIL
            }
          ]
        },

        // TODO: ultimately the following fields will depend on state (of action field selected)
        {
          colSpan: 12,
          id: ActionFieldIds.EMAIL,
          inputType: InputTypes.TEXT,
          instantErrorFeedback: true,
          placeholder: 'Recipient email address',
          required: true
        },
        {
          colSpan: 12,
          id: ActionFieldIds.SUBJECT,
          inputType: InputTypes.TEXT,
          instantErrorFeedback: true,
          placeholder: 'Subject',
          required: true
        },
        {
          colSpan: 8,
          id: ActionFieldIds.MAX_SEND_RATE,
          inputType: InputTypes.TEXT,
          instantErrorFeedback: true,
          placeholder: 'Max send rate',
          required: true
        },
        {
          colSpan: 4,
          id: ActionFieldIds.RATE_UNITS,
          inputType: InputTypes.SELECT,
          multi: false,
          instantErrorFeedback: true,
          required: true,
          options: [
            {
              label: 'Per minute',
              value: RateUnits.PER_MINUTE
            },
            {
              label: 'Per hour',
              value: RateUnits.PER_HOUR
            },
            {
              label: 'Per 4 hours',
              value: RateUnits.PER_4_HOURS
            },
            {
              label: 'Per 8 hours',
              value: RateUnits.PER_8_HOURS
            },
            {
              label: 'Per day',
              value: RateUnits.PER_DAY
            }
          ]
        }
      ]
    }

  ];
}

export const NewActionFormSpec = {

  formType: FormTypes.ALERTS_ACTION_FORM,

  fieldSectionsSelector: (state: Object) => {
    return fieldSectionsSelector(state.alerts);
  },

  initialChangeValues: [
    { field: ActionFieldIds.ENABLED, value: true },
    { field: ActionFieldIds.ACTION_TYPE, value: ActionTypes.SEND_EMAIL },
    { field: ActionFieldIds.RATE_UNITS, value: RateUnits.PER_HOUR }
  ],

  actionBarButtonsSelector: (state: Object) => {
    return [
      {
        text: 'Cancel',
        isPrimary: false,
        isDisabled: false,
        onClick: () => { browserHistory.push(ACTIONS_INDEX_PATHNAME); }
      },
      {
        text: 'Save action',
        isPrimary: true,
        isDisabled: !canSaveSelector(state),
        onClick: NewActionFormSpec.formSubmitHandler
      }
    ];
  },

  modalComponent: null,

  formMountedHandler: (props: Object) => {
    props.dispatch(alertsTriggers.request());
  },

  formSubmitHandler: (dispatch: Dispatcher) => {
    dispatch(alertsCreateAction());
  },

  canSaveSelector: canSaveSelector
};
