// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React from 'react';
import { browserHistory } from 'react-router';

import { formValueSelector } from 'redux-form';
import { InputTypes } from 'constants/management';
import { FormTypes } from 'types/form';

import type { TriggerFormState } from 'reducers/alerts/triggerForm';

import { TRIGGERS_INDEX_PATHNAME } from './index';

import {
  alertsCreateTriggerFromState,
  alertsActions
} from 'actions/alerts';

import { setVisibleModal } from 'actions/layout';

import TriggerSavedModal from 'components/alerts/TriggerSavedModal';

import {
  AlertSources,
  MetricTypes,
  ConditionTypes,
  TriggerFieldIds,
  SelectOptionListIds
} from 'types/alerts';
import type { Dispatcher } from 'types';


const generalFieldSection = {
  title: 'General',
  fields: [
    {
      colSpan: 12,
      id: TriggerFieldIds.TRIGGER_NAME,
      inputType: InputTypes.TEXT,
      instantErrorFeedback: true,
      placeholder: 'Trigger Name',
      required: true
    }
  ]
};

const alertSourceField = {
  colSpan: 12,
  id: TriggerFieldIds.ALERT_SOURCE,
  inputType: InputTypes.SELECT,
  instantErrorFeedback: false,
  locked: true,
  multi: false,
  placeholder: 'Component Type',
  required: true,
  options: [
    {
      label: 'Consumer group',
      value: AlertSources.CONSUMER_GROUP
    }
  ]
};

const bufferField = {
  colSpan: 4,
  id: TriggerFieldIds.BUFFER,
  inputType: InputTypes.TEXT,
  instantErrorFeedback: false,
  placeholder: 'Buffer (seconds)',
  required: true
};

const consumerGroupMetricOptions = [
  {
    label: 'Maximum latency (ms)',
    value: MetricTypes.MAX_LATENCY
  },
  {
    label: 'Average latency (ms)',
    value: MetricTypes.AVG_LATENCY
  },
  {
    label: 'Consumption difference',
    value: MetricTypes.CONSUMPTION_DIFF
  }
];

const consumerGroupConditionOptions = [
  {
    label: 'Greater than',
    value: ConditionTypes.GREATER_THAN
  },
  {
    label: 'Less than',
    value: ConditionTypes.LESS_THAN
  },
  {
    label: 'Equal to',
    value: ConditionTypes.EQUAL
  },
  {
    label: 'Not equal to',
    value: ConditionTypes.NOT_EQUAL
  }
];


export function fieldSectionsSelector(alertSource: string, triggerFormState: TriggerFormState) {

  if (!alertSource) {
    return [
      generalFieldSection,
      {
        title: 'Components',
        fields: [
          alertSourceField
        ]
      }
    ];
  }

  if (alertSource === AlertSources.CONSUMER_GROUP) {
    return [
      generalFieldSection,
      {
        title: 'Components',
        fields: [
          alertSourceField,
          {
            colSpan: 12,
            id: TriggerFieldIds.CONSUMER_GROUP_TARGET,
            inputType: InputTypes.SELECT,
            instantErrorFeedback: false,
            multi: false,
            placeholder: 'Consumer Group Name',
            required: true,
            options: triggerFormState.selectOptions[SelectOptionListIds.CONSUMER_GROUPS]
          }
        ]
      },
      {
        title: 'Criteria',
        fields: [
          {
            colSpan: 8,
            id: TriggerFieldIds.METRIC,
            inputType: InputTypes.SELECT,
            instantErrorFeedback: false,
            multi: false,
            placeholder: 'Metric',
            required: true,
            options: consumerGroupMetricOptions
          },
          {
            colSpan: 4,
            id: TriggerFieldIds.CONDITION,
            inputType: InputTypes.SELECT,
            instantErrorFeedback: false,
            multi: false,
            placeholder: 'Condition', // Equal To
            required: true,
            options: consumerGroupConditionOptions
          },
          {
            colSpan: 8,
            id: TriggerFieldIds.VALUE,
            inputType: InputTypes.TEXT,
            instantErrorFeedback: false,
            placeholder: 'Value',
            required: true
          },
          bufferField
        ]
      }
    ];
  }

  return [
    generalFieldSection,
    {
      title: 'Components',
      fields: [
        alertSourceField,
        {
          colSpan: 12,
          id: TriggerFieldIds.CONNECTOR_TARGET,
          inputType: InputTypes.SELECT,
          instantErrorFeedback: false,
          multi: false,
          placeholder: 'Connector(s)',
          required: true,
          options: triggerFormState.selectOptions[SelectOptionListIds.CONNECTORS]
        }
      ]
    },
    {
      title: 'Criteria',
      fields: [
        {
          colSpan: 8,
          id: TriggerFieldIds.METRIC,
          inputType: InputTypes.SELECT,
          instantErrorFeedback: false,
          multi: false,
          placeholder: 'Metric',
          required: true,
          options: [
            {
              label: 'Maximum latency (ms)',
              value: MetricTypes.MAX_LATENCY
            }
          ]
        },
        {
          colSpan: 4,
          id: TriggerFieldIds.CONDITION,
          inputType: InputTypes.SELECT,
          instantErrorFeedback: false,
          multi: false,
          placeholder: 'Condition', // Equal To
          required: true,
            options: [
              {
                label: 'Equal to',
                value: ConditionTypes.EQUAL
              }
            ]
        },
        {
          colSpan: 8,
          id: TriggerFieldIds.VALUE,
          inputType: InputTypes.TEXT,
          instantErrorFeedback: false,
          placeholder: 'Value',
          required: true
        },
        bufferField
      ]
    }
  ];
}

export function canSaveSelector(state: Object) {
  const getField = (fieldName) => formValueSelector(FormTypes.ALERTS_TRIGGER_FORM)(state, fieldName);

  const requiredFields = [TriggerFieldIds.TRIGGER_NAME, TriggerFieldIds.ALERT_SOURCE, TriggerFieldIds.METRIC, TriggerFieldIds.CONDITION, TriggerFieldIds.VALUE, TriggerFieldIds.BUFFER];
  if (requiredFields.filter(field => getField(field) == null).length > 0) {
    return false;
  }

  // todo: check trigger name for uniqueness.

  if (getField(TriggerFieldIds.ALERT_SOURCE) === AlertSources.CONSUMER_GROUP) {
    if (getField(TriggerFieldIds.CONSUMER_GROUP_TARGET) == null) {
      return false;
    }
  }

  if (getField(TriggerFieldIds.ALERT_SOURCE) === AlertSources.KAFKA_CONNECTOR) {
    if (getField(TriggerFieldIds.CONNECTOR_TARGET) == null) {
      return false;
    }

    if (!consumerGroupMetricOptions.find(metricOption => metricOption.value === getField(TriggerFieldIds.METRIC))) {
      return false;
    }

    if (!consumerGroupConditionOptions.find(conditionOption => conditionOption.value === getField(TriggerFieldIds.CONDITION))) {
      return false;
    }
  }

  return true;
}


export const NewTriggerFormSpec = {

  formType: FormTypes.ALERTS_TRIGGER_FORM,

  fieldSectionsSelector: (state: Object) => {
    return fieldSectionsSelector(
      formValueSelector(FormTypes.ALERTS_TRIGGER_FORM)(state, TriggerFieldIds.ALERT_SOURCE),
      state.alerts.triggerForm
    );
  },

  initialChangeValues: [
    { field: TriggerFieldIds.ALERT_SOURCE, value: AlertSources.CONSUMER_GROUP },
    { field: TriggerFieldIds.BUFFER, value: 60 }
  ],

  actionBarButtonsSelector: (state: Object) => {
    return [
      {
        text: 'Cancel',
        isPrimary: false,
        isDisabled: false,
        onClick: () => {
          browserHistory.push(TRIGGERS_INDEX_PATHNAME);
        }
      },
      {
        text: 'Save trigger',
        isPrimary: true,
        isDisabled: !canSaveSelector(state),
        onClick: NewTriggerFormSpec.formSubmitHandler
      }
    ];
  },

  // eslint-disable-next-line react/display-name
  modalComponent: <TriggerSavedModal />,

  formMountedHandler: (props: Object) => {
    props.dispatch(alertsActions.request());
  },

  formSubmitHandler: (dispatch: Dispatcher) => {
    dispatch(alertsCreateTriggerFromState());
    dispatch(setVisibleModal('ActionFormModal'));
  },

  canSaveSelector: canSaveSelector
};
