// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { browserHistory } from 'react-router';

import { encodeFieldName } from 'utils/connector';
import { setStepAccessible, submitForm } from 'actions/connectorForm';

import type { FormConfig, FormStates, FormState } from 'types/form';

import { FormTypes } from 'types/form';

export const NEW_SINK_PATHNAME = '/management/connect/sinks/new';
export const NEW_SOURCE_PATHNAME = '/management/connect/sources/new';
export const SINKS_INDEX_PATHNAME = '/management/connect/sinks';
export const SOURCES_INDEX_PATHNAME = '/management/connect/sources';

export const SINK = 'SINK';
export const SOURCE = 'SOURCE';

export const STEP1 = 'STEP1';
export const STEP2 = 'STEP2';
export const STEP3 = 'STEP3';

export const InputTypes = {
  FREEFORM_SELECT: 'FREEFORM_SELECT',
  NUMBER: 'NUMBER',
  PASSWORD: 'PASSWORD',
  SELECT: 'SELECT',
  SWITCH: 'SWITCH',
  TEXT: 'TEXT'
};

export type InputType = $Enum<typeof InputTypes>;

export const EncodedFieldNames = {
  CONNECTOR_CLASS: encodeFieldName('connector.class'),
  NAME: encodeFieldName('name'),
  TOPICS: encodeFieldName('topics')
};

export function validateNameAndClass(values: Object) {
  const errors = {};

  if (!values[EncodedFieldNames.CONNECTOR_CLASS]) {
    errors[EncodedFieldNames.CONNECTOR_CLASS] = 'Required';
  }
  if (!values[EncodedFieldNames.NAME]) {
    errors[EncodedFieldNames.NAME] = 'Required';
  }

  return errors;
}

export function validateTopics(values: Object) {
  const errors = {};

  if (!values[EncodedFieldNames.TOPICS]) {
    errors[EncodedFieldNames.TOPICS] = 'Required';
  }

  return errors;
}

const fieldSectionBase = {
  errors: [],
  dependents: [],
  group: null,
  instantErrorFeedback: false,
  locked: false,
  importance: 'HIGH',
  required: true,
  visible: true,
  value: ''
};

const FieldSectionFields = {
  ConnectorClass: {
    ...fieldSectionBase,
    order: -2,
    colSpan: 4,
    id: EncodedFieldNames.CONNECTOR_CLASS,
    inputType: InputTypes.SELECT,
    placeholder: 'Connection Class'
  },
  Name: {
    ...fieldSectionBase,
    order: -1,
    colSpan: 8,
    id: EncodedFieldNames.NAME,
    inputType: InputTypes.TEXT,
    instantErrorFeedback: true,
    placeholder: 'Connection Name'
  },
  Topics: {
    ...fieldSectionBase,
    order: 0,
    colSpan: 12,
    id: EncodedFieldNames.TOPICS,
    inputType: InputTypes.SELECT,
    multi: true,
    placeholder: 'Topics',
    value: []
  }
};

const FieldSteps = {
  ChooseTopics: {
    navLabel: 'Choose Topics',
    validate: validateTopics,
    fields: [
      EncodedFieldNames.TOPICS
    ],
    fieldSections: [
      {
        title: 'Which topics do you want to get data from?',
        fields: [
          FieldSectionFields.Topics
        ]
      }
    ]
  },
  SetupConnection: {
    navLabel: 'Setup Connection',
    validate: validateNameAndClass,
    fields: [
      EncodedFieldNames.CONNECTOR_CLASS,
      EncodedFieldNames.NAME
    ],
    fieldSections: [
      {
        title: 'How should we connect to your data?',
        fields: [
          FieldSectionFields.ConnectorClass,
          FieldSectionFields.Name
        ]
      }
    ]
  },
  TestAndVerify: {
    navLabel: 'Test & Verify',
    fields: [],
    fieldSections: []
  }
};

export const NewSinkFormData: FormConfig = {
  firstStep: STEP1,
  isEdit: false,
  lastStep: STEP3,
  pathname: NEW_SINK_PATHNAME,
  redirectPathname: SINKS_INDEX_PATHNAME,
  steps: {
    [STEP1]: {
      ...FieldSteps.ChooseTopics,
      nextStep: STEP2,
      onSubmit(values, dispatch) {
        dispatch(setStepAccessible(STEP2));
        browserHistory.push({
          pathname: NEW_SINK_PATHNAME,
          query: {
            step: STEP2
          }
        });
      }
    },
    [STEP2]: {
      ...FieldSteps.SetupConnection,
      zeroErrorsBeforeSubmit: true,

      // Topics was setup in STEP1
      fieldBlacklist: [EncodedFieldNames.TOPICS],
      nextStep: STEP3,
      previousStep: STEP1,
      onSubmit(values, dispatch) {
        dispatch(setStepAccessible(STEP3));
        browserHistory.push({
          pathname: NEW_SINK_PATHNAME,
          query: {
            step: STEP3
          }
        });
      }
    },

    [STEP3]: {
      ...FieldSteps.TestAndVerify,
      previousStep: STEP2,
      onSubmit(values, dispatch) {
        dispatch(submitForm(FormTypes.CONNECTOR_FORM));
      }
    }
  }
};

export const NewSourceFormData: FormConfig = {
  isEdit: false,
  redirectPathname: SOURCES_INDEX_PATHNAME,
  pathname: NEW_SOURCE_PATHNAME,
  firstStep: STEP1,
  lastStep: STEP2,
  steps: {
    [STEP1]: {
      ...FieldSteps.SetupConnection,
      zeroErrorsBeforeSubmit: true,
      nextStep: STEP2,
      onSubmit(values, dispatch) {
        dispatch(setStepAccessible(STEP2));
        browserHistory.push({
          pathname: NEW_SOURCE_PATHNAME,
          query: {
            step: STEP2
          }
        });
      }
    },
    [STEP2]: {
      ...FieldSteps.TestAndVerify,
      previousStep: STEP1,
      onSubmit(values, dispatch) {
        dispatch(submitForm(FormTypes.CONNECTOR_FORM));
      }
    }
  }
};

export const Edit: $Shape<FormConfig> = {
  isEdit: true,
  firstStep: STEP1,
  lastStep: STEP2,
  steps: {
    [STEP1]: {
      nextStep: STEP2,
      navLabel: 'Edit Connection',
      onSubmit(values, dispatch) {
        dispatch(setStepAccessible(STEP2));
        // TODO(@dylanharrington): create actions to move between steps
        // that will have access to non-window location in state
        browserHistory.push({
          pathname: window.location.pathname,
          query: {
            step: STEP2
          }
        });
      },
      fields: [
        EncodedFieldNames.CONNECTOR_CLASS,
        EncodedFieldNames.NAME
      ],
      fieldSections: [
        {
          title: '',
          fields: [
            { ...FieldSectionFields.ConnectorClass, locked: true },
            { ...FieldSectionFields.Name, locked: true }
          ]
        }
      ],
      zeroErrorsBeforeSubmit: true
    },
    [STEP2]: {
      ...FieldSteps.TestAndVerify,
      onSubmit(values, dispatch) {
        dispatch(submitForm(FormTypes.CONNECTOR_FORM));
      },
      previousStep: STEP1
    }
  }
};

export const EditSinkFormData: FormConfig = {
  ...Edit,
  redirectPathname: SINKS_INDEX_PATHNAME
};

export const EditSourceFormData: FormConfig = {
  ...Edit,
  redirectPathname: SOURCES_INDEX_PATHNAME
};

const createEmptySteps = steps => steps.reduce((accum, stepKey, index) => {
  accum[stepKey] = {
    canAccess: index === 0,
    errorsAreVisible: false,
    fieldSections: [],
    fields: []
  };
  return accum;
}, {});

const newSource: $Shape<FormState> = {
  type: SOURCE,
  currentPathname: NEW_SOURCE_PATHNAME,
  steps: createEmptySteps([STEP1, STEP2])
};

const newSink: $Shape<FormState> = {
  type: SINK,
  currentPathname: NEW_SINK_PATHNAME,
  steps: createEmptySteps([STEP1, STEP2, STEP3])
};

const edit: $Shape<FormState> = {
  currentStep: STEP1,
  steps: createEmptySteps([STEP1, STEP2])
};

export const FormConfigs: FormStates = {
  edit,
  [NEW_SINK_PATHNAME]: newSink,
  [NEW_SOURCE_PATHNAME]: newSource
};
