// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.

import { Router, Route, IndexRoute, IndexRedirect, Redirect } from 'react-router';
import { Provider } from 'react-redux';
import React, { Component } from 'react';

import { SINK, SOURCE, EditSinkFormData, EditSourceFormData, STEP1, NewSinkFormData, NewSourceFormData } from 'constants/management';
import App from 'components/App';
import { DataLayoutTypes } from 'types/monitoring/data';
import DataManager from 'components/monitoring/DataManager';
import { handleRouteEnterParamFiltering } from 'utils/navigation';

import Logout from 'components/Logout';
import NotFound from 'components/NotFound';
import { ConnectorIndex, ConnectorActionBar, EditForm, ManagementSubNav, ManagementFormContainer } from 'components/management';
import { SinksTopLinks } from 'components/management/sinks';
import { SourcesTopLinks } from 'components/management/sources';

import ClusterIndex from 'components/ClusterIndex';
import { EditClusterFormSpec } from 'constants/clusters';

import { Monitoring } from 'components/monitoring';
import MonitoringBody from 'components/monitoring/MonitoringBody';
import DeliveryLatencyCharts from 'components/monitoring/DeliveryLatencyCharts';
import DeliveryLatencyMemberCharts from 'components/monitoring/DeliveryLatencyMemberCharts';

import { SubNavLayout } from 'components/layout';
import ConnectLayout from 'components/ConnectLayout';

import * as Alerts from 'components/alerts';
import { ActionForm } from 'components/form';
import {
  NewTriggerFormSpec,
  EditTriggerFormSpec,
  NewActionFormSpec,
  EditActionFormSpec
} from 'constants/alerts';


function handleEnterMultiStepForm({ location }, replace) {
  if (!('step' in location.query)) {
    replace({
      ...location,
      query: {
        ...location.query,
        step: STEP1
      }
    });
  }
}

const multiStepFormProps = {
  onEnter: handleEnterMultiStepForm,
  queryParamsWhitelist: ['step']
};

export class AppContainer extends Component {
  props: {
    history: Object,
    routerKey: any,
    store: Object
  };

  render() {
    const {
      history,
      routerKey,
      store
    } = this.props;

    return (
      <Provider store={store}>
        <div style={{ height: '100%' }}>
          <Router history={history} key={routerKey}>
            <Route path="/logout" component={Logout} />

            <Route
              path="/"
              component={App}
              onChange={handleRouteEnterParamFiltering}
            >
              <IndexRedirect to="monitoring" />

              {/* redirect from 3.0.x -> 3.1.x paths */}
              <Redirect from="/kafka-connect" to="/management/connect" />
              <Redirect from="/kafka-connect/:one" to="/management/connect/:one" />
              <Redirect from="/kafka-connect/:one/:two" to="/management/connect/:one/:two" />
              <Redirect from="/kafka-connect/:one/:two/:three" to="/management/connect/:one/:two/:three" />

              <Route path="management" name="Management">
                <IndexRedirect to="connect" />
                <Route path="clusters" name="Clusters" component={SubNavLayout}>
                  <IndexRoute components={{ content: ClusterIndex }} />
                  <Route path=":clusterId">
                    <IndexRedirect to="edit" />
                    <Route
                      name="Edit Cluster"
                      path="edit"
                      components={{ content: ActionForm }}
                      routedFormData={EditClusterFormSpec}
                    />
                  </Route>
                </Route>
                <Route
                  component={ConnectLayout}
                  name="Kafka Connect"
                  queryParamsWhitelist={['connectClusterId']}
                  path="connect"
                >
                  <IndexRedirect to="sources" />
                  <Route path="sources" name="Sources">
                    <IndexRoute connectorType={SOURCE} components={{ subnav: ManagementSubNav, content: ConnectorIndex, topLinks: SourcesTopLinks }} />
                    <Route
                      {...multiStepFormProps}
                      components={{ content: ManagementFormContainer }}
                      name="New Source"
                      path="new"
                      routedFormData={NewSourceFormData}
                    />
                    <Route path=":connectorName">
                      <IndexRedirect to="edit" />
                      <Route
                        {...multiStepFormProps}
                        components={{
                          subnav: ManagementSubNav,
                          content: EditForm,
                          topLinks: ConnectorActionBar
                        }}
                        name="Edit Source"
                        path="edit"
                        routedFormData={EditSourceFormData}
                      />
                    </Route>
                  </Route>
                  <Route path="sinks" name="Sinks">
                    <IndexRoute connectorType={SINK} components={{ subnav: ManagementSubNav, content: ConnectorIndex, topLinks: SinksTopLinks }} />
                    <Route
                      {...multiStepFormProps}
                      components={{ content: ManagementFormContainer }}
                      name="New Sink"
                      path="new"
                      routedFormData={NewSinkFormData}
                    />
                    <Route path=":connectorName">
                      <IndexRedirect to="edit" />
                      <Route
                        {...multiStepFormProps}
                        components={{
                          subnav: ManagementSubNav,
                          content: EditForm,
                          topLinks: ConnectorActionBar
                        }}
                        name="Edit Sink"
                        path="edit"
                        routedFormData={EditSinkFormData}
                      />
                    </Route>
                  </Route>
                </Route>
              </Route>

              <Route
                component={SubNavLayout}
                name="Alerts"
                path="alerts"
              >
                <IndexRedirect to="overview" />

                <Route path="overview" name="Overview">
                  <IndexRedirect to="history" />
                  <Route path="history" name="History">
                    <IndexRoute components={{ subnav: Alerts.OverviewSubNav, content: Alerts.History }} />
                    <Route path=":timeguid" components={{ subnav: Alerts.OverviewSubNav, content: Alerts.History }} />
                  </Route>
                  <Route path="triggers" name="Triggers">
                    <IndexRoute components={{ subnav: Alerts.OverviewSubNav, content: Alerts.Triggers, topLinks: Alerts.TriggersTopLinks }} />
                    <Route
                      path=":triggerid/actions"
                      name="Add Actions"
                      components={{ content: Alerts.AddActions }}
                    />
                    <Route
                      path="new"
                      name="New Trigger"
                      components={{ content: ActionForm }}
                      routedFormData={NewTriggerFormSpec}
                    />
                    <Route
                      path=":triggerid/edit"
                      name="Edit Trigger"
                      components={{ content: ActionForm }}
                      routedFormData={EditTriggerFormSpec}
                    />
                  </Route>
                  <Route path="actions" name="Actions">
                    <IndexRoute components={{ subnav: Alerts.OverviewSubNav, content: Alerts.Actions, topLinks: Alerts.ActionsTopLinks }} />
                    <Route
                      path="new"
                      name="New Action"
                      components={{ content: ActionForm }}
                      routedFormData={NewActionFormSpec}
                    />
                    <Route
                      path=":actionid/edit"
                      name="Edit Action"
                      components={{ content: ActionForm }}
                      routedFormData={EditActionFormSpec}
                    />
                  </Route>
                </Route>
                <Route path="integration"
                  name="Integration"
                  components={{ subnav: Alerts.IntegrationSubNav, content: Alerts.RestAPI }}
                />
              </Route>

              {/* redirect from 3.0.x -> 3.1.x paths */}
              <Redirect from="/stream-monitoring/consumer-groups" to="/monitoring/consumer-groups" />
              <Redirect from="/stream-monitoring/consumer-groups/:id" to="/monitoring/consumer-groups/:id" />
              <Redirect from="/stream-monitoring/consumer-groups-topics" to="/monitoring/consumer-groups-topics" />
              <Redirect from="/stream-monitoring/consumer-groups-topics/:id" to="/monitoring/consumer-groups-topics/:id" />
              <Redirect from="/stream-monitoring/consumers" to="/monitoring/consumers" />
              <Redirect from="/stream-monitoring/consumers/:id" to="/monitoring/consumers/:id" />

              <Route
                component={Monitoring}
                name="Stream Monitoring"
                queryParamsWhitelist={['monitoringClusterId', 'fixed-end', 'fixed-start', 'rolling', 'growing']}
                path="monitoring"
              >
                <IndexRedirect to="consumer-groups" />
                <Route path="consumer-groups" name="Message Delivery">
                  <IndexRoute
                    routedDataLayoutType={DataLayoutTypes.CONSUMER_GROUPS}
                    components={{ bottom: MonitoringBody, children: DeliveryLatencyMemberCharts, top: DeliveryLatencyCharts, dataManager: DataManager }}
                  />
                  <Route
                    path=":id"
                    routedDataLayoutType={DataLayoutTypes.CONSUMER_GROUP}
                    components={{ bottom: MonitoringBody, children: DeliveryLatencyMemberCharts, top: DeliveryLatencyCharts, dataManager: DataManager }}
                  />
                </Route>
                <Route path="consumer-groups-topics" name="Message Delivery">
                  <IndexRoute
                    routedDataLayoutType={DataLayoutTypes.CONSUMER_GROUPS_TOPICS}
                    components={{ bottom: MonitoringBody, children: DeliveryLatencyMemberCharts, top: DeliveryLatencyCharts, dataManager: DataManager }}
                  />
                  <Route
                    path=":id"
                    routedDataLayoutType={DataLayoutTypes.CONSUMER_GROUP_TOPICS}
                    components={{ bottom: MonitoringBody, children: DeliveryLatencyMemberCharts, top: DeliveryLatencyCharts, dataManager: DataManager }}
                  />
                </Route>
                <Route path="consumers" name="Message Delivery">
                  <Route
                    path=":id"
                    routedDataLayoutType={DataLayoutTypes.CONSUMER}
                    components={{ bottom: MonitoringBody, children: DeliveryLatencyMemberCharts, top: DeliveryLatencyCharts, dataManager: DataManager }}
                  />
                </Route>
              </Route>

              <Route path="*" name="404 Not Found" component={NotFound} />

            </Route>
          </Router>
        </div>
      </Provider>
    );
  }
}

export default AppContainer;
