// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import type { Action } from '../../types';
import {
  ALERTS_RECEIVE_MONITORING_DATA,
  ALERTS_RECEIVE_HISTORY,
  ALERTS_PREPEND_NEW_HISTORY
} from 'actions/alerts';
import type {
  AlertsReceiveMonitoringDataAction,
  AlertsReceiveHistoryAction
} from 'actions/alerts';

import type { AlertSummary } from 'types/alerts';


export type HistoryItemState = {
  monitoringData: ?Object,
  alertInfo: ?AlertSummary
}

export type HistoryState = {
  main: Array<Object>,
  new: Array<Object>,
  item: HistoryItemState
};

const defaultState : HistoryState = {
  main: [],
  new: [],
  item: {
    monitoringData: null,
    alertInfo: null
  }
};

const preprocessHistoryItem = (item: Object) => {
  // TODO(mhowlett): doing the conversions to int here is is not robust against future changes to the schema.
  return {
    ...item,
    timestamp: Number(item.timestamp),
    monitoringTrigger: {
      ...item.monitoringTrigger,
      lagMs: Number(item.monitoringTrigger.lagMs),
      longValue: Number(item.monitoringTrigger.longValue)
    },
    triggers: item.triggers.map(trigger => ({
      longValue: Number(trigger.longValue),
      window: Number(trigger.window)
    }))
  };
};

function setHistory(state: HistoryState, action: AlertsReceiveHistoryAction): HistoryState {

  const data = action.payload.data.map(preprocessHistoryItem);

  switch (action.payload.spec.type) {

    case 'MAIN':

      if (action.payload.spec.itemId == null) {
        return {
          ...state,
          main: data
        };
      }

      return {
        ...state,
        main: state.main.concat(data.slice(1))
      };

    case 'NEW':
      // if there are no items in the main list, then might as well put these there.
      if (state.main == null || state.main.length === 0) {
        return {
          ...state,
          main: data
        };
      }

      const latestGuid = state.new.length > 0
        ? state.new[0].guid
        : state.main[0].guid;

      const latestLoc = data.findIndex(item => item.guid === latestGuid);

      // If the newest item in main isn't in the new list, then to avoid creating
      // a gap, replace the main list. this is not ideal, but if it ever happens
      // there are a lot of alerts triggering and the user has bigger problems.
      // Given the data available from the server, this is the best we can do here.
      if (latestLoc === -1) {
        return {
          ...state,
          main: data,
          new: []
        };
      }

      return {
        ...state,
        new: data.slice(0, latestLoc).concat(state.new)
      };

    case 'SPECIFIC':

      if (data.length !== 1) {
        return state;
      }

      return {
        ...state,
        item: {
          monitoringData: state.item.monitoringData,
          alertInfo: data[0]
        }
      };

    default:
      return state;
  }
}

function setMonitoringData(state: HistoryState, action: AlertsReceiveMonitoringDataAction): HistoryState {
  return {
    ...state,
    item: {
      alertInfo: state.item.alertInfo,
      monitoringData: action.payload.data
    }
  };
}


function prependNewHistory(state: HistoryState): HistoryState {
  return {
    ...state,
    main: [...state.new, ...state.main],
    new: []
  };
}

export function history(state: HistoryState = defaultState, action: Action<any, any>): HistoryState {
  switch (action.type) {
    case ALERTS_PREPEND_NEW_HISTORY:
      return prependNewHistory(state);
    case ALERTS_RECEIVE_HISTORY:
      return setHistory(state, action);
    case ALERTS_RECEIVE_MONITORING_DATA:
      return setMonitoringData(state, action);
    default:
      return state;
  }
}
