// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { groupBy, sortBy } from 'lodash';
import { LOCATION_CHANGE } from 'react-router-redux';

import * as ConnectorFormActions from 'actions/connectorForm';
import { FormConfigs, InputTypes } from 'constants/management';
import { encodeFieldName } from 'utils/connector';

import type { Action } from 'types';
import type {
  FetchEditSuccessAction,
  ShowErrorsAction,
  ValidateSuccessAction
} from 'actions/connectorForm';
import type { ConnectorPluginValidationConfig } from 'actions/connectorPlugins';
import type {
  FieldSectionField,
  FormState,
  FormStepKey,
  FormStateSteps
} from 'types/form';

function updateFieldOptions(state: FormState, action: Action<*, *>): FormState {
  const { field, options } = action.payload;

  return ({
    ...state,
    optionsByField: {
      ...state.optionsByField,
      [field]: options
    }
  });
}

function enableStepAccess(steps: FormStateSteps, stepKey: FormStepKey): FormStateSteps {
  if (steps[stepKey]) {
    return {
      ...steps,
      [stepKey]: {
        ...steps[stepKey],
        canAccess: true
      }
    };
  }
  return steps;
}

const definitionTypeToInputType = {
  BOOLEAN: InputTypes.SWITCH,
  CLASS: InputTypes.TEXT,
  DOUBLE: InputTypes.NUMBER,
  INT: InputTypes.NUMBER,
  LIST: InputTypes.FREEFORM_SELECT,
  LONG: InputTypes.NUMBER,
  PASSWORD: InputTypes.PASSWORD,
  SHORT: InputTypes.NUMBER,
  STRING: InputTypes.TEXT
};

const widthToColSpan = {
  DEFAULT: 4,
  FULL: 12,
  LONG: 6,
  MEDIUM: 4,
  SHORT: 2
};

// TODO(@dylanharrington): verify what is really necessary here since this is driving the core form types
const mapConfigToField = ({ value, definition }: ConnectorPluginValidationConfig): FieldSectionField => ({
  colSpan: widthToColSpan[definition.width] | widthToColSpan.DEFAULT,
  defaultValue: definition.default_value,
  dependents: definition.dependents,
  documentation: definition.documentation,
  errors: value.errors,
  group: definition.group,
  id: encodeFieldName(definition.name),
  importance: definition.importance,
  inputType: definitionTypeToInputType[definition.type],
  instantErrorFeedback: false,
  locked: false,
  options: value.recommended_values.map(value => ({ label: value, value })),
  order: definition.order,
  placeholder: definition.display_name,
  required: definition.required,
  value: value.value,
  visible: value.visible
});

function updateConnectorWithConfigs(
  state: FormState,
  validationConfigs: Array<ConnectorPluginValidationConfig>,
  orderedGroupNames: Array<string> = []
): FormState {

  const fields = validationConfigs.map((config: ConnectorPluginValidationConfig): FieldSectionField => {
    const field = mapConfigToField(config);

    // nothing currently can be assigned InputTypes.SELECT at this point
    if (field.inputType === InputTypes.FREEFORM_SELECT) {
      field.multi = true;
    }

    if (field.id === 'topics') {
      field.inputType = InputTypes.SELECT;
      field.colSpan = widthToColSpan.FULL;
    }

    if (field.id === 'topic') {
      field.inputType = InputTypes.SELECT;
      field.multi = false;
    }

    if (field.id === encodeFieldName('connector.class')) {
      field.inputType = InputTypes.SELECT;
      field.multi = false;
    }

    if (field.multi) {
      if (field.value && field.value.length) {
        field.value = field.value.split(',');
      } else {
        field.value = [];
      }
    }

    return field;
  });

  const visibleFields = fields.filter(({ visible }) => visible);

  const groupedAndSortedFields = groupBy(sortBy(visibleFields, 'order'), 'group');

  const fieldSections = [...orderedGroupNames, 'null'].map((groupName: string) => {
    if (groupName in groupedAndSortedFields) {
      return {
        title: groupName === 'null' ? 'General' : groupName,
        fields: groupedAndSortedFields[groupName]
      };
    }
  }).filter(fieldSection => typeof fieldSection !== 'undefined');

  const initialValues = fields.reduce((accum, field) => {
    if (field.value !== field.defaultValue) {
      accum[field.id] = field.value;
    }

    return accum;
  }, {});

  return {
    ...state,
    initialValues,
    steps: {
      ...state.steps,
      [state.currentStep]: {
        ...state.steps[state.currentStep],
        fieldSections
      }
    }
  };
}

function locationChange(state: FormState, action: Action<*, *>): FormState {
  const { pathname } = action.payload;

  if (pathname === state.currentPathname) {
    return {
      ...state,
      currentStep: action.payload.query.step
    };
  }

  if (pathname in FormConfigs) {
    return {
      ...state,
      ...FormConfigs[pathname]
    };
  }

  return initialState;
}

const submitSuccess = (state: FormState): FormState => ({
  ...state,
  completed: true
});

export const initialState: $Shape<FormState> = {
  errorCount: 0,
  completed: false,
  initialValues: {},
  isEdit: false,
  loadingConnector: false,
  optionsByField: {}
};

const setupEdit = (): FormState => ({
  ...initialState,
  ...FormConfigs.edit,
  currentPathname: window.location.pathname,
  loadingConnector: true,
  isEdit: true
});

function fetchEditSuccess(state: FormState, action: FetchEditSuccessAction): FormState {
  if (!state.isEdit) {
    return state;
  }

  const { configs, error_count, groups } = action.payload.validationData;
  const { type } = action.payload.connector;
  return {
    ...updateConnectorWithConfigs(state, configs, groups),
    type,
    errorCount: error_count,
    loadingConnector: false
  };
}

function showErrors(steps: FormStateSteps, action: ShowErrorsAction): FormStateSteps {
  const { step } = action.payload;
  if (steps[step]) {
    return {
      ...steps,
      [step]: {
        ...steps[step],
        errorsAreVisible: true
      }
    };
  }
  return steps;
}

function validateSuccess(state: FormState, action: ValidateSuccessAction): FormState {
  const { configs, error_count, groups } = action.payload.validationData;

  return {
    ...updateConnectorWithConfigs(state, configs, groups),
    errorCount: error_count
  };
}

export function reducer(state: FormState = initialState, action: Action<any, any>): FormState {
  switch (action.type) {
    case ConnectorFormActions.CONNECTOR_FORM_SUBMIT_SUCCESS:
      return submitSuccess(state, action);
    case ConnectorFormActions.CONNECTOR_FORM_VALIDATE_SUCCESS:
      return validateSuccess(state, action);
    case ConnectorFormActions.CONNECTOR_FORM_FETCH_EDIT_SUCCESS:
      return fetchEditSuccess(state, action);
    case ConnectorFormActions.LOAD_CONNECTOR_FOR_EDITING:
      return setupEdit();
    case ConnectorFormActions.STEP_SET_ACCESSIBLE:
      return {
        ...state,
        steps: enableStepAccess(state.steps, action.payload.step)
      };
    case LOCATION_CHANGE:
      return locationChange(state, action);
    case ConnectorFormActions.CONNECTOR_FORM_SEARCH_SUCCEEDED:
      return updateFieldOptions(state, action);
    case ConnectorFormActions.CONNECTOR_FORM_SHOW_ERRORS:
      return {
        ...state,
        steps: showErrors(state.steps, action)
      };
    default:
      return state;
  }
}

export default reducer;
