// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import {
  RECEIVE_DATA,
  CLEAR_DATA
} from 'actions/monitoring/data';

import type { Action } from 'types';

import {
  idNameFromDataLayoutType,
  memberIdNameFromDataLayoutType
} from 'utils/monitoring/data';

import { RequestTypes } from 'types/monitoring/data';

import { assert } from 'utils';


export function mergeArrays(array1: ?Array<number>, array2: ?Array<number>, startIndex: number): ?Array<number> {
  // assumptions not checked for:
  //   startIndex >= 0
  if (!array1) {
    if (!array2) {
      return null;
    }
    return Array(startIndex)
      .fill(null)
      .concat(array2);
  }
  const padding = startIndex - array1.length;
  const array2OrEmpty = array2 ? array2 : [];
  if (padding >= 0) {
    return array1
      .concat(Array(padding).fill(null))
      .concat(array2OrEmpty);
  } else if (padding < 0) {
    return array1
      .slice(0, padding)
      .concat(array2OrEmpty);
  }
}


export function mergeArrayCollection(
  data: ?Object,
  dataUpdate: ?Object,
  startIndex: number,
  dataUpdateArrayLength: number
): Object {

  // assumptions not checked for:
  //   data, dataUpdate format:
  //   {
  //     <name1>: [...],
  //     <name2>: [...]
  //   }
  //   or null
  //
  //   all arrays in data are of equal length and
  //   all arrays in dataUpdate are of equal length (but in general a different length to those in data)
  //
  //   startIndex >= 0
  //
  // note:
  //   dataUpdateArrayLength is a necessary parameter because it is valid for dataUpdate
  //   to be null or empty. In some cases there is enough information within this method
  //   that it could be calulcated, but this is not generally true.

  const result = {};

  if (dataUpdate) {
    Object.keys(dataUpdate).forEach(seriesName => {
      const array1 = data && data[seriesName]
        ? data[seriesName]
        : null;
      // $FlowFixMe
      const array2 = dataUpdate[seriesName];
      result[seriesName] = mergeArrays(array1, array2, startIndex);
    });
  }

  if (data) {
    Object.keys(data).forEach(seriesName => {
      if (!result[seriesName]) {
        const filler = Array(dataUpdateArrayLength).fill(null);
        // $FlowFixMe
        result[seriesName] = mergeArrays(data[seriesName], filler, startIndex);
      }
    });
  }

  return result;
}


export function constructMembers(
  members: ?Array<Object>,
  updateMembers: ?Array<Object>,
  memberIdNames: Array<string>
): Array<Object> {

  const membersOrEmpty = members ? members : [];

  if (!updateMembers || updateMembers.length === 0) {
    return [];
  }

  // Note: URI encoding here is used to ensure array -> string uniqueness.
  const keyAsString = (key: Object): string =>
    memberIdNames.map(name => encodeURIComponent(key[name])).join(' ');

  const memberIndices = !updateMembers
    ? {}
    : membersOrEmpty.reduce((accum, elem, i) => ({ ...accum, [keyAsString(elem)]: i }), {});

  return updateMembers.map(
    (updateMember) => {
      const key = keyAsString(updateMember);
      return memberIndices[key] == null
        ? updateMember
        : membersOrEmpty[memberIndices[key]];
    }
  );

}


export function receiveMemberList(state: ?Object, data: ?Object) {
  if (!data) {
    return state;
  }

  if (!state) {
    data.listEpoch = Date.now();
    delete data.startTimeMs;
    delete data.consumed;
    delete data.produced;
    return data;
  }

  const result = { ...state };

  result.sources = data.sources;
  result.destinations = data.destinations;

  const memberIdNames = memberIdNameFromDataLayoutType(data.type);
  result.members = constructMembers(state.members, data.members, memberIdNames);

  result.listEpoch = Date.now();

  return result;
}


export function dataTypeMismatch(state: Object, data: Object): boolean {
  if (data.type !== state.type) {
    return true;
  }

  const idName = idNameFromDataLayoutType(data.type);
  if (data[idName] !== state[idName]) {
    return true;
  }

  if (data.stepMs !== state.stepMs) {
    return true;
  }

  return false;
}


export function constructMessageDeliveryData(
  existingData: Object,
  newData: Object,
  newDataStartTimeMs: number,
  isFullPeriod: boolean,
  stepMs: number
) {

  const result = { ...newData };

  const startIndex = existingData
    ? (newDataStartTimeMs - existingData.startTimeMs) / stepMs
    : null;

  if (isFullPeriod || startIndex == null || startIndex < 0) {
    result.fullEpoch = Date.now();
    result.partialEpoch = Date.now();
    result.startTimeMs = newDataStartTimeMs;
  } else {
    result.fullEpoch = existingData.fullEpoch;
    result.partialEpoch = Date.now();
    result.startTimeMs = existingData.startTimeMs;
    result.consumed = mergeArrayCollection(
      existingData.consumed,
      newData.consumed,
      startIndex,
      newData.consumed.count.length
    );
    result.produced = mergeArrayCollection(
      existingData.produced,
      newData.produced,
      startIndex,
      newData.produced.count.length
    );
  }

  return result;
}


export function receiveMemberData(state: ?Object, data: ?Object, isFullPeriod: boolean) {

  assert(state != null, 'expecting state to be defined');

  if (data != null) {
    assert(data.members && data.members.length === 1, 'expecting exactly one member');
  }

  if (!state) {
    return null;
  }

  if (!data) {
    return state;
  }

  if (dataTypeMismatch(state, data)) {
    return state;
  }

  const member = data.members[0];

  const result = { ...state };

  // Note:
  // This is O(N). There is no avoiding this without using ImmutableJS
  // because although we could maintain a map of member props -> index,
  // we still need to copy every member of the array here (because if
  // a child of an object changes, that object needs to change ... up
  // to the root object because we need immutability). I don't think we
  // should do work at optimizing this unless it becomes a bottleneck,
  // which I think it probably won't.

  const newMembers = [];
  for (let i=0; i<state.members.length; ++i) {
    const sMember = state.members[i];

    if (sMember.partition === member.partition &&
        sMember.topic === member.topic &&
        sMember.group === member.group &&
        sMember.clientId === member.clientId
    ) {

      newMembers.push(
        constructMessageDeliveryData(
          sMember, member,
          data.startTimeMs, isFullPeriod, state.stepMs
        )
      );

    } else {
      newMembers.push(sMember);
    }

  }
  result.members = newMembers;

  return result;
}

export function receiveOverviewData(state: ?Object, data: ?Object, isFullPeriod: boolean) {
  assert(state != null, 'received overview data but data state is not defined.');
  if (!state) {
    return null;
  }

  if (!data) {
    return state;
  }

  if (dataTypeMismatch(state, data)) {
    return state;
  }

  const result = { ...state };
  result.overview = constructMessageDeliveryData(
    state.overview, data,
    data.startTimeMs, isFullPeriod, state.stepMs
  );

  return result;
}


export function receiveData(state: ?Object, payload: Object): ?Object {
  switch (payload.requestType) {
    case RequestTypes.MEMBER_LIST:
      return receiveMemberList(state, payload.data);
    case RequestTypes.MEMBER:
      return receiveMemberData(state, payload.data, payload.isFullPeriod);
    case RequestTypes.OVERVIEW:
      return receiveOverviewData(state, payload.data, payload.isFullPeriod);
    default:
      return state;
  }
}


export function clearData(): ?Object {
  return null;
}


export function data(state: ?Object = null, action: Action<*, *>): ?Object {
  switch (action.type) {
    case RECEIVE_DATA:
      return receiveData(state, action.payload);
    case CLEAR_DATA:
      return clearData();
    default:
      return state;
  }
}
