// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { browserHistory } from 'react-router';
import { call, put, fork, select } from 'redux-saga/effects';
import { takeEvery, takeLatest } from 'redux-saga';
import {
  ALERTS_REQUEST_ACTIONS,
  ALERTS_REQUEST_ACTION,
  ALERTS_CREATE_ACTION,
  ALERTS_UPDATE_ACTION,
  ALERTS_CREATE_ACTION_FROM_STATE,
  ALERTS_UPDATE_ACTION_FROM_STATE,
  ALERTS_DELETE_ACTION,
  alertsActions,
  alertsUpdateActionFromState
} from 'actions/alerts';
import type {
  AlertsRequestActionAction,
  AlertsDeleteActionAction,
  AlertsUpdateActionFromStateAction,
  AlertsUpdateActionAction
} from 'actions/alerts';
import { EnabledDisabled } from 'types';
import {
  RateUnits,
  ActionFieldIds
} from 'constants/alerts/newActionFormSpec';
import { change } from 'redux-form';
import { FormTypes } from 'types/form';
import { ActionTypes } from 'types/alerts';
import { getJson, postJson, putJson, deleteJson } from 'actions/util';
import { ACTIONS_INDEX_PATHNAME } from 'constants/alerts';
import {
  MILLISECONDS_ONE_MINUTE,
  MILLISECONDS_ONE_HOUR,
  MILLISECONDS_FOUR_HOURS,
  MILLISECONDS_EIGHT_HOURS,
  MILLISECONDS_ONE_DAY
} from 'utils/time';
import type { RateUnitType } from 'constants/alerts/newActionFormSpec';


export function* fetchActionsSaga(): Generator<*, *, *> {
  const actionsResponse: any = yield call(getJson, '/2.0/alerts/actions');
  const actions = actionsResponse.map(action => ({
    guid: action.guid,
    name: action.name,
    triggers: action.triggerGuid,
    triggerCount: action.triggerGuid.length,
    maxFrequencyPerDay: (action.maxSendRateNumerator * (24*60*60*1000) / action.intervalMs),
    status: action.enabled ? EnabledDisabled.ENABLED : EnabledDisabled.DISABLED
  }));
  yield put(alertsActions.receive(actions));
}

function calculateIntervalRateUnit(intervalMs: number, maxSendRateNumerator: number): { maxSendRateNumerator: number, rateUnits: RateUnitType } {
  switch (intervalMs) {
    case MILLISECONDS_ONE_MINUTE:
      return { maxSendRateNumerator, rateUnits: RateUnits.PER_MINUTE };
    case MILLISECONDS_ONE_HOUR:
      return { maxSendRateNumerator, rateUnits: RateUnits.PER_HOUR };
    case MILLISECONDS_FOUR_HOURS:
      return { maxSendRateNumerator, rateUnits: RateUnits.PER_4_HOURS };
    case MILLISECONDS_EIGHT_HOURS:
      return { maxSendRateNumerator, rateUnits: RateUnits.PER_8_HOURS };
    case MILLISECONDS_ONE_DAY:
      return { maxSendRateNumerator, rateUnits: RateUnits.PER_DAY };
    default:
      // should never occur, but make sane anyway.
      return {
        maxSendRateNumerator: maxSendRateNumerator * MILLISECONDS_ONE_DAY / intervalMs,
        rateUnits: RateUnits.PER_DAY
      };
  }
}

export function* fetchActionSaga(action: AlertsRequestActionAction): Generator<*, *, *> {
  const actionInfo: any = yield call(getJson, `/2.0/alerts/actions/${action.payload.id}`);

  const { rateUnits, maxSendRateNumerator } = calculateIntervalRateUnit(parseInt(actionInfo.intervalMs, 10), actionInfo.maxSendRateNumerator);

  yield put(change(FormTypes.ALERTS_ACTION_FORM, ActionFieldIds.ACTION_NAME, actionInfo.name));
  yield put(change(FormTypes.ALERTS_ACTION_FORM, ActionFieldIds.ENABLED, actionInfo.enabled));
  yield put(change(FormTypes.ALERTS_ACTION_FORM, ActionFieldIds.TRIGGERS, actionInfo.triggerGuid));
  yield put(change(FormTypes.ALERTS_ACTION_FORM, ActionFieldIds.ACTION_TYPE, ActionTypes.SEND_EMAIL));
  yield put(change(FormTypes.ALERTS_ACTION_FORM, ActionFieldIds.EMAIL, actionInfo.email.address));
  yield put(change(FormTypes.ALERTS_ACTION_FORM, ActionFieldIds.SUBJECT, actionInfo.email.subject));
  yield put(change(FormTypes.ALERTS_ACTION_FORM, ActionFieldIds.MAX_SEND_RATE, maxSendRateNumerator));
  yield put(change(FormTypes.ALERTS_ACTION_FORM, ActionFieldIds.RATE_UNITS, rateUnits));
}

export const RateUnitIntervals = {
  [RateUnits.PER_MINUTE]: MILLISECONDS_ONE_MINUTE,
  [RateUnits.PER_HOUR]: MILLISECONDS_ONE_HOUR,
  [RateUnits.PER_4_HOURS]: MILLISECONDS_FOUR_HOURS,
  [RateUnits.PER_8_HOURS]: MILLISECONDS_EIGHT_HOURS,
  [RateUnits.PER_DAY]: MILLISECONDS_ONE_DAY
};

function* getActionValues(): Generator<*, *, *> {
  const formDataValues: any = yield select(state => state.form[FormTypes.ALERTS_ACTION_FORM].values);
  const action = {
    name: formDataValues[ActionFieldIds.ACTION_NAME],
    enabled: formDataValues[ActionFieldIds.ENABLED],
    triggerGuid: formDataValues[ActionFieldIds.TRIGGERS],
    maxSendRateNumerator: formDataValues[ActionFieldIds.MAX_SEND_RATE],
    intervalMs: RateUnitIntervals[formDataValues[ActionFieldIds.RATE_UNITS]],
    email: {
      subject: formDataValues[ActionFieldIds.SUBJECT],
      address: formDataValues[ActionFieldIds.EMAIL]
    }
  };
  return action;
}

export function* createActionSaga(): Generator<*, *, *> {
  yield* createActionFromStateSaga();
  yield call(browserHistory.push, ACTIONS_INDEX_PATHNAME);
}

export function* createActionFromStateSaga(): Generator<*, *, *> {
  const values: any = yield* getActionValues();
  yield call(postJson, '/2.0/alerts/actions', values);
}

export function* updateActionSaga(action: AlertsUpdateActionAction): Generator<*, *, *> {
  yield* updateActionFromStateSaga(alertsUpdateActionFromState(action.payload.guid));
  yield call(browserHistory.push, ACTIONS_INDEX_PATHNAME);
}

export function* updateActionFromStateSaga(action: AlertsUpdateActionFromStateAction): Generator<*, *, *> {
  const values: any = yield* getActionValues();
  values.guid = action.payload.guid;
  yield call(putJson, `/2.0/alerts/actions/${action.payload.guid}`, values);
}

export function* deleteActionSaga(action: AlertsDeleteActionAction): Generator<*, *, *> {
  yield call(deleteJson, `/2.0/alerts/actions/${action.payload.guid}`);
  yield put(alertsActions.request());
  yield call(browserHistory.push, ACTIONS_INDEX_PATHNAME);
}

export function* actionsSaga(): Generator<*, *, *> {
  yield fork(function* () { yield* takeEvery(ALERTS_REQUEST_ACTIONS, fetchActionsSaga); });
  yield fork(function* () { yield* takeEvery(ALERTS_REQUEST_ACTION, fetchActionSaga); });
  yield fork(function* () { yield* takeLatest(ALERTS_CREATE_ACTION_FROM_STATE, createActionFromStateSaga); });
  yield fork(function* () { yield* takeLatest(ALERTS_UPDATE_ACTION_FROM_STATE, updateActionFromStateSaga); });
  yield fork(function* () { yield* takeLatest(ALERTS_DELETE_ACTION, deleteActionSaga); });
  yield fork(function* () { yield* takeLatest(ALERTS_CREATE_ACTION, createActionSaga); });
  yield fork(function* () { yield* takeLatest(ALERTS_UPDATE_ACTION, updateActionSaga); });
}

export default actionsSaga;
