// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import debug from 'debug';
import { call, put } from 'redux-saga/effects';
import { getJson } from 'actions/util';

import { setFormOptions } from 'actions/form';
import { alertsClusterIdConsumerGroupOptionsLoaded } from 'actions/alerts';

import type { FormChangeAction } from 'types/form';
import { FormTypes } from 'types/form';

import {
  AlertSources,
  SelectOptionListIds,
  TriggerFieldIds
} from 'types/alerts';

import { constructMonitoringDataRequest } from 'actions/monitoring/data';
import {
  RequestTypes,
  DataLayoutTypes
} from 'types/monitoring/data';
import { Rollups } from 'types/monitoring';
import { constructClusterPath } from 'sagas/clusters';

import type { ClusterId, ClustersResponse } from 'types/clusters';


const log = debug('controlcenter:sagas:alerts:newTrigger');

export function constructClusterIdConsumerGroupKey(clusterId: ClusterId, consumerGroup: string): string {
  return JSON.stringify({ clusterId: clusterId, consumerGroup: consumerGroup });
}

export function parseClusterIdConsumerGroupKey(key: string): Object {
  return JSON.parse(key);
}

export function* loadMonitoringClusterIdConsumerGroupsOptions(): Generator<*, *, *> {
  const clusterReqPath = constructClusterPath('MONITORING');

  try {
    const response: ClustersResponse | any = yield call(getJson, clusterReqPath);
    const { clusters } = response;

    const consumerGroupReqPath = (clusterId) => {
      const params = {
        startTimeMs: Date.now()-24*60*60*1000,
        stopTimeMs: Date.now(),
        rollup: Rollups.FIFTEEN_SECONDS,
        type: RequestTypes.MEMBER_LIST
      };
      return constructMonitoringDataRequest(clusterId, DataLayoutTypes.CONSUMER_GROUPS, params);
    };

    const consumerGroups = [];
    try {
      if (clusters == null || clusters.length === 0) {
        consumerGroups.push({
          clusterId: null,
          consumerGroups: yield call(getJson, consumerGroupReqPath(null))
        });
      } else {
        for (let i=0; i<clusters.length; ++i) {
          consumerGroups.push({
            clusterId: clusters[i],
            consumerGroups: yield call(getJson, consumerGroupReqPath(clusters[i].clusterId))
          });
        }
      }
    } catch (error) {
      log(`error looking consumer groups. status=${error.message}`);
    }

    const options = Array.prototype.concat.apply([],
      consumerGroups.map(cluster => {
        const clusterId = cluster.clusterId == null ? null : cluster.clusterId.clusterId;
        const displayName = cluster.clusterId == null ? null : cluster.clusterId.displayName;
        // $FlowFixMe - TODO(mhowlett): need better flow annotations on metrics data, but format changes are planned with new REST API.
        return cluster.consumerGroups.members.map(cg => ({
          label: `${cg.group} [${displayName}]`,
          value: clusterId === null || cg.group === null
            ? null
            : constructClusterIdConsumerGroupKey(clusterId, cg.group)
        }));
      })
    );

    return options;

  } catch (error) {
    log(`error looking up path=${clusterReqPath} status=${error.message}`);
  }
}

export function* alertsNewTriggerFormSaga(action: FormChangeAction): Generator<*, *, *> {
  const { meta: { field }, payload: value } = action;

  if (field !== TriggerFieldIds.ALERT_SOURCE) {
    return;
  }

  if (value === AlertSources.KAFKA_CONNECTOR) {
    log('unsupported alert source: CONNECTOR');
  } else if (value === AlertSources.CONSUMER_GROUP) {
    const options: any = yield loadMonitoringClusterIdConsumerGroupsOptions();
    yield put(setFormOptions(FormTypes.ALERTS_TRIGGER_FORM, SelectOptionListIds.CONSUMER_GROUPS, options));
    yield put(alertsClusterIdConsumerGroupOptionsLoaded());
  }
}

export default alertsNewTriggerFormSaga;
