// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { browserHistory } from 'react-router';
import { call, put, fork, select, take } from 'redux-saga/effects';
import { takeEvery, takeLatest } from 'redux-saga';
import {
  ALERTS_REQUEST_TRIGGERS,
  ALERTS_REQUEST_TRIGGER,
  ALERTS_PREFILL_NEW_TRIGGER,
  ALERTS_CREATE_TRIGGER_FROM_STATE,
  ALERTS_UPDATE_TRIGGER_FROM_STATE,
  ALERTS_UPDATE_TRIGGER,
  ALERTS_TRIGGER_ADD_ACTIONS_FROM_STATE,
  ALERTS_CLUSTERID_CONSUMER_GROUP_OPTIONS_LOADED,
  ALERTS_DELETE_TRIGGER,
  ALERTS_CREATE_FIRST_ACTION,
  alertsTriggers,
  alertsTriggerLastCreated,
  TriggerStateUpdateMethods,
  alertsUpdateTriggerFromState
} from 'actions/alerts';
import type {
  AlertsRequestTriggerAction,
  AlertsPrefillNewTriggerAction,
  AlertsUpdateTriggerFromStateAction,
  AlertsUpdateTriggerAction,
  AlertsTriggerAddActionsFromStateAction,
  AlertsDeleteTriggerAction
} from 'actions/alerts';
import {
  NEW_TRIGGER_PATHNAME,
  TRIGGERS_INDEX_PATHNAME,
  NEW_ACTION_PATHNAME
} from 'constants/alerts';
import { change } from 'redux-form';
import { FormTypes } from 'types/form';
import { MonitoringDataTypes } from 'types/monitoring';
import {
  MetricTypes,
  AlertSources,
  TriggerFieldIds
} from 'types/alerts';
import { selectItem } from 'components/withSelectionList/selectionListActions';
import { LIST_NAME_ACTIONS } from 'components/alerts/AddActions';
import { constructClusterIdConsumerGroupKey } from './newTrigger';
import { postJson, putJson, getJson, deleteJson } from 'actions/util';
import { fetchActionsSaga } from './actions';
import { selectors as selectionListSelectors } from 'components/withSelectionList';


export function* fetchTriggersSaga(): Generator<*, *, *> {

  const triggers: any = yield call(getJson, '/2.0/alerts/triggers');

  yield* fetchActionsSaga();
  const actions: any = yield select(state => state.alerts.actions);

  // TODO(mhowlett): make the below functional.

  const triggerCounts = {};
  for (let i=0; i<actions.length; ++i) {
    const actionTriggers = actions[i].triggers;
    for (let j=0; j<actionTriggers.length; ++j) {
      const guid = actionTriggers[j];
      if (triggerCounts[guid]) {
        triggerCounts[guid] = triggerCounts[guid] + 1;
      } else {
        triggerCounts[guid] = 1;
      }
    }
  }

  for (let i=0; i<triggers.length; ++i) {
    const trigger = triggers[i].guid;
    if (trigger in triggerCounts) {
      triggers[i].actionCount = triggerCounts[trigger];
    } else {
      triggers[i].actionCount = 0;
    }
  }

  yield put(alertsTriggers.receive(triggers));
}

export function* fetchTriggerSaga(action: AlertsRequestTriggerAction): Generator<*, *, *> {
  const reqPath = `/2.0/alerts/triggers/${action.payload.id}`;
  const triggerInfo: any = yield call(getJson, reqPath);
  // TODO(mhowlett): support other sources.
  triggerInfo.alertSource = AlertSources.CONSUMER_GROUP;

  if (action.payload.stateUpdateMethod === TriggerStateUpdateMethods.SELECTABLE_LIST) {

    yield* fetchActionsSaga();
    const actions: any = yield select(state => state.alerts.actions);
    const triggerGuid = triggerInfo.guid;

    // note that most of the time there will be nothing selected below because most of the time this route will be navigated to immediately after create trigger.
    // TODO(mhowlett): make functional
    for (let i=0; i<actions.length; ++i) {
      for (let j=0; j<actions[i].triggers.length; ++j) {
        if (actions[i].triggers[j] === triggerGuid) {
          yield put(selectItem(LIST_NAME_ACTIONS, actions[i].guid));
        }
      }
    }

  } else if (action.payload.stateUpdateMethod === TriggerStateUpdateMethods.REDUX_FORM) {

    yield put(change(FormTypes.ALERTS_TRIGGER_FORM, TriggerFieldIds.TRIGGER_NAME, triggerInfo.name));
    yield put(change(FormTypes.ALERTS_TRIGGER_FORM, TriggerFieldIds.ALERT_SOURCE, triggerInfo.alertSource));
    if (triggerInfo.alertSource === AlertSources.CONSUMER_GROUP) {
      yield take(ALERTS_CLUSTERID_CONSUMER_GROUP_OPTIONS_LOADED);
      // It's possible that the clusterId / consumerGroup option is old, and no longer reported by the server.
      // In this case, we'll assume it's still valid and add it to the list.
      yield put(change(FormTypes.ALERTS_TRIGGER_FORM, TriggerFieldIds.CONSUMER_GROUP_TARGET, constructClusterIdConsumerGroupKey(triggerInfo.clusterId, triggerInfo.group)));
    } else if (triggerInfo.alertSource === AlertSources.KAFKA_CONNECTOR) {
      yield put(change(FormTypes.ALERTS_TRIGGER_FORM, TriggerFieldIds.CONNECTOR_TARGET, triggerInfo.connector));
    }
    yield put(change(FormTypes.ALERTS_TRIGGER_FORM, TriggerFieldIds.METRIC, triggerInfo.metric));
    yield put(change(FormTypes.ALERTS_TRIGGER_FORM, TriggerFieldIds.CONDITION, triggerInfo.condition));
    yield put(change(FormTypes.ALERTS_TRIGGER_FORM, TriggerFieldIds.VALUE, triggerInfo.longValue));
    yield put(change(FormTypes.ALERTS_TRIGGER_FORM, TriggerFieldIds.BUFFER, triggerInfo.lagMs / 1000));
  }
}

export function* prefillSaga(action: AlertsPrefillNewTriggerAction): Generator<*, *, *> {
  const { consumerGroupName, clusterId, type } = action.payload;
  if (consumerGroupName == null) {
    return;
  }
  yield call(browserHistory.push, NEW_TRIGGER_PATHNAME);
  yield put(change(FormTypes.ALERTS_TRIGGER_FORM, TriggerFieldIds.TRIGGER_NAME, `${consumerGroupName} ${type === MonitoringDataTypes.DELIVERY ? 'consumption' : 'latency'}`));
  yield put(change(FormTypes.ALERTS_TRIGGER_FORM, TriggerFieldIds.ALERT_SOURCE, AlertSources.CONSUMER_GROUP));
  yield put(change(FormTypes.ALERTS_TRIGGER_FORM, TriggerFieldIds.CONSUMER_GROUP_TARGET, constructClusterIdConsumerGroupKey(clusterId, consumerGroupName)));
  if (type === MonitoringDataTypes.DELIVERY) {
    yield put(change(FormTypes.ALERTS_TRIGGER_FORM, TriggerFieldIds.METRIC, MetricTypes.CONSUMPTION_DIFF));
  } else {
    yield put(change(FormTypes.ALERTS_TRIGGER_FORM, TriggerFieldIds.METRIC, MetricTypes.AVG_LATENCY));
  }
}

function* getTriggerValues(): Generator<*, *, *> {
  const formDataValues: any = yield select(state => state.form[FormTypes.ALERTS_TRIGGER_FORM].values);

  const alertSource = formDataValues[TriggerFieldIds.ALERT_SOURCE];
  if (alertSource !== AlertSources.CONSUMER_GROUP) {
    return;
  }

  const consumerGroupTarget = JSON.parse(formDataValues[TriggerFieldIds.CONSUMER_GROUP_TARGET]);

  const values = {
    name: formDataValues[TriggerFieldIds.TRIGGER_NAME],
    clusterId: consumerGroupTarget.clusterId,
    group: consumerGroupTarget.consumerGroup,
    metric: formDataValues[TriggerFieldIds.METRIC],
    condition: formDataValues[TriggerFieldIds.CONDITION],
    longValue: formDataValues[TriggerFieldIds.VALUE],
    lagMs: formDataValues[TriggerFieldIds.BUFFER] * 1000
  };

  return values;
}

export function* createTriggerFromStateSaga(): Generator<*, *, *> {
  const values: any = yield* getTriggerValues();
  const trigger: any = yield call(postJson, '/2.0/alerts/triggers', values);
  yield put(alertsTriggerLastCreated(trigger));
}

export function* updateTriggerSaga(action: AlertsUpdateTriggerAction): Generator<*, *, *> {
  yield* updateTriggerFromStateSaga(alertsUpdateTriggerFromState(action.payload.guid));
  yield call(browserHistory.push, TRIGGERS_INDEX_PATHNAME);
}

export function* updateTriggerFromStateSaga(action: AlertsUpdateTriggerFromStateAction): Generator<*, *, *> {
  const values: any = yield* getTriggerValues();
  values.guid = action.payload.guid;
  yield call(putJson, `/2.0/alerts/triggers/${action.payload.guid}`, values);
}

export function* addActionsSaga(action: AlertsTriggerAddActionsFromStateAction): Generator<*, *, *> {
  const selectedActionsSet: any = yield select(
    state => selectionListSelectors.selectedItemsSelector(LIST_NAME_ACTIONS, state)
  );
  const selectedActions = Object.keys(selectedActionsSet);

  for (let i=0; i<selectedActions.length; ++i) {
    // TODO(mhowlett): possible race condition!
    const alertAction: any = yield call(getJson, `/2.0/alerts/actions/${selectedActions[i]}`);
    alertAction.triggerGuid.push(action.payload.triggerGuid);
    yield call(putJson, `/2.0/alerts/actions/${selectedActions[i]}`, alertAction);
  }

  yield call(browserHistory.push, TRIGGERS_INDEX_PATHNAME);
}

export function* deleteTriggerSaga(action: AlertsDeleteTriggerAction): Generator<*, *, *> {
  yield call(deleteJson, `/2.0/alerts/triggers/${action.payload.guid}`);
  yield put(alertsTriggers.request());
  yield call(browserHistory.push, TRIGGERS_INDEX_PATHNAME);
}

export function* createFirstActionSaga(): Generator<*, *, *> {
  yield put(alertsTriggers.request());
  yield call(browserHistory.push, NEW_ACTION_PATHNAME);
}

export function* triggersSaga(): Generator<*, *, *> {
  yield fork(function* () { yield* takeEvery(ALERTS_REQUEST_TRIGGERS, fetchTriggersSaga); });
  yield fork(function* () { yield* takeEvery(ALERTS_REQUEST_TRIGGER, fetchTriggerSaga); });
  yield fork(function* () { yield* takeEvery(ALERTS_PREFILL_NEW_TRIGGER, prefillSaga); });
  yield fork(function* () { yield* takeLatest(ALERTS_CREATE_TRIGGER_FROM_STATE, createTriggerFromStateSaga); });
  yield fork(function* () { yield* takeLatest(ALERTS_UPDATE_TRIGGER_FROM_STATE, updateTriggerFromStateSaga); });
  yield fork(function* () { yield* takeLatest(ALERTS_TRIGGER_ADD_ACTIONS_FROM_STATE, addActionsSaga); });
  yield fork(function* () { yield* takeLatest(ALERTS_DELETE_TRIGGER, deleteTriggerSaga); });
  yield fork(function* () { yield* takeLatest(ALERTS_CREATE_FIRST_ACTION, createFirstActionSaga); });
  yield fork(function* () { yield* takeLatest(ALERTS_UPDATE_TRIGGER, updateTriggerSaga); });
}

export default triggersSaga;
