// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import debug from 'debug';
import { call, fork, put } from 'redux-saga/effects';
import { takeLatest } from 'redux-saga';

import {
  deleteJson,
  getJson,
  putJson
} from 'actions/util';

import * as ConnectorActions from 'actions/connectors';

import type {
  Connector,
  ConnectorList,
  ConnectorStatus,
  DeleteConnectorAction,
  LoadConnectorAction,
  LoadConnectorStatusAction,
  SaveConnectorAction
} from 'actions/connectors';

const log = debug('controlcenter:sagas:management:connectors');

export function* deleteConnector(action: DeleteConnectorAction): Generator<*, *, *> {
  const { name } = action.payload;

  if (!window.confirm(`Are you sure you want to delete ${name}?`)) {
    return;
  }

  const reqPath = `/2.0/management/connect/connectors/${encodeURIComponent(name)}`;

  yield put(ConnectorActions.destroy.request(name));

  try {
    yield call(deleteJson, reqPath);
    yield put(ConnectorActions.destroy.success(name));
  } catch (error) {
    log(`error looking up path=${reqPath} status=${error.message}`);
    yield put(ConnectorActions.destroy.failure(name, error));
  }
}

export function* saveConnector(action: SaveConnectorAction): Generator<*, *, *> {
  const { config } = action.payload;
  const { name } = config;

  const reqPath = `/2.0/management/connect/connectors/${encodeURIComponent(name)}/config`;

  yield put(ConnectorActions.save.request(name));

  try {
    const connector: Connector | any = yield call(putJson, reqPath, config);
    yield put(ConnectorActions.save.success(name, connector));
  } catch (error) {
    log(`error looking up path=${reqPath} status=${error.message}`);
    yield put(ConnectorActions.save.failure(name, error));
  }
}

export function* fetchConnector(action: LoadConnectorAction): Generator<*, *, *> {
  const { name } = action.payload;
  const reqPath = `/2.0/management/connect/connectors/${encodeURIComponent(name)}`;

  yield put(ConnectorActions.fetch.request(name));

  try {
    const connector: Connector | any = yield call(getJson, reqPath);
    yield put(ConnectorActions.fetch.success(name, connector));
  } catch (error) {
    log(`error looking up path=${reqPath} status=${error.message}`);
    yield put(ConnectorActions.fetch.failure(name, error));
  }
}

export function* fetchConnectorList(): Generator<*, *, *> {
  const reqPath = '/2.0/management/connect/connectors';

  yield put(ConnectorActions.listFetch.request());

  try {
    const connectorList: ConnectorList | any = yield call(getJson, reqPath);
    yield put(ConnectorActions.listFetch.success(connectorList));
  } catch (error) {
    log(`error looking up path=${reqPath} status=${error.message}`);
    yield put(ConnectorActions.listFetch.failure(error));
  }
}

export function* pauseConnector(action: LoadConnectorAction): Generator<*, *, *> {
  const { name } = action.payload;
  const reqPath = `/2.0/management/connect/connectors/${encodeURIComponent(name)}/pause`;

  yield put(ConnectorActions.pause.request(name));

  try {
    const response: any = yield call(putJson, reqPath);
    yield put(ConnectorActions.pause.success(name, response));
  } catch (error) {
    log(`error looking up path=${reqPath} status=${error.message}`);
    yield put(ConnectorActions.pause.failure(name, error));
  }
}

export function* resumeConnector(action: LoadConnectorAction): Generator<*, *, *> {
  const { name } = action.payload;
  const reqPath = `/2.0/management/connect/connectors/${encodeURIComponent(name)}/resume`;

  yield put(ConnectorActions.resume.request(name));

  try {
    const response: any = yield call(putJson, reqPath);
    yield put(ConnectorActions.resume.success(name, response));
  } catch (error) {
    log(`error looking up path=${reqPath} status=${error.message}`);
    yield put(ConnectorActions.resume.failure(name, error));
  }
}

export function* loadConnectorStatus(action: LoadConnectorStatusAction): Generator<*, *, *> {
  const { name } = action.payload;
  const reqPath = `/2.0/management/connect/connectors/${encodeURIComponent(name)}/status`;

  yield put(ConnectorActions.status.request(name));

  try {
    const connectorStatus: ConnectorStatus | any = yield call(getJson, reqPath);
    yield put(ConnectorActions.status.success(name, connectorStatus));
  } catch (error) {
    log(`error looking up path=${reqPath} status=${error.message}`);
    yield put(ConnectorActions.status.failure(name, error));
  }
}

export function* deleteConnectorSaga(): Generator<*, *, *> {
  yield call(takeLatest, ConnectorActions.DELETE_CONNECTOR, deleteConnector);
}

export function* saveConnectorSaga(): Generator<*, *, *> {
  yield call(takeLatest, ConnectorActions.SAVE_CONNECTOR, saveConnector);
}

export function* loadConnectorSaga(): Generator<*, *, *> {
  yield call(takeLatest, ConnectorActions.LOAD_CONNECTOR, fetchConnector);
}

export function* loadConnectorListSaga(): Generator<*, *, *> {
  yield call(takeLatest, [
    ConnectorActions.CONNECTOR_DESTROY_SUCCESS,
    ConnectorActions.LOAD_CONNECTOR_LIST
  ], fetchConnectorList);
}

export function* loadConnectorStatusSaga(): Generator<*, *, *> {
  yield call(takeLatest, ConnectorActions.LOAD_CONNECTOR_STATUS, loadConnectorStatus);
}

export function* pauseConnectorSaga(): Generator<*, *, *> {
  yield call(takeLatest, ConnectorActions.PAUSE_CONNECTOR, pauseConnector);
}

export function* resumeConnectorSaga(): Generator<*, *, *> {
  yield call(takeLatest, ConnectorActions.RESUME_CONNECTOR, resumeConnector);
}

export function* connectorsSaga(): Generator<*, *, *> {
  yield fork(deleteConnectorSaga);
  yield fork(loadConnectorListSaga);
  yield fork(loadConnectorSaga);
  yield fork(loadConnectorStatusSaga);
  yield fork(pauseConnectorSaga);
  yield fork(resumeConnectorSaga);
  yield fork(saveConnectorSaga);
}

export default connectorsSaga;
