// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.

import { DataLayoutTypes } from 'types/monitoring/data';

import type {
  DataLayoutType,
  TimeSeries
} from 'types/monitoring/data';

// TODO(mhowlett): should probably be user configurable.
export const MAX_IGNORE_DELIVERY_DISCREPANCY_MS = 60000;

// a heuristic to decide when discrepancies between produced and consumed data
// should be considered to be an error or not. The returned index is the last
// index for which a discrepancy in the data should be considered an error.
export function computeIgnoreDeliveryDiscrepancyAfterIndex(
  produced: ?TimeSeries,
  consumed: ?TimeSeries,
  lastDate: ?Date
): number {
  if (
    !produced ||
    !produced.values ||
    !consumed ||
    !consumed.values ||
    produced.values.length === 0 ||
    consumed.values.length === 0
  ) {
    return Infinity;
  }

  // Math.ceil ensures consistency between calls where the same data may have been binned differently
  const maxAllowedDiffIdx = Math.ceil(MAX_IGNORE_DELIVERY_DISCREPANCY_MS / produced.stepMs);

  const lastIdx = lastDate == null
    ? produced.values.length - 1
    : Math.floor((lastDate - produced.startTimeMs) / produced.stepMs);
  if (lastIdx < 0) {
    return 0;
  }
  let idx = lastIdx;

  while (
    idx > lastIdx-maxAllowedDiffIdx &&
    (idx > produced.values.length - 1 || produced.values[idx] !== consumed.values[idx] || produced.values[idx] === null || consumed.values[idx] === null)
  ) {
    idx -= 1;
  }

  return Math.min(idx, produced.values.length - 1);
}

export function idNameFromDataLayoutType(dataLayoutType: DataLayoutType): ?string {
  switch (dataLayoutType) {
    case DataLayoutTypes.CONSUMER_GROUP:
    case DataLayoutTypes.CONSUMER_GROUP_TOPICS:
      return 'group';
    case DataLayoutTypes.CONSUMER:
      return 'clientId';
    default:
      return null;
  }
}

export function memberIdNameFromDataLayoutType(dataLayoutType: DataLayoutType): Array<string> {
  switch (dataLayoutType) {
    case DataLayoutTypes.CONSUMER_GROUPS:
      return ['group'];
    case DataLayoutTypes.CONSUMER_GROUPS_TOPICS:
      return ['topic'];
    case DataLayoutTypes.CONSUMER_GROUP:
      return ['clientId'];
    case DataLayoutTypes.CONSUMER_GROUP_TOPICS:
      return ['topic', 'partition'];
    case DataLayoutTypes.CONSUMER:
      return ['topic', 'partition'];
    default:
      return [];
  }
}

export function timeSeriesSlice(data: TimeSeries, startIdx: number, endIdx: number): TimeSeries {
  return {
    startTimeMs: data.startTimeMs + startIdx * data.stepMs,
    stepMs: data.stepMs,
    values: data.values.slice(startIdx, endIdx)
  };
}
