// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import React from 'react';
import { storiesOf, action } from '@kadira/storybook';
import { FlexColumn, SortDirection } from 'react-virtualized';

import withSelectionList from 'components/withSelectionList';
import { BigDataTable, CellActivity, CellContextMenu } from 'components/table';
import twoThousandRows from './mock-data/list.json';

import { connect, Provider } from 'react-redux';
import { browserHistory } from 'react-router';
import createStore from 'store/createStore';

import type { CellRenderer } from 'types/flex-table';

const rowData = Array.prototype.concat.apply([], new Array(500).fill(twoThousandRows));

const store = createStore({}, browserHistory);

// There are probably better ways to estimate here
const defaultPlaceholderRenderer = (/* cellRendererData: CellRenderer */) => (
  <div
    style={{
      height: '1em',
      width: `${10 + Math.random() * 80}%`,
      background: 'currentColor'
    }}
  />
);

const defaultCellRenderer = ({ cellData }: CellRenderer) => cellData;

const fancyCellRenderer = (renderer = defaultCellRenderer, placeholderRenderer = defaultPlaceholderRenderer) => (cellRendererData: CellRenderer) => cellRendererData.isScrolling
  ? placeholderRenderer(cellRendererData)
  : renderer(cellRendererData);


const contextMenuRenderer = ({ rowData }: CellRenderer) => (
  <CellContextMenu
    menuItems={[
      {
        icon: 'settingsSm',
        label: 'Settings',
        onActivate: action(`goto /connectors/${rowData.name}/edit`)
      },
      (rowData.status === 'RUNNING' ? {
        icon: 'pauseSm',
        label: 'Pause',
        onActivate: action('dispatch pause')
      } : {
        icon: 'playSm',
        label: 'Resume',
        onActivate: action('dispach resume')
      }),
      {
        icon: 'trashSm',
        label: 'Delete',
        onActivate: action('dispatch delete')
      }
    ]}
  />
);

const Table = (props) => fakeConnectorTable(props);

const SelectableList = withSelectionList(Table, '_id', 'list1');

const SelectedIds = connect(state => state.selectionList)(props => (
  <pre
    style={{
      background: 'gray',
      width: 400,
      maxHeight: 400,
      padding: 20
    }}
  >
    {JSON.stringify({ selectionList: props }, null, 2)}
  </pre>
));

const DependentList = connect((state, ownProps) => ({
  rowData: ownProps.rowData.filter(row => row['_id'] in state.selectionList.list1)
}))(Table);

const fakeConnectorTable = (options = {}) => {
  const props = {
    TableComponent: BigDataTable,
    bodyEventActions: {},
    flexColumnProps: {},
    headerEventActions: {},
    rowData,
    showActivityColumn: true,
    showMenuColumn: true,
    sortProps: {},
    useColumnGroups: false,
    useScrollPlaceholders: false,
    ...options
  };

  const cellRendererOptions = {};

  if (props.useScrollPlaceholders) {
    cellRendererOptions.cellRenderer = fancyCellRenderer();
  }

  const { TableComponent } = props;

  return (
    <TableComponent
      height={400}
      rowData={props.rowData}
      headerTableProps={{
        ...props.sortProps,
        ...props.headerEventActions
      }}
      bodyTableProps={{
        ...props.sortProps,
        ...props.bodyEventActions
      }}
    >
      {Object.keys(props.flexColumnProps).length > 0 && (
        <FlexColumn
          columnGroup={props.useColumnGroups ? 'Connector' : undefined}
          rowHeader={true}
          width={50}
          {...props.flexColumnProps}
        />
      )}
      <FlexColumn
        columnGroup={props.useColumnGroups ? 'Connector' : undefined}
        rowHeader={true}
        label="Index"
        dataKey="index"
        cellRenderer={({ rowIndex }) => rowIndex}
        width={50}
      />
      {props.showActivityColumn && (
        <FlexColumn
          columnGroup={props.useColumnGroups ? 'Connector' : undefined}
          rowHeader={true}
          label="Status"
          dataKey="status"
          width={90}
          cellRenderer={props.useScrollPlaceholders ? fancyCellRenderer(CellActivity) : CellActivity}
        />
      )}
      <FlexColumn
        columnGroup={props.useColumnGroups ? 'Connector' : undefined}
        rowHeader={true}
        label="Name"
        dataKey="name"
        width={100}
        {...cellRendererOptions}
      />
      {props.showMenuColumn && (
        <FlexColumn
          columnGroup={props.useColumnGroups ? 'Connector' : undefined}
          disableSort={true}
          rowHeader={true}
          dataKey="nothing"
          width={80}
          cellRenderer={props.useScrollPlaceholders ? fancyCellRenderer(contextMenuRenderer) : contextMenuRenderer}
        />
      )}
      <FlexColumn
        columnGroup={props.useColumnGroups ? 'Cluster' : undefined}
        label="Type"
        dataKey="type"
        width={60}
        {...cellRendererOptions}
      />
      <FlexColumn
        columnGroup={props.useColumnGroups ? 'Cluster' : undefined}
        label="Name"
        dataKey="cluster"
        width={100}
        {...cellRendererOptions}
      />
      <FlexColumn
        columnGroup={props.useColumnGroups ? 'Tasks' : undefined}
        label="Active"
        dataKey="activeTasks"
        width={100}
        {...cellRendererOptions}
      />
      <FlexColumn
        columnGroup={props.useColumnGroups ? 'Some Cool Group Name' : undefined}
        label="From topics"
        dataKey="topics"
        width={300}
        {...cellRendererOptions}
      />
      <FlexColumn
        columnGroup={props.useColumnGroups ? 'Events' : undefined}
        label="Created"
        dataKey="createdAt"
        width={160}
        {...cellRendererOptions}
      />
      <FlexColumn
        columnGroup={props.useColumnGroups ? 'Events' : undefined}
        label="Last Updated"
        dataKey="lastUpdated"
        width={160}
        {...cellRendererOptions}
      />
    </TableComponent>
  );
};

const headerEventActions = {
  onHeaderClick: action('headerTableProps.onHeaderClick'),
  onRowClick: action('headerTableProps.onRowClick'),
  onRowMouseOut: action('headerTableProps.onRowMouseOut'),
  onRowMouseOver: action('headerTableProps.onRowMouseOver')
};

const bodyEventActions = {
  onHeaderClick: action('bodyTableProps.onHeaderClick'),
  onRowClick: action('bodyTableProps.onRowClick'),
  onRowMouseOut: action('bodyTableProps.onRowMouseOut'),
  onRowMouseOver: action('bodyTableProps.onRowMouseOver')
};

storiesOf('BigDataTable: connector list', module)
  .add(`${rowData.length} items`, () => fakeConnectorTable({
  }))
  .add(`${rowData.length} items, columnGruops`, () => fakeConnectorTable({
    useColumnGroups: true
  }))
  .add(`${rowData.length} items, with event actions`, () => fakeConnectorTable({
    bodyEventActions,
    headerEventActions
  }))
  .add(`${rowData.length} items, columnGruops, with event actions`, () => fakeConnectorTable({
    useColumnGroups: true,
    bodyEventActions,
    headerEventActions
  }))
  .add('50 items', () => fakeConnectorTable({
    rowData: rowData.slice(0, 49)
  }))
  .add('50 items, with event actions', () => fakeConnectorTable({
    bodyEventActions,
    headerEventActions,
    rowData: rowData.slice(0, 49)
  }))
  .add('50 items, no menu column', () => fakeConnectorTable({
    rowData: rowData.slice(0, 49),
    showMenuColumn: false
  }))
  .add('50 items, no activity column', () => fakeConnectorTable({
    rowData: rowData.slice(0, 49),
    showActivityColumn: false
  }))
  .add('50 items, no activity or menu column', () => fakeConnectorTable({
    rowData: rowData.slice(0, 49),
    showActivityColumn: false,
    showMenuColumn: false
  }))
  .add('50 items, sorting buttons enabled, column groups', () => fakeConnectorTable({
    useColumnGroups: true,
    rowData: rowData.slice(0, 49),
    sortProps: {
      sort: action('sort'),
      sortBy: 'index',
      sortDirection: SortDirection.DESC
    }
  }))
  .add('50 items, sorting buttons enabled', () => fakeConnectorTable({
    rowData: rowData.slice(0, 49),
    sortProps: {
      sort: action('sort'),
      sortBy: 'index',
      sortDirection: SortDirection.DESC
    }
  }))
  .add(`${rowData.length} items, scroll placeholders`, () => fakeConnectorTable({
    useScrollPlaceholders: true
  }))
  .add('50 items, scroll placeholders', () => fakeConnectorTable({
    rowData: rowData.slice(0, 49),
    useScrollPlaceholders: true
  }))
  .add('selectable (wrapped w/ withSelectionList)', () => (
    <Provider store={store}>
      <div>
        <SelectableList
          rowData={rowData.slice(0, 49)}
        />
        <SelectedIds />
      </div>
    </Provider>
  ))
  .add('multiple tables', () => (
    <Provider store={store}>
      <div style={{ margin: 20, padding: 20, background: '#333' }}>
        <h1>Table One</h1>
        {fakeConnectorTable({
          rowData: rowData.slice(0, 49)
        })}

        <h1>Table Two</h1>
        {fakeConnectorTable({
          rowData: rowData.slice(50, 99)
        })}

        <h1>Table Three *and* Four</h1>
        <p>Note: Four depends on Three</p>
        <div style={{ display: 'inline-block', width: '50%', paddingRight: 10 }}>
          <SelectableList
            rowData={rowData.slice(0, 49)}
          />
        </div>
        <div style={{ display: 'inline-block', width: '50%', paddingLeft: 10 }}>
          <DependentList
            useColumnGroups={true}
            rowData={rowData.slice(0, 49)}
          />
        </div>
      </div>
    </Provider>
  ))
;
