// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.

import XMLHttpRequestMock from '../../mock/XMLHttpRequestMock';
import lolex from 'lolex';
import configureMockStore from 'redux-mock-store';
import thunk from 'redux-thunk';

import {
  ALLOW_DATA_FETCH,
  DECREMENT_CONNECTION_COUNT,
  DISALLOW_DATA_FETCH,
  INCREMENT_CONNECTION_COUNT
} from 'actions/monitoring/dataManager';

import { loadData, RECEIVE_DATA } from 'actions/monitoring/data';

import { SET_PAGE_SCROLL_INFO } from 'actions/monitoring/layout';

import {
  DataLayoutTypes,
  RequestTypes
} from 'types/monitoring/data';

const middlewares = [thunk];

describe('loadData action', () => {
  const successBody = {
    important: 'a good tradesman always blames his tools'
  };

  const successBody2 = {
    unimportant: '42'
  };

  const successBody3 = {
    freeLunch: 'missing'
  };

  let clock;
  let store;
  beforeEach(() => {
    clock = lolex.install();

    const clusterId = String(null);

    XMLHttpRequestMock.init(
      [
        {
          path: `/2.0/monitoring/${clusterId}/consumer_groups?startTimeMs=10000&stopTimeMs=20000&rollup=FIFTEEN_SECONDS&type=MEMBER_LIST`,
          status: 200,
          body: successBody
        },
        {
          path: `/2.0/monitoring/${clusterId}/consumer_group?startTimeMs=10000&stopTimeMs=20000&rollup=FIFTEEN_SECONDS&type=OVERVIEW&group=testConsumerGroup`,
          status: 200,
          body: successBody2
        },
        {
          path: `/2.0/monitoring/${clusterId}/consumer_groups?startTimeMs=10000&stopTimeMs=20000&rollup=FIFTEEN_SECONDS&type=MEMBER&memberGroup=other-group`,
          status: 200,
          body: successBody3
        },
        {
          path: `/2.0/monitoring/${clusterId}/consumer_groups?startTimeMs=10000&stopTimeMs=20000&rollup=FIFTEEN_SECONDS&type=OVERVIEW&group=testConsumerGroup2`,
          status: 404
        },
        {
          path: `/2.0/monitoring/${clusterId}/consumer_groups?startTimeMs=10000&stopTimeMs=20000&rollup=FIFTEEN_SECONDS&type=OVERVIEW&group=testConsumerGroup3`,
          status: 402
        },
        {
          path: `/2.0/monitoring/${clusterId}/consumer_groups?startTimeMs=10000&stopTimeMs=20000&rollup=FIFTEEN_SECONDS&type=OVERVIEW&group=testConsumerGroup4`,
          status: 500
        }
      ]
    );
    const mockStore = configureMockStore(middlewares);
    store = mockStore({
      routing: {
        locationBeforeTransitions: {
          query: {
            clusterId
          }
        }
      },
      monitoring: { status: { dataFetchInProgress: false } }
    });
  });

  afterEach(() => {
    XMLHttpRequestMock.restore();
    clock.uninstall();
  });

  it('200 (MEMBER_LIST) http ok behavior', () => {
    const thunk = loadData(
      DataLayoutTypes.CONSUMER_GROUPS,
      null, 10000, 20000, 'FIFTEEN_SECONDS',
      RequestTypes.MEMBER_LIST,
      null, // no extra arguments
      true  // isFullPeriod - irrelevant.
    );
    const expectedActions = [
      { type: INCREMENT_CONNECTION_COUNT, payload: null },
      { type: RECEIVE_DATA, payload: { requestType: RequestTypes.MEMBER_LIST, isFullPeriod: true, data: successBody } }
    ];
    return store.dispatch(thunk)
      .then(() => {
        expect(store.getActions().length).to.eql(4);
        expect(store.getActions().slice(0, 2)).to.eql(expectedActions);
        expect(store.getActions()[2].type).to.eql(SET_PAGE_SCROLL_INFO);
        expect(store.getActions()[3].type).to.eql(DECREMENT_CONNECTION_COUNT);
      });
  });

  it('200 (OVERVIEW) !isFullPeriod, with id', () => {
    const thunk = loadData(
      DataLayoutTypes.CONSUMER_GROUP,
      'testConsumerGroup', 10000, 20000, 'FIFTEEN_SECONDS',
      RequestTypes.OVERVIEW,
      null,
      false
    );
    const expectedActions = [
      { type: INCREMENT_CONNECTION_COUNT, payload: null },
      { type: RECEIVE_DATA, payload: { requestType: RequestTypes.OVERVIEW, isFullPeriod: false, data: successBody2 } }
    ];
    return store.dispatch(thunk)
      .then(() => {
        expect(store.getActions().length).to.eql(4);
        expect(store.getActions().slice(0, 2)).to.eql(expectedActions);
        expect(store.getActions()[2].type).to.eql(SET_PAGE_SCROLL_INFO);
        expect(store.getActions()[3].type).to.eql(DECREMENT_CONNECTION_COUNT);
      });
  });

  it('200 (MEMBER) with extraArguments', () => {
    const thunk = loadData(
      DataLayoutTypes.CONSUMER_GROUPS,
      null, 10000, 20000, 'FIFTEEN_SECONDS',
      RequestTypes.MEMBER,
      { memberGroup: 'other-group' },
      false
    );
    const expectedActions = [
      { type: INCREMENT_CONNECTION_COUNT, payload: null },
      { type: RECEIVE_DATA, payload: { requestType: RequestTypes.MEMBER, isFullPeriod: false, data: successBody3 } }
    ];
    return store.dispatch(thunk)
      .then(() => {
        expect(store.getActions().length).to.eql(4);
        expect(store.getActions().slice(0, 2)).to.eql(expectedActions);
        expect(store.getActions()[2].type).to.eql(SET_PAGE_SCROLL_INFO);
        expect(store.getActions()[3].type).to.eql(DECREMENT_CONNECTION_COUNT);
      });
  });

  it('500 (internal server error behavior)', () => {
    // internal server error.
    const thunk = loadData(
      DataLayoutTypes.CONSUMER_GROUP,
      'testConsumerGroup4', 10000, 20000, 'FIFTEEN_SECONDS',
      RequestTypes.OVERVIEW,
      null,
      true
    );
    const expectedActions = [
      { type: INCREMENT_CONNECTION_COUNT, payload: null },
      { type: DISALLOW_DATA_FETCH, payload: null },
      { type: DECREMENT_CONNECTION_COUNT, payload: null }
    ];
    return store.dispatch(thunk)
      .then(() => {
        expect(store.getActions()).to.eql(expectedActions);
        clock.tick(5000);
        expect(store.getActions()[3]).to.eql({ type: ALLOW_DATA_FETCH, payload: null });
      });
  });

  it('404', () => {
    // 404 happens when data not found. We'll create a test for that via a different consumer group name.
    const thunk = loadData(
      DataLayoutTypes.CONSUMER_GROUP,
      'testConsumerGroup2', 10000, 20000, 'FIFTEEN_SECONDS',
      RequestTypes.OVERVIEW,
      null,
      true
    );
    const expectedActions = [
      { type: INCREMENT_CONNECTION_COUNT, payload: null },
      { type: DISALLOW_DATA_FETCH, payload: null },
      { type: DECREMENT_CONNECTION_COUNT, payload: null }
    ];
    return store.dispatch(thunk)
      .then(() => {
        expect(store.getActions()).to.eql(expectedActions);
        clock.tick(5000);
        expect(store.getActions()[3]).to.eql({ type: ALLOW_DATA_FETCH, payload: null });
      });
  });

  it('402', () => {
    // 402 happens when license expired
    const thunk = loadData(
      DataLayoutTypes.CONSUMER_GROUP,
      'testConsumerGroup3', 10000, 20000, 'FIFTEEN_SECONDS',
      RequestTypes.OVERVIEW,
      null,
      true
    );
    const expectedActions = [
      { type: INCREMENT_CONNECTION_COUNT, payload: null },
      { type: DISALLOW_DATA_FETCH, payload: null },
      { type: DECREMENT_CONNECTION_COUNT, payload: null }
    ];
    return store.dispatch(thunk)
      .then(() => {
        expect(store.getActions()).to.eql(expectedActions);
        clock.tick(5000);
        expect(store.getActions()[3]).to.eql({ type: ALLOW_DATA_FETCH, payload: null });
      });
  });

});

