// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.

import XMLHttpRequestMock from '../mock/XMLHttpRequestMock';

import {
  encodeQueryData,
  getJson,
  makeAction,
  postJson,
  putJson
} from 'actions/util';

const MockEndPoints = {
  SUCCESS: 'http://localhost:9021/random360/success-endpoint',
  NOT_FOUND: 'http://localhost:9021/random1080/does-not-exist'
};

describe('actions/utils', () => {
  describe('makeAction', () => {
    const someValue = 'some cool value';
    const action1 = makeAction('action1', { someValue });
    const action2 = makeAction('action2', someValue);
    const action3 = makeAction('action3');

    it('creates an action with a type', () => {
      expect(action1.type).to.equal('action1');
      expect(action2.type).to.equal('action2');
      expect(action3.type).to.equal('action3');
    });

    it('creates an action with a payload', () => {
      expect(action1.payload.someValue).to.equal(someValue);
      expect(action2.payload).to.equal(someValue);
      expect(action3.payload).to.equal(undefined);
    });
  });

  describe('getJson', () => {

    const successBody = {
      data: 'hodor'
    };

    beforeEach(() => {
      XMLHttpRequestMock.init(
        [
          {
            path: MockEndPoints.NOT_FOUND,
            status: 404
          },
          {
            path: MockEndPoints.SUCCESS,
            status: 200,
            body: successBody
          }
        ]
      );
    });
    afterEach(XMLHttpRequestMock.restore);

    it('resolves json from an endpoint that exists', done => {
      getJson(MockEndPoints.SUCCESS)
        .then(json => {
          expect(XMLHttpRequestMock.calls().unmatched).to.equal(0);
          expect(XMLHttpRequestMock.calls().matched).to.equal(1);
          expect(XMLHttpRequestMock.called(MockEndPoints.SUCCESS)).to.be.true;
          expect(XMLHttpRequestMock.lastMethod()).to.eql('GET');
          expect(json).to.eql(successBody);
          done();
        })
        .catch(() => {
          expect.fail();
        });
    });

    it('rejects when an endpoint doesn\'t exist', done => {
      getJson(MockEndPoints.NOT_FOUND)
        .then(() => {
          expect.fail();
        })
        .catch(req => {
          expect(XMLHttpRequestMock.calls().unmatched).to.equal(0);
          expect(XMLHttpRequestMock.calls().matched).to.equal(1);
          expect(XMLHttpRequestMock.called(MockEndPoints.NOT_FOUND)).to.be.true;
          expect(XMLHttpRequestMock.lastMethod()).to.eql('GET');
          expect(req.status).to.equal(404);
          done();
        });
    });
  });


  describe('putJson', () => {

    const requestJson = {
      name: 'willis'
    };

    const successBody = {
      data: 'willis updated'
    };

    beforeEach(() => {
      XMLHttpRequestMock.init(
        [
          {
            path: MockEndPoints.NOT_FOUND,
            status: 404
          },
          {
            path: MockEndPoints.SUCCESS,
            status: 200,
            body: successBody
          }
        ]
      );
    });
    afterEach(XMLHttpRequestMock.restore);

    it('PUTs json to an endpoint that exists', done => {
      putJson(MockEndPoints.SUCCESS, requestJson)
        .then(json => {
          expect(XMLHttpRequestMock.calls().unmatched).to.equal(0);
          expect(XMLHttpRequestMock.calls().matched).to.equal(1);
          expect(XMLHttpRequestMock.called(MockEndPoints.SUCCESS)).to.be.true;
          expect(XMLHttpRequestMock.lastMethod()).to.eql('PUT');
          expect(json).to.eql(successBody);
          done();
        })
        .catch(() => {
          expect.fail();
        });
    });

    it('rejects when an endpoint doesn\'t exist', done => {
      putJson(MockEndPoints.NOT_FOUND, requestJson)
        .then(() => {
          expect.fail();
        })
        .catch(req => {
          expect(XMLHttpRequestMock.calls().unmatched).to.equal(0);
          expect(XMLHttpRequestMock.calls().matched).to.equal(1);
          expect(XMLHttpRequestMock.called(MockEndPoints.NOT_FOUND)).to.be.true;
          expect(XMLHttpRequestMock.lastMethod()).to.eql('PUT');
          expect(req.status).to.equal(404);
          done();
        });
    });
  });

  describe('postJson', () => {

    const requestJson = {
      name: 'hodor'
    };

    const successBody = {
      data: 'hodor created'
    };

    beforeEach(() => {
      XMLHttpRequestMock.init(
        [
          {
            path: MockEndPoints.NOT_FOUND,
            status: 404
          },
          {
            path: MockEndPoints.SUCCESS,
            status: 201,
            body: successBody
          }
        ]
      );
    });
    afterEach(XMLHttpRequestMock.restore);

    it('PUTs json to an endpoint that exists', done => {
      postJson(MockEndPoints.SUCCESS, requestJson)
        .then(json => {
          expect(XMLHttpRequestMock.calls().unmatched).to.equal(0);
          expect(XMLHttpRequestMock.calls().matched).to.equal(1);
          expect(XMLHttpRequestMock.called(MockEndPoints.SUCCESS)).to.be.true;
          expect(XMLHttpRequestMock.lastMethod()).to.eql('POST');
          expect(json).to.eql(successBody);
          done();
        })
        .catch(() => {
          expect.fail();
        });
    });

    it('rejects when an endpoint doesn\'t exist', done => {
      postJson(MockEndPoints.NOT_FOUND, requestJson)
        .then(() => {
          expect.fail();
        })
        .catch(req => {
          expect(XMLHttpRequestMock.calls().unmatched).to.equal(0);
          expect(XMLHttpRequestMock.calls().matched).to.equal(1);
          expect(XMLHttpRequestMock.called(MockEndPoints.NOT_FOUND)).to.be.true;
          expect(XMLHttpRequestMock.lastMethod()).to.eql('POST');
          expect(req.status).to.equal(404);
          done();
        });
    });
  });

  describe('encodeQueryData', () => {
    it('converts an object to query params formatted string', () => {
      expect(encodeQueryData({
        'this-is-amazing': 'truly'
      })).to.equal('this-is-amazing=truly');
    });

    it('converts objects with multiple properties to query param formatted string', () => {
      expect(encodeQueryData({
        'this-is-amazing': 'truly',
        'this-is-*SO*-dope': 'yah'
      })).to.equal('this-is-amazing=truly&this-is-*SO*-dope=yah');
    });

    it('works with non-standard characters', () => {
      expect(encodeQueryData({
        'damn&': 'daniel@example.com'
      })).to.equal('damn%26=daniel%40example.com');

      expect(encodeQueryData({
        '-中': '文'
      })).to.equal('-%E4%B8%AD=%E6%96%87');
    });
  });
});
