// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { expect } from 'chai';

import {
  calcHoverInfo,
  latencyFormatter,
  produceConsumeFormatter
} from 'components/charting/charts/hoverUtil';

describe('hover util', () => {
  const basicTimeScale = (x) => x instanceof Date ? x.getTime() : x;
  describe('calcHoverInfo', () => {
    it('handles largish times', () => {
      const oneHundredthScale = (x) => basicTimeScale(x) / 69;
      const hoverLabelInfo = calcHoverInfo(new Date(1462558332295.9487), 1462556416382, 5000, 400, oneHundredthScale);
      expect(hoverLabelInfo).to.not.be.null;
      if (!hoverLabelInfo) return;
      expect(hoverLabelInfo.ptX).to.eql(21196497572.463768);
      expect(hoverLabelInfo.idx).to.eql(383);
    });

    it('handles data at leading edge', () => {
      const hoverLabelInfo = calcHoverInfo(new Date(0), 0, 1, 100, basicTimeScale);
      expect(hoverLabelInfo).to.not.be.null;
      if (!hoverLabelInfo) return;
      expect(hoverLabelInfo.ptX).to.eql(0.5);
      expect(hoverLabelInfo.idx).to.eql(0);
    });

    it('handles data at trailing edge', () => {
      const hoverLabelInfo = calcHoverInfo(new Date(100), 0, 1, 100, basicTimeScale);
      expect(hoverLabelInfo).to.not.be.null;
      if (!hoverLabelInfo) return;
      expect(hoverLabelInfo.ptX).to.eql(100.5);
      expect(hoverLabelInfo.idx).to.eql(100);
    });

    it('handles data in interior', () => {
      const hoverLabelInfo = calcHoverInfo(new Date(42), 0, 1, 100, basicTimeScale);
      expect(hoverLabelInfo).to.not.be.null;
      if (!hoverLabelInfo) return;
      expect(hoverLabelInfo.ptX).to.eql(42.5);
      expect(hoverLabelInfo.idx).to.eql(42);
    });

    describe('handles bad input', () => {
      it('mouseOverTime too large returns null', () => {
        const hoverInfo = calcHoverInfo(new Date(101), 0, 1, 100, basicTimeScale);
        expect(hoverInfo).to.be.null;
      });

      it('mouseOverTime too small returns null', () => {
        const hoverInfo = calcHoverInfo(new Date(0), 1, 1, 100, basicTimeScale);
        expect(hoverInfo).to.be.null;
      });
    });

    describe('half time scale', () => {
      const halfTimeScale = (x) => basicTimeScale(x) / 2;
      it('maps even indices', () => {
        const hoverLabelInfo = calcHoverInfo(new Date(42), 0, 1, 100, halfTimeScale);
        expect(hoverLabelInfo).to.not.be.null;
        if (!hoverLabelInfo) return;
        expect(hoverLabelInfo.ptX).to.eql(21.25);
        expect(hoverLabelInfo.idx).to.eql(42);
      });

      it('maps odd indices', () => {
        const hoverLabelInfo = calcHoverInfo(new Date(69), 0, 1, 100, halfTimeScale);
        expect(hoverLabelInfo).to.not.be.null;
        if (!hoverLabelInfo) return;
        expect(hoverLabelInfo.ptX).to.eql(34.75);
        expect(hoverLabelInfo.idx).to.eql(69);
      });
    });

    describe('evenly divisible step sizing', () => {
      it('handles data at leading edge', () => {
        const hoverLabelInfo = calcHoverInfo(new Date(0), 0, 5, 100, basicTimeScale);
        expect(hoverLabelInfo).to.not.be.null;
        if (!hoverLabelInfo) return;
        expect(hoverLabelInfo.ptX).to.eql(2.5);
        expect(hoverLabelInfo.idx).to.eql(0);
      });

      it('handles data at trailing edge', () => {
        const hoverLabelInfo = calcHoverInfo(new Date(100), 0, 5, 100, basicTimeScale);
        expect(hoverLabelInfo).to.not.be.null;
        if (!hoverLabelInfo) return;
        expect(hoverLabelInfo.ptX).to.eql(102.5);
        expect(hoverLabelInfo.idx).to.eql(20);
      });

      it('handles data in interior', () => {
        const hoverLabelInfo = calcHoverInfo(new Date(42), 0, 5, 100, basicTimeScale);
        expect(hoverLabelInfo).to.not.be.null;
        if (!hoverLabelInfo) return;
        expect(hoverLabelInfo.ptX).to.eql(42.5);
        expect(hoverLabelInfo.idx).to.eql(8);
      });
    });

    describe('not evenly divisible step sizing', () => {
      it('handles data at leading edge', () => {
        const hoverLabelInfo = calcHoverInfo(new Date(0), 0, 17, 100, basicTimeScale);
        expect(hoverLabelInfo).to.not.be.null;
        if (!hoverLabelInfo) return;
        expect(hoverLabelInfo.ptX).to.eql(8.5);
        expect(hoverLabelInfo.idx).to.eql(0);
      });

      it('handles data at trailing edge', () => {
        const hoverLabelInfo = calcHoverInfo(new Date(100), 0, 17, 100, basicTimeScale);
        expect(hoverLabelInfo).to.not.be.null;
        if (!hoverLabelInfo) return;
        expect(hoverLabelInfo.ptX).to.eql(93.5);
        expect(hoverLabelInfo.idx).to.eql(5);
      });

      it('handles data in interior', () => {
        const hoverLabelInfo = calcHoverInfo(new Date(42), 0, 17, 100, basicTimeScale);
        expect(hoverLabelInfo).to.not.be.null;
        if (!hoverLabelInfo) return;
        expect(hoverLabelInfo.ptX).to.eql(42.5);
        expect(hoverLabelInfo.idx).to.eql(2);
      });
    });
  });

  describe('latencyFormatter', () => {
    const emptyTimeSeries = {
      startTimeMs: 0,
      stepMs: 1,
      values: []
    };
    describe('handles bad input', () => {
      const mouseOverTime = new Date(0);
      const emptyMultiTimeSeries = {
        startTimeMs: 0,
        stepMs: 1,
        values: {
          averageLatency: [],
          minLatency: [],
          maxLatency: []
        }
      };
      it('returns null with no multiSlotData', () => {
        const hoverLabelInfo = latencyFormatter(emptyTimeSeries, emptyTimeSeries, null, mouseOverTime, basicTimeScale, basicTimeScale);
        expect(hoverLabelInfo).to.be.null;
      });

      it('returns null with falsy stepMs on multiSlotData', () => {
        const falsyMultiTimeSeries = {
          startTimeMs: 0,
          stepMs: 0,
          values: {
            averageLatency: [],
            minLatency: [],
            maxLatency: []
          }
        };
        const hoverLabelInfo = latencyFormatter(emptyTimeSeries, emptyTimeSeries, falsyMultiTimeSeries, mouseOverTime, basicTimeScale, basicTimeScale);
        expect(hoverLabelInfo).to.be.null;
      });

      it('returns null with empty values multiSlotData', () => {
        const hoverLabelInfo = latencyFormatter(emptyTimeSeries, emptyTimeSeries, emptyMultiTimeSeries, mouseOverTime, basicTimeScale, basicTimeScale);
        expect(hoverLabelInfo).to.be.null;
      });

      it('returns null with mouseOverTime out of range', () => {
        const mouseOverTime = new Date(3943);
        const multiTimeSeriesData = {
          startTimeMs: 0,
          stepMs: 1,
          values: {
            averageLatency: [4, 5, 6, 2],
            minLatency: [1, 0, 4, 2],
            maxLatency: [7, 9, 100, 2]
          }
        };
        const hoverLabelInfo = latencyFormatter(emptyTimeSeries, emptyTimeSeries, multiTimeSeriesData, mouseOverTime, basicTimeScale, basicTimeScale);
        expect(hoverLabelInfo).to.be.null;
      });


    });
    it('extracts average latency info', () => {
      const mouseOverTime = new Date(2);
      const multiTimeSeriesData = {
        startTimeMs: 0,
        stepMs: 1,
        values: {
          averageLatency: [4, 5, 6, 2],
          minLatency: [1, 0, 4, 2],
          maxLatency: [7, 9, 100, 2]
        }
      };
      const hoverLabelInfo = latencyFormatter(emptyTimeSeries, emptyTimeSeries, multiTimeSeriesData, mouseOverTime, basicTimeScale, basicTimeScale);
      expect(hoverLabelInfo).to.not.be.null;
      if (!hoverLabelInfo) return;
      expect(hoverLabelInfo.ptX).to.eql(2.5);
      expect(hoverLabelInfo.ptY).to.eql(6);
    });
  });

  describe('produceConsumeFormatter', () => {
    const mouseOverTime = new Date(2);
    const emptyTimeSeries = {
      startTimeMs: 0,
      stepMs: 1,
      values: []
    };

    describe('handles bad input', () => {
      const mouseOverTime = new Date(0);
      it('returns null with no lineSlotData', () => {
        const hoverLabelInfo = produceConsumeFormatter(null, emptyTimeSeries, null, mouseOverTime, basicTimeScale, basicTimeScale);
        expect(hoverLabelInfo).to.be.null;
      });

      it('returns null with no areaSlotData', () => {
        const hoverLabelInfo = produceConsumeFormatter(emptyTimeSeries, null, null, mouseOverTime, basicTimeScale, basicTimeScale);
        expect(hoverLabelInfo).to.be.null;
      });

      it('returns null with empty values in lineSlotData', () => {
        const hoverLabelInfo = produceConsumeFormatter(emptyTimeSeries, emptyTimeSeries, null, mouseOverTime, basicTimeScale, basicTimeScale);
        expect(hoverLabelInfo).to.be.null;
      });

      it('returns null with mouseOverTime out of range', () => {
        const lineSlotData = {
          startTimeMs: 0,
          stepMs: 1,
          values: [4, 8, 9, 2]
        };
        const mouseOverTime = new Date(1003);
        const hoverLabelInfo = produceConsumeFormatter(lineSlotData, emptyTimeSeries, null, mouseOverTime, basicTimeScale, basicTimeScale);
        expect(hoverLabelInfo).to.be.null;
      });
    });

    const lineSlotData = {
      startTimeMs: 0,
      stepMs: 1,
      values: [4, 8, 9, 2]
    };
    it('handles exact match', () => {
      const areaSlotData = lineSlotData;
      const hoverLabelInfo = produceConsumeFormatter(lineSlotData, areaSlotData, null, mouseOverTime, basicTimeScale, basicTimeScale);
      expect(hoverLabelInfo).to.not.be.null;
      if (!hoverLabelInfo) return;
      expect(hoverLabelInfo.ptX).to.eql(2.5);
      expect(hoverLabelInfo.ptY).to.eql(9);
      expect(hoverLabelInfo.label).to.eql('100% (  9 of   9 messages)');
    });

    it('handles under consumption', () => {
      const areaSlotData = {
        startTimeMs: 0,
        stepMs: 1,
        values: [4, 8, 4, 2]
      };

      const hoverLabelInfo = produceConsumeFormatter(lineSlotData, areaSlotData, null, mouseOverTime, basicTimeScale, basicTimeScale);
      expect(hoverLabelInfo).to.not.be.null;
      if (!hoverLabelInfo) return;
      expect(hoverLabelInfo.ptX).to.eql(2.5);
      expect(hoverLabelInfo.ptY).to.eql(9);
      expect(hoverLabelInfo.label).to.eql('44% (  4 of   9 messages)');
    });

    it('handles over consumption', () => {
      const areaSlotData = {
        startTimeMs: 0,
        stepMs: 1,
        values: [4, 8, 16, 2]
      };
      const hoverLabelInfo = produceConsumeFormatter(lineSlotData, areaSlotData, null, mouseOverTime, basicTimeScale, basicTimeScale);
      expect(hoverLabelInfo).to.not.be.null;
      if (!hoverLabelInfo) return;
      expect(hoverLabelInfo.ptX).to.eql(2.5);
      expect(hoverLabelInfo.ptY).to.eql(9);
      expect(hoverLabelInfo.label).to.eql('178% ( 16 of   9 messages)');
    });

    it('handles massive over consumption', () => {
      const areaSlotData = {
        startTimeMs: 0,
        stepMs: 1,
        values: [4, 8, 160000, 2]
      };
      const hoverLabelInfo = produceConsumeFormatter(lineSlotData, areaSlotData, null, mouseOverTime, basicTimeScale, basicTimeScale);
      expect(hoverLabelInfo).to.not.be.null;
      if (!hoverLabelInfo) return;
      expect(hoverLabelInfo.ptX).to.eql(2.5);
      expect(hoverLabelInfo.ptY).to.eql(9);
      expect(hoverLabelInfo.label).to.eql('1777778% (160k of   9 messages)');
    });
  });

});
