// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { setStepAccessible, submitForm } from 'actions/connectorForm';
import {
  Edit,
  EncodedFieldNames,
  NewSinkFormData,
  NewSourceFormData,
  STEP1,
  STEP2,
  STEP3,
  validateNameAndClass,
  validateTopics
} from 'constants/management';
import { FormTypes } from 'types/form';

const { NAME, CONNECTOR_CLASS, TOPICS } = EncodedFieldNames;

describe('constants/management/index', () => {
  describe('validateNameAndClass', () => {
    it('validates connector class when it is present', () => {
      const values = {
        [CONNECTOR_CLASS]: 'a-nice-connector-class'
      };
      expect(validateNameAndClass(values)).to.not.have.key(CONNECTOR_CLASS);
      expect(validateNameAndClass(values)).to.have.key(NAME);
    });

    it('validates name when it is present', () => {
      const values = {
        [NAME]: 'dope-connector-name'
      };
      expect(validateNameAndClass(values)).to.not.have.key(NAME);
      expect(validateNameAndClass(values)).to.have.key(CONNECTOR_CLASS);
    });

    it('validates name and connector class when both are present', () => {
      const values = {
        [NAME]: 'dope-connector-name',
        [CONNECTOR_CLASS]: 'another-connector-class'
      };
      expect(validateNameAndClass(values)).to.be.empty;
    });

    it('invalidates name and topics when both are present', () => {
      const values = {};
      expect(validateNameAndClass(values)).to.have.all.keys(NAME, CONNECTOR_CLASS);
    });
  });

  describe('validateTopics', () => {
    it('validates topics when it is present', () => {
      const values = {
        [TOPICS]: 'totally-sweet-topic, super-cool-topic'
      };
      expect(validateTopics(values)).to.be.empty;
    });

    it('invalidates topics when it is empty or not present', () => {
      const values = {};
      expect(validateTopics(values)).to.have.key(TOPICS);

      const values2 = {
        [TOPICS]: ''
      };

      expect(validateTopics(values2)).to.have.key(TOPICS);
    });
  });

  describe('NewSinkFormData', () => {
    describe('step1', () => {
      it('onSubmit sets step2 accessible', () => {
        const dispatch = sinon.spy();
        const values = {};
        NewSinkFormData.steps[STEP1].onSubmit(values, dispatch);
        expect(dispatch).to.be.calledWithExactly(setStepAccessible(STEP2));
      });
    });

    describe('step2', () => {
      it('onSubmit sets step3 accessible', () => {
        const dispatch = sinon.spy();
        const values = {};
        NewSinkFormData.steps[STEP2].onSubmit(values, dispatch);
        expect(dispatch).to.be.calledWithExactly(setStepAccessible(STEP3));
      });
    });

    describe('step3', () => {
      it('onSubmit dispatches submitForm', () => {
        const dispatch = sinon.spy();
        const values = {};
        NewSinkFormData.steps[STEP3].onSubmit(values, dispatch);
        expect(dispatch).to.be.calledWithExactly(submitForm(FormTypes.CONNECTOR_FORM));
      });
    });
  });

  describe('NewSourceFormData', () => {
    describe('step1', () => {
      it('onSubmit sets step2 accessible', () => {
        const dispatch = sinon.spy();
        const values = {};
        NewSourceFormData.steps[STEP1].onSubmit(values, dispatch);
        expect(dispatch).to.be.calledWithExactly(setStepAccessible(STEP2));
      });
    });

    describe('step2', () => {
      it('onSubmit dispatches submitForm', () => {
        const dispatch = sinon.spy();
        const values = {};
        NewSourceFormData.steps[STEP2].onSubmit(values, dispatch);
        expect(dispatch).to.be.calledWithExactly(submitForm(FormTypes.CONNECTOR_FORM));
      });
    });
  });

  describe('Edit form config', () => {
    describe('step1', () => {
      it('onSubmit sets step2 accessible', () => {
        const dispatch = sinon.spy();
        const values = {};
        Edit.steps[STEP1].onSubmit(values, dispatch);
        expect(dispatch).to.be.calledWithExactly(setStepAccessible(STEP2));
      });
    });

    describe('step2', () => {
      it('onSubmit dispatches submitForm', () => {
        const dispatch = sinon.spy();
        const values = {};
        Edit.steps[STEP2].onSubmit(values, dispatch);
        expect(dispatch).to.be.calledWithExactly(submitForm(FormTypes.CONNECTOR_FORM));
      });
    });
  });
});
