// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.

import { LOCATION_CHANGE } from 'react-router-redux';

import {
  initialState,
  reducer
} from 'reducers/management/connectorForm';

import { GET } from '../../mock/connectorPlugins.json';

import { SINK, SOURCE, NEW_SINK_PATHNAME } from 'constants/management';

const baseConnector = {
  name: 'my-hdfs-sink-connector',
  config: {
    'connector.class': 'io.confluent.connect.hdfs.HdfsSinkConnector',
    'flush.size': '100',
    'hadoop.conf.dir': '/opt/hadoop/conf',
    'hadoop.home': '/opt/hadoop',
    'hdfs.url': 'hdfs://fakehost:9000',
    'rotate.interval.ms': '1000',
    'tasks.max': '1',
    topics: 'c3-monitoring-Group-INFO_TOPICS_IN_CLIENTTYPE-ONE_MINUTE-changelog,c3-monitoring-MonitoringStream-LOOKUP_CLIENTTYPE_TOPIC-ONE_HOUR-changelog',
    name: 'my-hdfs-sink-connector'
  },
  tasks: [],
  type: SINK
};


const currentStep = 'currentStep';

const baseFormState = {
  currentPathname: 'path/to/the/future',
  completed: false,
  currentStep,
  errorCount: 0,
  initialValues: {},
  isEdit: true,
  loadingConnector: false,
  optionsByField: {},
  steps: { [currentStep]: {} },
  type: SOURCE
};

describe('reducers/management/connectorForm', () => {
  describe('updateFieldOptions', () => {
    const state = {
      ...baseFormState
    };
    const newFieldKey = 'my-field';
    const newFieldVal = [];
    const action = {
      type: 'CONNECTOR_FORM_SEARCH_SUCCEEDED',
      payload: {
        field: newFieldKey,
        options: newFieldVal
      }
    };
    let newState;

    beforeEach(() => {
      newState = reducer(state, action);
    });

    it('updates key', () => {
      expect(newState.optionsByField[newFieldKey]).to.equal(newFieldVal);
    });

    it('does not mutate initialState', () => {
      expect(newState).to.not.equal(state);
    });
  });

  describe('enableStepAccess', () => {
    const stepKey = 'someStep';
    const action = {
      type: 'STEP_SET_ACCESSIBLE',
      payload: {
        step: stepKey
      }
    };

    it('enables a step if it exists and does not mutate', () => {
      const state = {
        ...baseFormState,
        steps: {
          ...baseFormState.steps,
          [stepKey]: {
            canAccess: false
          }
        }
      };

      expect(reducer(state, action).steps[stepKey].canAccess).to.equal(true);
    });

    it('returns initial state if key doesn\'t exist', () => {
      const state = {
        ...baseFormState
      };
      const newSteps = reducer(state, action);

      expect(newSteps).to.equal(newSteps);
    });
  });

  describe('locationChange', () => {
    const currentPathname = '/coolest-form-ever';
    const state = {
      ...baseFormState,
      currentPathname,
      currentStep: 'oldStep'
    };

    it('updates currentStep if pathname is same', () => {
      const newStep = 'newStep';
      const action = {
        type: LOCATION_CHANGE,
        payload: {
          pathname: currentPathname,
          query: {
            step: newStep
          }
        }
      };
      expect(reducer(state, action).currentStep).to.equal(newStep);
    });

    it('resets to initialState if pathname is not recognized', () => {
      const action = {
        type: LOCATION_CHANGE,
        payload: {
          pathname: '/lame-form',
          query: {
            step: 'newStep'
          }
        }
      };
      expect(reducer(state, action)).to.equal(initialState);
    });

    it('uses an existing config if pathname exists in FormConfigs', () => {
      const action = {
        type: LOCATION_CHANGE,
        payload: {
          type: LOCATION_CHANGE,
          pathname: NEW_SINK_PATHNAME,
          query: {
            step: 'newStep'
          }
        }
      };

      expect(reducer(state, action).type).to.equal(SINK);
    });
  });

  describe('submitSuccess', () => {
    const action = {
      type: 'CONNECTOR_FORM_SUBMIT_SUCCESS',
      payload: undefined
    };

    it('sets completed', () => {
      const state = {
        ...baseFormState,
        completed: false
      };

      expect(reducer(state, action).completed).to.equal(true);
    });
  });

  describe('setupEdit', () => {
    it('sets isEdit', () => {
      const action = {
        type: 'LOAD_CONNECTOR_FOR_EDITING',
        payload: undefined
      };
      const state = {
        ...baseFormState
      };
      expect(reducer(state, action).isEdit).to.equal(true);
    });
  });

  describe('fetchEditSuccess', () => {
    const { configs, error_count } = GET.HdfsSinkConnector;
    const action = {
      type: 'CONNECTOR_FORM_FETCH_EDIT_SUCCESS',
      payload: {
        name: baseConnector.name,
        connector: {
          ...baseConnector
        },
        validationData: {
          configs,
          error_count
        }
      }
    };

    it('does nothing when isEdit is false', () => {
      const state = {
        ...baseFormState,
        isEdit: false
      };
      expect(reducer(state, action)).to.equal(state);
    });

    it('sets correct errorCount', () => {
      const state = {
        ...baseFormState
      };
      expect(reducer(state, action).errorCount).to.equal(error_count);
    });

    it('extracts correct initialValues from HdfsSinkConnector config', () => {
      const state = {
        ...baseFormState
      };

      expect(reducer(state, action).initialValues).to.have.all.keys([
        'connector-period-class',
        'flush-period-size',
        'name',
        'topics'
      ]);
    });
  });

  describe('validateSuccess', () => {
    it('sets correct errorCount', () => {
      const { configs, error_count } = GET.HdfsSinkConnector;
      const state = {
        ...baseFormState,
        type: SINK
      };
      const action = {
        type: 'CONNECTOR_FORM_VALIDATE_SUCCESS',
        payload: {
          name: 'form_name',
          validationData: {
            configs,
            error_count
          }
        }
      };
      expect(reducer(state, action).errorCount).to.equal(error_count);
    });
  });
});
