// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.

import { expect } from 'chai';

import {
  UPDATE_END_TIME,
  UPDATE_FETCH_EXTENDED_END_TIME,
  SET_TIME_RANGE
} from 'actions/monitoring/timeframe';
import { RangeTypes } from 'constants/monitoring/timeframe';
import { timeframe } from 'reducers/monitoring/timeframe';

const initAction = { type: 'BLAH', payload: {} };
const baseTimeframeState = {
  startTimeMs: -1,
  stopTimeMs: -1,
  fetchExtendedStopTimeMs: null,
  rangeType: RangeTypes.ROLLING
};

describe('timeframe', () => {
  const initState = timeframe(undefined, initAction);

  it('has default state', () => {
    expect(initState).to.have.all.keys('startTimeMs', 'stopTimeMs', 'rangeType', 'fetchExtendedStopTimeMs');
  });

  describe('UPDATE_END_TIME', () => {
    it('handles empty payload', () => {
      expect(timeframe(baseTimeframeState, { type: UPDATE_END_TIME, payload: {} })).to.have.all.keys('startTimeMs', 'stopTimeMs', 'rangeType', 'fetchExtendedStopTimeMs');
    });

    describe('RangeTypes.ROLLING', () => {
      const rollingState = {
        startTimeMs: 0,
        stopTimeMs: 100,
        fetchExtendedStopTimeMs: null,
        rangeType: RangeTypes.ROLLING
      };
      it('updates start and stop time', () => {
        const action = {
          type: UPDATE_END_TIME,
          payload: {
            endTimeMs: 102
          }
        };
        const state = timeframe(rollingState, action);
        expect(state).to.exist;
        expect(state.startTimeMs).to.eql(2);
        expect(state.stopTimeMs).to.eql(102);
        expect(state.rangeType).to.eql(rollingState.rangeType);
      });
    });

    [RangeTypes.GROWING_DAY, RangeTypes.GROWING_WEEK, RangeTypes.GROWING_MONTH].forEach(rangeType => {
      describe(`RangeTypes.${rangeType}`, () => {
        it('only updates stop time', () => {
          const growingState = {
            startTimeMs: 0,
            stopTimeMs: 100,
            fetchExtendedStopTimeMs: null,
            rangeType: rangeType
          };
          const action = {
            type: UPDATE_END_TIME,
            payload: {
              endTimeMs: 102
            }
          };
          const state = timeframe(growingState, action);
          expect(state).to.exist;
          expect(state.startTimeMs).to.eql(0);
          expect(state.stopTimeMs).to.eql(102);
          expect(state.rangeType).to.eql(growingState.rangeType);
        });
      });
    });

    describe('RangeTypes.FIXED', () => {
      const fixedState = {
        startTimeMs: 0,
        stopTimeMs: 100,
        fetchExtendedStopTimeMs: null,
        rangeType: RangeTypes.FIXED
      };
      const action = {
        type: UPDATE_END_TIME,
        payload: {
          endTimeMs: 102
        }
      };
      it('doesnt change anything', () => {
        const state = timeframe(fixedState, action);
        expect(state).to.be.eql(fixedState);
      });
    });
  });

  describe('SET_TIME_RANGE', () => {
    it('handles empty payload', () => {
      const state = timeframe(baseTimeframeState, { type: SET_TIME_RANGE, payload: {} });
      expect(state).to.have.all.keys('startTimeMs', 'stopTimeMs', 'rangeType', 'fetchExtendedStopTimeMs');
    });

    it('ignores null', () => {
      const action = {
        type: SET_TIME_RANGE,
        payload: null
      };
      const state = timeframe(baseTimeframeState, action);
      expect(baseTimeframeState).to.eql(state);
    });

    it('adds everything', () => {
      const action = {
        type: SET_TIME_RANGE,
        payload: {
          startTimeMs: 100,
          stopTimeMs: 200,
          fetchExtendedStopTimeMs: null,
          rangeType: RangeTypes.ROLLING
        }
      };
      const state = timeframe(baseTimeframeState, action);
      expect(state).to.eql(action.payload);
    });

    it('doesnt add extra keys', () => {
      const action = {
        type: SET_TIME_RANGE,
        payload: {
          startTimeMs: 100,
          stopTimeMs: 200,
          fetchExtendedStopTimeMs: null,
          rangeType: RangeTypes.ROLLING,
          something: 'awful'
        }
      };
      const state = timeframe(baseTimeframeState, action);
      expect(state).to.have.all.keys('startTimeMs', 'stopTimeMs', 'rangeType', 'fetchExtendedStopTimeMs');
    });
  });

  describe('UPDATE_FETCH_EXTENDED_END_TIME', () => {
    it('fetchExtendedStopTimeMs is set, others unchanged', () => {
      const action = {
        type: UPDATE_FETCH_EXTENDED_END_TIME,
        payload: {
          endTimeMs: 300
        }
      };
      const state = timeframe(baseTimeframeState, action);
      expect(state).to.have.all.keys('startTimeMs', 'stopTimeMs', 'rangeType', 'fetchExtendedStopTimeMs');
      expect(state.fetchExtendedStopTimeMs).to.eql(300);
    });
  });
});
