// @flow
// (Copyright) [2016 - 2016] Confluent, Inc.
import { call, put, fork } from 'redux-saga/effects';
import { takeLatest } from 'redux-saga';

import * as Sagas from 'sagas/management/connectorPlugins';
import * as ConnectorPluginActions from 'actions/connectorPlugins';
import { GET } from '../../mock/connectorPlugins.json';

const { connectorPluginsSaga, validateConfig, validateConfigSaga } = Sagas;

describe('sagas/management/connectorPlugins', () => {
  const name = 'my-connector';
  const connectorResponse = {
    config: {
      'connector.class': 'some-connector-class',
      name
    },
    name,
    tasks: [],
    type: 'SINK'
  };
  const error = new Error('dope error');
  let generator;
  let next;

  describe('validateConfig', () => {
    const connectorConfig = connectorResponse.config;

    beforeEach(() => {
      const initialAction = ConnectorPluginActions.validateConfig(connectorConfig);
      generator = validateConfig(initialAction);
      next = generator.next();
    });

    it('should return fast without any "connector.class"', () => {

      // $FlowFixMe not really a problem, it's just a test
      const initialAction = ConnectorPluginActions.validateConfig({
        name: 'loosey goosey',
        config: {}
      });
      generator = validateConfig(initialAction);
      expect(
        generator.next()
      ).to.deep.equal({
        done: true,
        value: undefined
      });
    });

    it('should yield a request action', () => {
      expect(
        next.value
      ).to.deep.equal(
        put(ConnectorPluginActions.validate.request(connectorConfig.name, connectorConfig))
      );
    });

    it('should yield a failure action', () => {
      // move past request
      generator.next();

      expect(
        generator.throw(error).value
      ).to.deep.equal(
        put(ConnectorPluginActions.validate.failure(connectorConfig.name, error))
      );
    });

    it('should yield a success action', () => {
      const connectorPluginValidateResponse = GET.HdfsSinkConnector;
      // move past request
      generator.next();

      expect(
        generator.next(connectorPluginValidateResponse).value
      ).to.deep.equal(
        put(ConnectorPluginActions.validate.success(connectorResponse.name, connectorPluginValidateResponse))
      );
    });

    it('should finish after success or error', () => {
      // move past request
      generator.next();

      // move past failure and success
      generator.next();

      expect(
        generator.next()
      ).to.deep.equal({
        done: true,
        value: undefined
      });
    });
  });

  describe('validateConfigSaga', () => {
    beforeEach(() => {
      generator = validateConfigSaga();
      next = generator.next();
    });

    it('takes latest validate config action and starts `validateConfig`', () => {
      expect(
        next.value
      ).to.deep.equal(call(takeLatest, ConnectorPluginActions.CONNECTOR_PLUGINS_VALIDATE_CONFIG, validateConfig));
    });

    it('should finish', () => {
      next = generator.next();

      expect(
        next
      ).to.deep.equal({
        done: true,
        value: undefined
      });
    });
  });

  describe('connectorPluginsSaga', () => {
    beforeEach(() => {
      generator = connectorPluginsSaga();
      next = generator.next();
    });

    it('starts `validateConfigSaga`', () => {
      expect(
        next.value
      ).to.deep.equal(fork(validateConfigSaga));
    });

    it('should finish', () => {
      next = generator.next();

      expect(
        next
      ).to.deep.equal({
        done: true,
        value: undefined
      });
    });
  });
});
