package io.confluent.monitoring.clients.interceptor;

import com.google.common.collect.ImmutableMap;
import io.confluent.monitoring.common.MonitoringProducerDefaults;
import io.confluent.monitoring.common.TimeBucket;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.common.config.AbstractConfig;
import org.apache.kafka.common.config.ConfigDef;
import org.apache.kafka.common.config.ConfigException;

import java.util.Map;
import java.util.Properties;

public class MonitoringInterceptorConfig extends AbstractConfig {

  public static final String MONITORING_INTERCEPTOR_PREFIX = "confluent.monitoring.interceptor.";

  private static final ConfigDef config;

  public static final String TOPIC_CONFIG = MONITORING_INTERCEPTOR_PREFIX + "topic";
  public static final String DEFAULT_TOPIC_CONFIG = "_confluent-monitoring";
  public static final String TOPIC_CONFIG_DOC = "Topic on which monitoring data will be written.";

  public static final String PUBLISH_PERIOD = MONITORING_INTERCEPTOR_PREFIX + "publishMs";
  public static final Long DEFAULT_PUBLISH_PERIOD = TimeBucket.SIZE;
  public static final String PUBLISH_PERIOD_DOC =
      "Period the interceptor should use to publish messages to " + MONITORING_INTERCEPTOR_PREFIX +
      TOPIC_CONFIG;

  
  static {
    config = new ConfigDef()
        .define(
            TOPIC_CONFIG,
            ConfigDef.Type.STRING,
            DEFAULT_TOPIC_CONFIG,
            ConfigDef.Importance.HIGH,
            TOPIC_CONFIG_DOC
        ).define(
            PUBLISH_PERIOD,
            ConfigDef.Type.LONG,
            DEFAULT_PUBLISH_PERIOD,
            ConfigDef.Importance.HIGH,
            PUBLISH_PERIOD_DOC
        );
  }

  public MonitoringInterceptorConfig(Properties props) {
    super(config, props);
  }

  public MonitoringInterceptorConfig(Map<String, ?> clientConfigs) {
    super(config, clientConfigs);
  }

  public static void main(String[] args) {
    System.out.println(config.toHtmlTable());
  }

  /**
   * Order of preference:
   * -- config with prefix MONITORING_INTERCEPTOR_PREFIX from client config
   * -- from client config
   *
   * @param clientConfigs client configs
   */
  public static MonitoringInterceptorConfig getConfig(Map<String, ?> clientConfigs) {
    return new MonitoringInterceptorConfig(clientConfigs);
  }

  public static final Map<String, Object> PRODUCER_CONFIG_DEFAULTS = ImmutableMap.<String, Object>builder()
      .put(ProducerConfig.ACKS_CONFIG, MonitoringProducerDefaults.ACKS_CONFIG)
      .put(ProducerConfig.KEY_SERIALIZER_CLASS_CONFIG, MonitoringProducerDefaults.KEY_SERIALIZER_CLASS_CONFIG)
      .put(ProducerConfig.VALUE_SERIALIZER_CLASS_CONFIG, MonitoringProducerDefaults.VALUE_SERIALIZER_CLASS_CONFIG)
      .put(ProducerConfig.COMPRESSION_TYPE_CONFIG, MonitoringProducerDefaults.COMPRESSION_TYPE_CONFIG)
      .put(ProducerConfig.LINGER_MS_CONFIG, MonitoringProducerDefaults.LINGER_MS_CONFIG)
      .put(ProducerConfig.INTERCEPTOR_CLASSES_CONFIG, "")
      .build();

  public static Properties getProducerProperties(Map<String, ?> clientConfigs) {
    Properties props = new Properties();
    for (Map.Entry<String, ?> entry : PRODUCER_CONFIG_DEFAULTS.entrySet()) {
      props.put(entry.getKey(), entry.getValue());
    }
    for (Map.Entry<String, ?> entry : clientConfigs.entrySet()) {
      if (entry.getKey().startsWith(MONITORING_INTERCEPTOR_PREFIX)) {
        props.put(entry.getKey().substring(MONITORING_INTERCEPTOR_PREFIX.length()), entry.getValue());
      }
    }

      // we need to handle client_id special
    Object clientId = props.get(ProducerConfig.CLIENT_ID_CONFIG);
    if (clientId == null) {
      clientId = clientConfigs.get(ProducerConfig.CLIENT_ID_CONFIG);
      if (clientId != null) {
        props.put(ProducerConfig.CLIENT_ID_CONFIG, MONITORING_INTERCEPTOR_PREFIX + clientId);
      }
    }
    if (clientId == null) {
      throw new ConfigException("Expected " + ProducerConfig.CLIENT_ID_CONFIG);
    }

    // we need to handle bootstrap servers special
    Object bootstrap = props.get(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG);
    if (bootstrap == null) {
      bootstrap = clientConfigs.get(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG);
      if (bootstrap != null) {
        props.put(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG, bootstrap);
      }
    }
    if (bootstrap == null) {
      throw new ConfigException("Expected " + ProducerConfig.BOOTSTRAP_SERVERS_CONFIG);
    }

    return props;
  }
}
