// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.monitoring.clients.interceptor;

import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.clients.consumer.OffsetAndMetadata;
import org.apache.kafka.common.ClusterResource;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.record.TimestampType;
import org.easymock.EasyMock;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.powermock.api.easymock.PowerMock;
import org.powermock.core.classloader.annotations.PrepareForTest;
import org.powermock.modules.junit4.PowerMockRunner;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static org.junit.Assert.assertEquals;

@RunWith(PowerMockRunner.class)
@PrepareForTest(MonitoringInterceptor.class)
public class MonitoringConsumerInterceptorTest {

  private final String clientId = "test producer";
  private final String clusterId = "test cluster";
  private final String groupId = "test group";
  private final TopicPartition tp = new TopicPartition("test1", 0);
  private final ConsumerRecord<Long, Long>
      record =
      new ConsumerRecord<>(tp.topic(), tp.partition(), 0, System.currentTimeMillis(),
                           TimestampType.CREATE_TIME, 123, 1, 1, 1L, 2L);
  private final Map<String, Object> configs = new HashMap<>();

  private MonitoringInterceptor interceptor;
  private MonitoringConsumerInterceptor<Long, Long> consumerInterceptor;

  @Before
  public void setUp() throws Exception {
    interceptor = PowerMock.createMock(MonitoringInterceptor.class);
    consumerInterceptor = new MonitoringConsumerInterceptor<>();
    configs.put(CommonClientConfigs.CLIENT_ID_CONFIG, clientId);
    configs.put(ConsumerConfig.GROUP_ID_CONFIG, groupId);
  }

  private void expectConfigureConsumerInterceptor() {
    PowerMock.mockStatic(MonitoringInterceptor.class);
    EasyMock.expect(MonitoringInterceptor.createForConsumer(clusterId, groupId, configs))
        .andReturn(interceptor);

    interceptor.start();
    PowerMock.expectLastCall().once();
  }

  private Map<TopicPartition, List<ConsumerRecord<Long, Long>>> createMapWithOneRecord(
      TimestampType timestampType) {
    Map<TopicPartition, List<ConsumerRecord<Long, Long>>> recordsMap = new HashMap<>();
    List<ConsumerRecord<Long, Long>> tpRecords = new ArrayList<>();
    TopicPartition tp2 = new TopicPartition("test2", 0);
    ConsumerRecord<Long, Long> record2 = new ConsumerRecord<>(tp2.topic(), tp2.partition(), 0,
                                                              System.currentTimeMillis(),
                                                              timestampType, 456, 1, 1, 7L, 6L);
    tpRecords.add(record2);
    recordsMap.put(tp2, tpRecords);
    return recordsMap;
  }

  private void doTestOnConsumeOneRecordAndVerifyRecordsNotModified(TimestampType timestampType) {
    expectConfigureConsumerInterceptor();

    interceptor.recordMessageMetric(EasyMock.anyString(), EasyMock.anyInt(), EasyMock.anyLong(),
                                    EasyMock.anyInt(), EasyMock.anyLong(), EasyMock.anyLong());
    EasyMock.expectLastCall().once();

    PowerMock.replayAll();

    consumerInterceptor.configure(configs);
    consumerInterceptor.onUpdate(new ClusterResource(clusterId));

    ConsumerRecords<Long, Long>
        records =
        new ConsumerRecords<>(createMapWithOneRecord(timestampType));
    ConsumerRecords<Long, Long> interceptedRecords = consumerInterceptor.onConsume(records);

    assertEquals(records, interceptedRecords);

    PowerMock.verifyAll();
  }

  @Test
  public void testOnConsumeDoesNotModifyRecord() {
    doTestOnConsumeOneRecordAndVerifyRecordsNotModified(TimestampType.CREATE_TIME);
  }

  @Test
  public void testOnConsumeRecordsMetricsForLogAppendTime() {
    doTestOnConsumeOneRecordAndVerifyRecordsNotModified(TimestampType.LOG_APPEND_TIME);
  }

  @Test
  public void testOnConsumeRecordsMetricsCorrectly() {
    expectConfigureConsumerInterceptor();

    // TODO: when we actually have timestamp, size, and crc in ConsumerRecord, we then can match
    // actual parameters passed to recordMessageMetrics; but for now just testing that the method
    // was called
    interceptor.recordMessageMetric(EasyMock.anyString(), EasyMock.anyInt(), EasyMock.anyLong(),
                                    EasyMock.anyInt(), EasyMock.anyLong(), EasyMock.anyLong());
    EasyMock.expectLastCall().times(2);

    PowerMock.replayAll();

    consumerInterceptor.configure(configs);
    consumerInterceptor.onUpdate(new ClusterResource(clusterId));

    Map<TopicPartition, List<ConsumerRecord<Long, Long>>>
        recordsMap =
        createMapWithOneRecord(TimestampType.CREATE_TIME);
    // add one more list for another topic-partition
    List<ConsumerRecord<Long, Long>> tpRecords = new ArrayList<>();
    tpRecords.add(record);
    recordsMap.put(tp, tpRecords);

    ConsumerRecords<Long, Long> records = new ConsumerRecords<>(recordsMap);
    consumerInterceptor.onConsume(records);

    PowerMock.verifyAll();
  }

  @Test(expected = IllegalStateException.class)
  public void testOnConsumeBeforeConfigureFails() {
    ConsumerRecords<Long, Long>
        records =
        new ConsumerRecords<>(createMapWithOneRecord(TimestampType.CREATE_TIME));

    consumerInterceptor.onUpdate(new ClusterResource(clusterId));
    consumerInterceptor.onConsume(records);
  }

  @Test
  public void testCloseShutsDownInterceptor() {
    expectConfigureConsumerInterceptor();
    // expect interceptor shutdown() is called after close
    interceptor.shutdown();
    EasyMock.expectLastCall();

    PowerMock.replayAll();

    consumerInterceptor.configure(configs);
    consumerInterceptor.onUpdate(new ClusterResource(clusterId));
    consumerInterceptor.onConsume(ConsumerRecords.<Long, Long>empty());
    consumerInterceptor.close();

    PowerMock.verifyAll();
  }

  @Test
  public void testOnCommitDoesNothing() {
    // we just test by testing that calling that method before configure() does not throw an
    // exception
    consumerInterceptor.onCommit(new HashMap<TopicPartition, OffsetAndMetadata>());
  }

  @Test
  public void testOKToCloseBeforeConfigure() {
    consumerInterceptor.close();
    // testing that exception is not thrown
  }
}
