// (Copyright) [2016 - 2016] Confluent, Inc.
package io.confluent.monitoring.clients.interceptor;

import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.clients.producer.RecordMetadata;
import org.apache.kafka.common.ClusterResource;
import org.apache.kafka.common.TopicPartition;
import org.easymock.EasyMock;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.powermock.api.easymock.PowerMock;
import org.powermock.core.classloader.annotations.PrepareForTest;
import org.powermock.modules.junit4.PowerMockRunner;

import java.util.HashMap;
import java.util.Map;

import static org.junit.Assert.assertEquals;

@RunWith(PowerMockRunner.class)
@PrepareForTest(MonitoringInterceptor.class)
public class MonitoringProducerInterceptorTest {

  private final String clusterId = "test cluster";
  private final String clientId = "test producer";
  private final Map<String, Object> configs = new HashMap<>();
  private MonitoringInterceptor interceptor;
  private MonitoringProducerInterceptor<Long, Long> producerInterceptor;

  @Before
  public void setUp() throws Exception {
    interceptor = PowerMock.createMock(MonitoringInterceptor.class);
    producerInterceptor = new MonitoringProducerInterceptor<>();
    configs.put(CommonClientConfigs.CLIENT_ID_CONFIG, clientId);
  }

  private void expectConfigureAndStartProducerInterceptor() {
    PowerMock.mockStatic(MonitoringInterceptor.class);
    EasyMock.expect(MonitoringInterceptor.createForProducer(clusterId, configs)).andReturn(interceptor);

    interceptor.start();
    PowerMock.expectLastCall().once();
  }

  @Test
  public void testOnSendDoesNotModifyRecord() {
    // we don't need to configure monitoring producer interceptor for this
    ProducerRecord<Long, Long> record = new ProducerRecord<Long, Long>("test topic", 0, 1L, 2L);
    ProducerRecord<Long, Long> interceptedRecord = producerInterceptor.onSend(record);

    // verify that interceptedRecord is the same as record
    assertEquals(record, interceptedRecord);
  }

  @Test(expected = IllegalStateException.class)
  public void testOnAcknowledgementBeforeConfigureFails() {
    RecordMetadata metadata = new RecordMetadata(new TopicPartition("test", 0), 0, 0, 0, 0, 0, 0);
    producerInterceptor.onUpdate(new ClusterResource(clusterId));
    producerInterceptor.onAcknowledgement(metadata, null);
  }

  @Test(expected = IllegalStateException.class)
  public void testOnAcknowledgementBeforeOnUpdateFails() {
    RecordMetadata metadata = new RecordMetadata(new TopicPartition("test", 0), 0, 0, 0, 0, 0, 0);
    producerInterceptor.configure(configs);
    producerInterceptor.onAcknowledgement(metadata, null);
  }

  @Test
  public void testOnAcknowledgementDoesNothingIfNullMetadata() {
    expectConfigureAndStartProducerInterceptor();
    PowerMock.replayAll();

    // currently we do not record anything if error happened while producing (metadata is null)
    producerInterceptor.configure(configs);
    producerInterceptor.onUpdate(new ClusterResource(clusterId));
    producerInterceptor.onAcknowledgement(null, new Exception("Some exception"));

    PowerMock.verifyAll();
  }

  @Test
  public void testOnAcknowledgementRecordsMetricsCorrectly() {
    long checksum = 123;
    int keySize = 1;
    int valueSize = 3;

    expectConfigureAndStartProducerInterceptor();

    // TODO: when we actually have timestamp, size, and crc in recordmetadata, we then can match
    // actual parameters passed to recordMessageMetrics; but for now just testing that the method
    // was called
    long timestamp = System.currentTimeMillis();
    interceptor.recordMessageMetric(EasyMock.anyString(), EasyMock.anyInt(), EasyMock.anyLong(),
                                    EasyMock.anyInt(), EasyMock.anyLong(), EasyMock.anyLong());
    EasyMock.expectLastCall().once();

    PowerMock.replayAll();

    producerInterceptor.configure(configs);
    producerInterceptor.onUpdate(new ClusterResource(clusterId));

    RecordMetadata metadata = new RecordMetadata(new TopicPartition("test1", 1), 0, 0, timestamp,
                                                 checksum, keySize, valueSize);
    producerInterceptor.onAcknowledgement(metadata, null);

    PowerMock.verifyAll();
  }

  @Test
  public void testCloseShutsDownInterceptor() {
    expectConfigureAndStartProducerInterceptor();
    // expect interceptor shutdown() is called after close
    interceptor.shutdown();
    EasyMock.expectLastCall();

    PowerMock.replayAll();

    producerInterceptor.configure(configs);
    producerInterceptor.onUpdate(new ClusterResource(clusterId));
    producerInterceptor.onAcknowledgement(null, null);
    producerInterceptor.close();

    PowerMock.verifyAll();
  }

  @Test
  public void testOKToCloseBeforeConfigure() {
    producerInterceptor.close();
    // testing that exception is not thrown
  }
}
