from __future__ import print_function
import random
import math
import json
import sys
import time
import datetime


class MonitoringSimulator(object):
    """
        Session sequences ids per client are tracked by (cluster, topic, partition)
    """
    INTERVAL_SECS = 15

    def __init__(self, spec, start_time_unix, monitoring_topic_partitions, duration_mins,
                 generate_heartbeats):
        self.spec = spec
        self.start_window_unix = start_time_unix - start_time_unix % self.INTERVAL_SECS
        self.duration_mins = duration_mins
        self.monitoring_topic_partitions = monitoring_topic_partitions
        self.partitions = spec['partitions_per_topic'] * spec['topics']
        self.generate_heartbeats = generate_heartbeats

        assert self.partitions >= spec['partitions_per_producer']
        assert self.partitions % spec['partitions_per_producer'] == 0

        self.producers = self.get_producer_assignments()
        self.producer_sequences = self.get_initial_producer_sequences()
        self.consumers = self.get_consumer_assignments()
        self.consumer_sequences = self.get_initial_consumer_sequences()

    def get_initial_producer_sequences(self):
        """
            Producer sequence numbers are based on producer_id, cluster_id, topic_id, partition
        """
        return {
            (pid, tp): 0
            for pid, tp_set in enumerate(self.producers)
            for tp in tp_set
        }

    def get_initial_consumer_sequences(self):
        """
            Consumer sequence numbers are based on group_id, consumer_id, cluster_id, topic_id, partition
        """
        return {
            (consumer, tp): 0
            for consumer, tp_set in self.consumers.iteritems()
            for tp in tp_set
        }

    def get_topic_partition(self, partition_id):
        topic_id = partition_id // self.spec['partitions_per_topic']
        cluster_id = topic_id % self.spec['clusters']
        partition = partition_id % self.spec['partitions_per_topic']
        return (cluster_id, topic_id, partition)

    def get_consumer_assignments(self):
        """
            Assign consumers (group_id, consumer_id) to a set of cluster topic partitions (cluster_id, topic_id, partition)

            Assign topics to clusters in round-robin fashion

            Assign a single consumer group per topic and a consumer per partition.  Overall,
            consumer messages are not very important for throughput testing because
            they are a tiny fraction of producer messages.
        """
        consumers = {}
        for topic_id in range(self.spec['topics']):
            group_id = topic_id
            cluster_id = topic_id % self.spec['clusters']
            for partition in range(self.spec['partitions_per_topic']):
                consumer_id = partition
                consumers[(group_id, consumer_id)] = frozenset([(cluster_id, topic_id, partition)])
        return consumers

    def get_producer_assignments(self):
        """
            Assign producers (identified by index in producers array) to a set of topic partitions (cluster_id, topic_id, partition)

            Assign topics to clusters in round-robin fashion

            Assign producers to topic-partitions uniformly at random.  Each producer writes to spec['partitions_per_producer']
            consecutive partitions but we want to spread them evenly over all topics and partitions
        """
        producers = []
        for pid in range(self.spec['producers']):
            start_partition = int(math.floor(random.uniform(0, self.partitions)))
            partition_list = [p % self.partitions for p in range(start_partition, start_partition + self.spec['partitions_per_producer'])]
            topic_partitions = frozenset([self.get_topic_partition(p) for p in partition_list])
            producers.append(topic_partitions)
        return producers

    def print_times(self):
        end_window_unix = self.start_window_unix + self.duration_mins * 60 - self.INTERVAL_SECS
        start_window_dt = datetime.datetime.fromtimestamp(self.start_window_unix)
        end_window_dt = datetime.datetime.fromtimestamp(end_window_unix)
        ui_start_ms = self.start_window_unix * 1000
        ui_end_ms = (end_window_unix + self.INTERVAL_SECS) * 1000
        print("# Start first window: %s, start last window: %s" % (start_window_dt, end_window_dt))
        print("# C3 UI: http://localhost:9021/monitoring/consumer-groups?fixed-end=%s&fixed-start=%s" % (ui_end_ms, ui_start_ms))

    def create_producer_record(self, producer_id, window, timestamp, count, tp):
        (cluster_id, topic_id, partition) = tp
        sequence = self.producer_sequences[(producer_id, tp)]
        self.producer_sequences[(producer_id, tp)] += 1
        rec = {
            "clusterId": "cluster-%s" % str(cluster_id),
            "clientType": "PRODUCER",
            "clientId": "producer-id-%s" % producer_id,
            "session": "f58c1b2e-9474-4227-94a7-e03c32-%s" % producer_id,
            "sequence": str(sequence),
            "window": str(window),
            "timestamp": str(timestamp),
            "topic": "topic-%s" % topic_id,
            "partition": partition,
            "count": count,
            "aggregateBytes": "96",
            "aggregateCrc": -1.373341569E9,
            "samplePeriod": "15000",
            "minWindow": str(window),
            "maxWindow": str(window)
        }
        if sequence == 0:
            rec['clusterName'] = "Cluster%s" % cluster_id
        return rec

    def create_producer_records(self, interval_start_unix):
        """
            Generate two producer monitoring messages per wall clock tick (except on the first window)
        """
        topic_partition_counts = {self.get_topic_partition(p): {'produced_curr': 0, 'produced_prev': 0} for p in range(self.partitions)}
        curr_window = interval_start_unix * 1000
        prev_window = (interval_start_unix - self.INTERVAL_SECS) * 1000

        for producer_id, tp_set in enumerate(self.producers):
            for tp in tp_set:
                curr_count = int(math.floor(random.uniform(5, 100)))
                topic_partition_counts[tp]['produced_curr'] += curr_count
                curr_rec = self.create_producer_record(producer_id, curr_window, curr_window + 2, curr_count, tp)
                self.print_record(curr_rec)

                if interval_start_unix != self.start_window_unix:
                    prev_count = int(math.floor(random.uniform(5, 100)))
                    topic_partition_counts[tp]['produced_prev'] += prev_count
                    prev_rec = self.create_producer_record(producer_id, prev_window, curr_window + 2, prev_count, tp)
                    self.print_record(prev_rec)
        return topic_partition_counts

    def create_consumer_records(self, interval_start_unix, tp_counts):
        """
            Generate two consumer monitoring messages per wall clock tick (except on the first window)
        """
        curr_window = interval_start_unix * 1000
        prev_window = (interval_start_unix - self.INTERVAL_SECS) * 1000

        for consumer, tp_set in self.consumers.iteritems():
            for tp in tp_set:
                curr_rec = self.create_consumer_record(consumer, curr_window, curr_window + 3, tp_counts[tp]['produced_curr'], tp)
                self.print_record(curr_rec)

                if interval_start_unix != self.start_window_unix:
                    prev_rec = self.create_consumer_record(consumer, prev_window, curr_window + 3, tp_counts[tp]['produced_prev'], tp)
                    self.print_record(prev_rec)

    def create_consumer_record(self, consumer, window, timestamp, count, tp):
        (group_id, consumer_id) = consumer
        (cluster_id, topic_id, partition) = tp
        sequence = self.consumer_sequences[(consumer, tp)]
        self.consumer_sequences[(consumer, tp)] += 1
        rec = {
            "clusterId": "cluster-%s" % str(cluster_id),
            "clientId": "cg-%s-%s" % (group_id, consumer_id),
            "group": "cg-%s" % group_id,
            "session": "eb5af750-7eb8-%s-%s" % (group_id, consumer_id),
            "sequence": str(sequence),
            "window": str(window),
            "timestamp": str(timestamp),
            "topic": "topic-%s" % topic_id,
            "partition": partition,
            "count": str(count),
            "aggregateBytes": "232",
            "aggregateCrc": -2.025923529E9,
            "totalLatency": "299879",
            "minLatency": "7606",
            "maxLatency": "14085",
            "samplePeriod": "15000",
            "minWindow": str(window),
            "maxWindow": str(window)
        }
        if sequence == 0:
            rec['clusterName'] = "Cluster%s" % cluster_id
        return rec

    def create_heartbeat_records(self, interval_start_unix):
        """
            Generate a heartbeat message for each input partition
        """
        if self.generate_heartbeats is False:
            return

        for partition in range(self.monitoring_topic_partitions):
            rec = {
                "clientType": "CONTROLCENTER",
                "clientId": "0",
                "timestamp": str(interval_start_unix * 1000 + 78),
                "partition": -1.0,
                "samplePeriod": "-1",
                "type": "HEARTBEAT",
                "minWindow": "-1",
                "maxWindow": "-1"
            }
            self.print_record(rec, partition=partition)

    def print_record(self, rec, partition=None):
        partition = "null" if partition is None else str(partition)
        print("%s\t%s" % (partition, json.dumps(rec)))

    def generate(self):
        self.print_times()

        # Set the clock upfront before generating data
        self.create_heartbeat_records(self.start_window_unix - self.INTERVAL_SECS)

        for interval_start_unix in range(self.start_window_unix, self.start_window_unix + self.duration_mins * 60, self.INTERVAL_SECS):
            tp_counts = self.create_producer_records(interval_start_unix)
            self.create_consumer_records(interval_start_unix, tp_counts)
            self.create_heartbeat_records(interval_start_unix)

        self.print_times()


def main():
    now_unix = int(time.time())
    import argparse
    parser = argparse.ArgumentParser(description='C3 Interceptor Simulator')
    parser.add_argument('spec', help='Path to spec file.')
    parser.add_argument('--start-time-unix', dest='start_time_unix', type=int,
                        default=now_unix, help='Unix time to start simulation')
    parser.add_argument('monitoring_topic_partitions', type=int, help='Number of partitions for the monitoring topic')
    parser.add_argument('duration_mins', type=int, help='Number of minutes to simulate')
    parser.add_argument('--generate-heartbeats', dest='generate_heartbeats',
                        action='store_true', default=False, help='Generate heartbeat records')

    args = parser.parse_args()
    if args.spec == '-':
        spec = json.load(sys.stdin)
    else:
        with open(args.spec) as infile:
            spec = json.load(infile)

    simulator = MonitoringSimulator(spec, args.start_time_unix, args.monitoring_topic_partitions,
                                    args.duration_mins, args.generate_heartbeats)
    simulator.generate()

if __name__ == "__main__":
    main()
